/********************************************************************
*																	*
*		File :		ESPTYPES.CPP									*
*		Type :		source											*
*		Version :	1.0												*
*		Date :		26/07/2000										*
*		Author :	Marc MILEUR										*
*																	*
********************************************************************/


#include "espoir.h"


	// PRJLIST member functions definition
PRJLIST::PRJLIST ()
: _first (NULL)
{
}	// PRJLIST::PRJLIST


PRJLIST::~PRJLIST ()
{
}	// PRJLIST::~PRJLIST


void PRJLIST::insert (PROJECT * pProject)
{
	PRJCELL * newCell ;

	newCell = new PRJCELL (pProject, _first) ;
	_first = newCell ;
}	// PRJLIST::insert


void PRJLIST::suppress (PROJECT * pProject)
{
	PRJCELL *	currentCell ;
	PRJCELL *	nextCell ;
	BOOL		bContinue ;

	if (_first)
	{
		if (_first->_pProject == pProject)
		{
			nextCell = _first->_next ;
			delete _first ;
			_first = nextCell ;
		}
		else
		{
			currentCell = _first ;
			nextCell = _first->_next ;
			bContinue = TRUE ;
			while (!bContinue)
			{
				if (!nextCell)
					bContinue = FALSE ;
				else
				{
					if (nextCell->_pProject == pProject)
					{
						currentCell->_next = nextCell->_next ;
						delete nextCell ;
						bContinue = FALSE ;
					}
					else
					{
						currentCell = nextCell ;
						nextCell = nextCell->_next ;
					}
				}
			}
		}
	}
}	// PRJLIST::suppress


BOOL PRJLIST::exist (CHAR * szName)
{
	PRJCELL *	currentCell ;
	PRJCELL *	nextCell ;
	BOOL		bContinue ;

	if (_first)
	{
		if (!strcmp (szName, _first->_pProject->szName))
			return TRUE ;
		else
		{
			currentCell = _first ;
			nextCell = _first->_next ;
			bContinue = TRUE ;
			while (!bContinue)
			{
				if (!nextCell)
					return FALSE ;
				else
				{
					if (!strcmp (szName, nextCell->_pProject->szName))
						return TRUE ;
					else
					{
						currentCell = nextCell ;
						nextCell = nextCell->_next ;
					}
				}
			}
		}
	}
	return FALSE ;
}	// PRJLIST::existComp


PROJECT * PRJLIST::find (CHAR * szName)
{
	PRJCELL *	currentCell ;
	PRJCELL *	nextCell ;
	BOOL		bContinue ;

	if (_first)
	{
		if (!strcmp (szName, _first->_pProject->szName))
			return _first->_pProject ;
		else
		{
			currentCell = _first ;
			nextCell = _first->_next ;
			bContinue = TRUE ;
			while (!bContinue)
			{
				if (!nextCell)
					return NULL ;
				else
				{
					if (!strcmp (szName, nextCell->_pProject->szName))
						return nextCell->_pProject ;
					else
					{
						currentCell = nextCell ;
						nextCell = nextCell->_next ;
					}
				}
			}
		}
	}
	return NULL ;
}	// PRJLIST::findComp


	// PRJCELL member functions definition

PRJCELL::PRJCELL (PROJECT * pProject, PRJCELL * nextCell)
: _pProject (pProject), _next (nextCell)
{
}	// PRJCELL::PRJCELL


PRJCELL::~PRJCELL ()
{
	ProjectDestroy (_pProject) ;
}	// PRJCELL::~PRJCELL


	// project initialization
void ProjectInitialize (PROJECT * pProject, HWND hwnd)
{
	INT	i ;

	pProject->szName = new CHAR[BUFFER_SIZE] ;
	for (i = dat ; i < NBEDITFILE ; i ++)
	{
		(pProject->pdEditWnd)[i].hwnd = NULL ;
		(pProject->pdEditWnd)[i].pOfn = new OPENFILENAME ;
	}
	pProject->hwndRunProgress = NULL ;
	pProject->hwndProfile = NULL ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[dat].pOfn, ".dat Files (*.dat)\0*.dat\0All Files (*.*)\0*.*\0\0", "dat") ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[hkl].pOfn, ".hkl Files (*.hkl)\0*.hkl\0All Files (*.*)\0*.*\0\0", "hkl") ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[imp].pOfn, ".imp Files (*.imp)\0*.imp\0All Files (*.*)\0*.*\0\0", "imp") ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[ins].pOfn, ".ins Files (*.ins)\0*.ins\0All Files (*.*)\0*.*\0\0", "ins") ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[spf].pOfn, ".spf Files (*.spf)\0*.spf\0All Files (*.*)\0*.*\0\0", "spf") ;
	FileInitialize (hwnd, (pProject->pdEditWnd)[prf].pOfn, ".prf Files (*.prf)\0*.prf\0All Files (*.*)\0*.*\0\0", "prf") ;
	pProject->pDWThread = new DWORD ;
	pProject->hThread = NULL ;
}	// ProjectInitialization


	// project destroying
void ProjectDestroy (PROJECT * pProject)
{
	INT	i ;

	delete pProject->pDWThread ;
	for (i = dat ; i < NBEDITFILE ; i ++)
	{
		delete [] (pProject->pdEditWnd)[i].pOfn->lpstrFileTitle ;
		delete [] (pProject->pdEditWnd)[i].pOfn->lpstrFile ;
		delete (pProject->pdEditWnd)[i].pOfn ;
	}
	delete [] pProject->szName ;
	delete pProject ;
}	// ProjectDestroy


	// PROGRESSINFO member functions definition

PROGRESSINFO::PROGRESSINFO ()
: _szLine (new CHAR[LINE_SIZE]), _next (NULL)
{
	_szLine[0] = '\0' ;
}	// PROGRESSINFO::PROGRESSINFO


PROGRESSINFO::~PROGRESSINFO ()
{
	if (_next)
	{
		delete _szLine ;
		delete _next ;
	}
}	// PROGRESSINFO::~PROGRESSINFO


PROGRESSINFO::PROGRESSINFO (const PROGRESSINFO & PInfo)
: _szLine (new CHAR[LINE_SIZE]), _next (NULL)
{
	strcpy (_szLine, PInfo._szLine) ;
	if (PInfo._next)
	{
		PROGRESSINFO * next = new PROGRESSINFO ;
		* next = * (PInfo._next) ;
		_next = next ;
	}
	else
		_next = NULL ;
}	// PROGRESSINFO::PROGRESSINFO


PROGRESSINFO & PROGRESSINFO::operator= (const PROGRESSINFO &PInfo)
{
	if (this != & PInfo)
	{
		delete _next ;
		strcpy (_szLine, PInfo._szLine) ;
		if (PInfo._next)
		{
			PROGRESSINFO * next = new PROGRESSINFO ;
			* next = * (PInfo._next) ;
			_next = next ;
		}
		else
			_next = NULL ;
	}
	return * this ;
}	// PROGRESSINFO::operator=


CHAR * PROGRESSINFO::operator[] (LONG i) const
{
	if ((i < 0) || ((i > 0) && (!_next)))
		exit (0) ;
	if (!i)
		return _szLine ;
	else
		return (* _next)[i - 1] ;

}	// PROGRESSINFO::operator[] const


CHAR * & PROGRESSINFO::operator[] (LONG i)
{
	// CHAR *	szTemp ;

	if (i < 0)
		exit (0) ;
	if (!i)
		return _szLine ;
	else
	{
		if (!_next)
		{
			PROGRESSINFO * next = new PROGRESSINFO ;
			_next = next ;
		}
		return (* _next)[i - 1] ;
	}
}	// PROGRESSINFO::operator[]
