!
!     ESPOIR built into a mixed-language program
!
! 
      SUBROUTINE ESPOIR (PROGRESS, FILE)
*
*     Version 3.50  for Windows
*              with up to 6 independent objects
*              and many additional output files for drawing  
*
*     ESPOIR in french = HOPE in english 
*     something not to lose when dealing with structure determination 
*
************************************************************************ 
*
*     A Monte Carlo code for ab initio crystal structure determination
*     by fitting to "|Fobs|" (powder or single crystal) 
*
*     For more details see the documentation at 
*                   http://www.cristal.org/sdpd/espoir/ 
*             or    http://sdpd.univ-lemans.fr/sdpd/espoir/ 
*
*     The MC code was inspired from RMCA written by Malcolm Howe 
*     Last altered as such by 26th May 1992 (VERSION 2.11).
*               However, a more recent VERSION 3.04 is available 
*     See R.L. McGreevy, Nucl. Instr. and Meth. in Phys. Res. A354 
*          (1995) 1-16 for one of the most recent reviews 
*
*     Extensively modified by A. Le Bail - April 1999 for VERSION 0.90 
*                            alb@cristal.org 
*                            http://www.cristal.org/ 
*
*                            Rsidence Cristal - Appt 213 
*                            2, rue de Gasperi 
*                            72100 Le Mans 
*                            FRANCE
*
*    Modifications in VERSION 1.00 :
*     Space groups recognition added
*     Simulated annealing
*     Tolerance for events not improving the fit, annealed too 
*     No more constraints on distances or coordination 
*                                           May 1999 (A. Le Bail) 
*
*
*    Modifications in VERSION 2.00 :
*     Fit possible on a pseudo powder pattern regenerated from 
*        the extracted "|Fobs|" - this allows keeping the whole 
*        set of extracted structure factors, and speed is enhanced 
*        if compared to a fit on the true pattern. Moreover, the 
*        step is variable : indexed on the FWHM 
*     Option of minimal interatomic distances reinserted 
*     Special positions considered
*     Molecular replacement method by rotation + translation 
*        of a fragment
*
*
*                                          June 1999 (A. Le Bail) 
*
*
*    Modification in version 2.01 (November 1999) : 
*      Possibility of twinning by merohedry 
*             (2 domains only with 50% each) 
*     Removed from version 3 : not enough place in arrays...
*
*    Modification in version 2.02 (December 1999) : 
*      Possibility to use molecule global scattering factors 
*           application for instance to orientational disordering 
*           in C60 derivatives
*
*    Modification in cetsion 3.00 (March-April 2000) : 
*      Several objects (up to 4) are now possible.
*      An object can be either :
*        - a rigid molecule
*        - a group of independent atoms 
*
*
*    Modifications in version 3.5 (June-July 2000)  :
*      Torsion angles
*      GUI for Windows by Marc Mileur
*      various new output files for drawing
*            .xyz including animation, .wrl, .hgl, .ps, etc
*      etc, etc, etc
*
*
*                                     July 2000 (A. Le Bail) 
*
************************************************************************ 
*
*    Copyright (C) 1999,2000 Armel Le Bail 
*
* This program is free software; you can redistribute it and/or 
* modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License 
* along with this program; if not, write to the Free Software 
* Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*
*********************************************************************** 
*
*
*
C      use dfwin
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      INCLUDE 'esp2.inc'
	INTEGER(4) PROGRESS
	CHARACTER*120 BUFFER
      CHARACTER VERSION*8,NAM(N_T)*4,NAMC*4
      CHARACTER*(*) FILE
      CHARACTER*24 TEMPO
      CHARACTER STRSYM*50,SELECT1*80,PUFFER*7
      LOGICAL QEX
      LOGICAL FIRST,LOOPING,ACCEPTABLE
      PARAMETER(VERSION='3.50')
C
C
      CHARACTER*4 TABNC(128),TBXC(256),NX(N_O,N_A),NXB(N_O,N_A)
      CHARACTER*4 NAMB(N_T),NAMX(N_T),NAX(N_T),NAXB(N_T)
      CHARACTER*4 NXBB(6000),NAMT
C
C
      REAL MOVE(192,6),MOVEP(2,6),METRIC(3,3)
      INTEGER NCUM(N_O,N_T),NCUM1(N_O,N_T),SS,NTO(N_O),IGR2(7),IGR3(7)
      DIMENSION NAT(N_O,N_T),BOV(N_O),NOCC(N_O),NSPE(N_O)
      DIMENSION NPERM(N_O),MR(N_O),SROT(N_O),NAO(N_O),NTR(N_O),NRT(N_O)
      DIMENSION HNN(3,N_HKL),TEXT(20),XRYZ(5),XG(N_O,3),PREV(N_O,50)
      DIMENSION XL(N_O,5,N_A),XOLD(N_O,3,N_A),XLN(N_O,3,N_A)
      DIMENSION XLNN(N_O,3,N_A),XLM(N_O,3,N_A),XLS(N_O,3,N_A)
      DIMENSION TEMP(N_O,N_A),MTYP(N_O,N_A),B(N_T),FF(N_T)
      DIMENSION FOBS(N_HKL),FNUC(N_HKL),FNUCT(N_HKL),D(N_HKL)
      DIMENSION FC(N_T),DELFR(N_T),FA(N_T,10),FB(N_T,10)
      DIMENSION F(N_T,100),S(N_T,100),FX(N_T),SX(N_T),MP(N_T)
      DIMENSION AL(3,3),CELPRE(6),RCELLN(6)
      DIMENSION AL2(N_O,3,3),CELPR2(N_O,6),RCELLN2(N_O,6),V2(N_O)
      DIMENSION TM2(6),TM3(N_O,6),TM4(N_O,6),NCART(N_O)
      DIMENSION HH(3),IHH(3,N_HKL),FFF(N_T,N_HKL),IREFS(N_HKL)
      DIMENSION T(24,3),SM(24,3,3),IGR(76),PP1(N_O),PP2(N_O)
      DIMENSION RCUT(36),DELTA(N_T),NSP(N_O,N_A),NATOT(N_T)
      DIMENSION SS(24,3,3),TSS(24,3),IM(3,3),TS(3)
      DIMENSION NACC(N_O),NGEN(N_O),NTRIED(N_O),NACCP(N_O)
      DIMENSION NTRIEDP(N_O),NREJECT(N_O),DAMP(N_O),DAMP0(N_O)
      DIMENSION NGENP(N_O),NGENT(N_O),NPER(N_O),LASTPRINT(N_O)
      DIMENSION NGRAV(N_O),SHIFT(N_O,3),OMOVE(N_O),DELTAO(N_O)
      DIMENSION NXLL(6000),NZERO(N_O)
      DIMENSION XXX(6000),YYY(6000),ZZZ(6000)
      DIMENSION XX(6000),YY(6000),ZZ(6000)
      DIMENSION XB(6000),YB(6000),ZB(6000)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3)
      COMMON/TABCHR/TABNC,TBXC
      COMMON/TABLES/TBX(256,10),TBD(128,10),TABN(128) 
      COMMON/ANI/STST(3),NOB,NOBT(N_O),ISYST,TM1(6),NXB,ORI(3,3)
      COMMON/TORS/NBROT(N_O),NA1(N_O,500),NA2(N_O,500),NTOR(N_O,50,N_A)
C
C
      DATA XRYZ(1),XRYZ(2),XRYZ(3),XRYZ(4),XRYZ(5)
     */2.28962,1.93597,1.54056,0.70926,0.556363/ 
      DATA IGR2/'A','B','C','D','M','N','P'/
      DATA IGR3/'a','b','c','d','m','n','p'/
C
      PI2=6.2831853
      PI=114.59156  ! 360./3.1415926
      N1=1
      N2=2
C
C      PRINT 401
C401   FORMAT('  Entry file (no extension) ??',$)
C      READ 402,FILE
C      WRITE(*,*)FILE
402   FORMAT(A20)
      LFILE=LEN(FILE)
      DO WHILE(FILE(LFILE:LFILE).EQ.' ')
      LFILE=LFILE-1
      ENDDO
      LC0=19
      LCA=21
      IAN=18
      TEMPO=FILE(1:LFILE)//'.inp'
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 2600
      CALL FILEDEL(21,TEMPO)
2600  CALL OPEN_WRITE1(21,TEMPO)
      TEMPO=FILE(1:LFILE)//'.dat'
      CALL OPEN_READ1(19,TEMPO)
2603  READ(19,2601,ERR=2602)SELECT1
      IF(SELECT1(1:1).NE.'!')WRITE(21,2601)SELECT1
      GO TO 2603
2601  FORMAT(A80)
2602  CLOSE(LC0)
      CLOSE(LCA)
      TEMPO=FILE(1:LFILE)//'.inp'
      CALL OPEN_READ1(21,TEMPO)
      LPR=20
      TEMPO=FILE(1:LFILE)//'.imp'
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 403
      CALL FILEDEL(20,TEMPO)
403   CALL OPEN_WRITE1(20,TEMPO)
      LHKL=22
      TEMPO=FILE(1:LFILE)//'.hkl'
      CALL OPEN_READ1(22,TEMPO)
	WRITE(BUFFER,*)'ESPOIR version ',VERSION
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE(BUFFER,*)'Data file : ',FILE(1:LFILE)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,100) VERSION,FILE(1:LFILE)
      CALL DATN
C
C.....READ PROBLEM IDENTIFICATION
C
C      READ(LCA,*)FILE
      READ(LCA,500,ERR=3500) TEXT
500   FORMAT(20A4)
C
C.....INPUT CELL DIMENSIONS
C
      DO 829 I=1,3
      DO 829 J=1,3
829   AL(I,J)=0.0
      READ(LCA,*,ERR=3501) (CELPRE(I),I=1,6)
      WRITE(20,3050)(CELPRE(I),I=1,6)
3050  FORMAT(/' Cell parameters : ',3F9.4,3F10.3/)
      CALL DCELL(CELPRE,AL,RCELLN,V1)
      CALL MATRIX(CELPRE,RCELLN,TM1,TM2)
*
*------------------------------------------------------------------------- 
*     a,b,c Cell PARAMETERs
*
      A=CELPRE(1)
      BB=CELPRE(2)
      C=CELPRE(3)
C
C  Tolerated maximum difference with special position
C
      STST(1)=0.2/A
      STST(2)=0.2/BB
      STST(3)=0.2/C
C
*
*------------------------------------------------------------------------- 
*     Calculate METRIC matrix
*
      ALPH=CELPRE(4)*PI2/360.
      BET=CELPRE(5)*PI2/360.
      GAMM=CELPRE(6)*PI2/360.
      METRIC(1,1)=A**2
      METRIC(2,2)=BB**2
      METRIC(3,3)=C**2
      METRIC(1,2)=A*BB*COS(GAMM)
      METRIC(1,3)=A*C*COS(BET)
      METRIC(2,3)=BB*C*COS(ALPH)
      METRIC(2,1)=METRIC(1,2)
      METRIC(3,1)=METRIC(1,3)
      METRIC(3,2)=METRIC(2,3)
C
C  Cell for animation
C
      ORI(1,1)=TM1(1)
      ORI(1,2)=0.
      ORI(1,3)=0.
      ORI(2,1)=TM1(2)
      ORI(2,2)=TM1(4)
      ORI(2,3)=0.
      ORI(3,1)=TM1(3)
      ORI(3,2)=TM1(5)
      ORI(3,3)=TM1(6)
C
      WRITE(20,902)(RCELLN(I),I=1,6)
902   FORMAT(' Reciprocal cell parameters : ',3F9.4,3F10.3/) 
C
C.....READ SPACE GROUP
C
      READ(LCA,479,ERR=3502)(IGR(IR),IR=1,76)
  479 FORMAT(76A1)
C
C Change small characters in capital letters
C
      DO 903 IG=1,76
      DO 903 IG2=1,7
      IF(IGR(IG).EQ.IGR3(IG2))IGR(IG)=IGR2(IG2)
903   CONTINUE
      WRITE(LPR,478)(IGR(IR),IR=1,76)
  478 FORMAT(' SPACE GROUP : ',76A1/)
      CALL BURZ (0,ISYST,ISY,ISYB,IRL,SS,TSS,IGR)
      IF(ISYST.EQ.7)WRITE(20,3730)ISYST
      IF(ISYST.EQ.1)WRITE(20,3731)ISYST
      IF(ISYST.EQ.2)WRITE(20,3732)ISYST
      IF(ISYST.EQ.3)WRITE(20,3733)ISYST
      IF(ISYST.EQ.4)WRITE(20,3734)ISYST
      IF(ISYST.EQ.5)WRITE(20,3735)ISYST
3730  FORMAT('  BURZLAFF Crystal system : ',I3,' Triclinic'/)
3731  FORMAT('  BURZLAFF Crystal system : ',I3,' Monoclinic'/)
3732  FORMAT('  BURZLAFF Crystal system : ',I3,' Orthorhombic'/)
3733  FORMAT('  BURZLAFF Crystal system : ',I3,' Tetragonal'/)
3734  FORMAT('  BURZLAFF Crystal system : ',I3,' Hexagonal/Trigonal'/)
3735  FORMAT('  BURZLAFF Crystal system : ',I3,' Cubic'/)
C-----MAKE ARRAY FROM BURZ COMPATIBLE
      DO 125 I=1,3
      DO 125 IJ=1,3
      DO 125 IK=1,IRL
      SM(IK,IJ,I)=SS(IK,IJ,I)
      T(IK,IJ)=TSS(IK,IJ)
125   CONTINUE
C
C...READ
C              SLABDA = Wavelength
C              KXR (1 = neutron, >= 4 = RX) 
C              NA = atom number (total) 
C              KLL = number of different atoms - neutrons 
C              KMN = number of different atoms - X-rays 
C              NOB = number of different objects 
C                    either fragments or group of independent atoms 
C              KP : IF 0, fit on |F|s,
C                   IF 1, on profile
C
C  IPRI = code for printing Iobs/Icalc and Fobs/Fcalc at the END of a run 
C         and STRUVIR calculations for detecting polyhedra
C       IF = 0 : no printout
C       IF = 1 : printout
C       IF = 10 : same as 1, moreover, large nameN.xyz files are created
C                 for showing animation of accepted moves
C                 (N = test number). Each file will contain 
C                 several sets of positions
C
      READ(LCA,*,ERR=3503)SLABDA,KXR,NA,KLL,NOB,KP,IPRI 
C
C   Test if files for animation should be prepared
C
      IANIM=0
      IPRI2=IPRI
      IF(IPRI.GE.4)THEN
      IANIM=1
      IPRI=1
      ENDIF
C
C  Test for scattering factors parametrized or not 
C
      KKLL=0
      IF(KLL.LT.0)KKLL=1
      KLL=IABS(KLL)
C
C  If KP=1  read the U,V,W,STEPN
C        STEPN is the number of points used for describing 
C              the upper part of a peak, above the FWHM 
C              Use STEPN between 3 and 5 
C
      U=0.
      V=0.
      W=0.005
      STEPN=3.
      IF(KP.NE.1)GO TO 9750
      READ(LCA,*,ERR=3504)U,V,W,STEPN
      IF(STEPN.LT.3.)STEPN=3.
C
C-- Some tests on the program limits
C
9750  NUAT=N_O*N_A
      IF (NA.GT.NUAT) THEN
	WRITE (BUFFER,'(I4,A)') na,' atoms are given'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(BUFFER,'(A,I4)')'Arrays are only large enough for ',NUAT
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,*) na,' atoms are given'
      WRITE (20,*) 'Arrays are only large enough for ',NUAT
      RETURN
      ENDIF
      IF (KLL.GT.N_T) then
	WRITE (BUFFER,*) KLL,' atom types are given'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,*) 'Arrays are only large enough for ',N_T
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,*) KLL,' atoms types are given'
      WRITE (20,*) 'Arrays are only large enough for ',N_T
      RETURN
      ENDIF
      IF (NOB.GT.N_O) then
      WRITE (BUFFER,*) NOB,' different objects are given'
      CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (BUFFER,*) 'Arrays are only large enough for ',N_O
      CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,*) NOB,' different objects are given'
      WRITE (20,*) 'Arrays are only large enough for ',N_O
      RETURN
      ENDIF
      WRITE(20,5005)SLABDA
5005  FORMAT(' Wavelength : ',F9.5/)
      IF(KXR.LT.4)WRITE(20,*)'Neutron data'
      IF(KXR.GE.4)WRITE(20,*)'X-ray data'
      KMN=KLL
      IF(KXR.GE.4)KLL=0
      IF(KXR.LT.4)KMN=0
C   Wavelength identification
      DO 404 IX=1,5
      IF(SLABDA+0.1.GT.XRYZ(IX))GOTO 405 
404   CONTINUE
405   IXRAY=IX
      WRITE(20,*)
      IF(KP.EQ.0)WRITE(20,*)'The fit will be on the "|Fobs|"'
      IF(KP.EQ.1)WRITE(20,*)'The fit will be on the pseudo powder
     - pattern'
      IF(KP.EQ.1)WRITE(20,5011)U,V,W,STEPN
      IF(KP.NE.1)WRITE(20,5010)U,V,W,STEPN
5011  FORMAT('    with U, V, W, STEPN  = ',4F9.4) 
5010  FORMAT('    File .prf built with U, V, W, STEPN  = ',4F9.4) 
C
C.....Scattering factors retrieved by names 
C
        IF(KLL)406,407,406
406   READ(LCA,408,ERR=3505)(NAMB(I),I=1,KLL)
408   FORMAT(20A4)
      DO 409 I=1,KLL
      NAM(I)=NAMB(I)
C
C--- Change large characters in small
C
      NAMC=NAMB(I)
      CALL SMALLC(NAMC)
      NAMB(I)=NAMC
C
      DO 410 J=1,92
      IF(NAMB(I).EQ.TABNC(J)) GO TO 411
410   CONTINUE
      WRITE(LPR,412)NAMB(I)
412   FORMAT(34H SCATTERING LENGTHS NOT FOUND FOR ,A4)
      RETURN
411   FF(I)=TABN(J)
      B(I)=FF(I)
409   CONTINUE
407   IF(KMN)413,413,414
414   READ(LCA,408,ERR=3506)(NAMX(I),I=1,KMN)
      DO 415 I=1,KMN
      NAM(I)=NAMX(I)
C
C--- Change large characters in small
C
      NAMC=NAMX(I)
      CALL SMALLC(NAMC)
      NAMX(I)=NAMC
C
      DO 416 J=1,213
      IF(NAMX(I).EQ.TBXC(J)) GO TO 417
416   CONTINUE
      WRITE(LPR,418)NAMX(I)
418   FORMAT(' SCATTERING COEFFICIENTS NOT FOUND FOR ',A4)
      RETURN
417   FA(1,I)=TBX(J,1)
      FB(1,I)=TBX(J,2)
      FA(2,I)=TBX(J,3)
      FB(2,I)=TBX(J,4)
      FA(3,I)=TBX(J,5)
      FB(3,I)=TBX(J,6)
      FA(4,I)=TBX(J,7)
      FB(4,I)=TBX(J,8)
      FC(I)=TBX(J,9)
      IBX=INT(TBX(J,10)+0.5)
      DELFR(I)=TBD(IBX,IXRAY)
      B(I)=TBD(IBX,IXRAY+5)
415   CONTINUE
C
C If the scattering factors are given as couples Sintheta/lamda, f 
C
      IF(KKLL.EQ.0)GO TO 642
      DO 640 I=1,KMN
      READ(LCA,*,ERR=3507) MP(I),SX(I),FX(I),(S(I,L),F(I,L),L=1,MP(I))
      DO 641 LL=1,MP(I)
      S(I,LL)=S(I,LL)/SX(I)
      F(I,LL)=F(I,LL)*FX(I)
  641 CONTINUE
  640 CONTINUE
642   CONTINUE
C
C
C.....READ AND PRINT
C
413   NTYP=KLL
      IF(KLL.EQ.0)NTYP=KMN
C  NPAR = number of different atomic pairs
      NPAR=NTYP*(NTYP+1)/2
C  NCON = code for constraints on shortest interatomic distances 
C       IF = 1 : read RCUT below
C       IF = 0 : no constraints
      READ(LCA,*,ERR=3508)NCON
C
      IF(NCON.EQ.0)GO TO 9751
      READ(LCA,*,ERR=3509)(RCUT(IC),IC=1,NPAR)
      WRITE(20,5008)(RCUT(IC),IC=1,NPAR) 
5008  FORMAT(/' There will be minimal distance constraints :'/,
     -10X,10F8.3)
9751  CONTINUE
C
C  DELTA(i) is largest distance for moving atom-type i
      READ(LCA,*,ERR=3510) (DELTA(I),I=1,NTYP)
C  ANNEAL is the law on moves reduction 
C      MOVE=MOVE*(1.-NGEN/TIMELIM)**ANNEAL 
C  SIGMA is precision expected from data 
C  REJECT = accept a MOVE even if -REJECT < DELTAR < 0
      IF(IPRI2.GE.2)WRITE(20,5012)
5012  FORMAT(/' A .prf file will be created, and the
     - list of Fo/Fc will be given'/)
      WRITE(20,5013)NOB
5013  FORMAT(///' The model contains ',I3,' objects')
      READ(LCA,*,ERR=3511)ANNEAL,SIGMA,REJECT
C  print each iprint moves, up to timelin moves, saving each TIMESAV moves
      READ (LCA,*,ERR=3512) IPRINT,TIMELIM,TIMESAV 
C
C  IF after NSTART events, the R factor is larger than RMAX, then restart 
C
C  ICHI = 1 : test on Chi**2, ICHI=2 : test on R 
C
C  NTEST = number of runs
C
      READ (LCA,*,ERR=3513) NSTART,RMAX,ICHI,NTEST 
C
C--- Loop on different objects
C
      NTRTEST=0
      DO 2500 NO=1,NOB
C
C   NAO(NO) = Number of atoms in object NO 
C
      NAO(NO)=0
C
C
C  NOBT(NO) = Type of object
C           NOBT = 0 : group of independent atoms entered as fixed 
C           NOBT = 1 : group of independent atoms entered as random 
C           NOBT = 2 : fragment
C           NOBT = 3 : fragment with torsion angles
C           NOBT = 11 : group of independent atoms entered all at 0,0,0
C                      coordinates. Moves are accepted only if the rcut
C                      values are strictly respected.
C  NPERM(NO)       try permutations after NPERM moves 
C                  or try one translation after NPERM rotations 
C                      for a given object NO 
C
      READ(LCA,*,ERR=3514)NOBT(NO),NPERM(NO)
      NZERO(NO)=0
      IF(IABS(NOBT(NO)).EQ.11)NZERO(NO)=1
      NTO(NO)=0
      IF(IABS(NOBT(NO)).EQ.3)THEN
C  NTO is code for torsion angle action
      NTO(NO)=1
      IF(NOBT(NO).EQ.3)NOBT(NO)=2
      ENDIF
      IF(NCON.EQ.0.AND.NZERO(NO).EQ.1)THEN
      WRITE(20,2595)
	WRITE (BUFFER,*)'  You have choosen |NOBT|=11, then you should'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,*)'   select NCON=1 and give the RCUT values...'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      RETURN
2595    FORMAT('  You have chosen |NOBT|=11, then you should'
     1/'   select NCON=1 and give the RCUT values...')
      ENDIF
      IF(NOBT(NO).EQ.11)NOBT(NO)=1
      IF(NOBT(NO).EQ.-11)NOBT(NO)=-1
      NTR(NO)=0
C  IF NPERM < 0, Translation (no check) and then NPERM checked rotations
C  IF NPERM > 0, Translation (checked) and then NPERM checked rotations
C  IF NPERM = 0, Translation or rotation will be done sequentially
      IF(NPERM(NO).LT.0)NTR(NO)=-1
      IF(NTR(NO).EQ.-1)NTRTEST=1
      IF(NPERM(NO).GT.0)NTR(NO)=1
      IF(NPERM(NO).EQ.0)NTR(NO)=0
      NPERM(NO)=IABS(NPERM(NO))
C  Maximum rotation angle
      PP1(NO)=PI2
      PP2(NO)=PI2
C
C  IF NOBT < 0, then play on the maximum rotation angle
C               which is 360 in principle, but can be reduced
C
      IF(NOBT(NO).LT.0)THEN
       IF(NOBT(NO).EQ.-2)READ(LCA,*)PP1(NO)
       IF(NOBT(NO).EQ.-3)READ(LCA,*)PP1(NO),PP2(NO)
       PP1(NO)=6.28319/360*PP1(NO)
       PP2(NO)=6.28319/360*PP2(NO)
       NOBT(NO)=IABS(NOBT(NO))
      IF(NOBT(NO).EQ.3)NOBT(NO)=2
      ENDIF
C
C  IF NOBT = 1, then NTEST random starting configurations are tested 
C  IF NOBT = 0, then a non random starting configuration is entered 
C  IF NOBT = 2  then a rigid model is entered, translated and rotated 
C
C  If IPRI2 > 4 and NOBT(NO)=2, then IPRI=4
C       This for not entering in STRUVIR with a molecule
      IF(IPRI2.GT.4.AND.NOBT(NO).EQ.2)IPRI2=4
C
C  NAT(NO,I) = number of atoms of each type in that object
      READ(LCA,*,ERR=3515)(NAT(NO,I),I=1,NTYP)
      DO 2502 I=1,NTYP
2502  NAO(NO)=NAO(NO)+NAT(NO,I)
      IF(NAO(NO).GT.N_A)GO TO 3527
      SROT(NO)=FLOAT(NAO(NO))
C
C  BOV(NO)  = B Overall thermal parameter for that object 
C  NOCC(NO) = code for reading individual guessed occupation factors 
C        IF = 1 : the next line should give all the occup. factors 
C        IF = 0 : the program generates all occup = 1.
C  NSPE(NO) = code for general or special positions 
C        IF = 1 : read special positions codes NSP below 
C        IF = 0 : all atoms in general position, do not read anything
      READ(LCA,*,ERR=3516)BOV(NO),NOCC(NO),NSPE(NO)
      WRITE(20,5006)BOV(NO),NO
5006  FORMAT(/' Overall thermal parameter B =',F8.2,' for object',I3/)
      IF(NOCC(NO).EQ.0)GO TO 9748
      READ(LCA,*,ERR=3517)(XL(NO,5,I),I=1,NAO(NO))
      WRITE(20,5007)NO,(XL(NO,5,I),I=1,NAO(NO)) 
5007  FORMAT(' You think to have guessed the following occupation
     - factors for object :'I3/,10X,10F8.3) 
C
9748  DO 9746 I=1,NAO(NO)
9746  NSP(NO,I)=1
      IF(NSPE(NO).EQ.0)GO TO 9749
      READ(LCA,*,ERR=3518)(NSP(NO,I),I=1,NAO(NO))
      WRITE(20,5009)NO,(NSP(NO,I),I=1,NAO(NO)) 
5009  FORMAT(/' The special position codes for object'I3'
     - are :',/10X,10I4/)
C
C  MR(NO) is code for Molecular Replacement for object NO 
C
9749  MR(NO)=0
      IF(NOBT(NO).EQ.2)MR(NO)=NOBT(NO)
C
C If molecular replacement, read also the cell parameters 
C    in which the starting rigid model is described 
C
      NCART(NO)=0
      IF(MR(NO).EQ.0)GO TO 901
      READ(LCA,*,ERR=3519)(CELPR2(NO,I),I=1,6)
      IF(CELPR2(NO,1).EQ.0.)NCART(NO)=1
      IF(NCART(NO).EQ.0)GO TO 899
      CELPR2(NO,1)=100.
      CELPR2(NO,2)=100.
      CELPR2(NO,3)=100.
899   WRITE(20,900)(CELPR2(NO,I),I=1,6) 
900   FORMAT(/' Starting rigid model cell parameters : '/
     -20X,3F9.4,3F10.3/)
      CALL DCELL2(CELPR2,AL2,RCELLN2,V2,NO)
      WRITE(20,902)(RCELLN2(NO,I),I=1,6)
      CALL MATRIX2(CELPR2,RCELLN2,TM3,TM4,NO) 
901   CONTINUE
      IF(NOBT(NO).EQ.1)GO TO 2526
      WRITE(20,5000)
      DO 2525 I=1,NAO(NO)
      READ(LCA,*,ERR=3520)(XL(NO,J,I),J=1,3),XL(NO,5,I)
      WRITE(20,2527)(XL(NO,J,I),J=1,3),XL(NO,5,I) 
2525  CONTINUE
2526  CONTINUE
2527  FORMAT(20X,3F9.4,F10.2)
C
C--  End loop on objects
C
2500  CONTINUE
C
C--- Test on the total number of atoms
C
      NAVER=0
      DO 2503 NO=1,NOB
2503  NAVER=NAVER+NAO(NO)
      IF(NAVER.EQ.NA)GO TO 2504
	WRITE (BUFFER,*)'  Something wrong in the total atom number'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(20,2505)
2505  FORMAT(/'  Something wrong in the total atom number')
      RETURN
C
C.....BUILD AND PRINT OF SYMMETRY MATRICES C
2504  IABRA=ISYB
      ICENT=ISY+1
      NAG=IRL
C   IF BRAVAIS LATTICE R,A,B,C, THEN ADD SYMMETRY WITH TRANSLATION 
C                IABRA=5,6,7
C        P I F R A B C = 1 2 3 4 5 6 7
      GO TO(7199,7199,7199,6000,7100,7120,7130),IABRA
      GO TO 7199
6000  IRL=IRL*3
      DO 6001 IR=NAG+1,IRL-NAG
      DO 6001 I=1,3
      DO 6001 J=1,3
      IRR=IR-NAG
6001  SM(IR,I,J)=SM(IRR,I,J)
      DO 6002 IR=2*NAG+1,IRL
      DO 6002 I=1,3
      DO 6002 J=1,3
      IRR=IR-2*NAG
6002  SM(IR,I,J)=SM(IRR,I,J)
      DO 6003 IR=NAG+1,IRL-NAG
      IRR=IR-NAG
      T(IR,1)=T(IRR,1)+0.666667
      T(IR,2)=T(IRR,2)+0.333333
6003  T(IR,3)=T(IRR,3)+0.333333
      DO 6004 IR=2*NAG+1,IRL
      IRR=IR-2*NAG
      T(IR,1)=T(IRR,1)+0.333333
      T(IR,2)=T(IRR,2)+0.666667
6004  T(IR,3)=T(IRR,3)+0.666667
      GO TO 7199
7100  IRL=IRL*2
      DO 7001 IR=NAG+1,IRL
      DO 7001 I=1,3
      DO 7001 J=1,3
      IRR=IR-NAG
7001  SM(IR,I,J)=SM(IRR,I,J)
      DO 7102 IR=NAG+1,IRL
      IRR=IR-NAG
      T(IR,1)=T(IRR,1)
      T(IR,2)=T(IRR,2)+0.5
7102  T(IR,3)=T(IRR,3)+0.5
      GO TO 7199
7120  IRL=IRL*2
      DO 7002 IR=NAG+1,IRL
      DO 7002 I=1,3
      DO 7002 J=1,3
      IRR=IR-NAG
7002  SM(IR,I,J)=SM(IRR,I,J)
      DO 7122 IR=NAG+1,IRL
      IRR=IR-NAG
      T(IR,1)=T(IRR,1)+0.5
      T(IR,2)=T(IRR,2)
7122  T(IR,3)=T(IRR,3)+0.5
      GO TO 7199
7130  IRL=IRL*2
      DO 7131 IR=NAG+1,IRL
      DO 7131 I=1,3
      DO 7131 J=1,3
      IRR=IR-NAG
7131  SM(IR,I,J)=SM(IRR,I,J)
      DO 7132 IR=NAG+1,IRL
      IRR=IR-NAG
      T(IR,1)=T(IRR,1)+0.5
      T(IR,2)=T(IRR,2)+0.5
7132  T(IR,3)=T(IRR,3)
7199  CONTINUE
691   FORMAT(2X,I4)
      IF(ICENT.EQ.2) GOTO 690
      WRITE(LPR,301)
      GOTO 700
  690 WRITE(LPR,303)
  700 IF(IRL-2) 751,710,710
  710 WRITE(LPR,49)
      WRITE(LPR,1611)
      DO 740 IR=1,IRL
      WRITE(LPR,152) IR,((SM(IR,I,J),J=1,3),I=1,3),(T(IR,I),I=1,3)
  740 CONTINUE
  751 CONTINUE
C      IF(NCON.EQ.0)GO TO 9720
C.....
      IBRA=1   ! IF Bravais lattice P, A, B, C 
C
C  Extension of symmetry operators
C
      DO 760 IR=1,IRL
      DO 760 I=1,3
      TT(IR,I)=T(IR,I)
      DO 760 J=1,3
760   SMT(IR,I,J)=SM(IR,I,J)
      NAG=IRL
C   IF BRAVAIS LATTICE I, F,  THEN ADD SYMMETRY WITH TRANSLATION 
C                IABRA=2,3
C        P I F R A B C = 1 2 3 4 5 6 7
      GO TO(7200,6500,6600),IABRA
      GO TO 7200
6500  IRL=IRL*2
      DO 6501 IR=NAG+1,IRL
      DO 6501 I=1,3
      DO 6501 J=1,3
      IRR=IR-NAG
6501  SMT(IR,I,J)=SMT(IRR,I,J)
      DO 6502 IR=NAG+1,IRL
      IRR=IR-NAG
      TT(IR,1)=TT(IRR,1)+0.5
      TT(IR,2)=TT(IRR,2)+0.5
6502  TT(IR,3)=TT(IRR,3)+0.5
      IBRA=2
      IRL=IRL/2
      GO TO 7200
6600  IRL=IRL*4
      DO 6601 IR=NAG+1,IRL-2*NAG
      DO 6601 I=1,3
      DO 6601 J=1,3
      IRR=IR-NAG
6601  SMT(IR,I,J)=SMT(IRR,I,J)
      DO 6602 IR=2*NAG+1,IRL-NAG
      DO 6602 I=1,3
      DO 6602 J=1,3
      IRR=IR-2*NAG
6602  SMT(IR,I,J)=SMT(IRR,I,J)
      DO 6603 IR=3*NAG+1,IRL
      DO 6603 I=1,3
      DO 6603 J=1,3
      IRR=IR-3*NAG
6603  SMT(IR,I,J)=SMT(IRR,I,J)
      DO 6604 IR=NAG+1,IRL-2*NAG
      IRR=IR-NAG
      TT(IR,1)=TT(IRR,1)+0.5
      TT(IR,2)=TT(IRR,2)
6604  TT(IR,3)=TT(IRR,3)+0.5
      DO 6005 IR=2*NAG+1,IRL-NAG
      IRR=IR-2*NAG
      TT(IR,1)=TT(IRR,1)+0.5
      TT(IR,2)=TT(IRR,2)+0.5
6005  TT(IR,3)=TT(IRR,3)
      DO 6006 IR=3*NAG+1,IRL
      IRR=IR-3*NAG
      TT(IR,1)=TT(IRR,1)
      TT(IR,2)=TT(IRR,2)+0.5
6006  TT(IR,3)=TT(IRR,3)+0.5
      IBRA=4
      IRL=IRL/4
7200  CONTINUE
C
C
C  Add inversion center
C
C
      NAS=IRL*IBRA
      IF(ICENT.EQ.1)GO TO 782
      DO 780 II=1,NAS
      IR=NAS+II
      DO 780 I=1,3
      TT(IR,I)=TT(II,I)
      DO 780 J=1,3
780   SMT(IR,I,J)=0.
      DO 781 II=1,NAS
      IR=NAS+II
      DO 781 K=1,3
      DO 781 L=1,3
      SMT(IR,K,L)=-SMT(II,K,L)
781   CONTINUE
782   CONTINUE
C
C  NAS=IRL*IBRA*ICENT = total number of symmetry operators 
C
      NAS=IRL*IBRA*ICENT
      WRITE(20,*)' Total number of symmetry operators = ',NAS
      WRITE(LPR,*)'   Full list of symmetry operators :'
      WRITE(LPR,1611)
      DO 803 IR=1,NAS
      WRITE(LPR,152)IR,((SMT(IR,I,J),J=1,3),I=1,3),(TT(IR,I),I=1,3)
  803 CONTINUE
C
C  Print some data
C
9720  CALL LIST(TEXT,SIGMA,NTYP,NSTART,RMAX,NTEST,
     -DELTA,ANNEAL,IPRINT,TIMELIM,TIMESAV,REJECT,PROGRESS)
      WRITE (20,1100) NA,NTYP
C
C.....PRINT SCATTERING FACTORS
C
      IF(KKLL.EQ.1)GO TO 424
      IF(KXR-4) 419,425,425
 419  WRITE(LPR,51)
      WRITE(LPR,52) (B(I),I,I=1,KLL)
      IF(KMN) 424,424,423
 423  IF(KXR-4) 424,425,425
 425  WRITE(LPR,55)
      DO 426 I=1,KMN
      WRITE(LPR,427)NAMX(I),(FA(J,I),FB(J,I),J=1,4),FC(I),DELFR(I),B(I) 
426   CONTINUE
427   FORMAT(1X,A4,8F9.5/5X,3F9.5)
424   CONTINUE
C.....
C
C---  Test if some object should not move at all
C
C     XNOB is used to share the total number of moves
C          between objects
C
      XNOB=FLOAT(NOB)
      NOMOVE=0
      DO NO=1,NOB
      OMOVE(NO)=0.
      DO I=1,NTYP
      OMOVE(NO)=OMOVE(NO)+NAT(NO,I)*DELTA(I)
      ENDDO
      ENDDO
      TOTMOVE=0.
      DO NO=1,NOB
      TOTMOVE=TOTMOVE+OMOVE(NO)
      DELTAO(NO)=OMOVE(NO)/FLOAT(NAO(NO))
      IF(OMOVE(NO).EQ.0.)THEN
      NOMOVE=NOMOVE+1
      XNOB=XNOB-1.
	WRITE (BUFFER,*)'   WARNING : Object ',NO,' cannot move at all'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(20,2591)NO
      ENDIF
      IF(DELTAO(NO).NE.0.)THEN
      WRITE(20,2592)NO,DELTAO(NO)
      ENDIF
      ENDDO
      IF(TOTMOVE.EQ.0.)THEN
	WRITE (BUFFER,*)'   Nothing can be translated, are you sure ?'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(20,2528)
      RETURN
      ENDIF
2528  FORMAT(/,'   Nothing can be translated, are you sure ?',/)
2591  FORMAT(/,'   WARNING : Object ',I4,' cannot move at all',/)
2592  FORMAT(/,'   Object ',I4,' can be translated by ',F9.3,/)
C
C   BIG LOOP on NTEST
C
C
      JTEST=0
      DIFFM=2.
      DIFFM1=2.
8999  IF(NTEST.NE.0)JTEST=JTEST+1
      IF(JTEST.EQ.0)JTEST=1
	WRITE (BUFFER,*)'       '
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(A,I4)')'Test number : ',JTEST
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,101) JTEST
101   FORMAT(////1x,'Test number : ',I3)
      CALL DATN
C
C  If ianim=1, open file for animation
C
      IF(JTEST.GE.1000)WRITE(20,*)'Warning : more than 999 tests' 
      IF(IANIM.EQ.0)GO TO 3885
      IF(JTEST.LT.10)THEN
      WRITE(PUFFER, '(I1,A4)') JTEST,'.xyz'   
      TEMPO=FILE(1:LFILE)//PUFFER(1:5)
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 3886
      CALL FILEDEL(IAN,TEMPO)
3886  CALL OPEN_WRITE1(IAN,TEMPO)
      ENDIF
      IF(JTEST.GE.10.AND.JTEST.LT.100)THEN
      WRITE(PUFFER, '(I2,A4)') JTEST,'.xyz'   
      TEMPO=FILE(1:LFILE)//PUFFER(1:6)
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 3887
      CALL FILEDEL(IAN,TEMPO)
3887  CALL OPEN_WRITE1(IAN,TEMPO)
      ENDIF
      IF(JTEST.GE.100)THEN
      WRITE(PUFFER, '(I3,A4)') JTEST,'.xyz'   
      TEMPO=FILE(1:LFILE)//PUFFER(1:7)
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 3888
      CALL FILEDEL(IAN,TEMPO)
3888  CALL OPEN_WRITE1(IAN,TEMPO)
      ENDIF
3885  CONTINUE
*
*------------------------------------------------------------------------- 
*     Initialisation
*
      CALL ESP_INIT(ISEED,TIME_START)
*
*-------------------------------------------------------------------------
      TIME_USED=0.
C
C---  New loop on objects
C
C---  ITOT initialized to 0 (indice of all atoms in the cell)
      ITOT=0
C
      DO 2501 NO=1,NOB
      NREJECT(NO)=0
      I=0
      IF(MR(NO).NE.0.AND.JTEST.GE.2)GO TO 953 
C
C  Loop (Do 428) for XL = atomic coordinates 
C
      WRITE(20,5000)
5000  FORMAT(/' Starting coordinates x,y,z and occupation numbers'/)
      DO 428 L=1,NTYP
      NAX(L)(1:2)=NAM(L)(1:2)
      IF(NAX(L)(2:2).EQ.'+')NAX(L)(2:2)=' '
      IF(NAX(L)(2:2).EQ.'-')NAX(L)(2:2)=' '
      NAMT(1:2)=NAX(L)(1:2)
      CALL BIGC(NAMT)
      NAXB(L)(1:2)=NAMT(1:2)
      DO 428 K=1,NAT(NO,L)
      IF(NAT(NO,L).EQ.0)GO TO 428
      I=I+1
      IF(K.LE.9)WRITE(NX(NO,I),'(a,I1)')NAM(L)(1:2),K
      IF(K.GE.10)WRITE(NX(NO,I),'(a,I2)')NAM(L)(1:2),K
      IF(NX(NO,I)(2:2).EQ.'+')NX(NO,I)(2:2)=' '
      IF(NX(NO,I)(2:2).EQ.'-')NX(NO,I)(2:2)=' '
      NAMT(1:2)=NX(NO,I)(1:2)
      CALL BIGC(NAMT)
      NXB(NO,I)(1:2)=NAMT(1:2)
      IF(NX(NO,I)(2:2).NE.' ')GO TO 1576
      NX(NO,I)(2:2)=NX(NO,I)(3:3)
      NX(NO,I)(3:3)=NX(NO,I)(4:4)
      NX(NO,I)(4:4)=' '
1576  IF(JTEST.GT.1)GO TO 9000
C
C  Cartesian coordinates changed for a cubic cell with a=100A 
C
      IF(NCART(NO).EQ.0)GO TO 898
      XL(NO,1,I)=XL(NO,1,I)/100.
      XL(NO,2,I)=XL(NO,2,I)/100.
      XL(NO,3,I)=XL(NO,3,I)/100.
898   IF(MR(NO).EQ.0)GO TO 9000
C
C  Change coordinates of the rigid model in Cartesian coordinates 
C
C      WRITE(20,904)(XL(NO,J,I),J=1,3),XL(NO,5,I) 
C904   FORMAT(' Starting model  ',3F9.5,3X,F7.3)
      XL(NO,1,I)=TM3(NO,1)*XL(NO,1,I)+TM3(NO,2)*XL(NO,2,I)
     1+TM3(NO,3)*XL(NO,3,I)
      XL(NO,2,I)=TM3(NO,4)*XL(NO,2,I)+TM3(NO,5)*XL(NO,3,I)
      XL(NO,3,I)=TM3(NO,6)*XL(NO,3,I)
C      WRITE(20,905)(XL(NO,J,I),J=1,3),XL(NO,5,I) 
C905   FORMAT(' Cartesian      ',3F9.5,3X,F7.3) 
C
C  Refurbish the coordinates in the unknown structure cell 
C
      XL(NO,1,I)=TM2(1)*XL(NO,1,I)+TM2(2)*XL(NO,2,I)+TM2(3)*XL(NO,3,I)
      XL(NO,2,I)=TM2(4)*XL(NO,2,I)+TM2(5)*XL(NO,3,I)
      XL(NO,3,I)=TM2(6)*XL(NO,3,I)
C
C  We have now the rigid model inside the unknown
C
      GO TO 9001
9000  IF(MR(NO).NE.0) GO TO 400
      IF(NTEST.EQ.0) GO TO 400
C-- BUG corrected August 29, 2000
C      IF(NOBT(NO).EQ.0) GO TO 400
      IF(NOBT(NO).EQ.0) GO TO 9001
C
C  Identify the special position NSP(NO,I)
C
      CALL ESP_SPE(XL,I,NO,NZERO,ISEED,NSP(NO,I))
400   IF(NOCC(NO).EQ.1)GO TO 9001
C
C  XL(NO,5,I)= Occupation number (1=100% of the general position) 
C
      XL(NO,5,I)=1.
9001  MTYP(NO,I)=L
      XL(NO,4,I)=BOV(NO)
428   CONTINUE
      CLOSE (LCA,STATUS='DELETE')
953   IF(MR(NO).EQ.0) GO TO 993
      IF(JTEST.GT.1)GO TO 992
      DO 991 I=1,NAO(NO)
      XOLD(NO,1,I)=XL(NO,1,I)
      XOLD(NO,2,I)=XL(NO,2,I)
991   XOLD(NO,3,I)=XL(NO,3,I)
      GO TO 993
C
C  Restart from Xold for several tests of Molecular Replacement 
C
992   DO 994 I=1,NAO(NO)
      XL(NO,1,I)=XOLD(NO,1,I)
      XL(NO,2,I)=XOLD(NO,2,I)
994   XL(NO,3,I)=XOLD(NO,3,I)
993   I=0
      DO 1009 L=1,NTYP
      DO 1009 K=1,NAT(NO,L)
      IF(NAT(NO,L).EQ.0)GO TO 1009
      I=I+1
      WRITE(20,1010)NX(NO,I),(XL(NO,J,I),J=1,3),XL(NO,5,I) 
1009  CONTINUE
C      IF(NCON.EQ.0)GO TO 2501
C
C  Atoms really in the cell
C
      DO 800 L=1,NAO(NO)
      X=XL(NO,1,L)
      Y=XL(NO,2,L)
      Z=XL(NO,3,L)
      DO 800 K=1,NAS
      ITOT=ITOT+1
      IF(ITOT.GT.6000)GO TO 3521
      XLL(1,ITOT)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLL(2,ITOT)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLL(3,ITOT)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      NXLL(ITOT)=MTYP(NO,L)
C      WRITE(20,*)ITOT,(XLL(JJ,I),JJ=1,3) 
800   CONTINUE
C
C  Then we have to examine the molecule connectivity
C
      IF(NOBT(NO).NE.2.OR.JTEST.GE.2)GO TO 2501 
      CALL TORSION(XL,NX,NXB,METRIC,NAO,NO,IERR)
      IF(IERR.EQ.500)GO TO 3524
      IF(IERR.EQ.50)GO TO 3525
      IF(IERR.EQ.51)GO TO 3526
C
C
2501  CONTINUE
C
C---  End of loop on objects
C
C  NGEN   = number of generated moves
C  NTRIED = Number of tried moves
C  NACC   = Number of accepted moves
      DO 2530 NO=1,NOB
      NGEN(NO)=0
      NTRIED(NO)=0
2530  NACC(NO)=0
1010  FORMAT(A4,3F11.5,5X,F6.3)
C
C.....CYCLE THROUGH INPUT REFLECTIONS
C
      IF(JTEST.GE.2) GO TO 9500
      READ(LHKL,*,ERR=3522)NHKL
      IF (NHKL.GT.1000) then
	   WRITE (BUFFER,*) NHKL,' reflections are given'
	   CALL PROGRESSVIEW (PROGRESS, BUFFER)
	   WRITE (BUFFER,*) 'Arrays are only large enough for 1000'
	   CALL PROGRESSVIEW (PROGRESS, BUFFER)
         WRITE (20,*) NHKL,' reflections are given'
         WRITE (20,*) 'Arrays are only large enough for 1000'
         RETURN
        ENDIF
      SUM_F=0.
      DO 1960 I=1,NHKL
      READ(LHKL,*,ERR=3523)(IHH(J,I),J=1,3),FOBS(I)
      DO 300 KK=1,3
300   HH(KK)=IHH(KK,I)
      SUM_F=SUM_F+FOBS(I)
C     HH IS INDICES OF REFLECTION
      X=0.
      DO 1840 J=1,3
      DO 1840 K=J,3
 1840 X=AL(J,K)*HH(J)*HH(K)+X
C     X IS DSTAR**2 FOR REFLECTION HH
      X=SQRT(X)*0.5
C     X IS DSTAR/2 FOR HH
      FF(1)=0.
      IF(KMN)1950,1950,1920
1920  CONTINUE
      IF(KKLL.EQ.1)GO TO 2041
C  Parametrized scattering factors
      DO 2030 J=1,KMN
      FF(J)=0.
      DO 2030 K=1,4
      AR=-FB(K,J)*X**2
      FF(J)=FF(J)+FA(K,J)*EXP(AR)
2030  CONTINUE
      DO 2040 J=1,KMN
2040  FF(J)=FF(J)+DELFR(J)+FC(J)
      IF(KKLL.EQ.0)GO TO 1950
C  Couples of Sintheta/lambda and f
2041  DO 1940 J=1,KMN
      DO 1930 K=1,MP(J)
      K1=K-1
      IF(K1.EQ.0)K1=1
      IF(X.GT.S(J,K))GOTO 1930
C
C.....LINEAR INTERPOLATION OF SCATTERING FACTOR C
      FF(J)=(S(J,K1)-X)/(S(J,K1)-S(J,K))*(F(J,K)-
     2F(J,K1))+F(J,K1)
      GOTO 1940
 1930 CONTINUE
      FF(J)=F(J,MP(J))
 1940 CONTINUE
1950  DO 1952 J=1,NTYP
1952  FFF(J,I)=FF(J)
      IF(KKLL.EQ.0)GO TO 1960
      WRITE(20,*)'   '
      WRITE(20,*)'  h   k   l   Fobs  sintheta/L  Scattering Factors'
      WRITE(20,1962)(IHH(J,I),J=1,3),FOBS(I),X,(FF(J),J=1,NTYP) 
1960  CONTINUE
1962  FORMAT(3I4,F8.2,F10.2,8F6.1)
      CLOSE(LHKL)
C
CCC Normalization of |F| to 50. as a mean, arbitrarily
      XNHKL=NHKL
      HMIN=100.
      DO 1961 I=1,NHKL
1961  FOBS(I)=FOBS(I)/SUM_F*XNHKL*50.
      SUM_F=XNHKL*50.
C
C   9500 : Jump IF NTEST different from zero 
C
      FLOG =4.*ALOG(2.)
      SLOG =1./SQRT(FLOG)
9500  SUMCAL=0.
      SUM_FS=0.
C
C   Sum on hkl
C
      DO 1000 NM=1,NHKL
      FNUCT(NM)=0.
      IF(JTEST.GE.2) GO TO 111
      DO 105 I=1,3
105   HNN(I,NM)=IHH(I,NM)
C
      DO 109 IR=1,IRL
      TR(IR,NM)=0.0
      DO 108 I=1,3
      H(IR,I,NM)=0.0
      TR(IR,NM)=TR(IR,NM)+T(IR,I)*HNN(I,NM)
      DO 107 J=1,3
107   H(IR,I,NM)= SM(IR,J,I)*HNN(J,NM)+H(IR,I,NM) 
108   CONTINUE
109   CONTINUE
C
      SSS=0.
      DO 110 I=1,3
      DO 110 J=I,3
110   SSS=AL(I,J)*HNN(I,NM)*HNN(J,NM)+SSS 
C
C     SSS=1/d**2
C
      SSNN=0.25*SSS
C
C     Needs to calculate halfwidths and positions 
C
      D(NM)=SQRT(1./SSS)
      SINTH=SLABDA*SLABDA*SSS/4.
      COSTH=1.-SINTH
      TANTH=SQRT(SINTH/COSTH)
      THETA(NM)=ATAN(TANTH)*PI
      UU=U*TANTH*TANTH+V*TANTH+W
      HW(NM)=SQRT(UU)
      IF(HMIN.GT.HW(NM))HMIN=HW(NM)
      HW(NM)=HW(NM)*SLOG
      HW4(NM)=HW(NM)/SLOG*2.0
      BBB(NM)=HW(NM)*HW(NM)
C
111   DO 130 NO=1,NOB
      DO 130 I=1,NAO(NO)
      DO 120 IR=1,IRL
      ARG=TR(IR,NM)
      DO 70 J=1,3
70    ARG=ARG+H(IR,J,NM)*XL(NO,J,I)
      ARG=6.28319*ARG
      COSAR(NO,I,IR,NM)=COS(ARG)
120   SINAR(NO,I,IR,NM)=SIN(ARG)
C     TEMP=Temperature factor for atom I
      IF(JTEST.GE.2) GO TO 130
      TEMP(NO,I)=EXP(-XL(NO,4,I)*SSNN)
130   CONTINUE
C
C
      AV=0.
      BV=0.
      DO 170 NO=1,NOB
      DO 170 I=1,NAO(NO)
      SAI=0.
      SBI=0.
      DO 140 IR=1,IRL
      SAI=SAI+COSAR(NO,I,IR,NM)
      IF(ICENT.EQ.1)SBI=SINAR(NO,I,IR,NM)+SBI 
140   CONTINUE
      SAI=FLOAT(ICENT)*SAI
      IF(KXR.LT.4)GO TO 150
      IF(JTEST.GE.2) GO TO 161
      SNEXIX(NO,I,NM)=FFF(MTYP(NO,I),NM)*TEMP(NO,I)*XL(NO,5,I)
      SNXI(NO,I)=B(MTYP(NO,I))*XL(NO,5,I)*TEMP(NO,I)
      GOTO 161
150   IF(JTEST.GE.2) GO TO 552
      SNEXIN(NO,I)=B(MTYP(NO,I))*TEMP(NO,I)*XL(NO,5,I) 

552   AV=SNEXIN(NO,I)*SAI+AV
      BV=SNEXIN(NO,I)*SBI+BV
      IF(KXR.LT.4)GO TO 170
161   AV=SNEXIX(NO,I,NM)*SAI+AV
      BV=SNEXIX(NO,I,NM)*SBI+BV
      AVI=AV-SNXI(NO,I)*SBI
      AVJ=AV+SNXI(NO,I)*SBI
      BVI=BV-SNXI(NO,I)*SAI
      BVJ=BV+SNXI(NO,I)*SAI
170   CONTINUE
190   IF(KXR.LT.4)GOTO 200
      FNN=(AVI*AVI+AVJ*AVJ+BVI*BVI+BVJ*BVJ)
      GOTO 210
200   FNN=(AV*AV+BV*BV)
210   FNUC(NM)=SQRT(FNN)
      FNUCT(NM)=FNUCT(NM)+FNUC(NM)
      SUMCAL=SUMCAL+FNUC(NM)
1000  CONTINUE
C      WRITE(20,*)' KP, NHKL ',KP,NHKL
CKP      IF(KP.EQ.0.OR.KP.EQ.2)GO TO 999
      IF(JTEST.GE.2) GO TO 999
C  Variable step as a function of the FWHM 
C  The minimum step is FWHMmin/STEPN
      PO=0.01
      STEP=HMIN/STEPN
      POS(1)=0.01
      DO 550 I=2,16000
      TANTH=TAN(POS(I-1)/PI)
      UU=U*TANTH*TANTH+V*TANTH+W
      IF(UU.LE.0.)GO TO 551
      FWHM=SQRT(UU)
      ASTEP=STEP*FWHM/HMIN
      PO=PO+ASTEP
      IF(PO.GT.160.)GO TO 551
550   POS(I)=PO
551   NPTS=I-1
      CALL PROFILE1(NHKL,FOBS)
      SUM_Y=0.
      DO 998 I=N1,N2
998   SUM_Y=SUM_Y+YOBS(I)
999   DIFF=0.
      DO 1001 NM=1,NHKL
      FNUCT(NM)=FNUCT(NM)/SUMCAL*SUM_F
      IF(KP.EQ.1)GO TO 1001
      DIFF=DIFF+ABS(FOBS(NM)-FNUCT(NM))
      SUM_FS=SUM_FS+(FOBS(NM)-FNUCT(NM))**2 
1001  CONTINUE
      IF(KP.EQ.1)GO TO 980
      DIFFA=DIFF
      CALL PROFILE3(NHKL,FNUCT,SUM_FS,DIFFA)
      DIFFA=DIFFA/SUM_Y
      DIFF=DIFF/SUM_F
      GO TO 981
980   CALL PROFILE3(NHKL,FNUCT,SUM_FS,DIFF)
      DIFF=DIFF/SUM_Y
981   IF(JTEST.GE.2)GO TO 2593
	WRITE (BUFFER,*)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(A,F6.3)')'   Starting R value = ',DIFF
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(20,2529)DIFF
2529  FORMAT(/,'   Starting R value = ',F9.3,/)
2593  CONTINUE
*
*------------------------------------------------------------------------- 
*     Cumulated atom types
*
      DO NO=1,NOB
      NCUM(NO,1)=NAT(NO,1)
      ENDDO
      DO NO=1,NOB
      DO ITYPE=2,NTYP
         NCUM(NO,ITYPE)=NCUM(NO,ITYPE-1)+NAT(NO,ITYPE)
      ENDDO
      ENDDO
      DO NO=1,NOB
      NCUM1(NO,1)=NAT(NO,1)*NAS
      ENDDO
      DO NO=1,NOB
      DO ITYPE=2,NTYP
         NCUM1(NO,ITYPE)=NCUM1(NO,ITYPE-1)+NAT(NO,ITYPE)*NAS
      ENDDO
      ENDDO
C      WRITE(20,*)' NCUM, NCUM1 :',(NCUM(NO,I),I=1,NTYP), 
C     -(NCUM1(NO,I),I=1,NTYP)
*
*------------------------------------------------------------------------- 
*------------------------------------------------------------------------- 
*------------------------------------------------------------------------- 
*     This is where the loop begins.
*
      SAVE_TIME=MIN(TIMELIM,TIMESAV)
      DO 2531 NO=1,NOB
C  Counter for permutation or global translation
      NGENP(NO)=0
C  Counter for global rotation and torsion or atom move
      NGENT(NO)=0
C  Counter for accepted events
      NACCP(NO)=0
C  Counter for tried events
      NTRIEDP(NO)=0
C  Starter for Permutation or global translation
      NPER(NO)=-1
      LASTPRINT(NO)=NGEN(NO)-MOD(NGEN(NO),IPRINT)
C  Parameter for damping amplitude of events (moves, rotation)
      DAMP(NO)=1.
      DAMP0(NO)=1.
C  Has the gravity center to be recalculated ?
      NGRAV(NO)=0
C  Test if torsion or not
      NRT(NO)=1
C  Memory of the last applied torsion angle
      DO 2531 JTOR=1,50
      PREV(NO,JTOR)=0.
2531  CONTINUE
      DIFF0=1.
      FIRST=.TRUE.
      LOOPING=.TRUE.
      DO WHILE(LOOPING)
*
*------------------------------------------------------------------------- 
*
      ACCEPTABLE=.TRUE.
      CHISQ1=0.
      DIFF1=0.
      IF(.NOT.FIRST) THEN
*
*------------------------------------------------------------------------- 
*           Decide which object is concerned 
*
      NO=1
      IF(NOB.EQ.1)GO TO 2540
2590  CALL ESP_OBJ(ISEED,NOB,NO)
C
C---  Test if no move accepted on that object
C
      IF(OMOVE(NO).EQ.0.)GO TO 2590
*
*
2540  CONTINUE
      DO 963 I=1,NAO(NO)
      XLNN(NO,1,I)=XL(NO,1,I)
      XLNN(NO,2,I)=XL(NO,2,I)
963   XLNN(NO,3,I)=XL(NO,3,I)
C
C
2559   IF(NPERM(NO).EQ.0)GO TO 2560
       NPER(NO)=NPER(NO)+1
       IF(NTR(NO).EQ.0)GO TO 2560
       IF(NPER(NO).EQ.NPERM(NO)) GO TO 2999
C
C  Apply rotation along a rotatable bond ??
C
      IF(NTO(NO).EQ.0)GO TO 2560
      IF(NBROT(NO).EQ.0)GO TO 2560
C
C  Global rotation or rotation around a rotatable bond ?
C
      CALL ESP_TORROT(ISEED,2,NRT0)
C  If NRT0=1 : Global rotation
C  If NRT0=2 : Rotation around a rotatable bond
      NRT(NO)=NRT0
      IF(NRT(NO).EQ.1)GO TO 2560
C
C  Which torsion angle to be varied ?? (=NBRO)
C
      NBRT=NBROT(NO)
      IF(NBRT.EQ.1)THEN
      NBRO=1
      GO TO 2560
      ENDIF
      CALL ESP_TOR(ISEED,NBRT,NBRO)
*
*------------------------------------------------------------------------- 
*           Generate a random MOVE or a model rotation 
*
2560   NGEN(NO)=MOD(NGEN(NO)+1,10000000) 
*
*   Here is practiced some annealing
*
*   DAMP = Damping factor for reducing the move amplitude with 
*          the number of already generated moves (MOVE+PERM in fact) 
*
      IF(DAMP(NO).LT.0.01)GO TO 2800
      DAMP(NO)=(1.-(NGEN(NO)+NGENP(NO))*XNOB/TIMELIM)**ANNEAL
      IF(DAMP(NO).LT.0.5)DAMP0(NO)=2.*DAMP(NO)
2800  IF(MR(NO).EQ.0)GO TO 910
      IF(NRT(NO).EQ.2)GO TO 962
      CALL ROTAT(XLNN,XLN,NAO,ISEED,TM1,TM2,SROT,DAMP0,NGRAV,XG,NO,PP1)
      NGRAV(NO)=1
      GO TO 961
C
C  Rotate that @!$ damned torsion angle
C
962   CALL ROTORS(XLNN,XLN,NAO,ISEED,TM1,TM2,DAMP0,NBRO,NO,PP2,PREV)
      NGRAV(NO)=0
C
C
961   IF(NTR(NO).EQ.0)GO TO 2999
      GO TO 3001
910   CALL ESP_GENMOVE(NAO,ISEED,NCUM,DELTA,IMOVE,MOVE,
     -A,BB,C,DAMP,NSP,NO)
      CALL ESP_NEWPOS(XL,MOVE,IMOVE,NCON,NO)
            GO TO 3001
*
*------------------------------------------------------------------------- 
*           Generate a random permutation or a random translation 
*           Care IF an atom permute with itself 
*           Care IF an atom permute with another same atom-type 
*
2999  IF(MR(NO).EQ.0)GO TO 911
      IF(IABS(NTR(NO)).EQ.1)GO TO 965
      DO 964 I=1,NAO(NO)
      XLNN(NO,1,I)=XLN(NO,1,I)
      XLNN(NO,2,I)=XLN(NO,2,I)
964   XLNN(NO,3,I)=XLN(NO,3,I)
      GO TO 967
965   DO 966 I=1,NAO(NO)
      XLNN(NO,1,I)=XL(NO,1,I)
      XLNN(NO,2,I)=XL(NO,2,I)
966   XLNN(NO,3,I)=XL(NO,3,I)
967   CALL TRANSLA(XLNN,XLN,NAO,ISEED,DELTAO,A,BB,C,DAMP,NO)
            NGENP(NO)=NGENP(NO)+1
      GO TO 3001
911   CALL ESP_GENMOVEP(NAO,XL,MTYP,ISEED,NCUM,DELTA,IMOVE1,IMOVE2,
     -MOVEP,NSP,NO,PROGRESS)
         CALL ESP_NEWPOSP(XL,MOVEP,IMOVE1,IMOVE2,NO)
            NGENP(NO)=NGENP(NO)+1
      GO TO 3002
*
*------------------------------------------------------------------------- 
*           Check distances
*
3001  IF(NCON.EQ.0)GO TO 3002
      IF(MR(NO).EQ.0)GO TO 997
      CALL ESP_CHECKMR(XLN,NTYP,METRIC,RCUT,ACCEPTABLE,NOB,NAO,NO,
     1NXLL,MTYP)
      GO TO 3002
997   CALL ESP_CHECK(NTYP,METRIC,NCUM1,RCUT,
     -       IMOVE,MOVE,ACCEPTABLE,NOB,NAO,NO,NXLL,MTYP,NZERO) 
C         ENDIF
*---------------------------------------------------------------------- 
3002     NGENT(NO)=NGEN(NO)+NGENP(NO)
         IF (ACCEPTABLE.AND.NPER(NO).NE.NPERM(NO))
     1NTRIED(NO)=MOD(NTRIED(NO)+1,10000000)
         IF (NPER(NO).EQ.NPERM(NO)) NTRIEDP(NO)=NTRIEDP(NO)+1
         IF (ACCEPTABLE) THEN
         IF(NPER(NO).EQ.NPERM(NO)) GO TO 3003
         IF(MR(NO).NE.0)GO TO 3003
*
*---------------------------------------------------------------------- 
*           Calculate sums for chi-squared *
      CALL ESP_SUMSTOT(NHKL,FOBS,KXR,XL,IMOVE,MOVE,DIFF,
     -NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
      GO TO 3004
*---------------------------------------------------------------------- 
*           Calculate sums for chi-squared 
*
3003  IF(MR(NO).EQ.0)GO TO 920
      CALL ESP_SUMMR(NHKL,FOBS,KXR,XLN,
     -DIFF,NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
      GO TO 3004
920   CALL ESP_SUMSTOTP(NHKL,FOBS,KXR,XL,IMOVE1,IMOVE2,MOVEP,
     -DIFF,NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
       ENDIF
       ENDIF
*
*---------------------------------------------------------------------- 
*           Calculate chi-squared
*
3004  CALL ESP_CHISQ(SUM_FS,SIGMA,CHISQ,CHISQ1,DIFF,ICHI,DIFF1) 
*
*---------------------------------------------------------------------- 
*
*        Decide whether to accept MOVE
*
         IF (FIRST) THEN
            ACCEPTABLE=.FALSE.
            CHISQ0=CHISQ1
            DIFF0=DIFF1
         ELSE
*
*        Always accept translation if NTR(NO)=-1
*
       IF(NPER(NO).NE.NPERM(NO)) GO TO 2830
       IF(NTR(NO).EQ.-1)GO TO 3010
*
*
2830        PRAT=CHISQ0-CHISQ1
            IF(PRAT.GE.0.)GO TO 3010
*
*  Accept times to times an event that does not improve the fit 
*  on a criteria linked to DAMP
*
            REJ=REJECT*DAMP(NO)
            PDIFF=DIFF0-DIFF1
            IF(PDIFF.GE.0.)GO TO 3010
C  No tolerance on torsion angles for accepting events not improving
C         the fit
            IF(NRT(NO).EQ.2)THEN
            ACCEPTABLE=.FALSE.
            GO TO 3010
            ENDIF
C
            IF (PDIFF.GT.-REJ) PRAT=EXP(PDIFF/REJ)
            IF (PRAT.LT.RAN(ISEED)) ACCEPTABLE=.FALSE.
            IF (PDIFF.LE.-REJ) ACCEPTABLE=.FALSE.
            IF(ACCEPTABLE)NREJECT(NO)=NREJECT(NO)+1
         ENDIF
 3010  IF(ACCEPTABLE) NACC(NO)=MOD(NACC(NO)+1,10000000)
       IF(ACCEPTABLE.AND.NPER(NO).EQ.NPERM(NO))NACCP(NO)=NACCP(NO)+1 
          IF(DIFF0.LT.DIFFM)THEN
          DIFFM=DIFF0
          JTESTM=JTEST
          ENDIF 
*
*---------------------------------------------------------------------- 
*        Write summary
*
      IF(FIRST.OR.(ABS(NGENT(NO)-LASTPRINT(NO)).GE.IPRINT.AND.
     1ACCEPTABLE))
     -    THEN
      CALL ESPO_SUMMARY(NACC,NGEN,NTRIED,NACCP,NTRIEDP,
     -DIFF,CHISQ1,NREJECT,NCON,DAMP,MR,NOB,JTEST,DIFFM,JTESTM,PROGRESS)
            LASTPRINT(NO)=NGENT(NO)-MOD(NGENT(NO),IPRINT)
         ENDIF
*
         IF (ACCEPTABLE) THEN
        IF(NPER(NO).EQ.NPERM(NO)) GO TO 3005
         IF(MR(NO).NE.0)GO TO 3005
*---------------------------------------------------------------------- 
*           Make move
*
      CALL ESP_MAKEMOVE(XL,IMOVE,MOVE,NHKL,NCON,NAO,NO,IANIM)
            CHISQ0=CHISQ1
            DIFF0=DIFF1
       GO TO 3006

*---------------------------------------------------------------------- 
*           Make MOVE in case of accepted permutation 
*
3005  IF(MR(NO).EQ.0)GO TO 930
      CALL ESP_MAKEMR(XL,XLN,NCON,NAO,NO,IANIM)
      NGRAV(NO)=0
C  Put the memory of previous torsion angle to 0
      IF(NBRO.NE.0)PREV(NO,NBRO)=0
      GO TO 931
930   CALL ESP_MAKEMOVEP(XL,IMOVE1,IMOVE2,MOVEP,NHKL,NCON,NAO,NO) 
931         CHISQ0=CHISQ1
            DIFF0=DIFF1
3006      CONTINUE
         ENDIF
*   counter for permutation reset to zero
          IF(NPERM(NO).EQ.NPER(NO))NPER(NO)=0 
*
* Save if improvement
*
      IF(DIFF0.GT.DIFFM)GO TO 2881
      DO 2880 NONO=1,NOB
      DO 2879 I=1,NAO(NONO)
      DO J=1,3
      XLM(NONO,J,I)=XL(NONO,J,I)
      ENDDO
2879  CONTINUE
2880  CONTINUE
2881  CONTINUE
*
*
*
*---------------------------------------------------------------------- 
*        Check time
*
         TIME_USED=TIME_USED+1.
         IF (TIME_USED.LT.0.OR.TIME_USED.GE.SAVE_TIME) THEN
         SAVE_TIME=SAVE_TIME+TIMESAV
         ENDIF
*
*---------------------------------------------------------------------- 
*        Decide whether to continue program 
*
         IF ((TIME_USED.LT.0.).OR.(TIME_USED.GE.TIMELIM))
     -                 LOOPING=.FALSE.
         FIRST=.FALSE.
         IF(NGENT(NO).GE.NSTART) THEN
         IF(DIFF0.GE.RMAX) GO TO 8990
         ENDIF
      ENDDO
*
*        End of loop
*---------------------------------------------------------------------- 
*---------------------------------------------------------------------- 
*---------------------------------------------------------------------- 
C
C  Output of the final coordinates
C
      GO TO 8991
8990	WRITE (BUFFER,*)' Bad luck ! Try again.'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 8888
8991  CALL ESPO_SUMMARY(NACC,NGEN,NTRIED,NACCP,NTRIEDP,DIFF0,
     -CHISQ1,NREJECT,NCON,DAMP,MR,NOB,JTEST,DIFFM,JTESTM,PROGRESS) 
8888  WRITE(20,5001)
5001  FORMAT(/' Final coordinates x,y,z and occupation numbers'/)
      IF(NTRTEST.NE.1)GO TO 2900
      DO 2882 NO=1,NOB
      DO 2883 I=1,NAO(NO)
      DO J=1,3
      XL(NO,J,I)=XLM(NO,J,I)
      ENDDO
2883  CONTINUE
2882  CONTINUE
2900  NNPR=0
      IF(DIFF0.LE.DIFFM1)NNPR=1
      DO 2520 NO=1,NOB
      IF(NOBT(NO).NE.2)GO TO 2702
      DO J=1,3
      SHIFT(NO,J)=INT(XL(NO,J,1))
      DIF=XL(NO,J,1)-SHIFT(NO,J)
      IF(DIF.GT.1.)SHIFT(NO,J)=SHIFT(NO,J)+1.
      IF(DIF.LT.0.)SHIFT(NO,J)=SHIFT(NO,J)-1.
      ENDDO
2702  DO 1500 I=1,NAO(NO)
      DO J=1,3
      IF(NOBT(NO).EQ.2) GO TO 2700   
       XL(NO,J,I)=XL(NO,J,I)-INT(XL(NO,J,I))
       IF(XL(NO,J,I).GT.1.)XL(NO,J,I)=XL(NO,J,I)-1.
       IF(XL(NO,J,I).LT.0.)XL(NO,J,I)=XL(NO,J,I)+1.
       GO TO 2701
2700  XL(NO,J,I)=XL(NO,J,I)-SHIFT(NO,J)
2701  IF(NNPR.EQ.1)XLM(NO,J,I)=XL(NO,J,I)
      IF(NTRTEST.EQ.1)XLM(NO,J,I)=XL(NO,J,I)
      ENDDO
      WRITE(20,1010)NX(NO,I),(XL(NO,J,I),J=1,3),XL(NO,5,I) 
1500  CONTINUE
2520  CONTINUE
           IF(DIFF0.LT.DIFFM1)DIFFM1=DIFF0 
C  Close the .xyz file for animation
      IF(IANIM.EQ.1)CLOSE(IAN)
C
      IF(NNPR.EQ.0)GO TO 2655
*
*---------------------------------------------------------------------- 
*           Write out the .spf FILE for PLATON 
*
      IF(IPRI2.GE.1)THEN
      CALL OPEN_WRITE(24,FILE,'.spf')
      WRITE(24,501)(TEXT(I),I=1,20),(CELPRE(I),I=1,6),(IGR(IR),IR=1,36)
      DO 2521 NO=1,NOB
      DO 1502 I=1,NAO(NO)
      WRITE(24,502)NX(NO,I),(XLM(NO,J,I),J=1,3) 
1502  CONTINUE
2521  CONTINUE
      CLOSE(24)
      ENDIF
*
*---------------------------------------------------------------------- 
*           Write out the .ins FILE for SHELX 
*
      IF(IPRI2.GE.1)THEN
      CALL OPEN_WRITE(23,FILE,'.ins')
      WRITE(23,1561)(TEXT(I),I=1,20),SLABDA,(CELPRE(I),I=1,6)
      IBRA=IABRA
      IF(IABRA.EQ.3)IBRA=4
      IF(IABRA.EQ.4)IBRA=3
      IF(ICENT.EQ.1)IBRA=-IBRA
      WRITE(23,1571)IBRA
C
C  WRITE SYMMETRY OPERATORS
C
      IRLLL=IRL
      IF(IABRA.EQ.4)IRLLL=IRLLL/3
      IF(IABRA.EQ.5)IRLLL=IRLLL/2
      IF(IABRA.EQ.6)IRLLL=IRLLL/2
      IF(IABRA.EQ.7)IRLLL=IRLLL/2
      DO 1572 IR=2,IRLLL
        DO 1573 J=1,3
        TS(J)=T(IR,J)
        DO 1573 L=1,3
        IM(J,L)=INT(SM(IR,J,L))
 1573 CONTINUE
      CALL STRINGSYM(IM,TS,STRSYM)
      WRITE(23,'(A,A)')'SYMM ',STRSYM
 1572 CONTINUE
      WRITE(23,1566)(NAX(L),L=1,NTYP)
1568  CONTINUE
      DO L=1,NTYP
      NATOT(L)=0
      DO NO=1,NOB
      NATOT(L)=NATOT(L)+NAT(NO,L)
      ENDDO
      ENDDO
      WRITE(23,1569)(NATOT(L),L=1,NTYP)
      WRITE(23,1570)
      DO 2522 NO=1,NOB
      DO 1563 I=1,NAO(NO)
      OCCC=XL(NO,5,I)+10.
      WRITE(23,1562)NX(NO,I),MTYP(NO,I),(XLM(NO,J,I),J=1,3),OCCC 
1563  CONTINUE
2522  CONTINUE
      WRITE(23,1565)
1561  FORMAT('TITL ',20A4/'CELL 'F9.5,6F9.4/
     -'ZERR 1 0.001 0.001 0.001 0.0 0.0 0.0') 
1562  FORMAT(A4,1X,I2,4F11.5,'  10.05000') 
1565  FORMAT('HKLF 3'/'END')
1566  FORMAT('SFAC',8(X,A2))
1569  FORMAT('UNIT',8(X,I2))
1570  FORMAT('ACTA'/'L.S. 4'/'BOND'/'FMAP 2'/'PLAN 20'/
     -'WGHT    0.100000'/'FVAR       1.00') 
1571  FORMAT('LATT ',I2)
      CLOSE(23)
      ENDIF
2655  CALL DATN
      IF(NTEST.EQ.0)NTEST=JTEST
C      IF(IPRI.EQ.0.OR.JTEST.NE.NTEST) GO TO 250 
      IF(IPRI.EQ.0) GO TO 250 
*
      SUMCAL=0.
      DO 220 NM=1,NHKL
      FNUCT(NM)=0.
C
      DO 231 NO=1,NOB
      DO 231 I=1,NAO(NO)
      DO 232 IR=1,IRL
      ARG=TR(IR,NM)
      DO 233 J=1,3
233   ARG=ARG+H(IR,J,NM)*XLM(NO,J,I)
      ARG=6.28319*ARG
      COSAR(NO,I,IR,NM)=COS(ARG)
232   SINAR(NO,I,IR,NM)=SIN(ARG)
231   CONTINUE
      AV=0.
      BV=0.
      DO 224 NO=1,NOB
      DO 224 I=1,NAO(NO)
      SAI=0.
      SBI=0.
      DO 225 IR=1,IRL
      SAI=SAI+COSAR(NO,I,IR,NM)
      IF(ICENT.EQ.1)SBI=SINAR(NO,I,IR,NM)+SBI 
225   CONTINUE
      SAI=FLOAT(ICENT)*SAI
      IF(KXR.LT.4)GO TO 226
      GOTO 227
226   AV=SNEXIN(NO,I)*SAI+AV
      BV=SNEXIN(NO,I)*SBI+BV
      IF(KXR.LT.4)GO TO 224
227   AV=SNEXIX(NO,I,NM)*SAI+AV
      BV=SNEXIX(NO,I,NM)*SBI+BV
      AVI=AV-SNXI(NO,I)*SBI
      AVJ=AV+SNXI(NO,I)*SBI
      BVI=BV-SNXI(NO,I)*SAI
      BVJ=BV+SNXI(NO,I)*SAI
224   CONTINUE
      IF(KXR.LT.4)GOTO 228
      FNN=(AVI*AVI+AVJ*AVJ+BVI*BVI+BVJ*BVJ)
      GOTO 229
228   FNN=(AV*AV+BV*BV)
229   FNUC(NM)=SQRT(FNN)
      SUMCAL=SUMCAL+FNUC(NM)
      FNUCT(NM)=FNUCT(NM)+FNUC(NM)
220   CONTINUE
      IF(IPRI.EQ.0.OR.JTEST.NE.NTEST) GO TO 2650 
      WRITE(20,*)
      WRITE(20,*)'  h   k   l   Fcalc    Fobs     d(A)   2-theta   FWHM'
      WRITE(20,*)
2650  DIFF=0.
      DO 249 NM=1,NHKL
      FNUCT(NM)=FNUCT(NM)/SUMCAL*SUM_F
      DIFF=DIFF+ABS(FOBS(NM)-FNUCT(NM))
      IF(IPRI.EQ.0.OR.JTEST.NE.NTEST) GO TO 249 
      WRITE(LPR,1002)(IHH(K,NM),K=1,3),FNUCT(NM),FOBS(NM),D(NM),
     -THETA(NM),HW(NM)/SLOG
249   CONTINUE
      DIFF=DIFF/SUM_F
      IF(IPRI.EQ.0.OR.JTEST.NE.NTEST) GO TO 252 
      WRITE(20,*)' Final RF factor    = ',DIFF
CKP      IF(KP.EQ.0.OR.KP.EQ.2) GO TO 254
252   CALL PROFILE2(NHKL,FNUCT,SUM_FS,DIFF)
      DIFF=DIFF/SUM_Y
      IF(IPRI.EQ.0.OR.JTEST.NE.NTEST) GO TO 254 
      WRITE(20,*)' Final Rp(F) factor = ',DIFF 
*
*  .PRF FILE output
*
254   IF(NNPR.EQ.0)GO TO 250
      IF(IPRI2.GE.2)THEN
      TEMPO=FILE(1:LFILE)//'.prf'
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.not.QEX) go to 204
      CALL FILEDEL(12,TEMPO)
204   CALL OPEN_WRITE1(12,TEMPO)
      IVERS=8
      ZERO=0.
      WRITE(12,'(4F9.5,2I6,2X,20A4)')U,V,W,STEPN,N1,N2,TEXT
      WRITE(12,'(A,F9.4)') '3111   1.0000',ZERO
      THMAX=POS(N2)
      THMIN=POS(N1)
      ICN=NHKL
      DO 205 I=1,NHKL
      I1=IHH(1,I)
      I2=IHH(2,I)
      I3=IHH(3,I)
205   IREFS(I)=256*(256*(256*(8+1)+128+I1)+128+I2)+128+I3
      ICZ=NHKL
      NPTS=N2-N1+1
      WRITE(12,'(2F8.3,F8.5,I8)') THMAX,THMIN,STEP,IVERS
      AMDA1=SLABDA
      AMDA2=SLABDA
      NPAT1=1
      NVK=0
      WRITE(12,'(I3,I7,5f10.5)')NPAT1,NPTS,AMDA1,AMDA2,ZERO
      WRITE(12,'(8I5,8i3)')ICN,NVK
      WRITE(12,'(8(F7.0,1X))') (YOBS(J),J=N1,N2)
      WRITE(12,'(8(F7.0,1X))') (YCALC(J),J=N1,N2)
      WRITE(12,'(8I10)') (IREFS(J),J=1,ICZ)
      WRITE(12,'(10F8.3)')(THETA(J),J=1,ICZ)
      NEXCRG=0
      EXCRG=0
      WRITE(12,'(I8)') NEXCRG
      WRITE(12,'(2F8.2)') EXCRG,EXCRG
      CLOSE(12)
C
C  HPGL Output of the regenerated pattern
C
      CALL PRF2HGL(FILE,NHKL)
      ENDIF
C
C  Creation of the STRUVIR/STRUPLO namestru.dat file
C
      IF(IPRI2.GE.3)THEN
      TEMPO=FILE(1:LFILE)//'stru.dat'
      INQUIRE(FILE=TEMPO,EXIST=QEX)
      IF(.NOT.QEX) GO TO 2651
      CALL FILEDEL(13,TEMPO)
2651  CALL OPEN_WRITE1(13,TEMPO)
C
C   STRUVIR STRUVIR STRUVIR
C
C
      WRITE(13,3800)(TEXT(I),I=1,20),(CELPRE(I),I=1,6),(IGR(IR),IR=1,36)
3800  FORMAT('TITL ',20A4/'CELL '6F9.4/'SPGP '36A1) 
      CELLMIN=MIN(CELPRE(1),CELPRE(2),CELPRE(3))
      IF(CELLMIN.EQ.CELPRE(1))THEN
      WRITE(13,3801)
      GO TO 3799
      ENDIF
      IF(CELLMIN.EQ.CELPRE(2))THEN
      WRITE(13,3802)
      GO TO 3799
      ENDIF
      IF(CELLMIN.EQ.CELPRE(3))WRITE(13,3803)
3801  FORMAT('VIEW 1 0 0    0 0 0')
3802  FORMAT('VIEW 0 1 0    0 0 0')
3803  FORMAT('VIEW 0 0 1    0 0 0')
3799  IF(ICENT.EQ.2)WRITE(13,3804)
      IF(ICENT.EQ.1)WRITE(13,3805)
3804  FORMAT('PARA 1.00 1 1 1 1 1 1 2.50 0.00 0.01')
3805  FORMAT('PARA 1.00 0 1 1 1 1 1 2.50 0.00 0.01')
      DX=-2.5/A
      DY=-2.5/BB
      DZ=-2.5/C
      EX=1.-DX
      EY=1.-DY
      EZ=1.-DZ
      WRITE(13,3806)DX,EX,DY,EY,DZ,EZ
3806  FORMAT('XYZR ',3(2F5.2,1X))
      WRITE(13,3807)
3807  FORMAT('TETR 109.0 20.0  1.62 0.20')
      WRITE(13,3808)
3808  FORMAT('OCTA  90.0 20.0  1.97 0.43')
      WRITE(13,3809)
3809  FORMAT('PERS     70.00')
      WRITE(13,3810)
3810  FORMAT('FORM 4')
      WRITE(13,3815)(NAXB(L),L=1,NTYP)
3815  FORMAT('CIRC 0.2 ',8(A2,1X))
      IPOLYBZ=0
      DO 3838 L=1,NTYP
      IF(NAXB(L)(1:2).EQ.'O ')IPOLYBZ=1
      IF(NAXB(L)(1:2).EQ.'F ')IPOLYBZ=1
3838  CONTINUE
      IF(IPOLYBZ.EQ.0)GO TO 3837
      IBZRR=0
      DO 3835 L=1,NTYP
      IF(NAXB(L)(1:2).EQ.'O ')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'F ')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'N ')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'H ')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'P ')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'SI')GO TO 3835
      IF(NAXB(L)(1:2).EQ.'AL')GO TO 3835
      IBZRR=IBZRR+1
      IF(IBZRR.GT.5)GO TO 3837
      DBZRR=2.8
      IF(NAXB(L)(1:2).EQ.'K ')DBZRR=3.2
      IF(NAXB(L)(1:2).EQ.'BA')DBZRR=3.2
      IF(NAXB(L)(1:2).EQ.'CS')DBZRR=3.2
      IF(NAXB(L)(1:2).EQ.'RB')DBZRR=3.2
      IF(NAXB(L)(1:2).EQ.'PB')DBZRR=3.2
      WRITE(13,3836)DBZRR,NAXB(L)
3835  CONTINUE
3836  FORMAT('BZRR ',F5.2,1X,A2)
3837  WRITE(13,3811)
3811  FORMAT('END')
C
      K=0
      DO 3812 NO=1,NOB
      DO 3813 I=1,NAO(NO)
      DO 3820 J=1,3
      XLS(NO,J,I)=XLM(NO,J,I)
C-- Tests on positions supposed to be close to special positions
C
C   Very bad test... up too now
C
      IF(NOBT(NO).NE.1)GO TO 3820
      IF(ISYST.EQ.7)GO TO 3820   ! TRICLINIC
      IF(XLS(NO,J,I).LT.STST(J))XLS(NO,J,I)=0.0
      IF(ABS(XLS(NO,J,I)-0.25).LT.STST(J))XLS(NO,J,I)=0.25
      IF(ABS(XLS(NO,J,I)-0.5).LT.STST(J))XLS(NO,J,I)=0.5
      IF(ABS(XLS(NO,J,I)-0.75).LT.STST(J))XLS(NO,J,I)=0.75
      IF(ABS(XLS(NO,J,I)-1.00).LT.STST(J))XLS(NO,J,I)=0.0
      IF(ISYST.EQ.1)GO TO 3820   !  MONOCLINIC
      IF(ABS(XLS(NO,J,I)-0.125).LT.STST(J))XLS(NO,J,I)=0.125
      IF(ABS(XLS(NO,J,I)-0.375).LT.STST(J))XLS(NO,J,I)=0.375
      IF(ABS(XLS(NO,J,I)-0.625).LT.STST(J))XLS(NO,J,I)=0.625
      IF(ABS(XLS(NO,J,I)-0.875).LT.STST(J))XLS(NO,J,I)=0.875
      IF(ISYST.EQ.2)GO TO 3820   !  ORTHORHOMBIC
      IF(ABS(XLS(NO,J,I)-0.333333).LT.STST(J))XLS(NO,J,I)=0.333333
      IF(ABS(XLS(NO,J,I)-0.666667).LT.STST(J))XLS(NO,J,I)=0.666667
      IF(ABS(XLS(NO,J,I)-0.166667).LT.STST(J))XLS(NO,J,I)=0.166667
      IF(ABS(XLS(NO,J,I)-0.833333).LT.STST(J))XLS(NO,J,I)=0.833333
      IF(ABS(XLS(NO,J,I)-0.583333).LT.STST(J))XLS(NO,J,I)=0.583333
      IF(ABS(XLS(NO,J,I)-0.083333).LT.STST(J))XLS(NO,J,I)=0.083333
      IF(ABS(XLS(NO,J,I)-0.916667).LT.STST(J))XLS(NO,J,I)=0.916667
      IF(ABS(XLS(NO,J,I)-0.416667).LT.STST(J))XLS(NO,J,I)=0.416667
3820  CONTINUE
      K=K+1
      WRITE(13,3814)NXB(NO,I),K,(XLS(NO,J,I),J=1,3)
3813  CONTINUE
3812  CONTINUE
3814  FORMAT(A2,1X,I2,3X,3F9.5) 
      CLOSE(13)
      ENDIF
*
*---------------------------------------------------------------------- 
*           Write out the .xyz FILE for RASMOL or CHIME or etc.
*              the file will contain all atoms in the cell,
*              generated by symmetry operators, unless there is
*              a rigid body fragment
*
*
      IF(IPRI2.GE.3)THEN
      I=0
	NASSS=NAS
	DO 5499 NO=1,NOB
5499	IF(NOBT(NO).EQ.2)NASSS=1
      DO 5501 NO=1,NOB
      DO 5501 L=1,NAO(NO)
      X=XLS(NO,1,L)
      Y=XLS(NO,2,L)
      Z=XLS(NO,3,L)
      IF(NASSS.EQ.1) GO TO 5498   
       X=X-INT(X)
       IF(X.GT.1.)X=X-1.
       IF(X.LT.0.)X=X+1.
       Y=Y-INT(Y)
       IF(Y.GT.1.)Y=Y-1.
       IF(Y.LT.0.)Y=Y+1.
       Z=Z-INT(Z)
       IF(Z.GT.1.)Z=Z-1.
       IF(Z.LT.0.)Z=Z+1.
5498      DO 5501 K=1,NASSS
      I=I+1
      XA=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      YA=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      ZA=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      IF(NASSS.EQ.1) GO TO 5497   
       XA=XA-INT(XA)
       IF(XA.GT.1.)XA=XA-1.
       IF(XA.LT.0.)XA=XA+1.
       YA=YA-INT(YA)
       IF(YA.GT.1.)YA=YA-1.
       IF(YA.LT.0.)YA=YA+1.
       ZA=ZA-INT(ZA)
       IF(ZA.GT.1.)ZA=ZA-1.
       IF(ZA.LT.0.)ZA=ZA+1.
5497  XX(K)=XA
      YY(K)=YA
      ZZ(K)=ZA
      XB(I)=XA
      YB(I)=YA
      ZB(I)=ZA
C
C  Test for special positions
C
      IF(I.EQ.1.OR.K.EQ.1)GO TO 5510
      DO 5520 JJ=1,K-1
      IF (ABS(XX(JJ) - XX(K)-1.) .GE. 0.01)GO TO 5551
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5551
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5551
      I=I-1
      GOTO 5501
5551  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5552
      IF (ABS(YY(JJ) - YY(K)-1.) .GE. 0.01)GO TO 5552
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5552
      I=I-1
      GOTO 5501
5552  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5553
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5553
      IF (ABS(ZZ(JJ) - ZZ(K)-1.) .GE. 0.01)GO TO 5553
      I=I-1
      GOTO 5501
5553  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5520
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5520
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5520
      I=I-1
      GOTO 5501
5520  CONTINUE
C
C  Change coordinates in Cartesian coordinates 
C
5510  XXX(I)=TM1(1)*XB(I)+TM1(2)*YB(I)+TM1(3)*ZB(I)
      YYY(I)=TM1(4)*YB(I)+TM1(5)*ZB(I)
      ZZZ(I)=TM1(6)*ZB(I)
      NXBB(I)=NXB(NO,L)
5501  CONTINUE
C
C
      CALL OPEN_WRITE(24,FILE,'.xyz')
      WRITE(24,5503)I
      DO 5590 J=1,I
      WRITE(24,5504)NXBB(J),XXX(J),YYY(J),ZZZ(J)
5590  CONTINUE
5504  FORMAT(A2,2X,3F11.5)
5503  FORMAT(I4/)
      CLOSE(24)
      ENDIF
C
C  Call to the internal STRUVIR
C
      IF(IPRI2.GE.5)THEN
      TEMPO=FILE(1:LFILE)//'stru'
      NNPR2=0
      IF(JTEST.EQ.NTEST)NNPR2=1
C
C - Do not enter in STRUVIR if R > 25 % ???
C      IF(DIFFM1.GT.0.25)GO TO 250
C
      CALL STR2WRL(TEMPO,FILE,NNPR2,V1)
      ENDIF
*
*
C
C  End of the zone of conditional output
C
250   WRITE(20,5002)
5002  FORMAT(/' End of this test'////)
      IF(JTEST.EQ.NTEST)JTEST=0
      IF(JTEST.NE.0)GO TO 8999
*
*
501   FORMAT('TITL ',20A4/'CELL '6F9.4/'SPGR '36A1) 
502   FORMAT('ATOM ',A4,3F11.5)
1002  FORMAT(3I4,5F8.3)
   49 FORMAT(1H 30X,'AND THE EQUIV. POS. ARE:')
   51 FORMAT(35H NEUTRON NUCLEAR SCATTERING LENGTHS/)
   52 FORMAT(1H ,8(F7.3,5H = B(,I1,2H);))
   55 FORMAT(' REAL AND IM. X-RAY SCATTER. FACT.') 
100   FORMAT(/1x,'============================================
     -=================='/
     -        1x,' ESPOIR VERSION ',a,' by A. Le Bail - 2000 -
     -alb@cristal.org'/
     -        1x,'============================================
     -=================='//
     -        1x,'Using generic filename : ',A/)
  152 FORMAT(' POS.',I3,2X,9F5.1,3F8.4)
  301 FORMAT(//' THE STRUCT. IS NON CENTRIC,')
  303 FORMAT(//' THE STRUCT. IS CENTRIC,') 
1100  FORMAT(/1x,'Configuration contains ',i5,' atoms of ',i1,' types'/) 
1200  FORMAT(/' Using ',i4,' r points spaced at ',f5.3,
     - 'A up to ',f6.3,'A'/)
1300  FORMAT(' WARNING: ',i3,' pairs cause g',2i1,' to be ',
     - 'non-zero at ',f4.2,'A')
 1611 FORMAT(24X,15HROTATION MATRIX,15X,18HTRANSLATION VECTOR)
      GO TO 3600
3500  WRITE(20,*)'  Error reading the first line : TEXT'
	WRITE (BUFFER,*)'  Error reading the first line : TEXT'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3501  WRITE(20,*)'  Error reading the cell parameters, line 2'
	WRITE (BUFFER,*)'  Error reading the cell parameters, line 2'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3502  WRITE(20,*)'  Error reading the space group, line 3'
	WRITE (BUFFER,*)'  Error reading the space group, line 3'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3503  WRITE(20,*)'  Error reading lambda, etc, line 4'
	WRITE (BUFFER,*)'  Error reading lambda, etc, line 4'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3504  WRITE(20,*)'  Error reading U, V, W, step'
	WRITE (BUFFER,*)'  Error reading U, V, W, Step'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3505  WRITE(20,*)'  Error reading atom names (neutron)'
	WRITE (BUFFER,*)'  Error reading atom names (neutron)'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3506  WRITE(20,*)'  Error reading atom names (X-ray)'
	WRITE (BUFFER,*)'  Error reading atom names (X-ray)'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3507  WRITE(20,*)'  Error reading scattering factors'
	WRITE (BUFFER,*)'  Error reading scattering factors'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3508  WRITE(20,*)'  Error reading code for distance constraints'
	WRITE (BUFFER,*)'  Error reading code for distance constraints'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3509  WRITE(20,*)'  Error reading minimal distances'
	WRITE (BUFFER,*)'  Error reading minimal distances'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3510  WRITE(20,*)'  Error reading largest distance moves'
	WRITE (BUFFER,*)'  Error reading largest distance moves'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3511  WRITE(20,*)'  Error reading anneal, sigma, reject'
	WRITE (BUFFER,*)'  Error reading anneal, sigma, reject'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3512  WRITE(20,*)'  Error reading iprint, timelim, timesav'
	WRITE (BUFFER,*)'  Error reading iprint, timelim, timesav'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3513  WRITE(20,*)'  Error reading nstart, rmax, ichi, test'
	WRITE (BUFFER,*)'  Error reading nstart, rmax, ichi, test'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3514  WRITE(20,*)'  Error reading object type and perm, for object:',NO
      WRITE(BUFFER,*)'Error reading object type and perm, object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3515  WRITE(20,*)'  Error reading number of each atoms for object:',NO
	WRITE(BUFFER,*)'Error reading number of each atoms for object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3516  WRITE(20,*)'  Error reading Bov, Nocc, Nspe for object:',NO
	WRITE (BUFFER,*)'  Error reading Bov, Nocc, Nspe for object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3517  WRITE(20,*)'  Error reading occupancy factors for object:',NO
	WRITE(BUFFER,*)'Error reading occupancy factors for object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3518  WRITE(20,*)'  Error reading special positions for object:',NO
	WRITE(BUFFER,*)'Error reading special positions for object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3519  WRITE(20,*)
	WRITE(BUFFER,*)'Error reading cell parameters for object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3520  WRITE(20,*)
	WRITE(BUFFER,*)'Error reading atomic coordinates, object:',NO
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3521  WRITE(20,*)
	WRITE (BUFFER,*)'  More than 19200 atoms in the cell'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3522  WRITE(20,*)
	WRITE (BUFFER,*)'  Error reading the hkl number'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3523  WRITE(20,*)'  Error reading h, k, l, "|Fobs|"'
	WRITE (BUFFER,*)'  Error reading h, k, l, "|Fobs|"'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3524  WRITE(20,*)'More than 500 bonds in the TORSION subroutine'
      WRITE(BUFFER,*)'More than 500 bonds in the TORSION subroutine'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3525  WRITE(20,*)'More than 50 neighbours in the TORSION subroutine'
      WRITE(BUFFER,*)'More than 50 neighbours in the TORSION subroutine'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3526  WRITE(20,*)'More than 50 torsion angles'
      WRITE(BUFFER,*)'More than 50 torsion angles'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      GO TO 3600
3527  WRITE(20,*)'More than 60 atoms in an object'
      WRITE(BUFFER,*)'More than 60 atoms in an object'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
3600  CONTINUE
C      WRITE(*,3955)
C3955  FORMAT(//'  Type any character and a RETURN to continue : ',$)
C      READ(*,*)FEND
      CLOSE(20)
      CLOSE(28)
      RETURN
      END
*-----------------------------------------------------------------------
C	SUBROUTINE PROGRESSVIEW(PROGRESS,BUFFER)
C	INTEGER(4) PROGRESS
C	CHARACTER*79 BUFFER
C      WRITE(*,1)BUFFER
C1     FORMAT(A79)
C      RETURN
C      END
*-----------------------------------------------------------------------
      SUBROUTINE DATN
C
C Obviously, this subroutine is compiler specific...
C
C In case of problem, remove that subroutine and all 
C the CALL DATN inside the program.
C
C Date will be printed as follows :
C
C     10-Apr-2000   Hour:  0 Min: 29 Sec: 20 
C
C     dd-mmm-yyyy          hh     mm      ss 
C
C     DT_VALUES(3),MONTHS(DT_VALUES(2)),DT_VALUES(1), 
C     DT_VALUES(5),DT_VALUES(6),DT_VALUES(7) 
C
C
      CHARACTER*8 DA
      CHARACTER*10 TIME
      CHARACTER*5 ZONE
      INTEGER*4 DT_VALUES(8)    ! Return values from DATE_AND_TIME
      CHARACTER*3 MONTHS(12)    ! Month names
      DATA MONTHS /'Jan','Feb','Mar','Apr','May','Jun',
     1             'Jul','Aug','Sep','Oct','Nov','Dec'/

C
C Get current date values
C
       CALL DATE_AND_TIME (DA,TIME,ZONE,DT_VALUES) 
C
C Format date
C
      WRITE(20,10) DT_VALUES(3),
     1 MONTHS(DT_VALUES(2)),DT_VALUES(1),DT_VALUES(5),
     2 DT_VALUES(6),DT_VALUES(7)
10    FORMAT(1X,I2,'-',A3,'-',I4,5X,I2,' hour ',I2,' min ',I2,
     -' Sec ')
C
C  End of compiler specific subroutine
C
      RETURN
      END
      SUBROUTINE OPEN_READ1(UNIT,FILE)
      integer UNIT
      CHARACTER*(*) FILE
      OPEN (UNIT,FILE=FILE,STATUS='old')
      RETURN
      END
*
      SUBROUTINE OPEN_WRITE1(UNIT,FILE)
      integer UNIT
      CHARACTER*(*) FILE
      OPEN (UNIT,FILE=FILE,STATUS='NEW')
      RETURN
      END
*
      SUBROUTINE FILEDEL(UNIT,FILE)
      integer UNIT
      CHARACTER*(*) FILE
      OPEN (UNIT,FILE=FILE,STATUS='old')
      close(UNIT,STATUS='delete')
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
*     SUBROUTINE OPEN_WRITE OPENs a FILE for writing, filling in the 
*     supplied extension IF none is supplied with the FILE name.
*     Allows use of system specIFic facilities of OPEN statement.
*
      SUBROUTINE OPEN_WRITE(UNIT,FILE,extension)
      integer UNIT
      CHARACTER*(*) FILE,extension,temp*80
      temp=FILE
      i=index(FILE,'.')
      l=LEN(FILE)
      DO WHILE(FILE(l:l).EQ.' ')
         l=l-1
      ENDDO
      IF (i.EQ.0) temp=FILE(1:l)//extension
      OPEN (UNIT,FILE=temp,STATUS='old',err=10)
      close(UNIT,STATUS='delete')
10    OPEN (UNIT,FILE=temp,STATUS='NEW')
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
*     SUBROUTINE ESP_INIT must set the seed for the random number 
*     generator and obtain the current cpu clock reading in minutes 
*
      SUBROUTINE ESP_INIT(ISEED,TIME_START) 
C
C PORLIB is compiler spcecific part, introduced for 
C using the intrinsic function SECNDS(X) which returns the 
C  (time in seconds since midnight - X) 
C
C      USE PORTLIB
      USE PORTLIB
      X=0.
      ISEED=INT(GSECNDS(X))*2+1
      TIME_START=0.
*
*  run the random number generator N times
*  for avoiding effects of starting value
*
      N=ISEED/2000
      IF(N.LE.0)N=100
      DO 10 I=1,N
      BIDON=RAN(ISEED)
10    ISEED=ISEED/3
      ISEED=ISEED*2+1
      WRITE(20,*)' ISEED = ',ISEED
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE LIST(TEXT,SIGMA,NTYP,NSTART,RMAX,NTEST,
     -DELTA,ANNEAL,IPRINT,TIMELIM,TIMESAV,REJECT,PROGRESS) 
*
* Print some values and list to terminal 
*
      INCLUDE 'esp0.inc'
	INTEGER(4) PROGRESS
	CHARACTER*120 BUFFER
      DIMENSION DELTA(N_T)
      DIMENSION TEXT(20)
	WRITE (BUFFER,*)'Title of run : '
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(20A4)')TEXT
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,100) TEXT
      WRITE (20,140) SIGMA
      WRITE (20,141) REJECT
      DO I=1,NTYP
         WRITE (20,220) DELTA(I),I
      ENDDO
      WRITE (20,221) ANNEAL
      WRITE (20,260) IPRINT,TIMELIM,TIMESAV
      WRITE (20,261) NSTART,RMAX
      WRITE (20,263) NTEST
      RETURN
100   FORMAT(/1x,'Title of run : ',20A4) 
140   FORMAT(/1x,'Overall standard deviation is ',f7.4/) 
141   FORMAT(1x,'Accept a move even if -',F10.6,' > DELTA R > 0.'/) 
180   FORMAT(10x,i1,'-',i1,5x,f6.3,' A') 
220   FORMAT(' Maximum change in any coordinate is ',f8.3,
     - 'A for particles of type ',i2)
221   FORMAT(/' Move amplitudes will be progressively reduced by
     - MOVE=MOVE*(1.-NGEN/maxNGEN)**anneal'/,20X,'with anneal = ',F5.2) 
260   FORMAT(/' Writing summary every ',i10,' generated moves'/
     - ' Programme will run for ',f10.1,' moves saving every ',
     - f10.1,' moves'/)
261   FORMAT(' Restart after',i7,' events',
     - ' IF the R factor is still higher than ',f6.3/) 
263   FORMAT(' Number of different starting configurations to test :',
     -i5)
      END
*
*********************************************************************** 
*
      SUBROUTINE PROFILE1(NHKL,F)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp2.inc'
      DIMENSION F(N_HKL)
*  Sum at each point
100   continue
      DO 191 K=1,NPTS
      YOBS(K)=0.
*  SUM on all hkl
      KPOS=0
      DO 190 J=1,NHKL
      DELTA=POS(K)-THETA(J)
      IF(ABS(DELTA).GT.HW4(J))GO TO 190
      IF(KPOS.EQ.0)NHA(K)=J
      KPOS=1
      NHB(K)=J
      DELT=DELTA*DELTA
      OMEG=EXP(-DELT/BBB(J))/HW(J)
      YOBS(K)=YOBS(K)+OMEG*F(J)
190   CONTINUE
C      WRITE(20,*)POS(K),YOBS(K)
      IF(YOBS(K).LT.0.01)YOBS(K)=0.
191   CONTINUE
      DO 192 K=1,NPTS
192   IF(YOBS(K).NE.0.)GO TO 193
193   N1=K
      L=NPTS+1
      DO 194 K=1,NPTS
      L=L-1
194   IF(YOBS(L).NE.0.)GO TO 195
195   N2=L
      RETURN
      END
*
*********************************************************************** 
*
      SUBROUTINE PROFILE3(NHKL,F,SUM_FS,DIFF)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp2.inc'
      DIMENSION F(N_HKL),KK(N_HKL)
*  Sum at each point
      SUM_FS=0.
      DIFF=0.
      DO 189 J=1,NHKL
189   KK(J)=0
      DO 191 K=N1,N2
      IF(YOBS(K).EQ.0.)GO TO 191
      YCALC(K)=0.
*  SUM on all hkl
      DO 190 J=NHA(K),NHB(K)
      DELTA=POS(K)-THETA(J)
C      IF(ABS(DELTA).GT.HW4(J))GO TO 190
      KK(J)=KK(J)+1
      IF(KK(J).LE.100)GO TO 1000
      WRITE(20,*)'  More than 100 pts on the profile'
      WRITE(20,*)'    either change OMEGA array dimension'
      WRITE(20,*)'           or decrease NSTEP'
      RETURN
1000  DELT=DELTA*DELTA
      OMEGA(KK(J),J)=EXP(-DELT/BBB(J))/HW(J)
      YCALC(K)=YCALC(K)+OMEGA(KK(J),J)*F(J) 
190   CONTINUE
191   CONTINUE
      DO 192 K=N1,N2
      IF(YOBS(K).EQ.0.)GO TO 192
      SUM_FS=SUM_FS+(YOBS(K)-YCALC(K))**2
      DIFF=DIFF+ABS(YOBS(K)-YCALC(K))
192   CONTINUE
      RETURN
      END
*
*********************************************************************** 
*
      SUBROUTINE PROFILE2(NHKL,F,SUM_FS,DIFF)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp2.inc'
      DIMENSION F(N_HKL),KK(N_HKL)
*  Sum at each point
      SUM_FS=0.
      DIFF=0.
      DO J=1,NHKL
      KK(J)=0
      END DO
      DO K=N1,N2
      IF(YOBS(K).EQ.0.)GO TO 191
      YCALC(K)=0.
*  SUM on all hkl
      DO J=NHA(K),NHB(K)
      KK(J)=KK(J)+1
      YCALC(K)=YCALC(K)+OMEGA(KK(J),J)*F(J) 
      END DO
191   CONTINUE
      END DO
      DO K=N1,N2
      IF(YOBS(K).EQ.0.)GO TO 192
      SUM_FS=SUM_FS+(YOBS(K)-YCALC(K))**2
      DIFF=DIFF+ABS(YOBS(K)-YCALC(K))
192   CONTINUE
      END DO
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_GENMOVE(NAO,ISEED,NCUM,DELTA,IMOVE,MOVE,
     -A,BB,C,DAMP,NSP,NO)
      INCLUDE 'esp0.inc'
      REAL MOVE(192,6),DELTA(*)
      DIMENSION NSP(N_O,N_A),NAO(N_O),DAMP(N_O)
      INTEGER NCUM(N_O,N_T)
1000  IMOVE=INT(RAN(ISEED)*NAO(NO))+1
      MTYPE=1
      DO WHILE(IMOVE.GT.NCUM(NO,MTYPE))
         MTYPE=MTYPE+1
      ENDDO
      IF (DELTA(MTYPE).LE.0.) GO TO 1000
C  up to 168 included
      GO TO (1,2,3,4,5,4,7,5,9,10,11,11,7,3,3,3,3,3,3,3,7,11,
     +5,7,11,5,7,4,10,9,9,10,3,3,3,3,3,3,7,11,11,5,5,7,7,11,
     +11,5,5,7,11,5,3,3,3,7,7,11,11,5,5,3,3,3,3,3,3,3,3,
     +3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,5,11,7,
     +3,3,3,3,20,20,20,20,20,20,20,20,20,20,21,21,22,23,22,
     +22,22,7,11,20,20,3,3,3,5,5,3,3,3,3,3,3,3,3,22,22,22,22,
     +7,7,7,7,24,24,24,24,25,7,3,3,3,3,3,26,26,26,27,27,27,
     +26,26,27,27,26,3,3,3,3),NSP(NO,IMOVE) 
*162-    1/4,y,-y+1/2
*163-    1/8,y,-y+1/4
*164-    1/8,y,y+1/4
*165-    5/8,0,1/4
*166-    1/8,0,1/4
*167-    7/8,0,1/4
*168-    3/8,0,1/4
*  General position
1     MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  x,x,x : supposes a cubic symmetry
2     MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=MOVE(1,4)
      MOVE(1,6)=MOVE(1,4)
      RETURN
*  x,x,0, x,x,1/2...
20    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=MOVE(1,4)
      MOVE(1,6)=0.
      RETURN
*  x,x,z, ...
21    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=MOVE(1,4)
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  x,-x,0,...
22    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=-MOVE(1,4)
      MOVE(1,6)=0.
      RETURN
*  x,-x,z,...
23    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=-MOVE(1,4)
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  x,2x,0,...
24    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=2.*MOVE(1,4)
      MOVE(1,6)=0.
      RETURN
*  x,2x,z,...
25    MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=2.*MOVE(1,4)
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  0,y,y,...
26    MOVE(1,4)=0.
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=MOVE(1,5)
      RETURN
*  0,y,-y,...
27    MOVE(1,4)=0.
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=-MOVE(1,5)
      RETURN
*  0,0,0 *  1/2,0,0 or 0,1/2,0 or 0,0,1/2 or etc : no MOVE
3     GO TO 1000
*  x,0,1/4  *  x,0,0  *  x,1/2,1/2  *  x,0,1/2  *  x,1/2,0
7     MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=0.
      MOVE(1,6)=0.
      RETURN
*  1/2,y,0 *  0,y,0  *  1/2,y,1/2  *  0,y,1/2
11    MOVE(1,4)=0.
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=0.
      RETURN
*  0,1/2,z  *  0,0,z  *  1/2,1/2,z  *  1/2,0,z
5     MOVE(1,4)=0.
      MOVE(1,5)=0.
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  x,y,0
9     MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=0.
      RETURN
*  x,1/4,z *  x,0,z
4     MOVE(1,4)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/A*DAMP(NO)
      MOVE(1,5)=0.
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
*  0,y,z
10    MOVE(1,4)=0.
      MOVE(1,5)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/BB*DAMP(NO)
      MOVE(1,6)=2.*(RAN(ISEED)-0.5)*DELTA(MTYPE)/C*DAMP(NO)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_OBJ(ISEED,NOB,NO)
*
*  Find the object to be moved/permuted/rotated/translated...
*
      NO=INT(RAN(ISEED)*NOB)+1
C      IF(NO.GT.NOB)WRITE(20,*)'  NO > NOB !!!'
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_TOR(ISEED,NBRT,NBRO)
*
*  Find the torsion angle to be changed
*
      NBRO=INT(RAN(ISEED)*NBRT)+1
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_TORROT(ISEED,NBRT,NRT)
*
*  Decide for a global rotation or a simple torsion angle
*
      NRT=INT(RAN(ISEED)*NBRT)+1
      IF(NRT.EQ.3)NRT=2
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_GENMOVEP(NAO,XL,MTYP,ISEED,NCUM,DELTA,IMOVE1,
     -   IMOVE2,MOVEP,NSP,NO,PROGRESS)
      INCLUDE 'esp0.inc'
	INTEGER(4) PROGRESS
	CHARACTER*120 BUFFER
      REAL MOVEP(2,6),DELTA(*),XL(N_O,5,N_A)
      INTEGER NCUM(N_O,N_T),NSP(N_O,N_A),MTYP(N_O,N_A),NAO(N_O) 
*
*  Find two atoms to be permuted
*
      JN=0
10    IMOVE1=INT(RAN(ISEED)*NAO(NO))+1
      JN=JN+1
      IF(JN.GE.1000) GO TO 1000
      MTYPE=1
      DO WHILE(IMOVE1.GT.NCUM(NO,MTYPE))
         MTYPE=MTYPE+1
      ENDDO
      IF (DELTA(MTYPE).LE.0.) GO TO 10
20    IMOVE2=INT(RAN(ISEED)*NAO(NO))+1
      JN=JN+1
      IF(JN.GE.1000) GO TO 1000
      MTYPE=1
      DO WHILE(IMOVE2.GT.NCUM(NO,MTYPE))
         MTYPE=MTYPE+1
      ENDDO
      IF (DELTA(MTYPE).LE.0.) GO TO 20
      IF(MTYP(NO,IMOVE1).EQ.MTYP(NO,IMOVE2))GO TO 10
      IF(NSP(NO,IMOVE1).NE.NSP(NO,IMOVE2))GO TO 10
      MOVEP(1,4)=XL(NO,1,IMOVE2)
      MOVEP(1,5)=XL(NO,2,IMOVE2)
      MOVEP(1,6)=XL(NO,3,IMOVE2)
      MOVEP(2,4)=XL(NO,1,IMOVE1)
      MOVEP(2,5)=XL(NO,2,IMOVE1)
      MOVEP(2,6)=XL(NO,3,IMOVE1)
      RETURN
1000	WRITE (BUFFER,*)' Are you sure that permutations are possible ?'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      Write(20,*)' Are you sure that permutations are possible ?'
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_NEWPOS(XL,MOVE,IMOVE,NCON,NO)
      INCLUDE 'esp0.inc'
      REAL MOVE(192,6),XL(N_O,5,N_A)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3) 
*
      MOVE(1,1)=XL(NO,1,IMOVE)
      MOVE(1,2)=XL(NO,2,IMOVE)
      MOVE(1,3)=XL(NO,3,IMOVE)
      MOVE(1,4)=MOVE(1,1)+MOVE(1,4)+3.
      MOVE(1,5)=MOVE(1,2)+MOVE(1,5)+3.
      MOVE(1,6)=MOVE(1,3)+MOVE(1,6)+3.
      MOVE(1,4)=2.*(MOVE(1,4)/2.-INT(MOVE(1,4)/2.))-1.
      MOVE(1,5)=2.*(MOVE(1,5)/2.-INT(MOVE(1,5)/2.))-1.
      MOVE(1,6)=2.*(MOVE(1,6)/2.-INT(MOVE(1,6)/2.))-1.
      IF(NCON.EQ.1)GO TO 10
      RETURN
10    X=MOVE(1,1)
      Y=MOVE(1,2)
      Z=MOVE(1,3)
      DO 801 K=2,NAS
      MOVE(K,1)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      MOVE(K,2)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2) 
801   MOVE(K,3)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      X=MOVE(1,4)
      Y=MOVE(1,5)
      Z=MOVE(1,6)
      DO 802 K=2,NAS
      MOVE(K,4)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      MOVE(K,5)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2) 
802   MOVE(K,6)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_NEWPOSP(XL,MOVEP,IMOVE1,IMOVE2,NO)
      INCLUDE 'esp0.inc'
      REAL MOVEP(2,6),XL(N_O,5,N_A)
*
      MOVEP(1,1)=XL(NO,1,IMOVE1)
      MOVEP(1,2)=XL(NO,2,IMOVE1)
      MOVEP(1,3)=XL(NO,3,IMOVE1)
      MOVEP(2,1)=XL(NO,1,IMOVE2)
      MOVEP(2,2)=XL(NO,2,IMOVE2)
      MOVEP(2,3)=XL(NO,3,IMOVE2)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE TORSION(XL,NX,NXB,METRIC,NAO,NO,IERR)
      INCLUDE 'esp0.inc'
      CHARACTER*4 NX(N_O,N_A),NXB(N_O,N_A),A1(N_O,500),A2(N_O,500)
      REAL METRIC(3,3)
      INTEGER NAO(N_O),MTYP(N_O,N_A),K1(50),K2(50),NA3(N_O,500)
      DIMENSION XL(N_O,5,N_A),NBO(N_O),KNBO(N_O,500),NA0(N_O,N_A,50)
      DIMENSION NAT(N_O,N_A),NROT(N_O,2,50,N_A)
      COMMON/TORS/NBROT(N_O),NA1(N_O,500),NA2(N_O,500),NTOR(N_O,50,N_A)
C
C  NAT(NO,I) = number of neighbours for atom I in object NO
C  NA0(NO,I,(J=1,NAT(NO,I)))= neighbour atom number of neighbour J
C                             for atom I in object NO
C  NA1(NO,K)= atom number of first atom in bond K
C  NA2(NO,K)= atom number of second atom in bond K
C  A1(NO,K)= atom character identifier of the first atom in bond K
C  A2(NO,K)= atom character identifier of the second atom in bond K
C
      IERR=0
      DMAX=1.8
      DMAXH=1.2
      NHATOM=0
*
* Intra-object interatomic distances
*
      WRITE(20,20)
      NBO(NO)=0
* Sum on atoms in object NO
      DO 10 K=1,NAO(NO)
      IF(NXB(NO,K)(1:2).EQ.'H ')NHATOM=NHATOM+1
* Sum on other atoms in object NO
      DO 10 L=1,NAO(NO)
      IF(L.EQ.K)GO TO 10
*
            X=XL(NO,1,K)-XL(NO,1,L)+3.
            Y=XL(NO,2,K)-XL(NO,2,L)+3.
            Z=XL(NO,3,K)-XL(NO,3,L)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            DIST=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            DIST=SQRT(DIST)
            IF(DIST.LT.0.0001)GO TO 10
            IF (DIST.GT.DMAX)GO TO 10
C
C  Count bonds, but eliminate H bonds
C
      IF(NXB(NO,K)(1:2).EQ.'H ')GO TO 10
      IF(NXB(NO,L)(1:2).EQ.'H ')GO TO 10
*
      WRITE(20,21)NX(NO,K),NX(NO,L),DIST 
C
C  How many neighbours for atom K
C
      NAT(NO,K)=NAT(NO,K)+1
      IF(NAT(NO,K).GT.50)THEN
      IERR=50
      RETURN
      ENDIF
C
C  Which neighbour for atom K
C
      NA0(NO,K,NAT(NO,K))=L
      NBO(NO)=NBO(NO)+1
      IF(NBO(NO).GT.500)THEN
      IERR=500
      RETURN
      ENDIF
      A1(NO,NBO(NO))(1:4)=NX(NO,K)(1:4)
      A2(NO,NBO(NO))(1:4)=NX(NO,L)(1:4)
      NA1(NO,NBO(NO))=K
      NA2(NO,NBO(NO))=L
C  KNBO is test for rotatable bond
      KNBO(NO,NBO(NO))=1
C
C  Do not count pairs twice
C
      DO 9 JJ=1,NBO(NO)-1
      IF(A1(NO,NBO(NO)).EQ.A2(NO,JJ).AND.A2(NO,NBO(NO)).EQ.A1(NO,JJ))
     +THEN
      NBO(NO)=NBO(NO)-1
      GO TO 10
      ENDIF
9     CONTINUE
C
10    CONTINUE
      WRITE(20,22)NO,NBO(NO)
C
C  Find rotatable bonds
C
C  Find terminal bonds which are not rotatable
C      Both atoms in a non-terminal bond should have 
C      at least 2 neighbours
C
      WRITE(20,23)
      NTERM=0
      DO 101 I=1,NBO(NO)
      DO 99 J=1,NAO(NO)
      IF(A1(NO,I)(1:4).EQ.NX(NO,J)(1:4))THEN
      JJ=J
      IF(NAT(NO,JJ).EQ.1)GO TO 100
      ENDIF
      IF(A2(NO,I)(1:4).EQ.NX(NO,J)(1:4))THEN
      JJ=J
      IF(NAT(NO,JJ).EQ.1)GO TO 100
      ENDIF
99    CONTINUE
      GO TO 101
100   KNBO(NO,I)=0
      NTERM=NTERM+1
      WRITE(20,24)I,A1(NO,I),A2(NO,I)
101   CONTINUE
      IF(NTERM.EQ.0)WRITE(20,38)
C
C  Find cycles with 3, 4, 5, 6 or 7 atoms,
C          any bond in a cycle is not rotatable
C
      WRITE(20,36)
      IRING=0
C
C  First take atom A1 in bond I
      DO 200 K=1,NBO(NO)
C  No test on bonds already eliminated
      IF(KNBO(NO,I).EQ.0)GO TO 200
      I1=NA1(NO,K)
      I2=NA2(NO,K)
C
C  See neighbouring atoms of atom A2 in bond I
C
C  NAT(NO,I) = number of neighbours for atom I in object NO
C  NA0(NO,I,(J=1,NAT(NO,I)))= neighbour atom number of neighbour J
C                             for atom I in object NO
C
      DO 150 II2=1,NAT(NO,I2)
C  Possible atoms A3
      I3=NA0(NO,I2,II2)
      IF(I3.EQ.I1)GO TO 150
      DO 149 II3=1,NAT(NO,I3)
C  Possible atoms A4
      I4=NA0(NO,I3,II3)
      IF(I4.EQ.I2)GO TO 149
      IF(I4.EQ.I1)THEN
      WRITE(20,30)I1,I2,I3,NX(NO,I1),NX(NO,I2),NX(NO,I3)
      KNBO(NO,K)=0
      IRING=IRING+1
      ENDIF
      DO 148 II4=1,NAT(NO,I4)
C  Possible atoms A5
      I5=NA0(NO,I4,II4)
      IF(I5.EQ.I3)GO TO 148
      IF(I5.EQ.I1)THEN
      WRITE(20,29)I1,I2,I3,I4,NX(NO,I1),NX(NO,I2),
     +NX(NO,I3),NX(NO,I4)
      KNBO(NO,K)=0
      IRING=IRING+1
      ENDIF
      DO 147 II5=1,NAT(NO,I5)
C  Possible atoms A6
      I6=NA0(NO,I5,II5)
      IF(I6.EQ.I4)GO TO 147
      IF(I6.EQ.I2)GO TO 147
      IF(I6.EQ.I1)THEN
      WRITE(20,25)I1,I2,I3,I4,I5,NX(NO,I1),NX(NO,I2),
     +NX(NO,I3),NX(NO,I4),NX(NO,I5)
      KNBO(NO,K)=0
      IRING=IRING+1
      ENDIF
      DO 146 II6=1,NAT(NO,I6)
C  Possible atoms A7
      I7=NA0(NO,I6,II6)
      IF(I7.EQ.I5)GO TO 146
      IF(I7.EQ.I2)GO TO 146
      IF(I7.EQ.I3)GO TO 146
      IF(I7.EQ.I1)THEN
      WRITE(20,26)I1,I2,I3,I4,I5,I6,NX(NO,I1),NX(NO,I2),
     +NX(NO,I3),NX(NO,I4),NX(NO,I5),NX(NO,I6)
      KNBO(NO,K)=0
      IRING=IRING+1
      ENDIF
      DO 145 II7=1,NAT(NO,I7)
C  Possible atoms A8
      I8=NA0(NO,I7,II7)
      IF(I8.EQ.I6)GO TO 145
      IF(I8.EQ.I2)GO TO 145
      IF(I8.EQ.I3)GO TO 145
      IF(I8.EQ.I4)GO TO 145
      IF(I8.EQ.I1)THEN
      WRITE(20,31)I1,I2,I3,I4,I5,I6,I7,NX(NO,I1),NX(NO,I2),
     +NX(NO,I3),NX(NO,I4),NX(NO,I5),NX(NO,I6),NX(NO,I7)
      KNBO(NO,K)=0
      IRING=IRING+1
      ENDIF
145   CONTINUE
146   CONTINUE
147   CONTINUE
148   CONTINUE
149   CONTINUE
150   CONTINUE
200   CONTINUE
      IF(IRING.EQ.0)WRITE(20,39)
C
C  Remaining bonds are rotatable ?? Let us see :
C
      NBROT(NO)=0
      WRITE(20,27)
      DO 300 K=1,NBO(NO)
      IF(KNBO(NO,K).EQ.0)GO TO 300
      WRITE(20,28)K,A1(NO,K),A2(NO,K)
      NBROT(NO)=NBROT(NO)+1
      IF(NBROT(NO).GT.50)THEN
      IERR=51
      RETURN
      ENDIF
      NA1(NO,NBROT(NO))=NA1(NO,K)
      NA2(NO,NBROT(NO))=NA2(NO,K)
      A1(NO,NBROT(NO))(1:4)=A1(NO,K)(1:4)
      A2(NO,NBROT(NO))(1:4)=A2(NO,K)(1:4)
300   CONTINUE
      IF(NBROT(NO).EQ.0)WRITE(20,40)
C
C  Now the following have changed into :
C  NA1(NO,K)= atom number of first atom in rotatable bond K
C  NA2(NO,K)= atom number of second atom in rotatable bond K,
C   and A1 and A2 are the corresponding character identifiers,
C     and there are a total of NBROT rotatable bonds
C
C  Which side of the bond is the smallest and will be really rotated ?
C     Examination will not be perfect if exact numbering needs to see
C     at larger distance than the 10th neighbour. Anyway, this should
C     be sufficient to determine the smallest side.
C
      DO 400 K=1,NBROT(NO)
      I1=NA1(NO,K)
      I2=NA2(NO,K)
      DO 399 I=1,NAO(NO)
      NROT(NO,1,K,I)=0
      NROT(NO,2,K,I)=0
399   CONTINUE
C
C  NROT(NO,1,K,I) are code for determining if atom number I
C                 belongs to the side of atom 1 in the
C                 rotatable bond K
C  NROT(NO,2,K,I) are code for determining if atom number I
C                 belongs to the side of atom 2 in the
C                 rotatable bond K
C
      DO 450 II2=1,NAT(NO,I2)
      I3=NA0(NO,I2,II2)
      IF(I3.EQ.I1)GO TO 450
      NROT(NO,2,K,I3)=1
      DO 449 II3=1,NAT(NO,I3)
      I4=NA0(NO,I3,II3)
      IF(I4.EQ.I2)GO TO 449
      IF(I4.EQ.I1)GO TO 449
      NROT(NO,2,K,I4)=1
      DO 448 II4=1,NAT(NO,I4)
      I5=NA0(NO,I4,II4)
      IF(I5.EQ.I2)GO TO 448
      IF(I5.EQ.I1)GO TO 448
      NROT(NO,2,K,I5)=1
      DO 447 II5=1,NAT(NO,I5)
      I6=NA0(NO,I5,II5)
      IF(I6.EQ.I1)GO TO 447
      IF(I6.EQ.I2)GO TO 447
      NROT(NO,2,K,I6)=1
      DO 446 II6=1,NAT(NO,I6)
      I7=NA0(NO,I6,II6)
      IF(I7.EQ.I1)GO TO 446
      IF(I7.EQ.I2)GO TO 446
      NROT(NO,2,K,I7)=1
      DO 445 II7=1,NAT(NO,I7)
      I8=NA0(NO,I7,II7)
      IF(I8.EQ.I1)GO TO 445
      IF(I8.EQ.I2)GO TO 445
      NROT(NO,2,K,I8)=1
      DO 444 II8=1,NAT(NO,I8)
      I9=NA0(NO,I8,II8)
      IF(I9.EQ.I1)GO TO 444
      IF(I9.EQ.I2)GO TO 444
      NROT(NO,2,K,I9)=1
      DO 443 II9=1,NAT(NO,I9)
      I10=NA0(NO,I9,II9)
      IF(I10.EQ.I1)GO TO 443
      IF(I10.EQ.I2)GO TO 443
      NROT(NO,2,K,I10)=1
      DO 442 II10=1,NAT(NO,I10)
      I11=NA0(NO,I10,II10)
      IF(I11.EQ.I1)GO TO 442
      IF(I11.EQ.I2)GO TO 442
      NROT(NO,2,K,I11)=1
442   CONTINUE
443   CONTINUE
444   CONTINUE
445   CONTINUE
446   CONTINUE
447   CONTINUE
448   CONTINUE
449   CONTINUE
450   CONTINUE
C
      DO 550 II2=1,NAT(NO,I1)
      I3=NA0(NO,I1,II2)
      IF(I3.EQ.I2)GO TO 550
      NROT(NO,1,K,I3)=1
      DO 549 II3=1,NAT(NO,I3)
      I4=NA0(NO,I3,II3)
      IF(I4.EQ.I2)GO TO 549
      IF(I4.EQ.I1)GO TO 549
      NROT(NO,1,K,I4)=1
      DO 548 II4=1,NAT(NO,I4)
      I5=NA0(NO,I4,II4)
      IF(I5.EQ.I2)GO TO 548
      IF(I5.EQ.I1)GO TO 548
      NROT(NO,1,K,I5)=1
      DO 547 II5=1,NAT(NO,I5)
      I6=NA0(NO,I5,II5)
      IF(I6.EQ.I1)GO TO 547
      IF(I6.EQ.I2)GO TO 547
      NROT(NO,1,K,I6)=1
      DO 546 II6=1,NAT(NO,I6)
      I7=NA0(NO,I6,II6)
      IF(I7.EQ.I1)GO TO 546
      IF(I7.EQ.I2)GO TO 546
      NROT(NO,1,K,I7)=1
      DO 545 II7=1,NAT(NO,I7)
      I8=NA0(NO,I7,II7)
      IF(I8.EQ.I1)GO TO 545
      IF(I8.EQ.I2)GO TO 545
      NROT(NO,1,K,I8)=1
      DO 544 II8=1,NAT(NO,I8)
      I9=NA0(NO,I8,II8)
      IF(I9.EQ.I1)GO TO 544
      IF(I9.EQ.I2)GO TO 544
      NROT(NO,1,K,I9)=1
      DO 543 II9=1,NAT(NO,I9)
      I10=NA0(NO,I9,II9)
      IF(I10.EQ.I1)GO TO 543
      IF(I10.EQ.I2)GO TO 543
      NROT(NO,1,K,I10)=1
      DO 542 II10=1,NAT(NO,I10)
      I11=NA0(NO,I10,II10)
      IF(I11.EQ.I1)GO TO 542
      IF(I11.EQ.I2)GO TO 542
      NROT(NO,1,K,I11)=1
542   CONTINUE
543   CONTINUE
544   CONTINUE
545   CONTINUE
546   CONTINUE
547   CONTINUE
548   CONTINUE
549   CONTINUE
550   CONTINUE
C
C
C
400   CONTINUE
C
C  Compare the number of neighbors on each side of the rotatable bonds
C
      DO 600 K=1,NBROT(NO)
      K1(K)=0
      K2(K)=0
      DO 599 I=1,NAO(NO)
      IF(NROT(NO,1,K,I).EQ.1)K1(K)=K1(K)+1
      IF(NROT(NO,2,K,I).EQ.1)K2(K)=K2(K)+1
599   CONTINUE
      WRITE(20,32)K,K1(K),K2(K)  
600   CONTINUE   
C
C  Store the atoms on the smallest side that will be allowed to
C       rotate. NTOR(NO,K,I) defines the rotatable small part : atom
C       number I in object NO for torsion bond K will rotate if
C       NTOR(NO,K,I)=1.
C
      WRITE(20,34)NBROT(NO)
      DO 700 K=1,NBROT(NO)
      WRITE(20,35)K
C  Reverse the A1-A2 order in A2-A1 if K1<K2
C  because A2 is conventionally on the small rotating side
      IF(K1(K).LE.K2(K))THEN
      NA3(NO,K)=NA2(NO,K)
      NA2(NO,K)=NA1(NO,K)
      NA1(NO,K)=NA3(NO,K)
      ENDIF
      DO 700 I=1,NAO(NO)
      IF(K1(K).LE.K2(K))THEN
      NTOR(NO,K,I)=NROT(NO,1,K,I)
      IF(NTOR(NO,K,I).EQ.1)WRITE(20,33)NX(NO,I),A1(NO,K),A2(NO,K)
      ENDIF
      IF(K2(K).LT.K1(K))THEN
      NTOR(NO,K,I)=NROT(NO,2,K,I)
      IF(NTOR(NO,K,I).EQ.1)WRITE(20,33)NX(NO,I),A1(NO,K),A2(NO,K)
      ENDIF
700   CONTINUE
C
C  Find now the H atoms that will also rotate...
C
      IF(NHATOM.EQ.0)GO TO 1000
      DO 901 K=1,NBROT(NO)
      WRITE(20,41)K
      DO 900 KK=1,NAO(NO)
      IF(NXB(NO,KK)(1:2).EQ.'H ')GO TO 900
C  Take account of H atoms possibly linked to A2
      IF(KK.EQ.NA2(NO,K))GO TO 898
C
      IF(NTOR(NO,K,KK).NE.1)GO TO 900
898   DO 899 L=1,NAO(NO)
      IF(L.EQ.KK)GO TO 899
      IF(NXB(NO,L)(1:2).NE.'H ')GO TO 899
*
            X=XL(NO,1,KK)-XL(NO,1,L)+3.
            Y=XL(NO,2,KK)-XL(NO,2,L)+3.
            Z=XL(NO,3,KK)-XL(NO,3,L)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            DIST=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            DIST=SQRT(DIST)
            IF(DIST.LT.0.0001)GO TO 899
            IF (DIST.GT.DMAXH)GO TO 899
      NTOR(NO,K,L)=1
      WRITE(20,33)NX(NO,L),A1(NO,K),A2(NO,K)
899   CONTINUE
900   CONTINUE
901   CONTINUE
C
1000  CONTINUE
      WRITE(20,37)
C
C  Go back now with NBROT, NA1, NA2 and NTOR...
C
20    FORMAT(///'  Molecule connectivity, up to 1.8 Angstrom'/,
     +10X,' A1  A2   Distance'/)
21    FORMAT(10X,2(A4,1X),F9.5)
22    FORMAT(/'  Total number of non-H bonds in object ',I3,' : ',I4/)
23    FORMAT('  Terminal bonds : '/'  Bond number  A1   A2')
24    FORMAT(5X,I3,7X,2(A4,1X))
25    FORMAT('  5 atoms ring : ',5I3,2X,5(A4,1X))
26    FORMAT('  6 atoms ring : ',6I3,2X,6(A4,1X))
27    FORMAT(//'  Rotatable bonds : No   A1   A2')
28    FORMAT(18X,I4,3X,2(A4,1X))
29    FORMAT('  4 atoms ring : ',4I3,2X,4(A4,1X))
30    FORMAT('  3 atoms ring : ',3I3,2X,3(A4,1X))
31    FORMAT('  7 atoms ring : ',7I3,2X,7(A4,1X))
32    FORMAT('  Rotatable bond ',I3,' has',I3,' and',I3,' neighbours')
33    FORMAT('    Atom ',A4,' will rotate around bond ',A4,'-',A4)
34    FORMAT(//' The following ',I3,' groups of atoms can rotate, if you
     + wish'/'             (see the manual at the nobt parameter)')
35    FORMAT(/'    Group number ',I3)
36    FORMAT(//' Ring analysis '/
     +'         (the same ring may appear several times)')
37    FORMAT(//'  End of molecule connectivity analysis'//)
38    FORMAT(/'     No terminal bond found'/)
39    FORMAT(/'     No ring found'/)
40    FORMAT(/'     No rotatable bond found'/)
41    FORMAT(/'  The following H atoms will also rotate in group ',I3/)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_CHECK(NTYP,METRIC,NCUM,RCUT,IMOVE,MOVE,
     1ACCEPTABLE,NOB,NAO,NOO,NXLL,MTYP,NZERO)
      INCLUDE 'esp0.inc'
      REAL RCUT(*),METRIC(3,3),MOVE(192,6)
      INTEGER NCUM(N_O,N_T),NAO(N_O),MTYP(N_O,N_A),NZERO(N_O)
      LOGICAL ACCEPTABLE
      DIMENSION NXLL(6000),XLLL(3,6000)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3) 
*
*     Check whether the move
*     is acceptable according to the cut-off restrictions 
*
      IM=IMOVE
      IF((NOO-1).LE.0)GO TO 2
      DO 1 I=1,NOO-1
1     IM=IM+NAO(I)
2     IM=IM*NAS-NAS+1
      I=0
      DO 800 NO=1,NOB
      DO 800 L=1,NAO(NO)
      IF(NO.NE.NOO)GO TO 799
      IF(L.NE.IMOVE)GO TO 799
      X=MOVE(1,4)
      Y=MOVE(1,5)
      Z=MOVE(1,6)
799   DO 800 K=1,NAS
      I=I+1
      IF(NO.NE.NOO.AND.L.NE.IMOVE)GO TO 798
      XLLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      NXLL(I)=MTYP(NO,L)
      GO TO 800
798   XLLL(1,I)=XLL(1,I)
      XLLL(2,I)=XLL(2,I)
      XLLL(3,I)=XLL(3,I)
      NXLL(I)=MTYP(NO,L)
800   CONTINUE
C     WRITE(20,*)I,NOO,NOB
*
      I=0
* Sum on objects
      DO 10 NO=1,NOB
* Sum on atoms in object
      DO 10 K=1,NAO(NO)
* Sum on equivalent positions
      DO 10 KK=1,NAS
      I=I+1
         IF (I.NE.IM) THEN
          ITYPE=NXLL(I)
          JTYPE=NXLL(IM)
            I1=MIN(ITYPE,JTYPE)
            J1=MAX(ITYPE,JTYPE)
            IC=(I1-1)*(2*NTYP-I1)/2+J1
*
*---  Apply strict constraints if NOBT=11
*
*
         IF(NZERO(NOO).EQ.1)THEN
*
            X=XLLL(1,I)-MOVE(1,4)+3.
            Y=XLLL(2,I)-MOVE(1,5)+3.
            Z=XLLL(3,I)-MOVE(1,6)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            DNEW=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            DNEW=SQRT(DNEW)
              IF(DNEW.LT.0.0001)GO TO 10
            IF (DNEW.LT.RCUT(IC)) ACCEPTABLE=.FALSE.
C      WRITE(20,21)I,IM,JTYPE,ITYPE,IC,NTYP,NAS,IMOVE,DOLD,DNEW,RCUT(IC) 
C      WRITE(20,22)(XLLL(JJ,I),MOVE(1,JJ),MOVE(1,JJ+3),JJ=1,3)
            IF(.NOT.ACCEPTABLE)RETURN
*
         GO TO 10
         ENDIF
            X=XLLL(1,I)-MOVE(1,1)+3.
            Y=XLLL(2,I)-MOVE(1,2)+3.
            Z=XLLL(3,I)-MOVE(1,3)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            DOLD=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            DOLD=SQRT(DOLD)
*
* Avoid problems with special positions 
*
              IF(DOLD.LT.0.0001)GO TO 10
*
            X=XLLL(1,I)-MOVE(1,4)+3.
            Y=XLLL(2,I)-MOVE(1,5)+3.
            Z=XLLL(3,I)-MOVE(1,6)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            DNEW=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            DNEW=SQRT(DNEW)
              IF(DNEW.LT.0.0001)GO TO 10
            IF (DNEW.LT.RCUT(IC).AND.(DNEW.LT.DOLD)) ACCEPTABLE=.FALSE.
C      WRITE(20,21)I,IM,JTYPE,ITYPE,IC,NTYP,NAS,IMOVE,DOLD,DNEW,RCUT(IC) 
C      WRITE(20,22)(XLLL(JJ,I),MOVE(1,JJ),MOVE(1,JJ+3),JJ=1,3)
C21    FORMAT(8I4,3F9.5)
C22    FORMAT(9F9.5)
            IF(.NOT.ACCEPTABLE)RETURN
*
         ENDIF
10      CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_CHECKMR(XL,NTYP,METRIC,RCUT,ACCEPTABLE,NOB,
     1NAO,NOO,NXLL,MTYP)
      INCLUDE 'esp0.inc'
      REAL XL(N_O,3,N_A),RCUT(*),METRIC(3,3)
      INTEGER NAO(N_O),MTYP(N_O,N_A)
      LOGICAL ACCEPTABLE
      DIMENSION NXLL(6000),NCOD(6000),NOBI(6000),XLLL(3,6000)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3) 
*
*     Check whether the MOVE
*     is ACCEPTABLE according to the cut-off restrictions 
*
      J=0
      I=0
      DO 800 NO=1,NOB
      DO 800 L=1,NAO(NO)
C      WRITE(20,1000)XL(NO,1,L),XL(NO,2,L),XL(NO,3,L)
C1000  FORMAT(5X,3F12.5)
      IF(NO.NE.NOO)GO TO 799
      X=XL(NO,1,L)
      Y=XL(NO,2,L)
      Z=XL(NO,3,L)
799   DO 800 K=1,NAS
      I=I+1
      NOBI(I)=0
      IF(NO.NE.NOO)GO TO 798
         IF(K.EQ.1)THEN
           NOBI(I)=1
           J=J+1
           NCOD(J)=I
         ENDIF
      XLLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      NXLL(I)=MTYP(NO,L)
      GO TO 800
798   XLLL(1,I)=XLL(1,I)
      XLLL(2,I)=XLL(2,I)
      XLLL(3,I)=XLL(3,I)
      NXLL(I)=MTYP(NO,L)
800   CONTINUE
      I=0
* Sum on objects
      DO 10 NO=1,NOB
* Sum on atoms in object
      DO 10 K=1,NAO(NO)
* Sum on equivalent positions
      DO 10 KK=1,NAS
      I=I+1
* Sum on moving atoms
      DO 10 LL=1,J
* Avoid intra-fragment contact
        IF(NOBI(I).EQ.1)GO TO 10
*
         L=NCOD(LL)
         IF (I.NE.L) THEN
          ITYPE=NXLL(I)
          JTYPE=NXLL(L)
            I1=MIN(ITYPE,JTYPE)
            J1=MAX(ITYPE,JTYPE)
            IC=(I1-1)*(2*NTYP-I1)/2+J1
*
            X=XLLL(1,I)-XLLL(1,L)+3.
            Y=XLLL(2,I)-XLLL(2,L)+3.
            Z=XLLL(3,I)-XLLL(3,L)+3.
            X=2.*(X/2.-INT(X/2.))-1.
            Y=2.*(Y/2.-INT(Y/2.))-1.
            Z=2.*(Z/2.-INT(Z/2.))-1.
            D=METRIC(1,1)*X*X+METRIC(2,2)*Y*Y+METRIC(3,3)*Z*Z
     -       +2.*(METRIC(1,2)*X*Y+METRIC(1,3)*X*Z+METRIC(2,3)*Y*Z)
            D=SQRT(D)
*
* Avoid problems with special positions 
*
            IF(D.LT.0.001)GO TO 10
*
            IF (D.LT.RCUT(IC)) ACCEPTABLE=.FALSE.
            IF(.NOT.ACCEPTABLE)RETURN
*
         ENDIF
10      CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_SUMSTOT(NHKL,FOBS,KXR,XL,IMOVE,MOVE,DIFF,
     -NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      DIMENSION FOBS(N_HKL),XL(N_O,5,N_A),SNEWT(N_HKL),SNEW(N_HKL)
      DIMENSION NAO(N_O)
      REAL MOVE(192,6)
*
      IM=IMOVE
      XL(NO,1,IM)=MOVE(1,4)
      XL(NO,2,IM)=MOVE(1,5)
      XL(NO,3,IM)=MOVE(1,6)
      SUM_FS=0.
C
C  CALCULATION OF NEW |F|s
C
      SUMCAL=0.
C
C  Main loop
C
      DO 1000 NM=1,NHKL
      SNEWT(NM)=0.
      DO 121 IR=1,IRL
      ARG=TR(IR,NM)
      DO 72 J=1,3
72    ARG=ARG+H(IR,J,NM)*XL(NO,J,IM)
      ARG=6.28319*ARG
      COSAR(NO,IM,IR,NM)=COS(ARG)
121   SINAR(NO,IM,IR,NM)=SIN(ARG)
      AV=0.
      BV=0.
      DO 170 N=1,NOB
      DO 170 I=1,NAO(N)
      SAI=0.
      SBI=0.
      DO 140 IR=1,IRL
      SAI=SAI+COSAR(N,I,IR,NM)
      IF(ICENT.EQ.1)SBI=SINAR(N,I,IR,NM)+SBI 
140   CONTINUE
      SAI=FLOAT(ICENT)*SAI
      IF(KXR.LT.4)GO TO 150
      GOTO 161
150   AV=SNEXIN(N,I)*SAI+AV
      BV=SNEXIN(N,I)*SBI+BV
      IF(KXR.LT.4)GO TO 170
161   AV=SNEXIX(N,I,NM)*SAI+AV
      BV=SNEXIX(N,I,NM)*SBI+BV
      AVI=AV-SNXI(N,I)*SBI
      AVJ=AV+SNXI(N,I)*SBI
      BVI=BV-SNXI(N,I)*SAI
      BVJ=BV+SNXI(N,I)*SAI
170   CONTINUE
190   IF(KXR.LT.4)GOTO 200
      SNEW(NM)=(AVI*AVI+AVJ*AVJ+BVI*BVI+BVJ*BVJ)
      GOTO 210
200   SNEW(NM)=(AV*AV+BV*BV)
210   SNEW(NM)=SQRT(SNEW(NM))
      SUMCAL=SUMCAL+SNEW(NM)
      SNEWT(NM)=SNEWT(NM)+SNEW(NM)
1000  CONTINUE
      DIFF=0.
      XL(NO,1,IM)=MOVE(1,1)
      XL(NO,2,IM)=MOVE(1,2)
      XL(NO,3,IM)=MOVE(1,3)
      DO 1001 NM=1,NHKL
      SNEWT(NM)=SNEWT(NM)/SUMCAL*SUM_F
      IF(KP.EQ.1)GO TO 74
      SUM_FS=SUM_FS+(FOBS(NM)-SNEWT(NM))**2
      DIFF=DIFF+ABS(FOBS(NM)-SNEWT(NM)) 
74    DO 131 IR=1,IRL
      ARG=TR(IR,NM)
      DO 73 J=1,3
73    ARG=ARG+H(IR,J,NM)*XL(NO,J,IM)
      ARG=6.28319*ARG
      COSAR(NO,IM,IR,NM)=COS(ARG)
131   SINAR(NO,IM,IR,NM)=SIN(ARG)
1001  CONTINUE
      IF(KP.EQ.1)GO TO 1003
      DIFF=DIFF/SUM_F
      IF(KP.EQ.0.OR.KP.EQ.2)GO TO 1004
1003  CALL PROFILE2(NHKL,SNEWT,SUM_FS,DIFF)
      DIFF=DIFF/SUM_Y
1004  CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_SUMSTOTP(NHKL,FOBS,KXR,XL,IMOVE1,IMOVE2,MOVEP,
     -DIFF,NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      DIMENSION FOBS(N_HKL),XL(N_O,5,N_A),SNEWT(N_HKL),SNEW(N_HKL)
      DIMENSION NAO(N_O)
      REAL MOVEP(2,6)
*
      XL(NO,1,IMOVE1)=MOVEP(1,4)
      XL(NO,2,IMOVE1)=MOVEP(1,5)
      XL(NO,3,IMOVE1)=MOVEP(1,6)
      XL(NO,1,IMOVE2)=MOVEP(2,4)
      XL(NO,2,IMOVE2)=MOVEP(2,5)
      XL(NO,3,IMOVE2)=MOVEP(2,6)
      SUM_FS=0.
C
C  CALCULATION OF NEW |F|s
C
      SUMCAL=0.
C
C
      DO 1000 NM=1,NHKL
      SNEWT(NM)=0.
      DO 120 IR=1,IRL
      ARG=TR(IR,NM)
      DO 70 J=1,3
70    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE1)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE1,IR,NM)=COS(ARG)
120   SINAR(NO,IMOVE1,IR,NM)=SIN(ARG)
      DO 130 IR=1,IRL
      ARG=TR(IR,NM)
      DO 71 J=1,3
71    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE2)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE2,IR,NM)=COS(ARG)
130   SINAR(NO,IMOVE2,IR,NM)=SIN(ARG)
      AV=0.
      BV=0.
      DO 170 N=1,NOB
      DO 170 I=1,NAO(N)
      SAI=0.
      SBI=0.
      DO 140 IR=1,IRL
      SAI=SAI+COSAR(N,I,IR,NM)
      IF(ICENT.EQ.1)SBI=SINAR(N,I,IR,NM)+SBI 
140   CONTINUE
      SAI=FLOAT(ICENT)*SAI
      IF(KXR.LT.4)GO TO 150
      GOTO 161
150   AV=SNEXIN(N,I)*SAI+AV
      BV=SNEXIN(N,I)*SBI+BV
      IF(KXR.LT.4)GO TO 170
161   AV=SNEXIX(N,I,NM)*SAI+AV
      BV=SNEXIX(N,I,NM)*SBI+BV
      AVI=AV-SNXI(N,I)*SBI
      AVJ=AV+SNXI(N,I)*SBI
      BVI=BV-SNXI(N,I)*SAI
      BVJ=BV+SNXI(N,I)*SAI
170   CONTINUE
190   IF(KXR.LT.4)GOTO 200
      SNEW(NM)=(AVI*AVI+AVJ*AVJ+BVI*BVI+BVJ*BVJ)
      GOTO 210
200   SNEW(NM)=(AV*AV+BV*BV)
210   SNEW(NM)=SQRT(SNEW(NM))
      SUMCAL=SUMCAL+SNEW(NM)
      SNEWT(NM)=SNEWT(NM)+SNEW(NM)
1000  CONTINUE
      DIFF=0.
      XL(NO,1,IMOVE1)=MOVEP(1,1)
      XL(NO,2,IMOVE1)=MOVEP(1,2)
      XL(NO,3,IMOVE1)=MOVEP(1,3)
      XL(NO,1,IMOVE2)=MOVEP(2,1)
      XL(NO,2,IMOVE2)=MOVEP(2,2)
      XL(NO,3,IMOVE2)=MOVEP(2,3)
      DO 1001 NM=1,NHKL
      DO 121 IR=1,IRL
      ARG=TR(IR,NM)
      DO 72 J=1,3
72    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE1)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE1,IR,NM)=COS(ARG)
121   SINAR(NO,IMOVE1,IR,NM)=SIN(ARG)
      DO 131 IR=1,IRL
      ARG=TR(IR,NM)
      DO 73 J=1,3
73    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE2)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE2,IR,NM)=COS(ARG)
131   SINAR(NO,IMOVE2,IR,NM)=SIN(ARG)
      SNEWT(NM)=SNEWT(NM)/SUMCAL*SUM_F
      IF(KP.EQ.1)GO TO 1001
      SUM_FS=SUM_FS+(FOBS(NM)-SNEWT(NM))**2
      DIFF=DIFF+ABS(FOBS(NM)-SNEWT(NM)) 
1001  CONTINUE
      IF(KP.EQ.1)GO TO 1003
      DIFF=DIFF/SUM_F
      IF(KP.EQ.0.OR.KP.EQ.2)GO TO 1004
1003  CALL PROFILE2(NHKL,SNEWT,SUM_FS,DIFF)
      DIFF=DIFF/SUM_Y
1004  CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_SUMMR(NHKL,FOBS,KXR,XL,
     -DIFF,NAO,SUM_FS,SUM_F,SUM_Y,KP,NO,NOB)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      DIMENSION FOBS(N_HKL),XL(N_O,3,N_A),SNEW(N_HKL),SNEWT(N_HKL)
      DIMENSION NAO(N_O)
*
      SUM_FS=0.
C
C  CALCULATION OF NEW |F|s
C
      SUMCAL=0.
C
C
      DO 1000 NM=1,NHKL
      SNEWT(NM)=0.
      DO 120 I=1,NAO(NO)
      DO 120 IR=1,IRL
      ARG=TR(IR,NM)
      DO 70 J=1,3
70    ARG=ARG+H(IR,J,NM)*XL(NO,J,I)
      ARG=6.28319*ARG
      COSAR(NO,I,IR,NM)=COS(ARG)
120   SINAR(NO,I,IR,NM)=SIN(ARG)
      AV=0.
      BV=0.
      DO 170 N=1,NOB
      DO 170 I=1,NAO(N)
      SAI=0.
      SBI=0.
      DO 140 IR=1,IRL
      SAI=SAI+COSAR(N,I,IR,NM)
      IF(ICENT.EQ.1)SBI=SINAR(N,I,IR,NM)+SBI 
140   CONTINUE
      SAI=FLOAT(ICENT)*SAI
      IF(KXR.LT.4)GO TO 150
      GOTO 161
150   AV=SNEXIN(N,I)*SAI+AV
      BV=SNEXIN(N,I)*SBI+BV
      IF(KXR.LT.4)GO TO 170
161   AV=SNEXIX(N,I,NM)*SAI+AV
      BV=SNEXIX(N,I,NM)*SBI+BV
      AVI=AV-SNXI(N,I)*SBI
      AVJ=AV+SNXI(N,I)*SBI
      BVI=BV-SNXI(N,I)*SAI
      BVJ=BV+SNXI(N,I)*SAI
170   CONTINUE
190   IF(KXR.LT.4)GOTO 200
      SNEW(NM)=(AVI*AVI+AVJ*AVJ+BVI*BVI+BVJ*BVJ)
      GOTO 210
200   SNEW(NM)=(AV*AV+BV*BV)
210   SNEW(NM)=SQRT(SNEW(NM))
      SUMCAL=SUMCAL+SNEW(NM)
      SNEWT(NM)=SNEWT(NM)+SNEW(NM)
1000  CONTINUE
      DIFF=0.
      DO 1001 NM=1,NHKL
      SNEWT(NM)=SNEWT(NM)/SUMCAL*SUM_F
      IF(KP.EQ.1)GO TO 1001
      SUM_FS=SUM_FS+(FOBS(NM)-SNEWT(NM))**2
      DIFF=DIFF+ABS(FOBS(NM)-SNEWT(NM)) 
1001  CONTINUE
      IF(KP.EQ.1)GO TO 1003
      DIFF=DIFF/SUM_F
      IF(KP.EQ.0.OR.KP.EQ.2)GO TO 1004
1003  CALL PROFILE2(NHKL,SNEWT,SUM_FS,DIFF)
      DIFF=DIFF/SUM_Y
1004  CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_CHISQ(SUM_FS,SIGMA,CHISQ,CHISQ1,DIFF,ICHI,DIFF1) 
*
*
      GO TO (1,2),ICHI
1        CHISQ=SUM_FS
         CHISQ=CHISQ/SIGMA/SIGMA
         CHISQ1=CHISQ1+CHISQ
         DIFF1=DIFF1+DIFF
         GO TO 3
2        CHISQ=DIFF
         CHISQ1=CHISQ1+CHISQ
         DIFF1=DIFF1+DIFF
3     CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_MAKEMOVEP(XL,IMOVE1,IMOVE2,MOVEP,NHKL,NCON,NAO,NO)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      REAL MOVEP(2,6)
      DIMENSION XL(N_O,5,N_A),NAO(N_O)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3)
      XL(NO,1,IMOVE1)=MOVEP(1,4)
      XL(NO,2,IMOVE1)=MOVEP(1,5)
      XL(NO,3,IMOVE1)=MOVEP(1,6)
      IF(NCON.EQ.0)GO TO 10
      IM1=IMOVE1
      X=XL(NO,1,IM1)
      Y=XL(NO,2,IM1)
      Z=XL(NO,3,IM1)
      KOB=0
      IF(NO.EQ.1)GO TO 799
      DO 800 L=1,NO-1
800   KOB=KOB+NAO(L)*NAS
799   I=KOB+(IM1-1)*NAS
      DO 801 K=1,NAS
      I=I+1
      XLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3) 
801   CONTINUE
10    XL(NO,1,IMOVE2)=MOVEP(2,4)
      XL(NO,2,IMOVE2)=MOVEP(2,5)
      XL(NO,3,IMOVE2)=MOVEP(2,6)
      IF(NCON.EQ.0)GO TO 20
      IM2=IMOVE2
      X=XL(NO,1,IM2)
      Y=XL(NO,2,IM2)
      Z=XL(NO,3,IM2)
      I=KOB+(IM2-1)*NAS
      DO 803 K=1,NAS
      I=I+1
      XLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3) 
803   CONTINUE
C
C
20    DO 1000 NM=1,NHKL
      DO 121 IR=1,IRL
      ARG=TR(IR,NM)
      DO 72 J=1,3
72    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE1)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE1,IR,NM)=COS(ARG)
121   SINAR(NO,IMOVE1,IR,NM)=SIN(ARG)
      DO 131 IR=1,IRL
      ARG=TR(IR,NM)
      DO 73 J=1,3
73    ARG=ARG+H(IR,J,NM)*XL(NO,J,IMOVE2)
      ARG=6.28319*ARG
      COSAR(NO,IMOVE2,IR,NM)=COS(ARG)
131   SINAR(NO,IMOVE2,IR,NM)=SIN(ARG)
1000  CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_MAKEMOVE(XL,IMOVE,MOVE,NHKL,NCON,NAO,NO,IANIM)
      INCLUDE 'esp0.inc'
      INCLUDE 'esp1.inc'
      REAL MOVE(192,6)
      DIMENSION XL(N_O,5,N_A),NAO(N_O)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3)
      IM=IMOVE
      XL(NO,1,IM)=MOVE(1,4)
      XL(NO,2,IM)=MOVE(1,5)
      XL(NO,3,IM)=MOVE(1,6)
         DO I=1,3
           XL(NO,I,IM)=XL(NO,I,IM)-INT(XL(NO,I,IM))
           IF(XL(NO,I,IM).GT.1.)XL(NO,I,IM)=XL(NO,I,IM)-1.
           IF(XL(NO,I,IM).LT.0.)XL(NO,I,IM)=XL(NO,I,IM)+1.
         ENDDO
      IF(NCON.EQ.0)GO TO 10
      KOB=0
      IF(NO.EQ.1)GO TO 799
      DO 800 L=1,NO-1
800   KOB=KOB+NAO(L)*NAS
799   I=KOB+(IM-1)*NAS
      X=XL(NO,1,IM)
      Y=XL(NO,2,IM)
      Z=XL(NO,3,IM)
      DO 801 K=1,NAS
      I=I+1
      XLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
         DO J=1,3
           XLL(J,I)=XLL(J,I)-INT(XLL(J,I))
           IF(XLL(J,I).GT.1.)XLL(J,I)=XLL(J,I)-1.
           IF(XLL(J,I).LT.0.)XLL(J,I)=XLL(J,I)+1.
         ENDDO
801   CONTINUE
C
C
10    DO 1000 NM=1,NHKL
      DO 121 IR=1,IRL
      ARG=TR(IR,NM)
      DO 72 J=1,3
72    ARG=ARG+H(IR,J,NM)*XL(NO,J,IM)
      ARG=6.28319*ARG
      COSAR(NO,IM,IR,NM)=COS(ARG)
121   SINAR(NO,IM,IR,NM)=SIN(ARG)
1000  CONTINUE
      IF(IANIM.EQ.1)CALL ANIM(XL,NAO)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESP_MAKEMR(XL,XLN,NCON,NAO,NO,IANIM)
      INCLUDE 'esp0.inc'
      DIMENSION XL(N_O,5,N_A),XLN(N_O,3,N_A),NAO(N_O)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3)
      KOB=0
      IF(NO.EQ.1)GO TO 799
      DO 800 L=1,NO-1
800   KOB=KOB+NAO(L)*NAS
799   DO 900 IM=1,NAO(NO)
      XL(NO,1,IM)=XLN(NO,1,IM)
      XL(NO,2,IM)=XLN(NO,2,IM)
      XL(NO,3,IM)=XLN(NO,3,IM)
      IF(NCON.EQ.0)GO TO 900
      I=KOB+(IM-1)*NAS
      X=XL(NO,1,IM)
      Y=XL(NO,2,IM)
      Z=XL(NO,3,IM)
      DO 801 K=1,NAS
      I=I+1
      XLL(1,I)=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      XLL(2,I)=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      XLL(3,I)=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
         DO J=1,3
           XLL(J,I)=XLL(J,I)-INT(XLL(J,I))
           IF(XLL(J,I).GT.1.)XLL(J,I)=XLL(J,I)-1.
           IF(XLL(J,I).LT.0.)XLL(J,I)=XLL(J,I)+1.
         ENDDO
801   CONTINUE
900   CONTINUE
      IF(IANIM.EQ.1)CALL ANIM(XL,NAO)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ANIM(XLM,NAO)
      INCLUDE 'esp0.inc'
      CHARACTER*4 NXB(N_O,N_A),NXBB(6000)
      DIMENSION XLM(N_O,5,N_A),XLS(N_O,3,N_A),NAO(N_O)
      DIMENSION XXX(6000),YYY(6000),ZZZ(6000),NONO(6000)
      DIMENSION C(3,6000),KK(6000),SHIF(3,192),KSHIF(3,192)
      DIMENSION XX(6000),YY(6000),ZZ(6000),SHIFT(N_O,3)
      COMMON/XYZ/XLL(3,6000),NAS,NA,SMT(192,3,3),TT(192,3)
      COMMON/ANI/STST(3),NOB,NOBT(N_O),ISYST,TM1(6),NXB,ORI(3,3)
C
C  Shift atoms and fragments inside the cell
C
      DO 2520 NO=1,NOB
      IF(NOBT(NO).NE.2)GO TO 2702
      DO J=1,3
      SHIFT(NO,J)=INT(XLM(NO,J,1))
      DIF=XLM(NO,J,1)-SHIFT(NO,J)
      IF(DIF.GT.1.)SHIFT(NO,J)=SHIFT(NO,J)+1.
      IF(DIF.LT.0.)SHIFT(NO,J)=SHIFT(NO,J)-1.
      ENDDO
2702  DO 1500 I=1,NAO(NO)
      DO J=1,3
      IF(NOBT(NO).EQ.2) GO TO 2700   
       XLS(NO,J,I)=XLM(NO,J,I)-INT(XLM(NO,J,I))
       IF(XLS(NO,J,I).GT.1.)XLS(NO,J,I)=XLS(NO,J,I)-1.
       IF(XLS(NO,J,I).LT.0.)XLS(NO,J,I)=XLS(NO,J,I)+1.
       GO TO 2701
2700  XLS(NO,J,I)=XLM(NO,J,I)-SHIFT(NO,J)
2701  CONTINUE
      ENDDO
1500  CONTINUE
2520  CONTINUE
C
C
C
      DO 3812 NO=1,NOB
      DO 3813 I=1,NAO(NO)
      DO 3820 J=1,3
C      
C-- Tests on positions supposed to be close to special positions
C
C   Very bad test... up too now
C
      IF(NOBT(NO).NE.1)GO TO 3820
      IF(ISYST.EQ.7)GO TO 3820   ! TRICLINIC
      IF(XLS(NO,J,I).LT.STST(J))XLS(NO,J,I)=0.0
      IF(ABS(XLS(NO,J,I)-0.25).LT.STST(J))XLS(NO,J,I)=0.25
      IF(ABS(XLS(NO,J,I)-0.5).LT.STST(J))XLS(NO,J,I)=0.5
      IF(ABS(XLS(NO,J,I)-0.75).LT.STST(J))XLS(NO,J,I)=0.75
      IF(ABS(XLS(NO,J,I)-1.00).LT.STST(J))XLS(NO,J,I)=0.0
      IF(ISYST.EQ.1)GO TO 3820   !  MONOCLINIC
      IF(ABS(XLS(NO,J,I)-0.125).LT.STST(J))XLS(NO,J,I)=0.125
      IF(ABS(XLS(NO,J,I)-0.375).LT.STST(J))XLS(NO,J,I)=0.375
      IF(ABS(XLS(NO,J,I)-0.625).LT.STST(J))XLS(NO,J,I)=0.625
      IF(ABS(XLS(NO,J,I)-0.875).LT.STST(J))XLS(NO,J,I)=0.875
      IF(ISYST.EQ.2)GO TO 3820   !  ORTHORHOMBIC
      IF(ABS(XLS(NO,J,I)-0.333333).LT.STST(J))XLS(NO,J,I)=0.333333
      IF(ABS(XLS(NO,J,I)-0.666667).LT.STST(J))XLS(NO,J,I)=0.666667
      IF(ABS(XLS(NO,J,I)-0.166667).LT.STST(J))XLS(NO,J,I)=0.166667
      IF(ABS(XLS(NO,J,I)-0.833333).LT.STST(J))XLS(NO,J,I)=0.833333
      IF(ABS(XLS(NO,J,I)-0.583333).LT.STST(J))XLS(NO,J,I)=0.583333
      IF(ABS(XLS(NO,J,I)-0.083333).LT.STST(J))XLS(NO,J,I)=0.083333
      IF(ABS(XLS(NO,J,I)-0.916667).LT.STST(J))XLS(NO,J,I)=0.916667
      IF(ABS(XLS(NO,J,I)-0.416667).LT.STST(J))XLS(NO,J,I)=0.416667
3820  CONTINUE
3813  CONTINUE
3812  CONTINUE
*
*---------------------------------------------------------------------- 
*           Write out the .xyz FILE for WINMOL or XYZ2RGB or etc.
*              the file will contain all atoms in the cell,
*              generated by symmetry operators, 
*
*
      I=0
      DO 5501 NO=1,NOB
      DO 5501 L=1,NAO(NO)
      X=XLS(NO,1,L)
      Y=XLS(NO,2,L)
      Z=XLS(NO,3,L)
      IF(NOBT(NO).EQ.2) GO TO 5498   
       X=X-INT(X)
       IF(X.GT.1.)X=X-1.
       IF(X.LT.0.)X=X+1.
       Y=Y-INT(Y)
       IF(Y.GT.1.)Y=Y-1.
       IF(Y.LT.0.)Y=Y+1.
       Z=Z-INT(Z)
       IF(Z.GT.1.)Z=Z-1.
       IF(Z.LT.0.)Z=Z+1.
5498      DO 5501 K=1,NAS
      I=I+1
      XA=X*SMT(K,1,1)+Y*SMT(K,1,2)+Z*SMT(K,1,3)+TT(K,1)
      YA=X*SMT(K,2,1)+Y*SMT(K,2,2)+Z*SMT(K,2,3)+TT(K,2)
      ZA=X*SMT(K,3,1)+Y*SMT(K,3,2)+Z*SMT(K,3,3)+TT(K,3)
      IF(NOBT(NO).EQ.2) GO TO 5497   
       XA=XA-INT(XA)
       IF(XA.GT.1.)XA=XA-1.
       IF(XA.LT.0.)XA=XA+1.
       YA=YA-INT(YA)
       IF(YA.GT.1.)YA=YA-1.
       IF(YA.LT.0.)YA=YA+1.
       ZA=ZA-INT(ZA)
       IF(ZA.GT.1.)ZA=ZA-1.
       IF(ZA.LT.0.)ZA=ZA+1.
5497  XX(K)=XA
      YY(K)=YA
      ZZ(K)=ZA
      C(1,I)=XA
      C(2,I)=YA
      C(3,I)=ZA
      KK(I)=K
      NONO(I)=NOBT(NO)
      NXBB(I)=NXB(NO,L)
C
C  Test for special positions
C
      IF(NOBT(NO).EQ.2)GO TO 5510
      IF(I.EQ.1.OR.K.EQ.1)GO TO 5510
      DO 5520 JJ=1,K-1
      IF (ABS(XX(JJ) - XX(K)-1.) .GE. 0.01)GO TO 5551
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5551
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5551
      I=I-1
      GOTO 5501
5551  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5552
      IF (ABS(YY(JJ) - YY(K)-1.) .GE. 0.01)GO TO 5552
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5552
      I=I-1
      GOTO 5501
5552  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5553
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5553
      IF (ABS(ZZ(JJ) - ZZ(K)-1.) .GE. 0.01)GO TO 5553
      I=I-1
      GOTO 5501
5553  IF (ABS(XX(JJ) - XX(K)) .GE. 0.01)GO TO 5520
      IF (ABS(YY(JJ) - YY(K)) .GE. 0.01)GO TO 5520
      IF (ABS(ZZ(JJ) - ZZ(K)) .GE. 0.01)GO TO 5520
      I=I-1
      GOTO 5501
5520  CONTINUE
5510  CONTINUE
5501  CONTINUE
C
C  Put the NAS rigid fragments inside the cell
C
      JTEST=1
      DO 510 K=1,NAS
      DO 510 J=1,3
      KSHIF(J,K)=0
510   SHIF(J,K)=0.
      DO 520 JJ=1,I
      IF(NONO(JJ).NE.2)GO TO 520
      JTEST=0
      DO K=1,NAS
      DO J=1,3
      IF(KK(JJ).EQ.K)THEN
      SHIF(J,K)=SHIF(J,K)+C(J,JJ)
      KSHIF(J,K)=KSHIF(J,K)+1
      ENDIF
      ENDDO
      ENDDO
520   CONTINUE
C
      IF(JTEST.EQ.1)GO TO 1000
      DO K=1,NAS
      DO J=1,3
      IF(SHIF(J,K).LT.0.)SHIF(J,K)=SHIF(J,K)-FLOAT(KSHIF(J,K))
      SHIF(J,K)=FLOAT(INT(SHIF(J,K)/FLOAT(KSHIF(J,K))))
      ENDDO
      ENDDO
C
1000  DO 530 JJ=1,I
      DO 529 J=1,3
      IF(NONO(JJ).EQ.2) GO TO 700   
       C(J,JJ)=C(J,JJ)-INT(C(J,JJ))
       IF(C(J,JJ).GT.1.)C(J,JJ)=C(J,JJ)-1.
       IF(C(J,JJ).LT.0.)C(J,JJ)=C(J,JJ)+1.
       GO TO 701
700    DO K=1,NAS
       IF(KK(JJ).EQ.K)C(J,JJ)=C(J,JJ)-SHIF(J,K)
       ENDDO
701   CONTINUE
529   CONTINUE
530   CONTINUE
C
C
C
      WRITE(18,5503)I
      WRITE(18,5505)((ORI(II,JJ),II=1,3),JJ=1,3)
5505  FORMAT('La=',3F11.5,' Lb=',3F11.5,' Lc=',3F11.5,' t= 0ps')
      DO 5590 J=1,I
C
C  Change coordinates in Cartesian coordinates 
C
      XXX(J)=TM1(1)*(C(1,J)-0.5)+TM1(2)*(C(2,J)-0.5)+TM1(3)*(C(3,J)-0.5)
      YYY(J)=TM1(4)*(C(2,J)-0.5)+TM1(5)*(C(3,J)-0.5)
      ZZZ(J)=TM1(6)*(C(3,J)-0.5)
      WRITE(18,5504)NXBB(J),XXX(J),YYY(J),ZZZ(J)
5590  CONTINUE
5504  FORMAT(A2,2X,3F11.5)
5503  FORMAT(I4)
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE ESPO_SUMMARY(NACC,NGEN,NTRIED,NACCP,NTRIEDP,
     -DIFF,CHISQ1,NREJECT,NCON,DAMP,MR,NOB,JTEST,DIFFM,JTESTM,PROGRESS)
      INCLUDE 'esp0.inc'
	INTEGER(4) PROGRESS
	CHARACTER*120 BUFFER
      DIMENSION MR(N_O)
      DIMENSION NACC(N_O),NGEN(N_O),NTRIED(N_O),NACCP(N_O)
      DIMENSION NTRIEDP(N_O),NREJECT(N_O),DAMP(N_O)
	WRITE (BUFFER,*)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,*)
      DO 30 NO=1,NOB
      WRITE(BUFFER,'(A,I3,A,I3,A,F6.3,A,I3)')'  Object number',NO,' at t
     -est',JTEST,'.       Previous minimum R=',DIFFM,' at test',JTESTM
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE(20,2000)NO,JTEST,DIFFM,JTESTM
2000  FORMAT('  Object number',I3,' at test',I3,'.  Previous 
     -minimum R=',F5.3,' at test',I3)
      IF(NTRIEDP(NO).LE.1)NTRIEDP(NO)=0
      NACC(NO)=NACC(NO)-NACCP(NO)
      IF(NCON.EQ.1.AND.MR(NO).EQ.0)GO TO 10
      IF(MR(NO).NE.0)GO TO 20
	   WRITE (BUFFER,'(2(i7,a),g9.3,a,F5.3)') NACC(NO),' moves acc. ',
     -   NTRIED(NO),' tested; Chi**2=',CHISQ1,', R=',DIFF
	   CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(2(i7,a))') NACCP(NO),' perm. acc. ',NTRIEDP(NO),
     1' tested'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
         WRITE (20,'(2(i7,a),g9.3,a,F5.3)') NACC(NO),' moves acc. ',
     -    NTRIED(NO),' tested; Chi**2=',CHISQ1,', R=',DIFF
      WRITE (20,'(2(i7,a))') NACCP(NO),' perm. acc. ',NTRIEDP(NO),
     1' tested'
      WRITE(20,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
      NACC(NO)=NACC(NO)+NACCP(NO)
      GO TO 30
10    CONTINUE
	WRITE (BUFFER,'(3(i7,a),g9.3,a,F5.3)') NACC(NO),' moves acc. ',
     -    NGEN(NO),' gen. and ',NTRIED(NO),
     - ' tested; Chi**2=',CHISQ1,', R=',DIFF
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(2(i7,a))') NACCP(NO),' perm. acc. ',NTRIEDP(NO),
     1' tested'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,'(3(i7,a),g9.3,a,F5.3)') NACC(NO),' moves acc. ',
     -    NGEN(NO),' gen. and ',NTRIED(NO),
     - ' tested; Chi**2=',CHISQ1,', R=',DIFF
      WRITE (20,'(2(i7,a))') NACCP(NO),' perm. acc. ',NTRIEDP(NO),
     1' tested'
      WRITE(20,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
      NACC(NO)=NACC(NO)+NACCP(NO)
      GO TO 30
20    CONTINUE
	   WRITE (BUFFER,'(3(i7,a),g9.3,a,F5.3)') NACC(NO),' rot. acc. ',
     -    NGEN(NO),' gen. and ',NTRIED(NO),
     - ' tested; Chi**2=',CHISQ1,', R=',DIFF
	   CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(2(i7,a))') NACCP(NO),' trans. acc. ',NTRIEDP(NO),
     1' tested'
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
	WRITE (BUFFER,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
	CALL PROGRESSVIEW (PROGRESS, BUFFER)
      WRITE (20,'(3(i7,a),g9.3,a,F5.3)') NACC(NO),' rot. acc. ',
     -    NGEN(NO),' gen. and ',NTRIED(NO),
     - ' tested; Chi**2=',CHISQ1,', R=',DIFF
      WRITE (20,'(2(i7,a))') NACCP(NO),' trans. acc. ',NTRIEDP(NO),
     1' tested'
      WRITE(20,'(i7,a,F9.6)')NREJECT(NO),' events did not improve
     - the fit, DAMP =',DAMP(NO)
      NACC(NO)=NACC(NO)+NACCP(NO)
30    CONTINUE
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE DCELL(CELLN,AL,RCELLN,V)
      DIMENSION CELLN(6),AL(3,3),RCELLN(6),CELL(6)
      DEGRAD=3.14159/180.0
      DO 30 I=1,6
30    CELL(I)=CELLN(I)
      DO 31 I=1,3
      L=I+3
      IF(CELL(L)-90.0)32,33,32
33    CELL(L)=0.0
      GOTO 31
32    CELL(L)=COS(DEGRAD*CELL(L))
31    CONTINUE
      CALL TRCL(CELL,RCELLN,V)
C     RCELL IS THE RECIPROCAL CELL CONSTANTS
      DO 34 I=1,3
      AL(I,I)=RCELLN(I)*RCELLN(I)
      CALL PERM(I,J,K)
      IF(J-K)35,36,36
35    AL(J,K)=2.0*RCELLN(J)*RCELLN(K)*RCELLN(I+3)
      GOTO 34
36    AL(K,J)=2.0*RCELLN(J)*RCELLN(K)*RCELLN(I+3) 
34    CONTINUE
      DO 37 I=4,6
37    RCELLN(I)=ACOS(RCELLN(I))/DEGRAD
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE DCELL2(CELLN,AL,RCELLN,V,NO)
      INCLUDE 'esp0.inc'
      DIMENSION CELLN(N_O,6),AL(N_O,3,3),RCELLN(N_O,6),CELL(N_O,6)
      DIMENSION V(N_O)
      DEGRAD=3.14159/180.0
      DO 30 I=1,6
30    CELL(NO,I)=CELLN(NO,I)
      DO 31 I=1,3
      L=I+3
      IF(CELL(NO,L)-90.0)32,33,32
33    CELL(NO,L)=0.0
      GOTO 31
32    CELL(NO,L)=COS(DEGRAD*CELL(NO,L)) 
31    CONTINUE
      CALL TRCL2(CELL,RCELLN,V,NO)
C     RCELL IS THE RECIPROCAL CELL CONSTANTS
      DO 34 I=1,3
      AL(NO,I,I)=RCELLN(NO,I)*RCELLN(NO,I)
      CALL PERM(I,J,K)
      IF(J-K)35,36,36
35    AL(NO,J,K)=2.0*RCELLN(NO,J)*RCELLN(NO,K)*RCELLN(NO,I+3)
      GOTO 34
36    AL(NO,K,J)=2.0*RCELLN(NO,J)*RCELLN(NO,K)*RCELLN(NO,I+3) 
34    CONTINUE
      DO 37 I=4,6
37    RCELLN(NO,I)=ACOS(RCELLN(NO,I))/DEGRAD
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE TRCL(CELLN,RCELLN,V)
C     TRANSFORMS REAL CELL TO RECIPROCAL OR VICE VERSA 
C     INPUT CELL IS IN ARRAY CELL AS LENGTHS AND COSINES
      DIMENSION CELLN(6),RCELLN(6)
      DIMENSION SINA(3)
      ABC=1.0
      PROD=2.0
      V=-2.0
      DO 10 I=1,3
      L=I+3
      SINA(I)=1.0-CELLN(L)**2
      V=V+SINA(I)
      SINA(I)=SQRT(SINA(I))
      PROD=PROD*CELLN(L)
10    ABC=ABC*CELLN(I)
      V=ABC*SQRT(V+PROD)
C      V IS CELL VOLUME
C     PUT INVERTED CELL INTO RCELL
      DO 20 I=1,3
      CALL PERM(I,J,K)
      RCELLN(I)=CELLN(J)*CELLN(K)*SINA(I)/V
      L=I+3
20    RCELLN(L)=(CELLN(J+3)*CELLN(K+3)-CELLN(L))/(SINA(J)
     2 *SINA(K))
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE TRCL2(CELLN,RCELLN,V,NO) 
C     TRANSFORMS REAL CELL TO RECIPROCAL OR VICE VERSA 
C     INPUT CELL IS IN ARRAY CELL AS LENGTHS AND COSINES
      INCLUDE 'esp0.inc'
      DIMENSION CELLN(N_O,6),RCELLN(N_O,6),V(N_O)
        DIMENSION SINA(3)
      ABC=1.0
      PROD=2.0
      V(NO)=-2.0
      DO 10 I=1,3
      L=I+3
      SINA(I)=1.0-CELLN(NO,L)**2
      V(NO)=V(NO)+SINA(I)
      SINA(I)=SQRT(SINA(I))
      PROD=PROD*CELLN(NO,L)
10    ABC=ABC*CELLN(NO,I)
      V(NO)=ABC*SQRT(V(NO)+PROD)
C      V IS CELL VOLUME
C     PUT INVERTED CELL INTO RCELL
      DO 20 I=1,3
      CALL PERM(I,J,K)
      RCELLN(NO,I)=CELLN(NO,J)*CELLN(NO,K)*SINA(I)/V(NO)
      L=I+3
20    RCELLN(NO,L)=(CELLN(NO,J+3)*CELLN(NO,K+3)-CELLN(NO,L))/(SINA(J)
     2 *SINA(K))
      RETURN
      END
*
*----------------------------------------------------------------------- 
*
      SUBROUTINE PERM(I,J,K)
C     PERMS USEFUL COMBINATIONS OF INTEGERS IN THE RANGE 1 TO 3
      IF(I-2)10,20,30
10    J=2
      K=3
      RETURN
20    J=3
      K=1
      RETURN
30    J=1
      K=2
      RETURN
      END
*
********************************************************* 
*
      SUBROUTINE MATRIX(CELLN,RCELLN,TM2,TM1)
      DIMENSION CELLN(6),RCELLN(6),TM1(6),TM2(6)
      RAD=57.29578    ! 180./Pi
C
C *** CALCULATE ORTHOGONAL MATRIX ***
C
      SING=SIN(CELLN(6)/RAD)
      COSG=COS(CELLN(6)/RAD)
      TM1(1)=1./CELLN(1)
      TM1(2)=-COSG/(CELLN(1)*SING)
      TM1(3)=RCELLN(1)*COS(RCELLN(5)/RAD)
      TM1(4)=1./(CELLN(2)*SING)
      TM1(5)=RCELLN(2)*COS(RCELLN(4)/RAD)
      TM1(6)=RCELLN(3)
C
C *** CALCULATE INVERSE MATRIX ***
C
      TM2(1)=CELLN(1)
      TM2(2)=CELLN(2)*COSG
      TM2(3)=CELLN(3)*COS(CELLN(5)/RAD)
      TM2(4)=CELLN(2)*SING
      TM2(5)=-CELLN(3)*SIN(CELLN(5)/RAD)*COS(RCELLN(4)/RAD)
      TM2(6)=1/TM1(6)
        RETURN
      END
*
********************************************************* 
*
      SUBROUTINE MATRIX2(CELLN,RCELLN,TM2,TM1,NO)
      INCLUDE 'esp0.inc'
      DIMENSION CELLN(N_O,6),RCELLN(N_O,6),TM1(N_O,6),TM2(N_O,6)
      RAD=57.29578    ! 180./Pi
C
C *** CALCULATE ORTHOGONAL MATRIX ***
C
      SING=SIN(CELLN(NO,6)/RAD)
      COSG=COS(CELLN(NO,6)/RAD)
      TM1(NO,1)=1./CELLN(NO,1)
      TM1(NO,2)=-COSG/(CELLN(NO,1)*SING)
      TM1(NO,3)=RCELLN(NO,1)*COS(RCELLN(NO,5)/RAD)
      TM1(NO,4)=1./(CELLN(NO,2)*SING)
      TM1(NO,5)=RCELLN(NO,2)*COS(RCELLN(NO,4)/RAD)
      TM1(NO,6)=RCELLN(NO,3)
C
C *** CALCULATE INVERSE MATRIX ***
C
      TM2(NO,1)=CELLN(NO,1)
      TM2(NO,2)=CELLN(NO,2)*COSG
      TM2(NO,3)=CELLN(NO,3)*COS(CELLN(NO,5)/RAD)
      TM2(NO,4)=CELLN(NO,2)*SING
      TM2(NO,5)=-CELLN(NO,3)*SIN(CELLN(NO,5)/RAD)*COS(RCELLN(NO,4)/RAD)
      TM2(NO,6)=1/TM1(NO,6)
        RETURN
      END
*
********************************************************* *
      SUBROUTINE TRANSLA(XL,XLN,NAO,ISEED,DELTAO,A,BB,C,DAMP,NO)
      INCLUDE 'esp0.inc'
      DIMENSION XL(N_O,3,N_A),XLN(N_O,3,N_A),NAO(N_O),DAMP(N_O)
      REAL DELTAO(*)
      AMOVE1=2.*(RAN(ISEED)-0.5)*DELTAO(NO)/A*DAMP(NO)
      AMOVE2=2.*(RAN(ISEED)-0.5)*DELTAO(NO)/BB*DAMP(NO)
      AMOVE3=2.*(RAN(ISEED)-0.5)*DELTAO(NO)/C*DAMP(NO)
      DO 1 I=1,NAO(NO)
      XLN(NO,1,I)=XL(NO,1,I)+AMOVE1
      XLN(NO,2,I)=XL(NO,2,I)+AMOVE2
1     XLN(NO,3,I)=XL(NO,3,I)+AMOVE3
      RETURN
      END
*
************************************************************************ 
*
      SUBROUTINE ROTAT(XL,XLN,NAO,ISEED,TM1,TM2,S,DAMP,NGRAV,XG,NO,PI2)
      INCLUDE 'esp0.inc'
      DIMENSION XL(N_O,3,N_A),XLN(N_O,3,N_A),TM1(6),TM2(6),XG(N_O,3)
      DIMENSION NAO(N_O),S(N_O),NGRAV(N_O),DAMP(N_O),PI2(N_O)
C      PI=3.1415926
C      PI2=6.2831853 or less
C
C  Find current gravity center
C
      IF(NGRAV(NO).EQ.1) GO TO 101
      X=0.
      Y=0.
      Z=0.
      DO 100 I=1,NAO(NO)
      X=X+XL(NO,1,I)
      Y=Y+XL(NO,2,I)
100   Z=Z+XL(NO,3,I)
      XG(NO,1)=X/S(NO)
      XG(NO,2)=Y/S(NO)
      XG(NO,3)=Z/S(NO)
C
C  Change coordinates of the rigid model in Cartesian coordinates 
C
101   DO 1 I=1,NAO(NO)
      XA=XL(NO,1,I)-XG(NO,1)
      YA=XL(NO,2,I)-XG(NO,2)
      ZA=XL(NO,3,I)-XG(NO,3)
      XLN(NO,1,I)=TM1(1)*XA+TM1(2)*YA+TM1(3)*ZA
      XLN(NO,2,I)=TM1(4)*YA+TM1(5)*ZA
      XLN(NO,3,I)=TM1(6)*ZA
1     CONTINUE
C
C  Find 3 Euler angles at random
C
      PIDAMP=PI2(NO)*DAMP(NO)
C      T1=RAN(ISEED)*PI*DAMP(NO)
      T1=RAN(ISEED)*PIDAMP
      T2=RAN(ISEED)*PIDAMP
      T3=RAN(ISEED)*PIDAMP
      CT1=COS(T1)
      ST1=SIN(T1)
      CT2=COS(T2)
      ST2=SIN(T2)
      CT3=COS(T3)
      ST3=SIN(T3)
      R1=CT1*CT3-ST1*CT2*ST3
      R2=ST1*CT3+CT1*CT2*ST3
      R3=ST2*ST3
      R4=-ST1*CT2*CT3-CT1*ST3
      R5=-ST1*ST3+CT1*CT2*CT3
      R6=ST2*CT3
      R7=ST1*ST2
      R8=-CT1*ST2
      R9=CT2
      DO 20 I=1,NAO(NO)
C
C  Apply the rotation matrix
C
      XX=R1*XLN(NO,1,I)+R2*XLN(NO,2,I)+R3*XLN(NO,3,I)
      YY=R4*XLN(NO,1,I)+R5*XLN(NO,2,I)+R6*XLN(NO,3,I)
      ZZ=R7*XLN(NO,1,I)+R8*XLN(NO,2,I)+R9*XLN(NO,3,I)
      XLN(NO,1,I)=XX
      XLN(NO,2,I)=YY
      XLN(NO,3,I)=ZZ
C
C  Refurbish the coordinates in the unknown structure cell 
C
      XLN(NO,1,I)=TM2(1)*XLN(NO,1,I)+TM2(2)*XLN(NO,2,I)+
     1TM2(3)*XLN(NO,3,I)+XG(NO,1)
      XLN(NO,2,I)=TM2(4)*XLN(NO,2,I)+TM2(5)*XLN(NO,3,I)+XG(NO,2)
      XLN(NO,3,I)=TM2(6)*XLN(NO,3,I)+XG(NO,3) 
20    CONTINUE
C
C  We have now the NEW rotated rigid model inside the unknown 
C
      RETURN
      END
*
************************************************************************ 
*
      SUBROUTINE ROTORS(XL,XLN,NAO,ISEED,TM1,TM2,DAMP,NBRO,NO,PI2,PREV)
C
C  Applies rotation around the rotatable bonds
C
      INCLUDE 'esp0.inc'
      DIMENSION XL(N_O,3,N_A),XLN(N_O,3,N_A),TM1(6),TM2(6)
      DIMENSION NAO(N_O),DAMP(N_O),PI2(N_O),PREV(N_O,50)
      COMMON/TORS/NBROT(N_O),NA1(N_O,500),NA2(N_O,500),NTOR(N_O,50,N_A)
C      PI=3.1415926
C      PI2=6.2831853 or less
C
C  NBRO=torsion angle to be rotated : NTOR(NO,NBRO,I) I=1,NAO(NO)
C
C  Change coordinates of the rotatable part in Cartesian coordinates 
C
      DO 1 I=1,NAO(NO)
      IF(NTOR(NO,NBRO,I).NE.1)GO TO 1
      XA=XL(NO,1,I)
      YA=XL(NO,2,I)
      ZA=XL(NO,3,I)
      XLN(NO,1,I)=TM1(1)*XA+TM1(2)*YA+TM1(3)*ZA
      XLN(NO,2,I)=TM1(4)*YA+TM1(5)*ZA
      XLN(NO,3,I)=TM1(6)*ZA
1     CONTINUE
C
C  Change coordinates of the rotatable bond A1-A2 in Cartesian coordinates
C
      I=NA1(NO,NBRO)
      XA1=XL(NO,1,I)
      YA1=XL(NO,2,I)
      ZA1=XL(NO,3,I)
      XAC1=TM1(1)*XA1+TM1(2)*YA1+TM1(3)*ZA1
      YAC1=TM1(4)*YA1+TM1(5)*ZA1
      ZAC1=TM1(6)*ZA1
      I=NA2(NO,NBRO)
      XA2=XL(NO,1,I)
      YA2=XL(NO,2,I)
      ZA2=XL(NO,3,I)
      XAC2=TM1(1)*XA2+TM1(2)*YA2+TM1(3)*ZA2
      YAC2=TM1(4)*YA2+TM1(5)*ZA2
      ZAC2=TM1(6)*ZA2
C
C  Use A1 as the origin
C
      XAC2=XAC2-XAC1
      YAC2=YAC2-YAC1
      ZAC2=ZAC2-ZAC1
      DO 2 I=1,NAO(NO)
      IF(NTOR(NO,NBRO,I).NE.1)GO TO 2
      XLN(NO,1,I)=XLN(NO,1,I)-XAC1
      XLN(NO,2,I)=XLN(NO,2,I)-YAC1
      XLN(NO,3,I)=XLN(NO,3,I)-ZAC1
2     CONTINUE
C
C  Find the rotation angle at random
C       and go back to the original position with PREV
      PIDAMP=PI2(NO)*DAMP(NO)
      T1=(RAN(ISEED)-0.5)*PIDAMP-PREV(NO,NBRO)
      PREV(NO,NBRO)=T1
C
C  Find the direction cosines of the A1A2 rotation axis
C
      DIV=SQRT(XAC2**2+YAC2**2+ZAC2**2)
      DC1=XAC2/DIV
      DC2=YAC2/DIV
      DC3=ZAC2/DIV
C
C  Calculate matrix elements
C
      CT1=COS(T1)
      ST1=SIN(T1)
      CM=1.-CT1
      DC12=CM*DC1**2
      DC22=CM*DC2**2
      DC32=CM*DC3**2
      DC1DC2=CM*DC1*DC2
      DC1DC3=CM*DC1*DC3
      DC2DC3=CM*DC2*DC3
      R1=CT1+DC12
      R2=DC3*ST1+DC1DC2
      R3=-DC2*ST1+DC1DC3
      R4=-DC3*ST1+DC1DC2
      R5=CT1+DC22
      R6=DC1*ST1+DC2DC3
      R7=DC2*ST1+DC1DC3
      R8=-DC1*ST1+DC2DC3
      R9=CT1+DC32
      DO 20 I=1,NAO(NO)
C
C  Apply the rotation matrix
C
      IF(NTOR(NO,NBRO,I).NE.1)GO TO 19
      XX=R1*XLN(NO,1,I)+R2*XLN(NO,2,I)+R3*XLN(NO,3,I)
      YY=R4*XLN(NO,1,I)+R5*XLN(NO,2,I)+R6*XLN(NO,3,I)
      ZZ=R7*XLN(NO,1,I)+R8*XLN(NO,2,I)+R9*XLN(NO,3,I)
C
C  Reuse the old origin
C
      XLN(NO,1,I)=XX+XAC1
      XLN(NO,2,I)=YY+YAC1
      XLN(NO,3,I)=ZZ+ZAC1
C
C  Refurbish the coordinates in the unknown structure cell 
C
      XLN(NO,1,I)=TM2(1)*XLN(NO,1,I)+TM2(2)*XLN(NO,2,I)+
     1TM2(3)*XLN(NO,3,I)
      XLN(NO,2,I)=TM2(4)*XLN(NO,2,I)+TM2(5)*XLN(NO,3,I)
      XLN(NO,3,I)=TM2(6)*XLN(NO,3,I)
      GO TO 20
19    XLN(NO,1,I)=XL(NO,1,I)
      XLN(NO,2,I)=XL(NO,2,I)
      XLN(NO,3,I)=XL(NO,3,I)
20    CONTINUE
C
C  We have now the NEW rotated smallest part of the rotatable 
C                       bond inside the unknown 
C
      RETURN
      END
*
********************************************************* 
*
      SUBROUTINE BURZ (ISTRUV,ISYSTM,ISYMCE,IBRAVL,NG,SS,TS,NT) 
C-----ROUTINE TO INTERPRET HERMANN MAUGUIN SYMBOL FOR SPACE GROUP 
C-----THIS ROUTINE HAS BEEN ADAPTED FROM A PROGRAM SUPPLIED BY 
C-----PROF. BURZLAFF, UNIVERSITY OF ERLANGEN, GERMANY
      INTEGER E(3,3),ZEI(3,4),SS(24,3,3),SYS,N3(20)
      DIMENSION NT(76),IBRA(7),NMA(3),SH(3),ISYS(7)
      DIMENSION TS(24,3),TE(3)
      DATA ISYS/7,1,2,4,6,5,3/
      DATA IBRA/'P','A','B','C','F','I','R'/,NMA/3*0/,SH/3*0.25/
      DATA N3/'1','2','3','4',' ','6','A','B','C','D','M','N','P',
     1        ' ','/','-',4*' '/
 1    NBL=-1
      NG=1
      NS=1
      DO 2 I=1,24
      DO 2 J=1,3
      TS(I,J)=0.0
      DO 2 K=1,3
      E(J,K)=0
      SS(I,J,K)=0
 2    CONTINUE
      SS(1,1,1)=1
      SS(1,2,2)=1
      SS(1,3,3)=1
      DO 3 I=1,3
      TE(I)=0.0
      DO 3 J=1,4
 3    ZEI(I,J)=0
      DO 8 I=1,36
      KXK=NT(I)
CSTRUVIR      IF (KXK.NE.N3(17)) GO TO 4
      IF (KXK.NE.N3(5)) GO TO 4
      NLQ=0
      IC=0
      GO TO 8
 4    IF (NBL.GE.0) GO TO 7
      DO 5 J=1,7
      NBR=J
      IF (KXK.EQ.IBRA(J)) GO TO 6
 5    CONTINUE
      NBR=8
 6    NBL=0
      IF(KXK.EQ.IBRA(1))IBRAVL=1
      IF(KXK.EQ.IBRA(6))IBRAVL=2
      IF(KXK.EQ.IBRA(5))IBRAVL=3
      IF(KXK.EQ.IBRA(7))IBRAVL=4
      IF(KXK.EQ.IBRA(2))IBRAVL=5
      IF(KXK.EQ.IBRA(3))IBRAVL=6
      IF(KXK.EQ.IBRA(4))IBRAVL=7
      IF(ISTRUV.EQ.1)IBRAVL=NBR
      GO TO 8
 7    IF (NLQ.EQ.0) NBL=NBL+1
      IC=IC+1
      ZEI(NBL,IC)=KXK
      NLQ=1
      IF (KXK.EQ.N3(15)) NS=0
 8    CONTINUE
      DO 9 I=1,4
      IF (ZEI(2,I).NE.N3(3)) GO TO 9
      SYS=6
      GO TO 15
 9    CONTINUE
      DO 11 I=1,4
      IF (ZEI(1,I).NE.N3(3).AND.ZEI(1,I).NE.N3(6)) GO TO 10
      SYS=5
      GO TO 15
 10   IF (ZEI(1,I).NE.N3(4)) GO TO 11
      SYS=4
      GO TO 15
 11   CONTINUE
      IF (NBL.GT.1) GO TO 14
      IF (ZEI(1,1).NE.N3(1).AND.ZEI(1,1).NE.N3(16)) GO TO 12
      SYS=1
      GO TO 15
 12   SYS=2
      DO 13 I=1,4
 13   ZEI(2,I)=ZEI(1,I)
      ZEI(1,1)=N3(1)
      ZEI(3,1)=N3(1)
 14   SYS=3
      IF (ZEI(1,1).EQ.N3(1).OR.ZEI(2,1).EQ.N3(1)) SYS=2  
 15   GO TO (16,17,22,46,58,68), SYS
C-----TRICLINIC
 16   IF (ZEI(1,1).EQ.N3(16)) NS=0
      GO TO 76
C-----MONOCLINIC
 17   NG=2
      DO 18 I=1,3
 18   IF (ZEI(I,1).NE.N3(1)) IND=I
      ID=1
      IF (ZEI(IND,1).EQ.N3(2)) ID=-1
      DO 19 I=1,3
 19   SS(2,I,I)=SS(1,I,I)*ID
      SS(2,IND,IND)=-SS(2,IND,IND)
      DO 21 I=1,3
      IF (ZEI(I,1).EQ.N3(2).AND.ZEI(I,2).EQ.N3(1)) TS(2,I)=0.5
      DO 21 J=1,4
      IF (ZEI(I,J).EQ.N3(7)) TS(2,1)=0.5
      IF (ZEI(I,J).EQ.N3(8)) TS(2,2)=0.5
      IF (ZEI(I,J).EQ.N3(9)) TS(2,3)=0.5
      IF (ZEI(I,J).EQ.N3(12)) GO TO 20
      GO TO 21
 20   K=I+1
      IF (K.GT.3) K=K-3
      TS(2,K)=0.5
      TS(2,6-K-I)=0.5
 21   CONTINUE
      GO TO 76
C-----ORTHORHOMBIC
 22   NG=4
      IC=0
      IND=1
      IF (ZEI(1,1).NE.N3(2).AND.ZEI(2,1).NE.N3(2).AND.ZEI(3,1).NE.N3(2))
     1IND=-1
      IF (IND.EQ.-1) NS=0
      DO 24 I=1,3
      ID=1
      IF (ZEI(I,1).EQ.N3(2)) ID=-1
      DO 23 J=1,3
 23   SS(1+I,J,J)=SS(1,J,J)*ID*IND
 24   SS(1+I,I,I)=-SS(1+I,I,I)
      DO 28 I=1,3
      IF (ZEI(I,1).EQ.N3(2).AND.ZEI(I,2).EQ.N3(1)) TS(1+I,I)=0.5
      DO 28 J=1,4
      IF (ZEI(I,J).EQ.N3(7)) TS(1+I,1)=0.5
      IF (ZEI(I,J).EQ.N3(8)) TS(1+I,2)=0.5
      IF (ZEI(I,J).EQ.N3(9)) TS(1+I,3)=0.5
      IF (ZEI(I,J).EQ.N3(12).OR.ZEI(I,J).EQ.N3(10)) GO TO 25
      GO TO 28
 25   K=I+1
      IF (K.GT.3) K=K-3
      IF (ZEI(I,J).EQ.N3(10)) GO TO 26
      TS(1+I,K)=0.5
      TS(1+I,6-K-I)=0.5
      GO TO 28
 26   IC=1
      IF (NS.EQ.1) GO TO 27
      TS(1+I,K)=0.25
      TS(1+I,6-K-I)=0.25
      GO TO 28
 27   TS(1+I,1)=0.25
      TS(1+I,2)=0.25
      TS(1+I,3)=0.25
 28   CONTINUE
      IF (IC.EQ.1) GO TO 76
      IF (NS.EQ.1) GO TO 36
      DO 29 I=1,3
      K=1+I
      IF (K.GT.3) K=K-3
      TC=TS(1+K,I)+TS(1+6-I-K,I)
      IF (TC.EQ.1.0) TC=0.0
 29   TS(1+I,I)=TC
C-----SPECIAL TREATMENT OF C M M A, C M C A, I M M A
      IF (NBR.EQ.1.OR.NBR.EQ.5) GO TO 76
      MA=0
      DO 31 I=1,3
      DO 30 J=1,4
 30   IF (ZEI(I,J).EQ.N3(11)) NMA(I)=1
 31   MA=MA+NMA(I)
      IF (NBR.EQ.6.AND.MA.EQ.2) GO TO 33
      IF (MA.EQ.0.OR.MA.EQ.3.OR.NBR.EQ.6) GO TO 76
      DO 32 I=1,3
      IF (NMA(NBR-1).EQ.1) GO TO 76
 32   SH(NBR-1)=0.
C-----ORIGIN SHIFT
 33   DO 35 I=1,NG
      DO 35 J=1,3
      DO 34 K=1,3
      ID=1
      IF (J.NE.K) ID=0
 34   TS(I,J)=TS(I,J)+(ID-SS(I,J,K))*SH(K)
      IF (TS(I,J).GT.1.) TS(I,J)=TS(I,J)-1.
 35   IF (TS(I,J).LT.1.0) TS(I,J)=TS(I,J)+1.
      GO TO 76
 36   IC=0
      DO 37 I=1,3
 37   IF (SS(1+I,1,1)*SS(1+I,2,2)*SS(1+I,3,3).EQ.-1) IC=1
      IF (IC.EQ.1) GO TO 41
      TC=TS(2,1)+TS(3,2)+TS(4,3)
      IF (TC.EQ.0.0) GO TO 76
      DO 40 I=1,3
      K=I+1
      IF (K.GT.3) K=K-3
      IF (TC.GT.0.5) GO TO 38
      IF (TS(1+I,I).EQ.0.0) GO TO 40
      M=I-1
      IF (M.EQ.0) M=M+3
      TS(1+M,I)=0.5
      GO TO 40
 38   IF (TC.GT.1.0) GO TO 39
      IF (TS(1+I,I).NE.0.0) GO TO 40
      L=K+1
      IF (L.GT.3) L=L-3
      TS(1+K,L)=0.5
      TS(1+L,K)=0.5
      GO TO 40
 39   TS(1+I,K)=0.5
 40   CONTINUE
      GO TO 76
 41   DO 42 I=1,3
 42   IF (SS(1+I,1,1)*SS(1+I,2,2)*SS(1+I,3,3).EQ.1) ID=I
      DO 45 I=1,3
      TC=TS(2,I)+TS(3,I)+TS(4,I)
      IF (TC.EQ.0.0.OR.TC.EQ.1.0) GO TO 45
      IF (ZEI(1,1).EQ.N3(11).AND.ZEI(2,1).EQ.N3(12).OR.ZEI(2,1).EQ.
     1N3(11).AND.ZEI(3,1).EQ.N3(12).OR.ZEI(3,1).EQ.N3(11).AND.ZEI(1,1)
     2.EQ.N3(12)) GO TO 44
      DO 43 J=1,3
      IF (ID.EQ.J) GO TO 43
      IF (TS(1+J,I).EQ.0.5) GO TO 43
      TS(1+J,I)=0.5
 43   CONTINUE
      GO TO 45
 44   K=I-1
      IF (K.EQ.0) K=K+3
      TS(1+K,I)=0.5
 45   CONTINUE
      GO TO 76
C-----TETRAGONAL
 46   NG=4
      IF (NBL.EQ.3) NG=8
      SS(2,1,2)=-1
      SS(2,2,1)=1
      SS(2,3,3)=1
      M=ZEI(1,1)
      N=ZEI(1,2)
      DO 47 I=1,3
      DO 47 J=1,3
 47   IF (M.EQ.N3(16)) SS(2,I,J)=-SS(2,I,J)
      IF (M.EQ.N3(16)) GO TO 50
      IF (N.EQ.N3(1)) TS(2,3)=0.25
      IF (N.EQ.N3(2)) TS(2,3)=0.5
      IF (N.EQ.N3(3)) TS(2,3)=0.75
      IF (ZEI(1,3).EQ.N3(12).OR.(ZEI(1,4).EQ.N3(12).AND.NBL.EQ.3))
     1TS(2,1)=0.5
      IF ((ZEI(1,4).EQ.N3(12).AND.NBL.EQ.1).OR.(N.EQ.N3(1).AND.NS.EQ.1.
     1AND.NBR.EQ.6)) TS(2,2)=0.5
      IF (N.EQ.N3(1).AND.NS.EQ.0.AND.NBR.EQ.6) GO TO 48
      IF (ZEI(2,2).EQ.N3(1).OR.(ZEI(1,4).NE.N3(12).AND.ZEI(2,1).EQ.
     1N3(12).AND.ZEI(3,1).EQ.N3(11))) GO TO 49
      GO TO 50
 48   TS(2,1)=0.25
      TS(2,2)=0.75
      IF (NBL.EQ.1) TS(2,1)=0.75
      IF (NBL.EQ.1) TS(2,2)=0.25
      GO TO 50
 49   TS(2,1)=0.5
      TS(2,2)=0.5
 50   SS(3,1,1)=-1
      SS(3,2,2)=-1
      SS(3,3,3)=1
      TS(3,1)=SS(2,1,2)*TS(2,2)+TS(2,1)
      TS(3,2)=SS(2,2,1)*TS(2,1)+TS(2,2)
      TS(3,3)=SS(2,3,3)*TS(2,3)+TS(2,3)
      DO 51 I=1,3
 51   IF (NBR.EQ.6.AND.TS(3,1).EQ.0.5.AND.TS(3,2).EQ.0.5.AND.TS(3,3).EQ.
     10.5) TS(3,I)=0.0
      DO 52 I=1,3
      TS(4,I)=TS(2,I)
      DO 52 J=1,3
      TS(4,I)=TS(4,I)+SS(2,I,J)*TS(3,J)
      DO 52 K=1,3
 52   SS(4,I,J)=SS(4,I,J)+SS(2,I,K)*SS(3,K,J)
      IF (NBL.EQ.1) GO TO 76
      IF (NS.EQ.0) GO TO 56
      M=ZEI(2,1)
      N=ZEI(3,1)
      IF (M.NE.N3(2).AND.N.NE.N3(2)) GO TO 55
      IF (M.EQ.N3(2).AND.N.EQ.N3(2)) GO TO 54
      IF (M.EQ.N3(2)) GO TO 53
      IF (M.EQ.N3(9).OR.M.EQ.N3(12)) TE(3)=0.5
      E(1,1)=-1
      E(2,2)=1
      E(3,3)=1
      IF (M.NE.N3(12).AND.M.NE.N3(8)) GO TO 57
      TE(1)=0.5
      TE(2)=0.5
      GO TO 57
 53   E(1,1)=1
      E(2,2)=-1
      E(3,3)=-1
      IF (N.EQ.N3(9)) TE(3)=0.5
      IF (N.EQ.N3(10)) TE(3)=0.25
      IF (N.EQ.N3(10)) TE(2)=0.5
      IF (ZEI(2,2).NE.N3(1)) GO TO 57
      TE(1)=0.5
      TE(2)=0.5
      GO TO 57
 54   E(1,2)=1
      E(2,1)=1
      E(3,3)=-1
      IF (ZEI(2,2).NE.0.OR.NBR.EQ.6.OR.ZEI(1,2).EQ.0) GO TO 57
      IF (ZEI(1,2).EQ.N3(1)) TE(3)=0.75
      IF (ZEI(1,2).EQ.N3(2)) TE(3)=0.5
      IF (ZEI(1,2).EQ.N3(3)) TE(3)=0.25
      GO TO 57
 55   M=ZEI(2,1)
      E(1,1)=-1
      E(2,2)=1
      E(3,3)=1
      IF (M.EQ.N3(9).OR.M.EQ.N3(12)) TE(3)=0.5
      IF (M.EQ.N3(12).OR.M.EQ.N3(8)) TE(1)=0.5
      IF (M.EQ.N3(12).OR.M.EQ.N3(8)) TE(2)=0.5
      GO TO 57
 56   E(1,1)=-1
      E(2,2)=1
      E(3,3)=1
      IF (ZEI(2,1).EQ.N3(9).OR.ZEI(2,1).EQ.N3(12)) TE(3)=0.5
      IF (ZEI(2,1).EQ.N3(8).OR.ZEI(2,1).EQ.N3(12)) TE(2)=0.5
      IF (ZEI(2,1).EQ.N3(8).OR.ZEI(2,1).EQ.N3(12)) TE(1)=0.5
      IF (ZEI(1,3).EQ.N3(12).OR.ZEI(1,4).EQ.N3(12)) TE(1)=TE(1)+0.5  
 57   NE=4
      GO TO 74
C-----HEXAGONAL
 58   NG=3
      M=ZEI(1,1)
      N=ZEI(1,2)
      IF (M.EQ.N3(16).AND.N.EQ.N3(3)) NS=0
      IF (M.EQ.N3(6)) GO TO 61
      SS(2,1,2)=-1
      SS(2,2,1)=1
      SS(2,2,2)=-1
      SS(2,3,3)=1
      IF (N.EQ.N3(1)) TS(2,3)=1.0/3.0
      IF (N.EQ.N3(2)) TS(2,3)=2.0/3.0
      SS(3,1,1)=-1
      SS(3,2,1)=-1
      SS(3,1,2)=1
      SS(3,3,3)=1
      TS(3,3)=2.0*TS(2,3)
      IF (TS(3,3).GE.1.0) TS(3,3)=TS(3,3)-1.0
      IF (NBL.EQ.1.AND.N.NE.N3(6)) GO TO 76
      IF (N.NE.N3(6)) GO TO 60
      NG=NG+NG
      DO 59 I=1,3
      DO 59 J=1,3
      DO 59 K=1,3
      SS(3+I,J,K)=SS(I,J,K)
 59   SS(3+I,3,3)=-1
 60   IF (NBL.EQ.1) GO TO 76
      IF (ZEI(2,1).NE.N3(9).AND.ZEI(3,1).NE.N3(9)) GO TO 63
      TS(4,3)=0.5
      TS(5,3)=0.5
      TS(6,3)=0.5
      GO TO 63
 61   NG=NG+NG
      SS(2,1,1)=1
      SS(2,1,2)=-1
      SS(2,2,1)=1
      SS(2,3,3)=1
      IF (N.EQ.N3(1)) TS(2,3)=1.0/6.0
      IF (N.EQ.N3(2)) TS(2,3)=2.0/6.0
      IF (N.EQ.N3(3)) TS(2,3)=3.0/6.0
      IF (N.EQ.N3(4)) TS(2,3)=4.0/6.0
      IF (N.EQ.N3(5)) TS(2,3)=5.0/6.0
      DO 62 I=1,4
      DO 62 J=1,3
      TS(2+I,J)=TS(2,J)
      DO 62 K=1,3
      TS(2+I,J)=TS(2+I,J)+SS(2,J,K)*TS(1+I,K)
      IF (TS(2+I,J).GT.1.0) TS(2+I,J)=TS(2+I,J)-1.0
      DO 62 L=1,3
 62   SS(2+I,J,K)=SS(2+I,J,K)+SS(2,J,L)*SS(1+I,L,K)
      IF (NBL.EQ.1) GO TO 76
 63   NG=NG+NG
      M=ZEI(2,1)
      N=ZEI(3,1)
      IF (M.EQ.N3(1)) GO TO 65
      IF (M.EQ.N3(2)) GO TO 64
      E(1,2)=-1
      E(2,1)=-1
      E(3,3)=1
      IF (M.EQ.N3(9)) TE(3)=0.5
      GO TO 67
 64   E(1,2)=1
      E(2,1)=1
      E(3,3)=-1
      TE(3)=2.0*TS(2,3)
C-----GROUP P 31 I 2 AND P 32 I 2
      IF (ZEI(1,1).EQ.N3(3).AND.(ZEI(1,2).EQ.N3(1).OR.ZEI(1,2).EQ.
     1N3(2))) TE(3)=0.
      GO TO 67
 65   IF (N.EQ.N3(2)) GO TO 66
      E(1,2)=1
      E(2,1)=1
      E(3,3)=1
      IF (N.EQ.N3(9)) TE(3)=0.5
      GO TO 67
 66   E(1,2)=-1
      E(2,1)=-1
      E(3,3)=-1
      TE(3)=2.0*TS(2,3)
      IF (TE(3).GT.1.0) TE(3)=TE(3)-1.0 
 67   NE=6
      IF (ZEI(1,1).EQ.N3(3).OR.(ZEI(1,2).EQ.N3(3).AND.ZEI(1,1).EQ.
     1N3(16))) NE=3
      GO TO 74
C-----CUBIC
 68   NG=12
      IF (NBL.EQ.3) NG=24
      IF (ZEI(1,1).NE.N3(2).AND.ZEI(1,1).NE.N3(4).AND.ZEI(1,1).NE.
     1N3(16)) NS=0
      DO 69 I=1,3
      DO 69 J=1,3
      SS(1+I,J,J)=1
      IF (I.EQ.J) GO TO 69
      SS(1+I,J,J)=-1
      IF (ZEI(1,1).EQ.N3(12)) TS(1+I,J)=0.5
      IF (ZEI(1,1).EQ.N3(10)) TS(1+I,J)=0.25 
 69   CONTINUE
      IF (ZEI(1,1).NE.N3(7).AND.ZEI(3,1).NE.N3(10).AND.ZEI(1,2).NE.N3(3)
     1.AND.ZEI(1,2).NE.N3(1).OR.NBR.EQ.5) GO TO 71
      DO 70 I=1,3
      TS(1+I,I)=0.5
      K=I+1
      IF (K.EQ.4) K=1
 70   TS(1+I,K)=0.5
 71   DO 72 I=1,4
      DO 72 J=1,3
      DO 72 K=1,3
      L=J+1
      IF (L.EQ.4) L=1
      M=J-1
      IF (M.EQ.0) M=3
      SS(4+I,J,K)=SS(I,L,K)
      SS(8+I,J,K)=SS(I,M,K)
      TS(4+I,J)=TS(I,L)
 72   TS(8+I,J)=TS(I,M)
      IF (NG.EQ.12) GO TO 76
      NE=12
      E(1,2)=1
      E(2,1)=1
      E(3,3)=1
      IF (ZEI(3,1).EQ.N3(2)) E(3,3)=-1
      IF (ZEI(3,1).EQ.N3(9)) TE(3)=0.5
      DO 73 I=1,3
      IF (ZEI(3,1).EQ.N3(12).OR.ZEI(1,2).EQ.N3(2)) TE(I)=0.5 
 73   IF (ZEI(3,1).EQ.N3(10).OR.ZEI(1,2).EQ.N3(1).OR.ZEI(1,2).EQ.N3(3))
     1TE(I)=0.25
      IF (ZEI(1,2).EQ.N3(1).AND.NBR.EQ.1) TE(1)=0.75
      IF ((ZEI(1,2).NE.N3(1).OR.NBR.NE.6).AND.(ZEI(1,2).NE.N3(3).OR.NBR
     1.NE.1)) GO TO 74
      TE(2)=0.75
      TE(3)=0.75
 74   DO 75 I=1,NE
      DO 75 J=1,3
      TS(NE+I,J)=TE(J)
      DO 75 K=1,3
      TS(NE+I,J)=TS(NE+I,J)+E(J,K)*TS(I,K)
      DO 75 L=1,3
 75   SS(NE+I,J,K)=SS(NE+I,J,K)+E(J,L)*SS(I,L,K) 
 76   DO 81 I=1,NG
      DO 81 J=1,3
 77   IF (TS(I,J).GE.1.0) GO TO 78
      GO TO 79
 78   TS(I,J)=TS(I,J)-1.0
      GO TO 77
 79   IF (TS(I,J).LT.0.0) GO TO 80
      GO TO 81
 80   TS(I,J)=TS(I,J)+1.0
      GO TO 79
 81   CONTINUE
      ISYSTM=ISYS(SYS)
      ISYMCE=1-NS
      RETURN
      END
C
C***************************************************************** 
C
      SUBROUTINE ESP_SPE(XL,I,NO,NZERO,ISEED,NSP)
C
C  Subroutine for special positions
C
      INCLUDE 'esp0.inc'
      DIMENSION XL(N_O,5,N_A),NZERO(N_O)
      GO TO (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,
     +23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,
     +44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,
     +65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,
     +86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,
     +105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,
     +121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,
     +137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,
     +153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168)
     +,NSP
*
* 1 -    x,y,z general position
* 2 -    x,x,x
* 6 -    x,0,z
* 9 -    x,y,0
* 10-    0,y,z
* 13-    x,0,0
* 12-    0,y,0
* 8 -    0,0,z
* 3 -    0,0,0
* 14-    1/2,0,0
* 15-    0,1/2,0
* 16-    0,0,1/2
* 17-    1/2,1/2,1/2
* 18-    1/2,1/2,0
* 19-    1/2,0,1/2
* 20-    0,1/2,1/2
* 21-    x,1/2,1/2
* 22-    1/2,y,1/2
* 23-    1/2,1/2,z
* 24-    x,0,1/2
* 25-    0,y,1/2
* 26-    1/2,0,z
* 27-    x,1/2,0
* 11-    1/2,y,0
* 5 -    0,1/2,z
* 28-    x,1/2,z
* 29-    1/2,y,z
* 30-    x,y,1/2
* 4 -    x,1/4,z
* 31-    x,y,1/4
* 32-    1/4,y,z
* 33-    1/4,1/4,0
* 34-    1/4,0,1/4
* 35-    0,1/4,1/4
* 36-    1/4,1/4,1/2
* 37-    1/4,1/2,1/4
* 38-    1/2,1/4,1/4
* 7 -    x,0,1/4
* 39-    x,1/4,0
* 40-    0,y,1/4
* 41-    1/4,y,0
* 42-    0,1/4,z
* 43-    1/4,0,z
* 44-    x,1/2,1/4
* 45-    x,1/4,1/2
* 46-    1/2,y,1/4
* 47-    1/4,y,1/2
* 48-    1/2,1/4,z
* 49-    1/4,1/2,z
* 50-    x,1/4,1/4
* 51-    1/4,y,1/4
* 52-    1/4,1/4,z
* 53-    1/4,1/4,1/4
* 55-    3/4,3/4,3/4
* 56-    x,1/4,3/4
* 57-    x,3/4,1/4
* 58-    1/4,y,3/4
* 59-    3/4,y,1/4
* 60-    1/4,3/4,z
* 61-    3/4,1/4,z
* 54-    1/4,1/4,3/4
* 62-    3/4,1/4,1/4
* 63-    1/4,3/4,1/4
* 64-    0,1/2,1/4
* 65-    0,1/4,1/2
* 66-    1/2,0,1/4
* 67-    1/4,0,1/2
* 68-    1/2,1/4,0
* 69-    1/4,1/2,0
* 70-    1/2,1/2,1/4
* 71-    1/2,1/4,1/2
* 72-    1/4,1/2,1/2
* 73-    1/4,0,0
* 74-    0,1/4,0
* 75-    0,0,1/4
* 76-    3/4,1/4,1/2
* 77-    1/2,3/4,1/4
* 78-    1/4,1/2,3/4
* 79-    3/4,1/2,1/4
* 80-    1/2,1/4,3/4
* 81-    1/4,3/4,1/2
* 82-    1/4,3/4,0
* 83-    0,1/4,3/4
* 84-    3/4,0,1/4
* 85-    1/4,0,3/4
* 86-    0,3/4,1/4
* 87-    3/4,1/4,0
* 88-    0,1/2,1/4
* 89-    1/4,0,1/2
* 90-    1/2,1/4,0
* 91-    0,1/4,1/2
* 92-    1/4,1/2,0
* 93-    1/2,0,1/4
* 94-    1/8,1/8,z
* 95-    1/8,y,1/8
* 96-    x,1/8,1/8
* 97-    1/8,1/8,5/8
* 98-    1/8,1/8,1/8
* 99-    0,1/4,5/8
*100-    0,1/4,1/8
*101-    x,x,0
*102-    x,x,1/2
*103-    x,x,3/8
*104-    x,x,5/8
*105-    x,x,1/4
*106-    x,x,3/4
*107-    x,x+1/2,1/4
*108-    x,x+1/2,1/2
*109-    x,x+1/2,0
*110-    x,x+1/2,3/4
*111-    x,x,z
*112-    x,x+1/2,z
*113-    x,-x+1/2,1/4
*114-    x,-x,z
*115-    x,-x,1/2
*116-    x,-x,0
*117-    x,-x,1/4
*118-    x,1/4,1/8
*119-    1/4,y,1/8
*120-    x,x+1/4,7/8
*121-    x,x+1/4,1/8
*122-    0,3/4,1/8
*123-    0,1/4,3/8
*124-    0,1/4,1/8
*125-    1/3,2/3,z
*126-    2/3,1/3,z
*127-    2/3,1/3,1/2
*128-    2/3,1/3,0
*129-    2/3,1/3,1/4
*130-    1/3,2/3,1/2
*131-    1/3,2/3,0
*132-    1/3,2/3,1/4
*133-    1/3,2/3,3/4
*134-    2/3,1/3,3/4
*135-    x,-x,5/6
*136-    x,-x,1/3
*137-    x,-x,1/6
*138-    x,-x,2/3
*139-    x,0,5/6
*140-    x,0,1/3
*141-    x,0,1/6
*142-    x,0,2/3
*143-    x,2x,1/4
*144-    x,2x,3/4
*145-    x,2x,1/2
*146-    x,2x,0
*147-    x,2x,z
*148-    x,1/8,1/8
*149-    1/8,1/8,1/8
*150-    5/8,5/8,5/8
*151-    7/8,7/8,7/8
*152-    3/8,3/8,3/8
*153-    7/8,1/8,1/8
*154-    1/2,y,y
*155-    0,y,y
*156-    1/4,y,y
*157-    1/2,y,-y
*158-    0,y,-y
*159-    1/4,y,-y
*160-    1/4,y,y+1/2
*161-    1/2,y,y+1/2
*162-    1/4,y,-y+1/2
*163-    1/8,y,-y+1/4
*164-    1/8,y,y+1/4
*165-    5/8,0,1/4
*166-    1/8,0,1/4
*167-    7/8,0,1/4
*168-    3/8,0,1/4
*  x,y,z general position
1     IF(NZERO(NO).EQ.1)THEN
      XL(NO,1,I)=0.
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.
      GO TO 400
      ENDIF
      XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,x,x
2     XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=XL(NO,1,I)
      GO TO 400
*  0,0,0
3     XL(NO,1,I)=0.
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=XL(NO,1,I)
      GO TO 400
*  x,1/4,z
4     XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  0,1/2,z
5     XL(NO,1,I)=0.
      XL(NO,2,I)=0.5
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,0,z
6     XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  0,0,z
8     XL(NO,1,I)=0.
      XL(NO,2,I)=0.
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,y,0
9     XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.
      GO TO 400
*  0,y,z
10    XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/2,y,0
11    XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.
      GO TO 400
*  0,y,0
12    XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.
      GO TO 400
*  x,0,0
13    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.
      GO TO 400
*  1/2,0,0
14    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.
      GO TO 400
*  0,1/2,0
15    XL(NO,1,I)=0.
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.
      GO TO 400
*  0,0,1/2
16    XL(NO,1,I)=0.
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/2,1/2
17    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/2,0
18    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.
      GO TO 400
*  1/2,0,1/2
19    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.5
      GO TO 400
*  0,1/2,1/2
20    XL(NO,1,I)=0.
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.5
      GO TO 400
*  x,1/2,1/2
21    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,y,1/2
22    XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/2,z
23    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.5
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,0,1/2
24    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.5
      GO TO 400
*  0,y,1/2
25    XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,0,z
26    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,1/2,0
27    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.
      GO TO 400
*  x,1/2,z
28    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.5
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/2,y,z
29    XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,y,1/2
30    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5
      GO TO 400
*  x,y,1/4
31    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,y,z
32    XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/4,1/4,0
33    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.
      GO TO 400
*  1/4,0,1/4
34    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  0,1/4,1/4
35    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,1/4,1/2
36    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/4,1/2,1/4
37    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  1/2,1/4,1/4
38    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.25
      GO TO 400
*  x,0,1/4
7     XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  x,1/4,0
39    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.
      GO TO 400
*  0,y,1/4
40    XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,y,0
41    XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.0
      GO TO 400
*  0,1/4,z
42    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/4,0,z
43    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,1/2,1/4
44    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  x,1/4,1/2
45    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,y,1/4
46    XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,y,1/2
47    XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/4,z
48    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/4,1/2,z
49    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,1/4,1/4
50    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,y,1/4
51    XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,1/4,z
52    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.25
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/4,1/4,1/4
53    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,1/4,3/4
54    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.75
      GO TO 400
*  3/4,3/4,3/4
55    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.75
      GO TO 400
*  x,1/4,3/4
56    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.75
      GO TO 400
*  x,3/4,1/4
57    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,y,3/4
58    XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.75
      GO TO 400
*  3/4,y,1/4
59    XL(NO,1,I)=0.75
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,3/4,z
60    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.75
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  3/4,1/4,z
61    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.25
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  3/4,1/4,1/4
62    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.25
      GO TO 400
      GO TO 400
*  1/4,3/4,1/4
63    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.25
      GO TO 400
*  0,1/2,1/4
64    XL(NO,1,I)=0.
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  0,1/4,1/2
65    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,0,1/4
66    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,0,1/2
67    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/4,0
68    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.
      GO TO 400
*  1/4,1/2,0
69    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.
      GO TO 400
*  1/2,1/2,1/4
70    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  1/2,1/4,1/2
71    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/4,1/2,1/2
72    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.5
      GO TO 400
*  1/4,0,0
73    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.
      GO TO 400
*  0,1/4,0
74    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.
      GO TO 400
*  0,0,1/4
75    XL(NO,1,I)=0.
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  3/4,1/4,1/2
76    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,3/4,1/4
77    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,1/2,3/4
78    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.75
      GO TO 400
*  3/4,1/2,1/4
79    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  1/2,1/4,3/4
80    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.75
      GO TO 400
*  1/4,3/4,1/2
81    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.5
      GO TO 400
*  1/4,3/4,0
82    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.
      GO TO 400
*  0,1/4,3/4
83    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.75
      GO TO 400
*  3/4,0,1/4
84    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,0,3/4
85    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.75
      GO TO 400
*  0,3/4,1/4
86    XL(NO,1,I)=0.
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.25
      GO TO 400
*  3/4,1/4,0
87    XL(NO,1,I)=0.75
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.0
      GO TO 400
*  0,1/2,1/4
88    XL(NO,1,I)=0.
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  1/4,0,1/2
89    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.5
      GO TO 400
*  1/2,1/4,0
90    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.0
      GO TO 400
*  0,1/4,1/2
91    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.5
      GO TO 400
*  1/4,1/2,0
92    XL(NO,1,I)=0.25
      XL(NO,2,I)=0.5
      XL(NO,3,I)=0.
      GO TO 400
*  1/2,0,1/4
93    XL(NO,1,I)=0.5
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  1/8,1/8,z
94    XL(NO,1,I)=0.125
      XL(NO,2,I)=0.125
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  1/8,y,1/8
95    XL(NO,1,I)=0.125
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.125
      GO TO 400
*  x,1/8,1/8
96    XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.125
      GO TO 400
*  1/8,1/8,5/8
97    XL(NO,1,I)=0.125
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.625
      GO TO 400
*  1/8,1/8,1/8
98    XL(NO,1,I)=0.125
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.125
      GO TO 400
*  0,1/4,5/8
99    XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.625
      GO TO 400
*  0,1/4,1/8
100   XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.125
      GO TO 400
*  x,x,0
101   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.
      GO TO 400
*  x,x,1/2
102   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.5
      GO TO 400
*  x,x,3/8
103   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.375
      GO TO 400
*  x,x,5/8
104   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.625
      GO TO 400
*  x,x,1/4
105   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.25
      GO TO 400
*  x,x,3/4
106   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=0.75
      GO TO 400
*  x,x+1/2,1/4
107   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  x,x+1/2,1/2
108   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.5
      XL(NO,3,I)=0.5
      GO TO 400
*  x,x+1/2,0
109   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.5
      XL(NO,3,I)=0.
      GO TO 400
*  x,x+1/2,3/4
110   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.5
      XL(NO,3,I)=0.75
      GO TO 400
*  x,x,z
111   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,x+1/2,z
112   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.5
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,-x+1/2,1/4
113   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)+0.5
      XL(NO,3,I)=0.25
      GO TO 400
*  x,-x,z
114   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,-x,1/2
115   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.5
      GO TO 400
*  x,-x,0
116   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.
      GO TO 400
*  x,-x,1/4
117   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.25
      GO TO 400
*  x,1/4,1/8
118   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.125
      GO TO 400
*  1/4,y,1/8
119   XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.125
      GO TO 400
*  x,x+1/4,7/8
120   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.25
      XL(NO,3,I)=0.875
      GO TO 400
*  x,x+1/4,1/8
121   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=XL(NO,1,I)+0.24
      XL(NO,3,I)=0.125
      GO TO 400
*  0,3/4,1/8
122   XL(NO,1,I)=0.
      XL(NO,2,I)=0.75
      XL(NO,3,I)=0.125
      GO TO 400
*  0,1/4,3/8
123   XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.375
      GO TO 400
*  0,1/4,1/8
124   XL(NO,1,I)=0.
      XL(NO,2,I)=0.25
      XL(NO,3,I)=0.125
      GO TO 400
*  1/3,2/3,z
125   XL(NO,1,I)=0.3333333
      XL(NO,2,I)=0.6666667
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  2/3,1/3,z
126   XL(NO,1,I)=0.6666667
      XL(NO,2,I)=0.3333333
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  2/3,1/3,1/2
127   XL(NO,1,I)=0.6666667
      XL(NO,2,I)=0.3333333
      XL(NO,3,I)=0.5
      GO TO 400
*  2/3,1/3,0
128   XL(NO,1,I)=0.6666667
      XL(NO,2,I)=0.3333333
      XL(NO,3,I)=0.
      GO TO 400
*  2/3,1/3,1/4
129   XL(NO,1,I)=0.6666667
      XL(NO,2,I)=0.3333333
      XL(NO,3,I)=0.25
      GO TO 400
*  1/3,2/3,1/2
130   XL(NO,1,I)=0.3333333
      XL(NO,2,I)=0.6666667
      XL(NO,3,I)=0.5
      GO TO 400
*  1/3,2/3,0
131   XL(NO,1,I)=0.3333333
      XL(NO,2,I)=0.6666667
      XL(NO,3,I)=0.
      GO TO 400
*  1/3,2/3,1/4
132   XL(NO,1,I)=0.3333333
      XL(NO,2,I)=0.6666667
      XL(NO,3,I)=0.25
      GO TO 400
*  1/3,2/3,3/4
133   XL(NO,1,I)=0.3333333
      XL(NO,2,I)=0.6666667
      XL(NO,3,I)=0.75
      GO TO 400
*  2/3,1/3,3/4
134   XL(NO,1,I)=0.6666667
      XL(NO,2,I)=0.3333333
      XL(NO,3,I)=0.75
      GO TO 400
*  x,-x,5/6
135   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.8333333
      GO TO 400
*  x,-x,1/3
136   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.3333333
      GO TO 400
*  x,-x,1/6
137   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.1666667
      GO TO 400
*  x,-x,2/3
138   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=-XL(NO,1,I)
      XL(NO,3,I)=0.6666667
      GO TO 400
*  x,0,5/6
139   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.8333333
      GO TO 400
*  x,0,1/3
140   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.3333333
      GO TO 400
*  x,0,1/6
141   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.1666667
      GO TO 400
*  x,0,2/3
142   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.6666667
      GO TO 400
*  x,2x,1/4
143   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=2.*XL(NO,1,I)
      XL(NO,3,I)=0.25
      GO TO 400
*  x,2x,3/4
144   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=2.*XL(NO,1,I)
      XL(NO,3,I)=0.75
      GO TO 400
*  x,2x,1/2
145   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=2.*XL(NO,1,I)
      XL(NO,3,I)=0.5
      GO TO 400
*  x,2x,0
146   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=2.*XL(NO,1,I)
      XL(NO,3,I)=0.
      GO TO 400
*  x,2x,z
147   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=2.*XL(NO,1,I)
      XL(NO,3,I)=RAN(ISEED)
      GO TO 400
*  x,1/8,1/8
148   XL(NO,1,I)=RAN(ISEED)
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.125
      GO TO 400
*  1/8,1/8,1/8
149   XL(NO,1,I)=0.125
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.125
      GO TO 400
*  5/8,5/8,5/8
150   XL(NO,1,I)=0.625
      XL(NO,2,I)=0.625
      XL(NO,3,I)=0.625
      GO TO 400
*  7/8,7/8,7/8
151   XL(NO,1,I)=0.875
      XL(NO,2,I)=0.875
      XL(NO,3,I)=0.875
      GO TO 400
*  3/8,3/8,3/8
152   XL(NO,1,I)=0.375
      XL(NO,2,I)=0.375
      XL(NO,3,I)=0.375
      GO TO 400
*  7/8,1/8,1/8
153   XL(NO,1,I)=0.875
      XL(NO,2,I)=0.125
      XL(NO,3,I)=0.125
      GO TO 400
*  1/2,y,y
154   XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=XL(NO,3,I)
      GO TO 400
*  0,y,y
155   XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=XL(NO,3,I)
      GO TO 400
*  1/4,y,y
156   XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=XL(NO,3,I)
      GO TO 400
*  1/2,y,-y
157   XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=-XL(NO,3,I)
      GO TO 400
*  0,y,-y
158   XL(NO,1,I)=0.
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=-XL(NO,3,I)
      GO TO 400
*  1/4,y,y
159   XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=-XL(NO,3,I)
      GO TO 400
*  1/4,y,1/2+y
160   XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5+XL(NO,3,I)
      GO TO 400
*  1/2,y,1/2+y
161   XL(NO,1,I)=0.5
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=0.5+XL(NO,3,I)
      GO TO 400
*  1/4,y,-y+1/2
162   XL(NO,1,I)=0.25
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=-XL(NO,3,I)+0.5
      GO TO 400
*  1/8,y,-y+1/4
163   XL(NO,1,I)=0.125
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=-XL(NO,3,I)+0.25
      GO TO 400
*  1/8,y,y+1/4
164   XL(NO,1,I)=0.125
      XL(NO,2,I)=RAN(ISEED)
      XL(NO,3,I)=XL(NO,3,I)+0.25
      GO TO 400
*  5/8,0,1/4
165   XL(NO,1,I)=0.625
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  1/8,0,1/4
166   XL(NO,1,I)=0.125
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  7/8,0,1/4
167   XL(NO,1,I)=0.875
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
      GO TO 400
*  3/8,0,1/4
168   XL(NO,1,I)=0.375
      XL(NO,2,I)=0.
      XL(NO,3,I)=0.25
400   CONTINUE
      RETURN
      END
C***************************************************************** 
C
C The following subroutine is only used in order to build the SHELX 
C     .ins output file.
C
C*****************************************************************
      SUBROUTINE STRINGSYM(IM,TS,STRSYM)
      Character STRSYM*50,sigx(3)*1,xlab(3)*1,xyzt(3)*15,auxc*15
      Character STRING1*15,STRING2*15
      dimension IM(3,3),TS(3),ino(3),lni(3)
      data xlab/'X','Y','Z'/
      do i=1,3
       auxc=' '
       LENx=1
       do j=1,3
        ino(j)=0
        IF(IM(i,j).EQ.1) sigx(j)='+'
        IF(IM(i,j).EQ.-1) sigx(j)='-'
        IF(IM(i,j).EQ.0) ino(j)=1
       END DO
       do j=1,3
        IF(LENx.lt.1) LENx=1
        IF(ino(j).EQ.0) then
         IF(j.EQ.1.and.sigx(j).EQ.'+') then
           STRING1=auxc(1:LENx)//xlab(j)
           CALL ATRIM(STRING1,STRING2)
           auxc=STRING2
          else
           STRING1=auxc(1:LENx)//sigx(j)//xlab(j)
           CALL ATRIM(STRING1,STRING2)
           auxc=STRING2
         ENDIF
        ENDIF
        CALL ALEN_TRIM(auxc,LENx)
       END DO
       IF(abs(TS(i)).gt.0.0001) then
       WRITE(xyzt(i),'(a,a,f6.4,a)') auxc(1:LENx),'+',TS(i),','
       else
       WRITE(xyzt(i),'(a,a)') auxc(1:LENx),','
       ENDIF
        CALL ALEN_TRIM(xyzt(i),lni(i))
      END do
      STRSYM=xyzt(1)(1:lni(1))//xyzt(2)(1:lni(2))//xyzt(3)(1:lni(3)-1)
      RETURN
      END
*
*
      SUBROUTINE ATRIM(STRING1,STRING2) 
C         ATRIM returns the argument with trailing blanks 
C                       removed
C
      Character string1*15,string2*15
      J=0
      DO 1 lfile=1,15
      IF(string1(lfile:lfile).eq.' ')GO TO 1
      J=J+1
      string2(J:J)=string1(lfile:lfile) 
1     CONTINUE
      DO 2 K=J+1,15
2     string2(K:K)=' '
      RETURN
      END
*
*
      SUBROUTINE ALEN_TRIM(STRING,L)
C         ALEN_TRIM returns the length of the character 
C                           argument without counting trailing 
C                           blank characters.
      Character string*15
      L=0
      DO 1 J=1,15
      IF(string(J:J).eq.' ')GO TO 1
      L=L+1
1     CONTINUE
      RETURN
      END
*
*
*
C      function gran(seed)
C! machine dependent RAN
C      integer seed
C      gran=ran(seed)   ! replace ran by your specific call
C      return
C      end
*
      function gsecnds(sek)
! machine dependent SECNDS
! replace this line by your specific call
C
C PORLIB is compiler spcecific part, introduced for 
C using the intrinsic function SECNDS(X) which returns the 
C  (time in seconds since midnight - X) 
C
      USE PORTLIB
      gsecnds=SECNDS(sek)
      return
      end
*
      SUBROUTINE SMALLC(NAM)
      CHARACTER*4 NAM
      DO 1 I=1,2
      IF(NAM(I:I).EQ.'A')NAM(I:I)='a'
      IF(NAM(I:I).EQ.'B')NAM(I:I)='b'
      IF(NAM(I:I).EQ.'C')NAM(I:I)='c'
      IF(NAM(I:I).EQ.'D')NAM(I:I)='d'
      IF(NAM(I:I).EQ.'E')NAM(I:I)='e'
      IF(NAM(I:I).EQ.'F')NAM(I:I)='f'
      IF(NAM(I:I).EQ.'G')NAM(I:I)='g'
      IF(NAM(I:I).EQ.'H')NAM(I:I)='h'
      IF(NAM(I:I).EQ.'I')NAM(I:I)='i'
      IF(NAM(I:I).EQ.'J')NAM(I:I)='j'
      IF(NAM(I:I).EQ.'K')NAM(I:I)='k'
      IF(NAM(I:I).EQ.'L')NAM(I:I)='l'
      IF(NAM(I:I).EQ.'M')NAM(I:I)='m'
      IF(NAM(I:I).EQ.'N')NAM(I:I)='n'
      IF(NAM(I:I).EQ.'O')NAM(I:I)='o'
      IF(NAM(I:I).EQ.'P')NAM(I:I)='p'
      IF(NAM(I:I).EQ.'Q')NAM(I:I)='q'
      IF(NAM(I:I).EQ.'R')NAM(I:I)='r'
      IF(NAM(I:I).EQ.'S')NAM(I:I)='s'
      IF(NAM(I:I).EQ.'T')NAM(I:I)='t'
      IF(NAM(I:I).EQ.'U')NAM(I:I)='u'
      IF(NAM(I:I).EQ.'V')NAM(I:I)='v'
      IF(NAM(I:I).EQ.'W')NAM(I:I)='w'
      IF(NAM(I:I).EQ.'X')NAM(I:I)='x'
      IF(NAM(I:I).EQ.'Y')NAM(I:I)='y'
      IF(NAM(I:I).EQ.'Z')NAM(I:I)='z'
1     CONTINUE
      RETURN
      END
*
      SUBROUTINE BIGC(NAM)
      CHARACTER*4 NAM
      DO 1 I=1,2
      IF(NAM(I:I).EQ.'a')NAM(I:I)='A'
      IF(NAM(I:I).EQ.'b')NAM(I:I)='B'
      IF(NAM(I:I).EQ.'c')NAM(I:I)='C'
      IF(NAM(I:I).EQ.'d')NAM(I:I)='D'
      IF(NAM(I:I).EQ.'e')NAM(I:I)='E'
      IF(NAM(I:I).EQ.'f')NAM(I:I)='F'
      IF(NAM(I:I).EQ.'g')NAM(I:I)='G'
      IF(NAM(I:I).EQ.'h')NAM(I:I)='H'
      IF(NAM(I:I).EQ.'i')NAM(I:I)='I'
      IF(NAM(I:I).EQ.'j')NAM(I:I)='J'
      IF(NAM(I:I).EQ.'k')NAM(I:I)='K'
      IF(NAM(I:I).EQ.'l')NAM(I:I)='L'
      IF(NAM(I:I).EQ.'m')NAM(I:I)='M'
      IF(NAM(I:I).EQ.'n')NAM(I:I)='N'
      IF(NAM(I:I).EQ.'o')NAM(I:I)='O'
      IF(NAM(I:I).EQ.'p')NAM(I:I)='P'
      IF(NAM(I:I).EQ.'q')NAM(I:I)='Q'
      IF(NAM(I:I).EQ.'r')NAM(I:I)='R'
      IF(NAM(I:I).EQ.'s')NAM(I:I)='S'
      IF(NAM(I:I).EQ.'t')NAM(I:I)='T'
      IF(NAM(I:I).EQ.'u')NAM(I:I)='U'
      IF(NAM(I:I).EQ.'v')NAM(I:I)='V'
      IF(NAM(I:I).EQ.'w')NAM(I:I)='W'
      IF(NAM(I:I).EQ.'x')NAM(I:I)='X'
      IF(NAM(I:I).EQ.'y')NAM(I:I)='Y'
      IF(NAM(I:I).EQ.'z')NAM(I:I)='Z'
1     CONTINUE
      RETURN
      END
*
************************************************************* 
*
      SUBROUTINE PRF2HGL(FILE,NHKL)
C	PROGRAMME DE TRACE ECRAN DES .PRF DE YOUNG ou des .DAT
C
      INCLUDE 'esp0.inc'
      INCLUDE 'esp2.inc'
C      COMMON/YCALC/OMEGA(100,N_HKL),YCALC(N_PTS),YOBS(N_PTS),
C     -POS(N_PTS),NHA(N_PTS),NHB(N_PTS),BBB(N_HKL),HW4(N_HKL),
C     -THETA(N_HKL),HW(N_HKL),U,V,W,NPTS,N1,N2
      character*(*) file
      character*24 tmp
      logical qex
	COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
	DIMENSION ST(16000),X(16000),YOO(16000),YCC(16000)
	DIMENSION OB(10),CA(10),YDD(16000),KKK(10)
	DIMENSION YO(16000),YC(16000),YD(16000)
	COMMON/TRAN/M0,NT,DPX,DPY,DVX,DVY,NINT,VIT
C	PRINT 7501
C7501  format('  entry filename (no extension) ??',$)
C      read 2,file
C2     format(A20)
C4     format(A24)
      lfile=len(file)
      do while (file(lfile:lfile).eq.' ')
      lfile=lfile-1
      enddo
	IOK=14
      tmp=file(1:lfile)//'.hgl'
      inquire(file=tmp,exist=qex)
      if(qex.eq..FALSE.) go to 10
C      print *,'warning, this old file will be deleted :',tmp
      call filedel(iok,tmp)
10    call open_write1(iok,tmp)
C	TYPE 8000
C8000	FORMAT('  .PRF (1) OR .DAT (2) ?'$)
C	READ *,NT
C	IF(NT.EQ.2) GO TO 8001
C	IF(NT.EQ.1)GO TO 4500
C8001	CLOSE(26)
C      tmp=file(1:lfile)//'.dat'
C      call open_read1(26,tmp)
C      READ(26,*)NBMES,TH0,STP
C	KI=0
C	IJK=NBMES/10+0.5
C	DO 9060 I=1,IJK
C	READ(26,*)(YO(KI+J),J=1,10)
C	KI=KI+10
C9060	CONTINUE
C	DO 9090 I=1,NBMES
C	YD(I)=0.
C9090	YC(I)=0.
C9080	GO TO 8010
C4500	CLOSE(26)
C      tmp=file(1:lfile)//'.prf'
C      call open_read1(26,tmp)
C      READ(26,4501)
C	READ(26,*)THMAX,TH0,STP
      THMAX=POS(N2)
      TH0=POS(N1)
C      WRITE(IOK,*)'  THMAX, TH0 ',THMAX,TH0
C
C---  STP may be not constant in ESPOIR...
C
C4501	FORMAT(/)
C	READ(26,*)NPHASE,NBMES
      NBMES=NPTS
C4502	FORMAT(I3,I5)
C	READ(26,*)IC1,IC2
      IC1=NHKL
C      WRITE(IOK,*)'  NBMES, IC1 ',NBMES,IC1
C4503	FORMAT(2I5)
	NBB=NBMES
C	READ(26,*)(YO(J),J=1,NBMES)
C	READ(26,*)(YC(J),J=1,NBMES)
      J=0
      DO 550 I=N1,N2
      J=J+1
      YO(J)=YOBS(I)
      YC(J)=YCALC(I)
550   CONTINUE
C      WRITE(IOK,4504)(YO(J),J=1,NBMES)
C      WRITE(IOK,4504)(YC(J),J=1,NBMES)
4504	FORMAT(8(F7.0,1X))
	DO 4505 KK=1,NBMES
4505	YD(KK)=YO(KK)-YC(KK)
	NT=1
C8010	CLOSE(26)
C
C See the characters
C
C      DO 1000 I=1,300
C      WRITE(20,*)I
C      CALL INBUF2(I)
C1000  CONTINUE
C      RETURN
C
C
C
C
C
C  FACTEUR ECHELLE =0.9 NE JAMAIS CHANGER
	VFT=0.9
	IPAPER=4
	CALL HPON
	CALL PS(IPAPER)
	CALL SP(1)
C  DISPLACE 1.5 CM TOWARD UP AND RIGHT
C	TYPE 8022
C8022	FORMAT('   SPEED ?? (0. A 127.9)'$)
C	READ *,VIT
      VIT=20.
	CALL VS(VIT)
	DPX=670.
	DPY=670.
C	IF(NT.EQ.2)GO TO 3500
      J=N1-1
	DO 90 I=1,NBMES
C	SST=I
C90	ST(I)=TH0+(SST-1.)*STP
      J=J+1
90	ST(I)=POS(J)
3500	NB1=1
	KK=0
199	YMAX=0.
	YMAX1=0.
	YMIN1=0.
C	TYPE 200,ST(1),ST(NBMES)
C200	FORMAT('  XMIN ET XMAX : ',2F10.3)
C	TYPE 201
C201	FORMAT('  XMIN , XMAX ET NBRE INTERVALLES ? ',$)
C	READ *,XMIN,XMAX,NINT
C	TYPE 300
C300	FORMAT('  DRAWING SIZE MAXI : 24(X)*17(Y) CM '/
C     1'  SIZE X ET Y ?',$)
C	READ *,DVX,DVY
      XMIN=ST(1)
      XMAX=ST(NBMES)
      IXMIN=INT(XMIN)/10
      IXMAX=INT(XMAX)/10+1
      NINT=(IXMAX-IXMIN)
      XMIN=10.*FLOAT(IXMIN)
      XMAX=10.*FLOAT(IXMAX)
C      WRITE(IOK,*)
C     +'IXMIN,IXMAX,NINT,XMIN,XMAX',IXMIN,IXMAX,NINT,XMIN,XMAX
      DVX=24.
      DVY=17.
	DVX=DVX*11040./27.46/.9+DPX
	DVY=DVY*7721./19.21/.9+DPY
      K1=1
      K2=NBMES
	DO 202 I=1,NBMES
	IF(XMIN.GE.ST(I))K1=I
202	IF(XMAX.GE.ST(I))K2=I
	M0=K2-K1+1
C      WRITE(IOK,*)'  K1, K2, M0',K1,K2,M0
	IF(NT.EQ.4)NT=2
	IF(NT.EQ.2)YMIN1=YO(K1)
	DO 104 I=1,M0
	J=K1+I-1
	X(I)=ST(J)
	YOO(I)=YO(J)
	YCC(I)=YC(J)
	YDD(I)=YD(J)
	YMIN1=AMIN1(YMIN1,YDD(I))
C	IF(NT.EQ.1)YMIN1=AMIN1(YMIN1,YDD(I))
C	IF(NT.EQ.2)YMIN1=AMIN1(YMIN1,YOO(I))
	YMAX1=AMAX1(YMAX1,YDD(I))
	YMAX=AMAX1(YMAX,YCC(I))
104	YMAX=AMAX1(YMAX,YOO(I))
	XMIN=X(1)
	XMAX=X(M0)
	YYMIN1=ABS(YMIN1)
	YYMIN1=-AMAX1(YYMIN1,YMAX1)
C	IF(NT.EQ.2)GO TO 8050
C	TYPE 8031
C8031	FORMAT('  CHANGE OF YMAX ?? YES=1,NO=0',$)
C	READ *,NCYMAX
C	IF(NCYMAX.EQ.0)GO TO 8030
C	TYPE 8032,YMAX
C8032	FORMAT(' OLD YMAX = ',E14.7,' NEW YMAX  ??',$)
C	READ *,YMAX
C	DO 8033 I=1,M0
C	IF(YOO(I).GT.YMAX)YOO(I)=YMAX
C8033	IF(YCC(I).GT.YMAX)YCC(I)=YMAX
C8050	IF(NT.EQ.2)YYMIN1=YMIN1
C	IF(NT.EQ.1)GO TO 8030
C	TYPE 8020,YYMIN1,YMAX
C8020	FORMAT('  YMIN ET YMAX :',2E14.7)
C	TYPE 8021
C8021	FORMAT('  YMIN ET YMAX ??',$)
C	READ *,YYMIN1,YMAX
8030	CALL TRACE(YOO,YCC,YDD,X,XMIN,XMAX,YYMIN1,YMAX)
	CALL HPOFF
      CLOSE(14)
	RETURN
	END
	SUBROUTINE TRACE(YOO,YCC,YDD,X,XMIN,XMAX,YMIN1,YMAX)
	DIMENSION YOO(16000),YCC(16000),YDD(16000),X(16000)
	COMMON/TRAN/M0,NT,DPX,DPY,DVX,DVY,NINT,VIT
	COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
	YMIN=YMIN1*2.
	NTT=1
	NTTT=1
	IF(NT.EQ.2)YMIN=YMIN1
C	TYPE 3
C3	FORMAT('Iobs as .(1) line (2) +(3) x(4) circ(5) ??',$)
C	READ *,NTT
C	IF(NT.EQ.1)GO TO 2
C	TYPE 4
C4	FORMAT('  TRACE AXIS Y=0 ? YES=1,NO=0 ',$)
C	READ *,NTTT
      NTT=4
2	CALL GRA(XMIN,XMAX,YMIN,YMAX,X,YOO,NTT)
	IF(NTTT.EQ.0)GO TO 5
	YY=-YMIN/(YMAX-YMIN)*(DVY-DPY)+DPY
	CALL GRID(YY)
C5	IF(NT.EQ.2)GO TO 1
5	CALL SP(2)
	CALL GRA(XMIN,XMAX,YMIN,YMAX,X,YCC,2)
	YMIN=YMIN1
	YMAX=YMAX-YMIN
	CALL SP(1)
	CALL GRA(XMIN,XMAX,YMIN,YMAX,X,YDD,2)
	YY=-YMIN/(YMAX-YMIN)*(DVY-DPY)+DPY
	CALL GRID(YY)
1	CALL CADRE(XMIN,XMAX)
	RETURN
	END
	SUBROUTINE GRID(DY)
	COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
	COMMON/TRAN/M0,NT,DPX,DPY,DVX,DVY,NINT,VIT
	DX=DVX
	CALL PLOTHGL(DX,DY,3)
	DX=DPX
	CALL PLOTHGL(DX,DY,2)
	RETURN
	END
	SUBROUTINE GRA(X1,X2,Y1,Y2,E,Y,N)
	DIMENSION E(16000),Y(16000)
	COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
	COMMON/TRAN/M0,NT,DPX,DPY,DVX,DVY,NINT,VIT
      IF(N.EQ.1)GO TO 20
      IF(N.EQ.2)GO TO 20
C      PRINT 21
C21    FORMAT('  Half size of the cross/circ in pixels ?',$)
C      READ *,CS
      CS=2.
20	DO 11 I=1,M0
	YIY=((Y(I)-Y1)/(Y2-Y1))*(DVY-DPY)+DPY
	XIX=((E(I)-X1)/(X2-X1))*(DVX-DPX)+DPX
	GO TO (1,2,3),N
C  Dots
1	CALL PLOTHGL(XIX-1.,YIY,3)
      IF(Y(I).EQ.0.)GO TO 11
      CALL PLOTHGL(XIX+1.,YIY,2)
      CALL PLOTHGL(XIX,YIY+1.,3)
      CALL PLOTHGL(XIX,YIY-1.,2)
	GO TO 11
C  Line
2	IF(I.EQ.1)CALL PLOTHGL(XIX,YIY,3)
 	IF(Y(I).EQ.0.)CALL PLOTHGL(XIX,YIY,3)
      IF(Y(I).EQ.0.)GO TO 11
	CALL PLOTHGL(XIX,YIY,2)
	GO TO 11
C Cross +
3	CALL PLOTHGL(XIX-CS,YIY,3)
      IF(Y(I).EQ.0.)GO TO 11
      CALL PLOTHGL(XIX+CS,YIY,2)
      CALL PLOTHGL(XIX,YIY+CS,3)
      CALL PLOTHGL(XIX,YIY-CS,2)
      GO TO 11
C Cross x
4	CALL PLOTHGL(XIX-CS,YIY+CS,3)
      IF(Y(I).EQ.0.)GO TO 11
      CALL PLOTHGL(XIX+CS,YIY-CS,2)
      CALL PLOTHGL(XIX+CS,YIY+CS,3)
      CALL PLOTHGL(XIX-CS,YIY-CS,2)
      GO TO 11
C Circ 
5	CALL PLOTHGL(XIX-CS,YIY,3)
      IF(Y(I).EQ.0.)GO TO 11
      CALL PLOTHGL(XIX,YIY+CS,2)
      CALL PLOTHGL(XIX+CS,YIY,2)
      CALL PLOTHGL(XIX,YIY-CS,2)
      CALL PLOTHGL(XIX-CS,YIY,2)
      GO TO 11
11	CONTINUE
	RETURN
	END
	SUBROUTINE CADRE(XMIN,XMAX)
	COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
	COMMON/TRAN/M0,NT,DPX,DPY,DVX,DVY,NINT,VIT
      DIMENSION XXX(20),XLABX(20)
	CALL PLOTHGL(DPX,DPY,3)
	CALL PLOTHGL(DPX,DVY,2)
	CALL PLOTHGL(DVX,DVY,2)
	CALL PLOTHGL(DVX,DPY,2)
	CALL PLOTHGL(DVX,DPY,3)
C  TRACE DES GRADUATIONS AXE X
	XINT=NINT-1.   
	DINT=(DVX-DPX)/(XMAX-XMIN)*10.
	DX=10.
	XLAB=FLOAT(INT(XMAX/10.)*10)
      XX=DVX-DINT/10.*(XMAX-XLAB)
	CALL CARAC
	CALL PLOTHGL(XX,DPY,2)
	CALL PLOTHGL(XX,DPY+100.,2)
	CALL PLOTHGL(XX,DPY,2)
	CALL PLOTHGL(XX-50.,DPY-100.,3)
	CALL LB(XLAB)
 	CALL PLOTHGL(XX,DPY,3)
	DO 1 I=2,NINT-1
	XX=XX-DINT
	XLAB=XLAB-DX
	CALL PLOTHGL(XX,DPY,2)
	CALL PLOTHGL(XX,DPY+100.,2)
	CALL PLOTHGL(XX,DPY,2)
	CALL PLOTHGL(XX-50.,DPY-100.,3)
	CALL LB(XLAB)
1	CALL PLOTHGL(XX,DPY,3)
	CALL PLOTHGL(DPX,DPY,2)
	RETURN
	END
	SUBROUTINE CARAC
C  SPECIFICATION DES VALEURS PAR DEFAUT
C  DF BLANC ;
	CALL INBUF(68)  !  D
	CALL INBUF(70)  !  F
	CALL INBUF(32)  !  BLANC
	CALL INBUF(59)  !  ;
C  CHOIX DU JEU DE CARACTERES : 0
C  CS BLANC 0 ;
	CALL INBUF(67)   ! C
	CALL INBUF(83)   ! S
	CALL INBUF(32)   !  BLANC
	CALL INBUF(48)   ! 0
	CALL INBUF(59)   !  ;
C  DEFINITION DU TERMINATEUR DE LABEL
C  DT BLANC ETX
	CALL INBUF(68)  !  D
	CALL INBUF(84)  !  T
C	CALL INBUF(32)  !  BLANC
	CALL INBUF(3)   !  ETX
	CALL INBUF(59)  !  ;
C	CALL INBUF(13)  !  CR
	RETURN
	END
	SUBROUTINE CP(X,Y)
C   DEPLACEMENT PLUME VERS BAS (2)
C  CP BLANC  0.,-2.;
	CALL INBUF(67)  !  C
	CALL INBUF(80)  !  P
	CALL INBUF(32)  !  BLANC
	CALL TRANSL(X,1)  !  (X,1)
	CALL INBUF(44)    !  ,
	CALL TRANSL(Y,1)  !  (Y,1)
	CALL INBUF(59)  !  ;
      CALL INBUF(13)
	RETURN
	END
	SUBROUTINE LB(X)
C  INSERTION DE LABEL CHIFFRE
C  LB BLANK X ETX ;
	CALL INBUF(80)  !  P
	CALL INBUF(85)  !  U
	CALL INBUF(32)  !  BLANC
	CALL INBUF(59)  !  ;
C
	CALL CP(-1.,-1.)
C
	CALL INBUF(76)  !  L
	CALL INBUF(66)  !  B
	CALL TRANSL(X,1)  !  (X,1)
	CALL INBUF(3)  !  ETX
	CALL INBUF(59)  !  ;
	CALL INBUF(80)  !  P
	CALL INBUF(85)  !  U
	CALL INBUF(32)  !  BLANC
	CALL INBUF(59)  !  ;
	RETURN
	END
C
C  Routines for HPGL 7475 output
C
      SUBROUTINE HPON
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75   
      NEWLIN=1    
      LREC=0
C     PLOTTER ON    <ESC>.(
      CALL INBUF(27)
      CALL INBUF(46)
      CALL INBUF(40)
      CALL INBUF(13)
C     ESC.@:CR    
      CALL INBUF(27)    
      CALL INBUF(46)    
      CALL INBUF(64)    
      CALL INBUF(58)    
      CALL INBUF(13)    
C     ESC.I121;;17:CR   
      CALL INBUF(27)    
      CALL INBUF(46)    
      CALL INBUF(73)    
      CALL ITRANS(121)  
      CALL INBUF(59)    
      CALL INBUF(59)    
      CALL ITRANS(17)   
      CALL INBUF(58)    
      CALL INBUF(13)    
C     ESC.M:CR    
      CALL INBUF(27)    
      CALL INBUF(46)    
      CALL INBUF(77)    
      CALL INBUF(58)    
      CALL INBUF(13)    
C     ESC.N;19:CR 
      CALL INBUF(27)    
      CALL INBUF(46)    
      CALL INBUF(78)    
      CALL INBUF(59)    
      CALL ITRANS(19)   
      CALL INBUF(58)    
      CALL INBUF(13)    
      RETURN
	END
C	-------------------------------
      SUBROUTINE HPOFF 
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75   
C     EINE POSITION ZURUECK, DANN ;PU;SP0;CR
      LREC=LREC-1
      CALL INBUF(59)
      CALL INBUF(80)
      CALL INBUF(85)
      CALL INBUF(59)
      CALL INBUF(83)
      CALL INBUF(80)
      CALL INBUF(48)
      CALL INBUF(59)
      CALL INBUF(13)
C     PLOTTER OFF    <ESC>.)
      CALL INBUF(27)
      CALL INBUF(46)
      CALL INBUF(41)
      CALL INBUF(13)  
      RETURN
	END
C
C	--------------------------
C
      SUBROUTINE PLOTHGL(X,Y,IPEN)     
      IMPLICIT INTEGER(A-Z)   
      REAL X,Y,VFT
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      DATA LPX/16158/LPY/11040/   
C  SCALING WITH VFT
      IX=NINT(X*VFT)    
      IY=NINT(Y*VFT)    
C  TEST AT LIMITS
C     IF(IX.LT.0.OR.IX.GT.LPX.OR.IY.LT.0.OR.IY.GT.LPY) GOTO 100   
  200 IF (NEWLIN) 210,220,231 
C     NEWLIN = 1  STARTING A NEW COMMAND LINE 
  220 IF (IPEN-IPENA) 230,230,230 !!!!!!!!!!!!!!!!230,240,230  
C     CHANGE ACCORDING TO IPEN
  230 LREC=LREC-1 
      CALL INBUF(59)
	CALL INBUF(13)    
  231 NEWLIN=0    
      GOTO (250,260) IPEN-1   
C     PD BLANK
  250 CALL INBUF(80)    
      CALL INBUF(68)    
      CALL INBUF(32)    
      GOTO 240    
C     PU BLANK
  260 CALL INBUF(80)    
      CALL INBUF(85)    
      CALL INBUF(32)    
C     X-COORDINATE  ,  Y-COORDINATE  ,
  240 CALL ITRANS(IX)   
      CALL INBUF(44)    
      CALL ITRANS(IY)   
      CALL INBUF(44)    
      IPENA=IPEN  
      IF(LREC-105) 210,290,290
  290 LREC=LREC-1 
      CALL INBUF(59)    
      CALL INBUF(13)    
      NEWLIN=1    
  210 CONTINUE    
      RETURN
      END   
C
C
C
      SUBROUTINE PS(IPAPER)
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      IF(IPAPER-3) 104,102,104
C     IPAPER=3 'FORMAT DIN A3 ' PS3;IN;     
102   CALL INBUF(80)    
      CALL INBUF(83)    
      CALL INBUF(51)    
      CALL INBUF(59)    
      CALL INBUF(73)    
      CALL INBUF(78)    
      CALL INBUF(59)    
      CALL INBUF(13)
      LPX=16158   
      LPY=11040   
      GOTO 70
C     IPAPER=4 'FORMAT DIN A4 '   PS4;IN;     
104   CALL INBUF(80)    
      CALL INBUF(83)    
      CALL INBUF(52)    
      CALL INBUF(59)    
      CALL INBUF(73)    
      CALL INBUF(78)    
      CALL INBUF(59)    
      CALL INBUF(13)
      LPX=11040   
      LPY=7721    
70    RETURN
      END   
C
C	---------------------------
C
	SUBROUTINE SP(NR)
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75   
C	INST. SP (Select PEN)
      CALL INBUF(83)    
      CALL INBUF(80)    
      CALL ITRANS(NR) 
      CALL INBUF(59)    
      CALL INBUF(13)
      CALL INBUF(80)
      CALL INBUF(85)
      CALL INBUF(32)
      IPENA=3
      RETURN
      END   
C
C
	SUBROUTINE VS(VEL)
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75   
C	INST. VS (Vitesse)
      CALL INBUF(86)    
      CALL INBUF(83)    
      CALL TRANSL(VEL,2)
      CALL INBUF(59)    
      CALL INBUF(13)
      RETURN
	END
C
C	-----------------
C
      SUBROUTINE INBUF(IN)    
      IMPLICIT INTEGER(A-Z)   
      CHARACTER CH(121) 
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      LREC=LREC+1 
      CH(LREC)=CHAR(IN) 
      IF(IN-13) 10,20,10
   20 WRITE(IOK,40) (CH(K), K=1,LREC)    
   40 FORMAT(120A1)     
      LREC=0
   10 CONTINUE    
      RETURN
      END   
C
C	-----------------
C
      SUBROUTINE INBUF2(IN)    
      IMPLICIT INTEGER(A-Z)   
      CHARACTER CH(121) 
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      LREC=LREC+1 
      CH(LREC)=CHAR(IN) 
C      IF(IN-13) 10,20,10
   20 WRITE(IOK,40) (CH(K), K=1,LREC)    
   40 FORMAT(120A1)     
      LREC=0
   10 CONTINUE    
      RETURN
      END   
C
C
C
      SUBROUTINE ITRANS(INPUT)
      CHARACTER BL,PU(5),PUFFER*5   
      DATA BL/' '/
      WRITE(PUFFER, '(I5)') INPUT   
      READ(PUFFER, '(5A1)') (PU(K), K=1,5)
      DO 10 K=1,5 
      IF (PU(K).EQ.BL) GOTO 10
      IN=ICHAR(PU(K))   
      CALL INBUF(IN)    
   10 CONTINUE    
      RETURN
      END   
C
C
C
      SUBROUTINE TRANSL(INPUT,N)    
      REAL INPUT  
      CHARACTER   BL,PU(11),PUFFER*11     
      DATA BL/' '/
      WRITE(PUFFER, '(F11.4)') INPUT
      READ(PUFFER, '(11A1)') (PU(K), K=1,11)    
      DO 10 K=1,6 
      IF (PU(K).EQ.BL) GOTO 10
      IN= ICHAR(PU(K))  
      CALL INBUF(IN)    
   10 CONTINUE    
      IF (N.GE.1) THEN  
      CALL INBUF(46)    
      DO 20 K=8,N+7     
      IN=ICHAR(PU(K))   
      CALL INBUF(IN)    
   20 CONTINUE    
      ENDIF 
      RETURN
      END   
*
************************************************************* 
*
      SUBROUTINE STR2WRL(FILE,TEMPO,NNPR2,V1)
C
C
C             STR2WRL originates from STRUVIR,
C
C             and STRUVIR originates from :
C
C
C ************************************************************
C ************************************************************
C ***                                                      ***
C ***                   STRUPLO'84                         ***
C ***                                                      ***
C ***             A FORTRAN PLOTPROGRAM                    ***
C ***                                                      ***
C ***             FOR CRYSTAL STRUCTURE                    ***
C ***                                                      ***
C ***                  ILLUSTRATIONS                       ***
C ***                                                      ***
C ***                                                      ***
C ***  AUTHOR: REINHARD X. FISCHER                         ***
C ***          INSTITUT FUER MINERALOGIE                   ***
C ***          DER UNIVERSITAET                            ***
C ***          AM HUBLAND                                  ***
C ***          D-8700 WUERZBURG, FRG                       ***
C ***                                                      ***
C ************************************************************
C ************************************************************
C
C              MORE OR LESS COMPLETE HISTORY :
C
C VERSION   1986  (6000 ATOMS, 1000 reduced to 500 POLYHEDRA)
C
C Written by R. X. Fischer (J. Appl. Cryst. (1985). 18, 258-262).
C Augmented by A. Le Lirzin, 1989 (COLO, PRES, BZRR)
C and by D. Kassner, 1990 (RES, NRES in PARA, incorporation of
C Burzlaff routine in SPGR, logic of VIEW and ROTA, addressing
C in SHAD, reorganization of COMMON-block structure).
C
C Received at MSU 17Aug92, via Tim McCarthy from Giaquinta at MIT.
C M.S.U. writeup by Donald L. Ward from the original by W. H. Baur.
C MSU version: 09-Sep-1992, Donald L. Ward
C
C Received at Laboratoire des Fluorures March 1996 from D.L. Ward
C
C Augmented by A. Le Bail, 1996,
C           Laboratoire des Fluorures, CNRS URA 449
C           Universite du Maine
C           72017 Le Mans Cedex
C           FRANCE
C           Email armel@fluo.univ-lemans.fr
C           http://fluo.univ-lemans.fr:8001/
C           April 1996
C                  HPGL output (STRUVIR.HGL)
C                  VRML output (PO.WRL and WI.WRL)
C
C Augmented by M. Koennnecke, 1996,
C                  POVRAY output (STRUPLO.POV)
C
C    PROGRAM STRUPLO	
      character*(*) file,tempo
      character*24 temp
      CHARACTER*28 CR
      CHARACTER TIT*40,BL*4
C      CHARACTER  IDATE*9, ITIME*8
      LOGICAL LOG,LOG2,LOG3,LOG4,QEX
      EQUIVALENCE (YC35,YC(35))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(48,3,4),
     *YF(900),MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
      COMMON/MODI/ID,PA1,PA2,AR1,AR2,OR1,OR2,RES,FAC,NRES,IBUR
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      COMMON/WRL/AID,NAID,CART,IO5,IO7,IO8,NCIRC,ICOL,NCOL,NPT
      DIMENSION IYA(4,6000)
      DIMENSION NAID(6000),CART(6000,3),NCIRC(6000),NCOL(20,2)
      DIMENSION ICOL(6000)
      INTEGER AID(6000,2)
      EQUIVALENCE (IYA(1,1),YA(1,1))
      PARAMETER (IYAQ=1000,IYAQ1=10)
*BER**DEC nicht definiert --> DEC=0 setzen***************************
        DEC=0.
*********************************************************************
C
*BER  CALL TINIT
C      WRITE(6,2000)
C2000  FORMAT('$Eingabefile: ')
C
C  MODIF ALB: initializations deplaced from the BLOCK DATA
C
      CALL ZERO
C
C     FILE  DECLARATIONS
C
      IO1=13
      IOK=IO4
      IO5=15
      IO7=16    
      IO8=17
      NSTOP=0
C
C      PRINT 6000
C6000  FORMAT(' Entry file (without .dat extension ) ? ',$)
C      READ 5000,file
C5000  format(A20)
      lfile=len(file)
      do while (file(lfile:lfile).eq.' ')
         lfile=lfile-1
      enddo
      temp=file(1:lfile)//'.dat'
      call open_read1(IO1,temp)
      temp=file(1:lfile)//'.imp'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.) GO TO 2010
      CALL FILEDEL(IO2,TEMP)
2010  OPEN (IO2,FILE=TEMP,FORM='FORMATTED',STATUS='NEW')
      lfile=len(TEMPO)
      do while (TEMPO(lfile:lfile).eq.' ')
         lfile=lfile-1
      enddo
      TEMP=TEMPO(1:LFILE)//'wi.wrl'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.) GO TO 2013
      CALL FILEDEL(IO5,TEMP)
2013  OPEN (IO5,FILE=TEMP,STATUS='NEW')
      TEMP=TEMPO(1:LFILE)//'po.wrl'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.) GO TO 2014
      CALL FILEDEL(IO7,TEMP)
2014  OPEN (IO7,FILE=TEMP,STATUS='NEW')
C
C     Get time/date information, write welcome message
C
C      CALL DATE (IDATE)
C      CALL TIME (ITIME)
C      WRITE (6, 2001) IDATE, ITIME
CESPOIR      WRITE (6, 2001)
2001  FORMAT (/,' STRUVIR - (version beta-3 September-96)')
*BER*******
C     OPEN (1,FILE='TEMP')

*BER*****************************************************************
CESPOIR      WRITE (6,2002)
C2002  FORMAT(' Output listing      : STRUVIR.IMP'/
C     *' HPGL file           : STRUVIR.HGL'/
C     *' File for PSPLOT.EXE : STRUPLO.XYZ'/
C     *' VRML wireframe file : WI.WRL'/
C     *'      polyhedra file : PO.WRL'/
C     *' POVRAY file         : STRUPLO.POV')
C
      CALL READ
      LOG2=MI(45).EQ.0
      LOG3=YC(29).LT.-10..OR.MI(45).NE.2
C
C *** CALCULATE ORTHOGONAL MATRIX ***
C
      YD(1)=SIN(YC(26)*YC35)
      YD(2)=SIN(YC(27)*YC35)
      YD(3)=SIN(YC(28)*YC35)
      YD(4)=COS(YC(26)*YC35)
      YD(5)=COS(YC(27)*YC35)
      YD(6)=COS(YC(28)*YC35)
      YD(7)=(YD(4)*YD(5)-YD(6))/(YD(1)*YD(2))
      YC(37)=YC(23)*YD(2)*SQRT(1.-YD(7)*YD(7))
      YC(38)=-YC(23)*YD(2)*YD(7)
      YC(39)=YC(24)*YD(1)
      YC(40)=YC(23)*YD(5)
      YC(41)=YC(24)*YD(4)
      YC(42)=YC(25)
C
C *** CALCULATE INVERSE MATRIX ***
C
      YC(48)=1./YC(42)
      YC(46)=1./YC(40)
      YC(43)=1./YC(37)
      YC(47)=-(YC(41)*YC(48))/YC(40)
      YC(44)=-(YC(38)*YC(43)*YC(46))
      YC(45)=-YC(43)*(YC(38)*YC(47)+YC(39)*YC(48))
      IF (PP(1).GT.-1E10) CALL VIEW
C
C *** CALCULATE RANGE FOR ATOMS OUTSIDE THE UNITCELL ***
C
    1 IF (LOG2)THEN
        DO 2 I=1,24,3
    2   CALL CORTH
        YD(1)=YC(56)*YC(60)-YC(57)*YC(59)
        YD(2)=YC(57)*YC(58)-YC(55)*YC(60)
        YD(3)=YC(55)*YC(59)-YC(56)*YC(58)
        YD(4)=YC(53)*YC(60)-YC(54)*YC(59)
        YD(5)=YC(54)*YC(58)-YC(52)*YC(60)
        YD(6)=YC(52)*YC(59)-YC(53)*YC(58)
        YD(7)=YC(53)*YC(57)-YC(54)*YC(56)
        YD(8)=YC(54)*YC(56)-YC(52)*YC(57)
        YD(9)=YC(52)*YC(56)-YC(53)*YC(55)
        YD(10)=SQRT(YD(1)*YD(1)+YD(2)*YD(2)+YD(3)*YD(3))
        YD(11)=SQRT(YD(4)*YD(4)+YD(5)*YD(5)+YD(6)*YD(6))
        YD(12)=SQRT(YD(7)*YD(7)+YD(8)*YD(8)+YD(9)*YD(9))
        C73=AMAX1(YC(9)+YC(10),YC(17)+YC(18))
        YD(4)=ABS(C73*YD(10)/(YD(1)*YC(37)+YD(2)*YC(38)+YD(3)*YC(40)))
        YD(7)=ABS(C73*YD(11)/(YD(5)*YC(39)+YD(6)*YC(41)))
        YD(8)=ABS(C73*YD(12)/(YD(9)*YC(42)))
        YC(1)=-YD(4)
        YC(3)=-YD(7)
        YC(5)=-YD(7)
        YC(2)=YD(4)+1.
        YC(4)=YD(7)+1.
        YC(6)=YD(8)+1.
        WRITE (IO2,1000) (YC(J),J=1,6)
      ELSE
        WRITE (IO2,1013) (YC(J),J=1,6)
      ENDIF
C
C *** SET -X,-Y,-Z OPERATOR ***
C
    4 IF (MI(44).EQ.1)THEN
        DO 6 I=1,MI(5)
        L=I+MI(5)
        DO 6 J=1,3
        DO 6 K=1,4
    6   YY(L,J,K)=-YY(I,J,K)
        MI(5)=L
        WRITE (IO2,1014)
      ELSE
        WRITE (IO2,1015)
      ENDIF
C
C *** PREPARE DO LOOP PARAMETERS ***
C
    8 DO 9 I=1,6
        MO(I)=IFIX(YC(I)-SIGN(1E-4,YC(I))+10.)
9     CONTINUE
C
C *** READ ATOMIC PARAMETERS ***
C
      DO 42 I=9,6000
      GOTO (10,18,19,21,20),MI(9)
   10 K=0
   11 READ (IO1,1004,END=43) (MN(J),J=1,80)
      IF (MN(1).EQ.MM( 1).AND.MN(2).EQ.MM( 1)) GOTO 11
      IF (MN(1).EQ.MM(41).AND.MN(2).EQ.MM(28)) GOTO 11
      IF (MN(1).EQ.MM(24).AND.MN(2).EQ.MM(34)) GOTO 11
      IF (MN(1).EQ.MM(27).AND.MN(2).EQ.MM(34)) GOTO 43
      IF (MN(1).NE.MM(21).OR. MN(2).NE.MM(40)) GOTO 13
      MN(3)=MM(1)
      MN(4)=MM(1)
      CALL READN
      DO 12 J=1,30
      IF (YD(J).LT.-1E20) GOTO 11
      K=K+1
   12 YF(K)=YD(J)
      GOTO 11
   13 DO 14 J=19,44
   14 IF (MN(1).EQ.MM(J)) GOTO 15
      GOTO 11
   15 MN(81)=MN(1)
      MN(82)=MN(2)
      CALL READN
      MN(3)=IFIX(YD(1))
      DO 17 J=3,5
      IF (YD(J).LE.15.)THEN
        YD(J-1)=YD(J)
      ELSE
        K=IFIX((YD(J)+5.)*0.1)
        YD(J-1)=(YD(J)-FLOAT(K*10))*YF(K)
      ENDIF
   17 CONTINUE
      MN(1)=MN(81)
      MN(2)=MN(82)
      GOTO 22
   18 READ (IO1,1002,END=43) (MN(J),J=1,3) , (YD(J),J=2,4)
C      WRITE(20,1002) (MN(J),J=1,3) , (YD(J),J=2,4)
      GOTO 22
   19 READ (IO1,1003,END=43) (MN(J),J=1,3) , (YD(J),J=2,4)
C      WRITE(20,1003) (MN(J),J=1,3) , (YD(J),J=2,4)
      GOTO 22
   20 READ (IO1, MH ,END=43) (MN(J),J=1,3) , (YD(J),J=2,4)
C      WRITE(20,MH) (MN(J),J=1,3) , (YD(J),J=2,4)
      GOTO 22
   21 READ (IO1,1004,END=43) (MN(J),J=1,80)
C      WRITE(20,1004) (MN(J),J=1,80)
      N1=MN(1)
      N2=MN(2)
      CALL READN
      MN(1)=N1
      MN(2)=N2
      MN(3)=IFIX(YD(1))
   22 IYA(1,MI(8)+1)=IPACK(1)
C
C *** STORE ATOM NAME ***
C
      IF (MI(24).GT.0)THEN
        DO 24 J=1,MI(24)
   24   IF (MN(3) .EQ. MI(J+24)) GOTO 26
      ENDIF
      MI(24)=MI(24)+1
      MI(MI(24)+24)=MN(3)
C
C *** GENERATE EQUIVALENT POSITIONS ***
C
   26 ITEMP=IYA(1,MI(8)+1)
      MN(2)=MI(8)
      DO 41 J=1,MI(5)
      DO 40 K=1,3*MI(4),3
      MI(8)=MI(8)+1
      IYA(1,MI(8))=ITEMP
      DO 27 L=1,3
      YA(L+1,MI(8))=YY(J,L,1)+YB(K+L+11)+
     *YY(J,L,2)*YD(2)+YY(J,L,3)*YD(3)+YY(J,L,4)*YD(4)
   27 CONTINUE
C
C *** REDUCE COORDINATES TO RANGE OF UNIT CELL ***
C
      DO 31 L=2,4
      YA(L,MI(8))=AMOD(YA(L,MI(8)),1.)
      IF(YA(L,MI(8)).LT.0.)YA(L,MI(8))=YA(L,MI(8))+1.
   31 CONTINUE
C	WRITE(IO2,*)'LIMITS',MI(8),(YA(KK,MI(8)),KK=2,4)
C
C *** CHECK SPECIAL POSITIONS ***
C
C      WRITE(IO2,*)'MI(5), MI(4), J, K = ',MI(5),MI(4),J,K
C      IF (J.EQ.1 .OR. K.GT.1) GOTO 36
      IF (J.EQ.1) GOTO 36
C      WRITE(IO2,*)' MN(2), MI(8)',MN(2),MI(8)
      DO 35 L=MN(2),MI(8)-1
      IF (ABS(YA(2,L) - YA(2,MI(8))-1.) .GE. 0.001)GO TO 351
      IF (ABS(YA(3,L) - YA(3,MI(8))) .GE. 0.001)GO TO 351
      IF (ABS(YA(4,L) - YA(4,MI(8))) .GE. 0.001)GO TO 351
C      WRITE(IO2,*)' ATOM suppressed: L,MI(8)= ',L,MI(8)
C	WRITE(IO2,*)'MI(8)',MI(8),(YA(KK,MI(8)),KK=2,4)
C	WRITE(IO2,*)'L    ',L,(YA(KK,L),KK=2,4)
      MI(8)=MI(8)-1
      GOTO 41
351   IF (ABS(YA(2,L) - YA(2,MI(8))) .GE. 0.001)GO TO 352
      IF (ABS(YA(3,L) - YA(3,MI(8))-1.) .GE. 0.001)GO TO 352
      IF (ABS(YA(4,L) - YA(4,MI(8))) .GE. 0.001)GO TO 352
C      WRITE(IO2,*)' ATOM suppressed: L,MI(8)= ',L,MI(8)
C	WRITE(IO2,*)'MI(8)',MI(8),(YA(KK,MI(8)),KK=2,4)
C	WRITE(IO2,*)'L    ',L,(YA(KK,L),KK=2,4)
      MI(8)=MI(8)-1
      GOTO 41
352   IF (ABS(YA(2,L) - YA(2,MI(8))) .GE. 0.001)GO TO 353
      IF (ABS(YA(3,L) - YA(3,MI(8))) .GE. 0.001)GO TO 353
      IF (ABS(YA(4,L) - YA(4,MI(8))-1.) .GE. 0.001)GO TO 353
C      WRITE(IO2,*)' ATOM suppressed: L,MI(8)= ',L,MI(8)
C	WRITE(IO2,*)'MI(8)',MI(8),(YA(KK,MI(8)),KK=2,4)
C	WRITE(IO2,*)'L    ',L,(YA(KK,L),KK=2,4)
      MI(8)=MI(8)-1
      GOTO 41
353   IF (ABS(YA(2,L) - YA(2,MI(8))) .GE. 0.001)GO TO 35
      IF (ABS(YA(3,L) - YA(3,MI(8))) .GE. 0.001)GO TO 35
      IF (ABS(YA(4,L) - YA(4,MI(8))) .GE. 0.001)GO TO 35
C      WRITE(IO2,*)' ATOM suppressed: L,MI(8)= ',L,MI(8)
C	WRITE(IO2,*)'MI(8)',MI(8),(YA(KK,MI(8)),KK=2,4)
C	WRITE(IO2,*)'L    ',L,(YA(KK,L),KK=2,4)
      MI(8)=MI(8)-1
      GOTO 41
   35 CONTINUE
C
C *** CHECK LIMITS FOR ALL ATOMS ***
C
   36 IF (YA(2,MI(8)) .LT. YC(1) .OR. YA(2,MI(8)) .GT. YC(2)) GOTO 37
      IF (YA(3,MI(8)) .LT. YC(3) .OR. YA(3,MI(8)) .GT. YC(4)) GOTO 37
      IF (YA(4,MI(8)) .LT. YC(5) .OR. YA(4,MI(8)) .GT. YC(6)) GOTO 37
      GOTO 38
   37 IYA(1,MI(8))=-ABS(IYA(1,MI(8)))
      GOTO 40
C
C *** CHECK LIMITS FOR CIRCLES ***
C
   38 IF (LOG3) GOTO 40
      IF (YA(2,MI(8)) .LT. YC(29) .OR. YA(2,MI(8)) .GT. YC(30)) GOTO 39
      IF (YA(3,MI(8)) .LT. YC(31) .OR. YA(3,MI(8)) .GT. YC(32)) GOTO 39
      IF (YA(4,MI(8)) .LT. YC(33) .OR. YA(4,MI(8)) .GT. YC(34)) GOTO 39
      GOTO 40
   39 IYA(1,MI(8))=(IYA(1,MI(8))/IYAQ1)*IYAQ1+5
   40 CONTINUE
   41 CONTINUE
   42 CONTINUE
C
C *** GENERATE ATOMS OUTSIDE THE UNIT CELL ***
C
   43 MI(36)=MI(8)
C
C -- Test on the number of atoms inside the cell...
C        if too much, return...
C
      MI8=MI(8)-8
      VM=FLOAT(MI8)
      ATV=VM/V1
      WRITE(IO2,*)'Volume =',V1,'Number of atoms inside =',MI8
      IF(ATV.GT.0.5)THEN
      WRITE(IO2,*)'That is too much !!'
      GO TO 1091
      ENDIF
C
C
      DO 47 I=9,MI(36)
      ITEMP=ABS(IYA(1,I)/IYAQ1)*IYAQ1
      YD(2)=YA(2,I)
      YD(3)=YA(3,I)
      YD(4)=YA(4,I)
      DO 47 J=MO(1),MO(2)
      DO 47 K=MO(3),MO(4)
      DO 47 L=MO(5),MO(6)
      IF (J.EQ.10 .AND. K.EQ.10 .AND. L.EQ.10) GOTO 47
      MI(8)=MI(8)+1
      IYA(1,MI(8))=ITEMP
      YA(2,MI(8))=YD(2)+FLOAT(J)-10.
      YA(3,MI(8))=YD(3)+FLOAT(K)-10.
      YA(4,MI(8))=YD(4)+FLOAT(L)-10.
C
C *** CHECK LIMITS ***
C
      IF (YA(2,MI(8)) .LT. YC(1) .OR. YA(2,MI(8)) .GT. YC(2)) GOTO 44
      IF (YA(3,MI(8)) .LT. YC(3) .OR. YA(3,MI(8)) .GT. YC(4)) GOTO 44
      IF (YA(4,MI(8)) .LT. YC(5) .OR. YA(4,MI(8)) .GT. YC(6)) GOTO 44
      GOTO 45
   44 MI(8)=MI(8)-1
      GOTO 47
C
C *** CHECK LIMITS FOR CIRCLES ***
C
   45 IF (LOG3) GOTO 47
      IF (YA(2,MI(8)) .LT. YC(29) .OR. YA(2,MI(8)) .GT. YC(30)) GOTO 46
      IF (YA(3,MI(8)) .LT. YC(31) .OR. YA(3,MI(8)) .GT. YC(32)) GOTO 46
      IF (YA(4,MI(8)) .LT. YC(33) .OR. YA(4,MI(8)) .GT. YC(34)) GOTO 46
      GOTO 47
   46 IYA(1,MI(8))=(IYA(1,MI(8))/IYAQ1)*IYAQ1+5
   47 CONTINUE
C
C *** OMIT ATOMS ***
C
      I=356
   48 IF (I.GE.MI(11))GO TO 54
      I=I+1
      IF (MN(I+1).LT.0)THEN
        MN(1)= MN(I)
        MN(2)=-MN(I+1)
        I=I+1
      ELSE
        MN(1)=MN(I)
        MN(2)=MN(I)
      ENDIF
      DO 53 J=MN(1),MN(2)
   53 IYA(1,J)=-IYA(1,J)
      GO TO 48
C
C *** ADD ATOMS ***
C
   54 IF (MI(2).GT.0)THEN
        DO 56 I=156,MI(10),4
        MI(8)=MI(8)+1
        IYA(1,MI(8))=ABS(IYA(1,MN(I)))
        YA(2,MI(8))=YA(2,MN(I))+FLOAT(MN(I+1))-5.
        YA(3,MI(8))=YA(3,MN(I))+FLOAT(MN(I+2))-5.
   56   YA(4,MI(8))=YA(4,MN(I))+FLOAT(MN(I+3))-5.
      ENDIF
C
C *** IF NO OUTLINES: ORIGIN ATOMS ARE NOT GRAPHED (INCREASING PLOTTER SCALE)
C
      IF(MI(48).EQ.0)THEN
        DO 57 I=1,8
   57   IYA(1,I)=-ABS(IYA(1,I))
      ENDIF
C
      YD( 7)= 1E5
      YD( 9)= 1E5
      YD( 8)=-1E5
      YD(10)=-1E5
      WRITE (IO2,1009)
      WRITE (IO2,1007)
C MODIF ALB
      NPT=MI(8)
C 
      K=0
      DO 66 I=1,MI(8)
      IF (I.EQ.9)THEN
        WRITE (IO2,1010)
        WRITE (IO2,1007)
      ENDIF
      IF (I.EQ.1+MI(36))THEN
        WRITE (IO2,1011)
        WRITE (IO2,1007)
        K=1
      ENDIF
      II=I
      CALL UNPACK(II)
C
C *** CALCULATE ORTHOGONAL COORDINATES ***
C
      YD(1)=YA(2,I)
      YD(2)=YA(3,I)
      YD(3)=YA(4,I)
      CALL CORTH
C
C *** ROTATE COORDINATES ***
C
      YA(2,I)=YB( 4)*YD(4)+YB( 5)*YD(5)+YB( 6)*YD(6)
      YA(3,I)=YB( 7)*YD(4)+YB( 8)*YD(5)+YB( 9)*YD(6)
      YA(4,I)=YB(10)*YD(4)+YB(11)*YD(5)+YB(12)*YD(6)
C
C     IF (I.LE.MI(36) .AND. I.GT.8 .AND. IO2.NE.6)
C    *  WRITE (IO2,1005) I,(MH(J),J=1,3),(YD(J),J=1,3),(YA(J,I),J=2,4)
C
C *** FIND MINIMUM AND MAXIMUM FOR SCALECALCULATION ***
C
      IF (IYA(1,I).GT.0)THEN
        YD( 7)=AMIN1(YD( 7),YA(2,I))
        YD( 8)=AMAX1(YD( 8),YA(2,I))
        YD( 9)=AMIN1(YD( 9),YA(3,I))
        YD(10)=AMAX1(YD(10),YA(3,I))
C
        DEC=AMAX1(DEC,YA(4,I))
C
        WRITE (IO2,1005) I,(MH(J),J=1,3),(YD(J),J=1,3),(YA(J,I),J=2,4)
C MODIF ALB
      AID(I,1)=MH(1)
      AID(I,2)=MH(2)
      NAID(I)=MH(3)
      CART(I,1)=YA(2,I)
      CART(I,2)=YA(3,I)
      CART(I,3)=YA(4,I)
C
      ELSE
        WRITE (IO2,1006) I,(MH(J),J=1,3),(YD(J),J=1,3),(YA(J,I),J=2,4)
C MODIF ALB
      AID(I,1)=MH(1)
      AID(I,2)=MH(2)
      NAID(I)=MH(3)
      CART(I,1)=YA(2,I)
      CART(I,2)=YA(3,I)
      CART(I,3)=YA(4,I)
C
      ENDIF
   66 CONTINUE
C MODIF ALB
C
C  Color attribution for VRML output
C
      J=1
      ICOL(9)=1
      NCOL(1,1)=AID(9,1)
      NCOL(1,2)=AID(9,2)
      DO 660 I=10,MI(8)
      JJ=J
      DO 658 K=1,JJ
      IF(AID(I,1).EQ.NCOL(K,1).AND.AID(I,2).EQ.NCOL(K,2))GO TO 670
658   CONTINUE
      J=J+1
      NCOL(J,1)=AID(I,1)
      NCOL(J,2)=AID(I,2)
670   CONTINUE
C      IF(J.GT.20)PRINT *,'More than 20 different atom names = STOP !'
      IF(J.GT.20)WRITE(20,*)'More than 20 different atom names = STOP !'
      IF(J.GT.20)RETURN
      DO 659 K=1,J
      IF(AID(I,1).EQ.NCOL(K,1).AND.AID(I,2).EQ.NCOL(K,2))ICOL(I)=K
659   CONTINUE
660   CONTINUE
CESPOIR      PRINT *,'Number of different atoms is ',J
C
C  End of color attribution
C
      DO 67 I=1,80
   67 MN(I)=-1
C
C *** MOVE ORIGIN TO CENTER OF THE PLOT ***
C
      YB(10)=YD( 7)
      YB(11)=YD( 8)
      YB(12)=YD( 9)
      YB(13)=YD(10)
      YB(18)=(YB(10)+YB(11))*0.5
      YB(19)=(YB(12)+YB(13))*0.5
      DO 500 I=1,MI(8)
      YA(2,I)=YA(2,I)-YB(18)
      YA(3,I)=YA(3,I)-YB(19)
  500 YA(4,I)=YA(4,I)-DEC
      IF (MI(7).LE.0)GO TO 80
C
C *** STORE ATOM NUMBERS TO BE GRAPHED AS CIRCLES ***
C
C *** A) GIVEN AS INTEGERS ***
C
      I=456
      J=1
      L=0
   69 I=I+1
      IF (MN(I).LE.0)THEN
        J=J+1
        GOTO 78
      ENDIF
      IF (MN(I+1).LE.0 .AND. MN(I+1).GE.-5000)THEN
        MN(1)=MN(I)
        MN(2)=-MN(I+1)
      ELSE
        MN(1)=MN(I)
        MN(2)=MN(I)
        J=J+1
      ENDIF
      I=I+1
      DO 77 K=MN(1),MN(2)
      IYA(1,K)=IYA(1,K)+J
      L=L+1
      IF (L.GE.3000)THEN
        WRITE (IO3,1000)
        WRITE (IO2,1000)
        RETURN
      ENDIF
   77 MO(L)=K
   78 IF (I-MI(7).LT.456)GO TO 69
      MI(7)=L
   80 IF (MI(38).LE.0)GO TO 86
C
C *** B) GIVEN AS ATOM NAMES ***
C
      DO 85 I=9,MI(8)
C MODIF ALB
      NCIRC(I)=0
C
      IF (IYA(1,I).LE.0.)GO TO 85
      MN(1)=IYA(1,I)/IYAQ1
      MN(2)=IYA(1,I)/IYAQ
      IF (IYA(1,I).NE.MN(1)*IYAQ1) GOTO 85
      DO 83 J=557,MI(38)+556
      MN(3)=MN(J)/100
      IF (MN(3).EQ.MN(2))THEN
        MI(7)=MI(7)+1
        MO(MI(7))=I
C  MODIF ALB
C  STRUVIR MARK ON CIRCLES TO BECOME SPHERES
               NCIRC(I)=I
C
        IYA(1,I)=IYA(1,I)+MN(J)-MN(3)*10*IYAQ1
        GO TO 85
      ENDIF
   83 CONTINUE
   85 CONTINUE
   86 IF (IABS(MI(7))+IABS(MI(38)).LE.0)GO TO 91
C
C *** SORT CIRCLES ***
C
      DO 90 I=1,MI(7)-1
      K=0
      YD(1)=DIST(MO(I))
      DO 88 J=I+1,MI(7)
      IF(YD(1).GE.DIST(MO(J)))THEN
        YD(1)=DIST(MO(J))
        K=J
      ENDIF
   88 CONTINUE
      IF (K.GT.0)THEN
        L    =MO(I)
        MO(I)=MO(K)
        MO(K)=L
      ENDIF
   90 CONTINUE
   91 KK=0
      LL=0
      DO 94 I=9,MI(8)
      IF(I.EQ.MI(36)+1) KK=0
      IF (MI(39)+MI(40).EQ.0) GOTO 94
C
C *** CALCULATE INTERATOMIC DISTANCES AND ANGLES ***
C
      J=I
      LOG=J.GT.MI(36)
      IF (LOG.AND.LOG2) GOTO 95
      LOG4=IABS(IYA(1,I)/IYAQ1).EQ.IABS(IYA(1,I-1)/IYAQ1)
      LOG=LOG4.OR.LOG
      DO 203 JJ=1,5
      IF(IABS(IYA(1,I)/IYAQ).EQ.IFIX(YP(JJ,1)))THEN
        IF(.NOT.LOG4) KK=KK+1
        IF(LOG.AND.IYA(1,I).LE.0) GOTO 94
        LL=LL+1
C MODIF ALB: no more than 500 zarbi polyhedra
      IF(LL.GT.500)WRITE(20,1050)
      IF(LL.GT.500)RETURN
C
        MV(LL,13)=I
        MV(LL,14)=KK
        GOTO 207
      ENDIF
  203 CONTINUE
  207 IF (.NOT.LOG)THEN
        CALL UNPACK(J)
        WRITE (IO2,1008) J,MH(1),MH(2),MH(3)
      ENDIF
      CALL ANDI(J,II,NN,LOG)
      IF(NSTOP.EQ.1)RETURN
*BER**
      IF(LL.LE.0) GOTO 208
      IF(MV(LL,13).EQ.I)THEN
        CALL FIND(LL,NN,JJ,LOG)
        IF(NSTOP.EQ.1)RETURN
        GOTO 94
      ENDIF
208   CONTINUE
C
C *** SEARCH POLYHEDRA ***
C
      IF (IYA(1,I)     .LE.0. ) GOTO 94
      IF (IYA(1,I)/IYAQ.EQ.459) GOTO 94
      IF (IYA(1,I)/IYAQ.EQ. 81) GOTO 94
      CALL SEARCH(J,II,MAXPOL)
C
C  Inelegant solution if more than 500 polyhedra
C
      IF(MAXPOL.EQ.1)GO TO 1091
C
   94 CONTINUE
C
C   95 IF ((MI(37)+LL).GT.0)CALL WRITE(LL)
   95 CALL WRITE(LL,TEMPO)
C
C  MODIF ALB
CESPOIR        WRITE(6,961)
961     FORMAT(' Do you want .HGL and .XYZ output (Yes=0/No=1) ?')
CESPOIR        READ(5,*)ITEST
      ITEST=0
C	IF(ITEST.EQ.1) RETURN
      lfile=len(TEMPO)
      do while (TEMPO(lfile:lfile).eq.' ')
         lfile=lfile-1
      enddo
      TEMP=TEMPO(1:LFILE)//'ps.scr'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.) GO TO 2011
      CALL FILEDEL(9,TEMP)
2011  OPEN (9,FILE=TEMP,STATUS='NEW')
      TEMP=TEMPO(1:LFILE)//'stru.hgl'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.) GO TO 2012
      CALL FILEDEL(IO4,TEMP)
2012  OPEN (IO4,FILE=TEMP,STATUS='NEW')
C
      YB(14)=ABS(YB(10)-YB(11))
      YB(15)=ABS(YB(12)-YB(13))
      YB(16)=YB(14)*0.5
      YB(17)=YB(15)*0.5
      YB(18)=(YB(10)+YB(11))*0.5
      YB(19)=(YB(12)+YB(13))*0.5
      IF (MI(46).LE.0)GO TO 109
C
C *** GENERATE PLOT ***
C
C
C *** SET FRAME ***
C
      PA1=MAX(PA1,AR1)
      PA2=MAX(PA2,AR2)
*BER  CALL FRAME(PA1,PA2)
C     CALL PLOT(0.0,0.0,3)
C     CALL PLOT(AR1,0.0,2)
C     CALL PLOT(AR1,AR2,2)
C     CALL PLOT(0.0,AR2,2)
C     CALL PLOT(0.0,0.0,2)
C
C *** PLOT TITLE ***
C
*BER  IF(TIT.NE.' ')THEN
*BER    CALL SYMBOL(1.,PA2-.7,0.4,TIT,0.,40)
*BER    AR2=MIN(AR2,PA2-1.)
*BER  ENDIF
C
C *** CALCULATE SCALE FACTOR ***
C
C     YC(75)=AMIN1(AR1/YB(14),AR2/YB(15))*FAC*0.9
      YC(75)=AMIN1(AR1/YB(14),AR2/YB(15))*FAC
      C75=AMIN1(1024./YB(14),750./YB(15))*FAC*0.9
      YC(77)=PA1*0.5
      YC(78)=PA2*0.5
C
*BER**WRITE SCALE FACTOR TO STRUPLO.XYZ******
      WRITE(9,7771) YC(75)   
7771  FORMAT(F10.4)
***************************************************
C  
       WRITE (IO2,1016) PA1,PA2,YC(75),YC(77)+OR1,YC(78)+OR2,
     * (YB(10)-YB(18))*YC(75),(YB(11)-YB(18))*YC(75),
     * (YB(12)-YB(19))*YC(75),(YB(13)-YB(19))*YC(75),
     *  YB(14)        *YC(75), YB(15)        *YC(75)
C
*BER**Titel, Frame auf Datei 9 schreiben****************
       WRITE(9,1030) TIT
       WRITE(9,1031) YC(77),YC(78)
1030   FORMAT(A80)
1031   FORMAT(2(F10.4,1X))
**********************************************************************
C
C *** SET ORIGIN ***
C
*     CALL PLOT(YC(77)+OR1,YC(78)+OR2,-3)
C
C *** SET SCALE FACTOR
C
*BER  CALL FACTOR(YC(75))
C
C *** START PLOT ***
C
*BER  CALL TIMER(1,LAST_TIMER)
      CALL DRAW
      IF(NSTOP.EQ.1)GO TO 1091
  109 CALL HPOFF
      CLOSE(9)
      CALL PSPLOT(TEMPO)
1091  CLOSE(IO1)
      CLOSE(IO2)
      CLOSE(IO4)
      CLOSE(7)
      CLOSE(IO5)
      CLOSE(IO7)
      RETURN
 1000 FORMAT (///,1X,'DEFAULT RANGE FOR GENERATING ATOMS IS X:',F8.4,1X,
     &'TO',F7.4,/,39X,'Y:',F8.4,' TO',F7.4,/,39X,'Z:',F8.4,' TO',F7.4)
C1001 FORMAT (2A1,I2,5X,3F10.5)
 1002 FORMAT (2A1,I2,26X,3F10.6)
 1003 FORMAT (2A1,I2,3F10.6)
 1004 FORMAT (80A1)
 1005 FORMAT ('(',I4,')',1X,2A1,I3,3F10.5,1X,3F10.5)
 1006 FORMAT ('(',I4,')',1X,2A1,I3,3F10.5,1X,3F10.5,1X,'NOT G.')
 1007 FORMAT ('NO',6X,'ATOM',3X,'X/A',7X,'Y/B',7X,'Z/C',8X,'XO',8X,
     &'YO',8X,'ZO',/,'--------------------------------------------------
     &-----------------------------')
 1008 FORMAT (///,1X,'DISTANCES TO ATOM (',I4,')',1X,2A1,I2,//,3X,'NO',
     &5X,'NAME',10X,'DIST.'/,1X,'-----------------------------')
 1009 FORMAT (///,1X,'ORIGIN SPECIFICATION',//)
 1010 FORMAT (///,1X,'FRACTIONAL AND ORTHOGONAL COORDINATES OF ATOMS INS
     &IDE THE UNIT CELL',//)
 1011 FORMAT (///,1X,'FRACTIONAL AND ORTHOGONAL COORDINATES OF ATOMS OUT
     &SIDE THE UNIT CELL',//)
 1013 FORMAT (///,1X,'GIVEN RANGE FOR GENERATING ATOMS IS X:',F8.4,1X,
     &'TO',F7.4,/,37X,'Y:',F8.4,' TO',F7.4,/,37X,'Z:',F8.4,' TO',F7.4)
 1014 FORMAT (///,1X,'THE CRYSTAL STRUCTURE IS CENTROSYMMETRIC WITH CENT
     &RE IN THE ORIGIN')
 1015 FORMAT (///,1X,'THERE IS NO CENTRE IN THE ORIGIN OF THE CRYSTAL ST
     &RUCTURE')
 1016 FORMAT (///' THE PLOT IS SCALED INTO A FRAME OF X=',F8.2,
     *'CM AND Y=',F8.2,'CM'/
     *' THE SCALE FACTOR IS:',F7.3/
     *' THE ORIGIN IS SHIFTED TO X=',F7.2,'CM AND Y=',F7.2,'CM'//
     *' PARAMETERS AFTER SCALING ARE:'/
     *' MINIMUM AND MAXIMUM VALUES IN X=',F6.2,'CM AND',F6.2,'CM'/
     *' MINIMUM AND MAXIMUM VALUES IN Y=',F6.2,'CM AND',F6.2,'CM'/
     *' LENGTH AND WIDTH OF THE DRAWING ARE X=',F7.2,'CM AND Y=',F7.2)
 1050 FORMAT ('  No more than 500 bizarre polyhedra is allowed')
C1017 FORMAT (1X,'4 ',I1,';')
C1018 FORMAT (1X,'ONLY 3000 ATOMS CAN BE REPRESENTED AS CIRCLES, JOB IS
C    &TERMINATED')
C1020 FORMAT (1X,'VS ',F4.2,';')
C1026 FORMAT (1X,4(I6,' '))
C1027 FORMAT (1X,4(I4,' '))
C1028 FORMAT (1X,5(I6,' '))
      END
*BER**Function AMAX1 und AMIN1 definieren****************************
*
      FUNCTION AMAX1(WERT1,WERT2)
      IF(WERT1.GT.WERT2) THEN
        AMAX1=WERT1
      ELSE
        AMAX1=WERT2
      ENDIF
      RETURN
      END
*
      FUNCTION AMIN1(WERT1,WERT2)
      IF(WERT1.LT.WERT2) THEN
        AMIN1=WERT1
      ELSE
        AMIN1=WERT2
      ENDIF
      RETURN
      END
*********************************************************************
      SUBROUTINE ANDI(I,II,L,LOG)
      EQUIVALENCE (YC36,YC(36))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(106),YZ(100,5),
     *YE(200,5),MH(20),MI(11198),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      LOGICAL LOG
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
C
      L=0
      MH(17)=MH(1)
      MH(18)=MH(2)
      MH(19)=MH(3)
      DO 4 J=9,MI(8)
      IF (IYA(1,J).LE.0 .AND. LOG) GOTO 4
      IF (J.EQ.I) GO TO 4
      DO 3 K=2,4
    3 IF (ABS(YA(K,I)-YA(K,J)).GT.YC(89)) GOTO 4
      L=L+1
      YZ(L,1)=FLOAT(J)
      YZ(L,2)=YA(2,I)-YA(2,J)
      YZ(L,3)=YA(3,I)-YA(3,J)
      YZ(L,4)=YA(4,I)-YA(4,J)
      YZ(L,5)=SQRT(YZ(L,2)*YZ(L,2)+
     *             YZ(L,3)*YZ(L,3)+
     *             YZ(L,4)*YZ(L,4))
      IF (YZ(L,5).GT.YC(89)) L=L-1
C  -- Out if L = 30...
      IF(L.GE.30)GO TO 500
C
    4 CONTINUE
500   II=0
      CALL SORT1(L)
      IF (LOG) GOTO 6
      DO 5 I1=1,L
      I2=I1
      MH(13)=IFIX(YZ(I1,1))
      CALL UNPACK(MH(13))
    5 WRITE (IO2,1001) MH(13),(MH(I2),I2=1,3),YZ(I1,5)
      WRITE (IO2,1002) I,MH(17),MH(18),MH(19)
    6 DO 13 J=1,L-1
      IF (YZ(J,5).GT.YC(90)) GO TO 13
      DO 12 K=J+1,L
      IF (YZ(K,5).GT.YC(90)) GO TO 13
      II=II+1
      IF(II.GT.200.OR.II.LE.0)THEN
      NSTOP=1
      GO TO 600
      ENDIF
      IF(J.GT.100.OR.J.LE.0)THEN
      NSTOP=1
      GO TO 600
      ENDIF
      IF(K.GT.100.OR.K.LE.0)THEN
      NSTOP=1
      GO TO 600
      ENDIF
      YE(II,1)=YZ(J,1)
      YE(II,2)=YZ(K,1)
      MH(8)=IFIX(YE(II,2))
      CALL UNPACK(MH(8))
      MH(4)=MH(1)
      MH(5)=MH(2)
      MH(6)=MH(3)
      MH(7)=IFIX(YE(II,1))
      CALL UNPACK(MH(7))
      IF (ABS(YZ(J,5)*YZ(K,5)).GE.1E-8)THEN
        ARCO=(YZ(J,2)*YZ(K,2)+YZ(J,3)*YZ(K,3)+YZ(J,4)*YZ(K,4))/
     &  (YZ(J,5)*YZ(K,5))
        IF(ARCO.GT. 1.)ARCO= 1.
        IF(ARCO.LT.-1.)ARCO=-1.
      ELSE
        ARCO=1.
        WRITE (IO2,1003)
      ENDIF
      YE(II,3)=ACOS(ARCO)*YC36
      YE(II,4)=YZ(J,5)
      YE(II,5)=YZ(K,5)
      IF (LOG) GOTO 12
      YB(1)=SQRT((YA(2,MH(7))-YA(2,MH(8)))**2+
     *           (YA(3,MH(7))-YA(3,MH(8)))**2+
     *           (YA(4,MH(7))-YA(4,MH(8)))**2)
C MODIF ALB : T format gave problems on my PC...
C     WRITE (IO2,1004) (MH(I1),I1=1,8) , (YE(II,I1),I1=3,5) , YB(1)
      WRITE (IO2,1004)MH(7),MH(1),MH(2),MH(3),MH(8),MH(4),MH(5),MH(6),
     &(YE(II,I1),I1=3,5) , YB(1)
C
      IF(IYA(1,IFIX(YE(II,1))).LT.0 .OR.
     *   IYA(1,IFIX(YE(II,2))).LT.0) II=II-1
C
C --  Out if II = 30, enough....
C
      IF(II.GE.30)GO TO 600
   12 CONTINUE
   13 CONTINUE
600   CONTINUE
      RETURN
 1001 FORMAT (1X,'(',I4,')',3X,2A1,I3,5X,F10.5)
 1002 FORMAT (///,1X,'DISTANCES AND ANGLES AROUND CENTRAL ATOM CA (',I4,
     &')',1X,2A1,I2,//,3X,'ATOM 1',13X,'ATOM 2',7X,'ANGLE',6X,'A1-CA',5X
     &,'A2-CA',5X,'A1-A2',/,'-------------------------------------------
     &-----------------------------')
 1003 FORMAT (' !!! WARNING !!! THE ARGUMENT IN ACOS IS GREATER THAN',
     *' 1., THE ANGLE IS SET TO 0.0,'/
     *' CHECK DISTANCE TABLE FOR DOUBLE ENTRY')
C 1004 FORMAT (1X,T1,'(',T8,2A1,I2,T20,'(',T27,2A1,I2,T2,I4,')'
C     &,T21,I4,')',T14,'-CA-',T34, 4F10.5)
 1004 FORMAT (1X,'(',I4,')',1X,2A1,I2,2X,'-CA-',2X,'(',I4,')',
     &2A1,I2,4X,4F10.5)
      END
C-----------------------------------------------------------------------
      SUBROUTINE ZERO
      CHARACTER TIT*40,BL*4
      CHARACTER MM*4
      CHARACTER*28 CR
      INTEGER BX(27)
      DIMENSION IYA(4,6000)
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(1626),MI(104),MM(226),
     *MZ2(10844),MZ(24),MU(14735),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      COMMON/MODI/ID,PA1,PA2,AR1,AR2,OR1,OR2,RES,FAC,NRES,IBUR
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
      COMMON/BZR/BX
      EQUIVALENCE(IYA,YA)
C
C DATA for STRUVIR
C
      MM(1)=' '
      MM(2)='+'
      MM(3)='0'
      MM(4)='1'
      MM(5)='2'
      MM(6)='3'
      MM(7)='4'
      MM(8)='5'
      MM(9)='6'
      MM(10)='7'
      MM(11)='8'
      MM(12)='9'
      MM(13)='.'
      MM(14)='-'
      MM(15)=';'
      MM(16)=','
      MM(17)='('
      MM(18)=')'
      MM(19)='P'
      MM(20)='I'
      MM(21)='F'
      MM(22)='C'
      MM(23)='B'
      MM(24)='A'
      MM(25)='R'
      MM(26)='D'
      MM(27)='E'
      MM(28)='G'
      MM(29)='H'
      MM(30)='J'
      MM(31)='K'
      MM(32)='L'
      MM(33)='M'
      MM(34)='N'
      MM(35)='O'
      MM(36)='Q'
      MM(37)='S'
      MM(38)='T'
      MM(39)='U'
      MM(40)='V'
      MM(41)='W'
      MM(42)='X'
      MM(43)='Y'
      MM(44)='Z'
      MM(45)='TI'
      MM(46)='CE'
      MM(47)='SY'
      MM(48)='SP'
      MM(49)='FO'
      MM(50)='AD'
      MM(51)='OM'
      MM(52)='RO'
      MM(53)='TE'
      MM(54)='OC'
      MM(55)='CI'
      MM(56)='XY'
      MM(57)='BO'
      MM(58)='EN'
      MM(59)='PA'
      MM(60)='FI'
      MM(61)='MO'
      MM(62)='VI'
      MM(63)='SH'
      MM(64)='CO'
      MM(65)='PE'
      MM(66)='BZ'
      MM(67)='  '
      MM(68)='  '
      MM(69)='  '
      MM(70)='  '
      DO 20 I=71,226
20    MM(I)=' '
      MZ(1)=1
      MZ(2)=2
      MZ(3)=5
      MZ(4)=3
      MZ(5)=1
      MZ(6)=4
      MZ(7)=6
      MZ(8)=8
      MZ(9)=7
      MZ(10)=2
      MZ(11)=5
      MZ(12)=3
      MZ(13)=2
      MZ(14)=5
      MZ(15)=3
      MZ(16)=1
      MZ(17)=4
      MZ(18)=6
      MZ(19)=8
      MZ(20)=7
      MZ(21)=4
      MZ(22)=6
      MZ(23)=8
      MZ(24)=7
      DO 18 I=1,27
18    BX(I)=0
      CR='(C) REINHARD X. FISCHER 1986'
      BL='    '
      IO1=13
      IO2=4
      IO3=20
      IO4=7
      DO 14 I=1,104
14    MI(I)=1
      MI(1)=0
      MI(2)=0
      MI(3)=0
      MI(6)=0
      MI(7)=0
      MI(8)=8
      MI(9)=4
      MI(10)=156
      MI(11)=356
      DO 15 I=12,24
15    MI(I)=0
      MI(25)=4403
      DO 16 I=26,33
16    MI(I)=0
      DO 17 I=35,38
17    MI(I)=0
      MI(42)=0
      MI(43)=0
      MI(45)=0
      MI(49)=2
      MI(50)=2
      MI(51)=2
      MI(52)=2
      MI(53)=7
      MI(54)=7
      MI(55)=7
      MI(56)=7
      MI(57)=2
      MI(58)=2
      MI(59)=3
      MI(60)=2
      MI(61)=3
      MI(62)=3
      MI(63)=4
      MI(64)=5
      MI(65)=3
      MI(66)=3
      MI(67)=5
      MI(68)=4
      MI(69)=3
      MI(70)=3
      MI(71)=4
      MI(72)=4
      MI(73)=5
      MI(74)=4
      MI(75)=6
      MI(76)=6
      MI(77)=4
      MI(78)=5
      MI(79)=6
      MI(80)=6
      MI(81)=4
      MI(82)=5
      MI(83)=5
      MI(84)=5
      DO 13 I=1,6
13    YC(I)=-1E30
      YC(7)=109.
      YC(8)=20.
      YC(9)=1.62
      YC(10)=0.2
      YC(11)=15.
      DO 12 I=12,1626
12    YC(I)=0.
      YC(15)=90.
      YC(16)=20.
      YC(17)=1.97
      YC(18)=0.4
      YC(19)=15.
      YC(23)=10.
      YC(24)=10.
      YC(25)=10.
      YC(26)=90.
      YC(27)=90.
      YC(28)=90.
      YC(29)=-1.
      YC(30)=-1.
      YC(31)=-1.
      YC(32)=-1.
      YC(33)=-1.
      YC(34)=-1.
      YC(35)=1.745329E-02
      YC(36)=57.29578
      YC(37)=1.
      YC(40)=1.
      YC(42)=1.
      YC(43)=1.
      YC(46)=1.
      YC(48)=1.
      YC(52)=1.
      YC(56)=1.
      YC(60)=1.
      YC(61)=1.
      YC(62)=1.
      YC(64)=1.
      YC(66)=1.
      YC(68)=1.
      YC(69)=1.
      YC(70)=1.
      YC(71)=1.
      YC(72)=1.
      YC(73)=28.8
      YC(74)=18.7
      YC(75)=1.
      YC(76)=100.
      YC(77)=40.
      YC(78)=15.
      DO 10 I=79,88
10    YC(I)=1.
      YC(89)=3.2
      YC(90)=4.
      DO 11 I=91,120
11    YC(I)=-1E30
      YC(275)=1.
      YC(467)=1.
      YC(659)=1.
      DO 9 I=1,25
9     YB(I)=0.
      YB(4)=1.
      YB(8)=1.
      YB(12)=1.
      YB(25)=1E30
      DO 8 I=1,70
8     YP(I)=0.
      DO 7 I=1,50
      DO 7 J=1,3
7     NH(I,J)=0
      NH(1,1)=1
      NH(1,2)=3000
      NH(1,3)=9
      IS=1
      DO 6 I=1,10
6     PP(I)=-1E30
      ID=1
      PA1=0.
      PA2=0.
      AR1=28.8
      AR2=21.6
      OR1=0.
      OR2=0.
      RES=0.01
      FAC=1.
      NRES=0
      IBUR=1
      DO 1 I=1,14735
1     MU(I)=0
      DO 2 I=1,300
2     MZ2(I)=0
      DO 3 I=301,7844
3     MZ2(I)=-1
      DO 4 I=7845,10844
4     MZ2(I)=0
      DO 5 I=1,4
      DO 5 J=1,6000
5     YA(I,J)=0.
      YA(1,1)=439010.
      YA(1,2)=439020.
      YA(2,2)=1.
      YA(1,3)=439030.
      YA(3,3)=1.
      YA(1,4)=439040.
      YA(4,4)=1.
      YA(1,5)=439050.
      YA(2,5)=1.
      YA(3,5)=1.
      YA(1,6)=439060.
      YA(2,6)=1.
      YA(4,6)=1.
      YA(1,7)=439070.
      YA(3,7)=1.
      YA(4,7)=1.
      YA(1,8)=439080.
      YA(2,8)=1.
      YA(3,8)=1.
      YA(4,8)=1.
      RETURN
      END
C-----------------------------------------------------------------------------
C-----------------------------------------------------------------------------
      SUBROUTINE BZRR(JJ,II)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(106),YZ(100,5),
     *YE(200,5),MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/BZR/BX,BY,BZ
      INTEGER BY(13),BZ(4),BX(10)
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DIMENSION W(30)
C      LOGICAL IND
      LOGICAL LOG
      I=1
      IF(JJ.EQ.1) GOTO 41
      DO 42 I=1,JJ-1
      IF(DIST(MV(I,13)).LE.DIST(MV(JJ,13))) GOTO 42
      DO 43 M=1,14
      DO 44 K=JJ,I,-1
   44 MV(K+1,M)=MV(K,M)
      MV(I,M)=MV(JJ+1,M)
   43 MV(JJ+1,M)=0
      GOTO 41
   42 CONTINUE
   41 N=I
   45 W(27)=0
      W(28)=0
      W(29)=0
      DO 1 J=1,3
      IF(MV(N,J).LE.0)GO TO 1
      W(27)=W(27)+YA(2,MV(N,J))*(1./3.)
      W(28)=W(28)+YA(3,MV(N,J))*(1./3.)
      W(29)=W(29)+YA(4,MV(N,J))*(1./3.)
    1 MU(MV(N,14),2,J)=J
      MU(MV(N,14),2,7)=3
      MU(MV(N,14),1,3)=1
      W(2)=YA(2,MV(N,13))
      W(3)=YA(3,MV(N,13))
      W(4)=YA(4,MV(N,13))
      DO 2 I=4,MIN(II,12)
      IF(MV(N,I).EQ.0) GOTO 37
      M=0
      W(5)=YA(2,MV(N,I))
      W(6)=YA(3,MV(N,I))
      W(7)=YA(4,MV(N,I))
      DO 3 J=2,MU(MV(N,14),1,3)+1
      DO 5 L=2,4
      W(1)=YA(L,MV(N,MU(MV(N,14),J,1)))
      DO 4 K=2,3
C *** W de 8 a 13 definis comme 2-1 et 3-1 ***
    4 W(3*K+L)=YA(L,MV(N,MU(MV(N,14),J,K)))-W(1)
      W(12+L)=W(3+L)-W(1)
      W(15+L)=W(25+L)-W(1)
    5 CONTINUE
      W(20)=W( 9)*W(13)-W(10)*W(12)
      W(21)=W(10)*W(11)-W( 8)*W(13)
      W(22)=W( 8)*W(12)-W( 9)*W(11)
C *** W(23) represente le det (2-1,3-1,i-1) ****
C *** W(24) represente le det avec le barycentre ***
      W(23)=W(20)*W(14)+W(21)*W(15)+W(22)*W(16)
      IF(MU(MV(N,14),1,3).NE.1) GOTO 202
      IF(ABS(W(23)).GE.1E-4) GOTO 203
      DO 300 K=1,MU(MV(N,14),J,7)
      L=MOD(K,MU(MV(N,14),J,7))+1
      M=MOD(K-2+MU(MV(N,14),J,7),MU(MV(N,14),J,7))+1
      DO 301 NN=2,4
      W(6+NN)=YA(NN,MV(N,MU(MV(N,14),J,L)))
     *       -YA(NN,MV(N,MU(MV(N,14),J,K)))
      W(9+NN)=YA(NN,MV(N,MU(MV(N,14),J,M)))
     *       -YA(NN,MV(N,MU(MV(N,14),J,K)))
  301 W(12+NN)=W(3+NN)-YA(NN,MV(N,MU(MV(N,14),J,K)))
      W(17)=W( 9)*W(13)-W(10)*W(12)
      W(18)=W(10)*W(11)-W( 8)*W(13)
      W(19)=W( 8)*W(12)-W( 9)*W(11)
      W(11)=W( 9)*W(16)-W(10)*W(15)
      W(12)=W(10)*W(14)-W( 8)*W(16)
      W(13)=W( 8)*W(15)-W( 9)*W(14)
      W(20)=W(17)*W(11)+W(18)*W(12)+W(19)*W(13)
      IF(W(20).LT.0) GOTO 302
  300 CONTINUE
      WRITE (IO2,*) 'impossible d inserer un atome a une face plate'
      WRITE (IO3,*) 'impossible d inserer un atome a une face plate'
      RETURN
  302 DO 303 M=MU(MV(N,14),J,7),L,-1
  303 MU(MV(N,14),J,M+1)=MU(MV(N,14),J,M)
      MU(MV(N,14),J,L)=I
      MU(MV(N,14),J,7)=MU(MV(N,14),J,7)+1
      MU(MV(N,14),1,2)=4
      GOTO 100
  203 DO 204 K=1,MU(MV(N,14),2,7)
  204 BX(K)=1
      BZ(1)=2
      CALL MERGE(MV(N,14))
      CALL SPLIT(MV(N,14),I,.TRUE.)
      IF(NSTOP.EQ.1)RETURN
      GOTO 100
  202 IF(ABS(W(23)).LE.1E-4) GOTO 25
      W(24)=W(20)*W(17)+W(21)*W(18)+W(22)*W(19)
      W(23)=W(23)*W(24)
      IF(W(23).GE.0) GOTO 3
   25 M=M+1
      IF(M.LE.0.OR.M.GT.4)THEN
      NSTOP=1
      RETURN
      ENDIF
      BZ(M)=J
    3 CONTINUE
      LOG=.FALSE.
      CALL MERGE(MV(N,14))
      IF(BY(11).NE.0) GOTO 747
      RETURN
  747 DO 30 J=1,BY(11)
      K=MOD(J-2+BY(11),BY(11))+1
      L=MOD(J,BY(11))+1
      DO 31 M=2,4
      W(1  )=YA(M,MV(N,BY(J)))
      W(6+M)=YA(M,MV(N,BY(K)))-W(1)
      W(9+M)=YA(M,MV(N,BY(L)))-W(1)
   31 W(12+M)=W(3+M)-W(1)
      DO 35 M=2,4
   35 IF(ABS(W(9+M)-W(6+M)).GE.1E-8) GOTO 36
      W(30)=(W(5)-W(2))**2+(W(6)-W(3))**2+(W(7)-W(4))**2
      MU(MV(N,14),1,1)=IFIX(SQRT(W(30))*999)
      GOTO 37
   36 W(21)=W(9)*W(13)-W(10)*W(12)
      W(22)=W(10)*W(11)-W(8)*W(13)
      W(23)=W(8)*W(12)-W(9)*W(11)
      W(24)=W(21)*W(14)+W(22)*W(15)+W(23)*W(16)
      IF(ABS(W(24)).LE.1E-4) GOTO 32
      BX(J)=1
      GOTO 30
   32 W(24)=W( 9)*W(16)-W(10)*W(15)
      W(25)=W(10)*W(14)-W( 8)*W(16)
      W(26)=W( 8)*W(15)-W( 9)*W(14)
      W( 1)=W(21)*W(24)+W(22)*W(25)+W(23)*W(26)
      IF(W(1).GE.0) GOTO 30
      BX(J)=1
      BX(K)=1
C **** 34 remplace par 30 *******
   30 CONTINUE
      CALL SPLIT(MV(N,14),I,LOG)
  100 DO 50 K=1,10
   50 BX(K)=0
      DO 51 K=1,4
   51 BZ(K)=0
      DO 52 K=1,13
   52 BY(K)=0
      W(27)=((I-1)*W(27)+W(5))/I
      W(28)=((I-1)*W(28)+W(6))/I
      W(29)=((I-1)*W(29)+W(7))/I
    2 CONTINUE
   37 MU(MV(N,14),1,2)=I-1
      RETURN
      END
C-----------------------------------------------------------------------------
      SUBROUTINE CHECK(J)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(582),MF(600),MG(68),MI(48),
     *MK(36),ML(20),MM(70),MN(1000,8),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C      INTEGER L,KP,KN
CNOUSEINTEGER Z(28)
C
C *** CHECK OVERLAPS WITH POLYHEDRA ***
C
      MG( 4)=0
      MG( 5)=0
      MG(11)=0
      K=MG(1)
      IF(J.EQ.1) K=K-1
      IF(K.NE.0)THEN
        DO 3 I=1,K
        P=     PFAKT(MN(I,1))
      IF(NSTOP.EQ.1)RETURN
        YB(4)=P*YA(2,MN(I,1))
        YB(5)=P*YA(3,MN(I,1))
        YB(7)=FLOAT(MN(I,8))*1E-3
        YB(8)=SQRT((YB(1)-YB(4))**2+(YB(2)-YB(5))**2)
        IF(YB(8)-YB(3).LT.YB(7))THEN
          MG(4)=MG(4)+1
          MC(MG(4))=I
        ENDIF
    3   CONTINUE
      ENDIF
C
C *** CHECK OVERLAPS WITH CIRCLES ***
C
      K=MG(2)
      IF (J.EQ.2) K=K-1
      IF (K.NE.0)THEN
        DO 6 I=1,K
        P=     PFAKT(MO(I))
      IF(NSTOP.EQ.1)RETURN
        YB(4)=P*YA(2,MO(I))
        YB(5)=P*YA(3,MO(I))
        YB(7)=P*RAD (MO(I))
        YB(8)=SQRT((YB(1)-YB(4))**2+(YB(2)-YB(5))**2)
        IF(YB(8).LT.YB(3)+YB(7))THEN
          MG(5)=MG(5)+1
          MD(MG(5))=I
        ENDIF
    6   CONTINUE
      ENDIF
C
C *** CHECK OVERLAPS WITH BIZARRE POLYHEDRA ***
C
      K=MG(10)
      IF(J.EQ.3) K=K-1
      IF(K.NE.0)THEN
        DO 9 I=1,K
        P=     PFAKT(MV(I,13))
      IF(NSTOP.EQ.1)RETURN
        YB(4)=P*YA(2,MV(I,13))
        YB(5)=P*YA(3,MV(I,13))
        YB(7)=P*FLOAT(MU(MV(I,14),1,1))*1E-3
        YB(8)=SQRT((YB(4)-YB(1))**2+(YB(5)-YB(2))**2)
        IF(YB(8)-YB(3).LT.YB(7))THEN
          MG(11)=MG(11)+1
          ME(MG(11))=I
        ENDIF
    9   CONTINUE
      ENDIF
      RETURN
      END
C---------------------------------------------------------------------------
      SUBROUTINE CORTH
      COMMON/ALL1/YA(4,6000),YP(95),YC(90),YD(40),YY(1496),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      YD(4)=YC(37)*YD(1)
      YD(5)=YC(38)*YD(1)+YC(39)*YD(2)
      YD(6)=YC(40)*YD(1)+YC(41)*YD(2)+YC(42)*YD(3)
      RETURN
      END
      FUNCTION DIST(I)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(1651),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
           IF(I.GT.6000)THEN
           WRITE(20,*)' I > 6000 in Function DIST !'
           GO TO 1
           ENDIF
C
      IF(YB(25).GT.1000.)THEN
        DIST=-YA(4,I)
      ELSE
        DIST=SQRT(YA(2,I)**2+YA(3,I)**2+(YA(4,I)-YB(25))**2)
      ENDIF
1     CONTINUE
      RETURN
      END
C---------------------------------------------------------------------------
      SUBROUTINE DRAW
      EQUIVALENCE (YC35,YC(35))
      EQUIVALENCE (YC36,YC(36))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(10,2),MM(70),MN(1000,8),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/MODI/ID,PA1,PA2,AR1,AR2,OR1,OR2,RES,FAC,NRES,IBUR
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
      CHARACTER TIT*40,BL*4
      INTEGER L,KL,KP,MA1(28)
      EQUIVALENCE (MA(1,1),MA1(1))
      LOGICAL LOG,IND
      DIMENSION TBSICO(91)
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
      PARAMETER (IYAQ=1000)
C
C *** OUTPUT DEV = XY-PLOTTER HP 7475
C
      CALL HPON
      VFT=AMIN1(11040./1024.,7721./750.)
C      PM3=PM3/VFT
C      PM4=PM4/VFT
      IPAPER=4
C      IF(PM1.GT.12000..OR.PM2.GT.8401.)IPAPER=3
      CALL PS(IPAPER)
      CALL SP(1)
      VIT=20.
      CALL VS(VIT)
C
C ************************************************************
C ***                                                      ***
C *** START MAIN ROUTINE FOR DRAWING CIRCLES AND POLYHEDRA ***
C ***                                                      ***
C ************************************************************
C
      IF(NSTOP.EQ.1)RETURN
      DO 1 I=1,91
    1 TBSICO(I)=SIN(YC35*5.*FLOAT(I-37))
      MG(1)=0
      MG(2)=0
      MG(10)=0
      YD(23)=0.
      YD(24)=0.
      IF (MI(48).GT.0)THEN
        KL=1
        CALL NEWPEN(KL)
      CALL SP(KL)
        CALL OUTLIN
      ENDIF
      IF (MI(34).EQ.0)THEN
        DO 100 I=1,MI(37)-1
        DO 100 J=I+1,MI(37)
        DD=(YA(2,MN(I,1))-YA(2,MN(J,1)))**2+
     *     (YA(3,MN(I,1))-YA(3,MN(J,1)))**2+
     *     (YA(4,MN(I,1))-YA(4,MN(J,1)))**2
        IF (DD.LE.16.)THEN
          P=PFAKT(MN(I,1))
          IF(NSTOP.EQ.1)RETURN
          XP=YA(2,MN(I,1))*C75+512.
          YPP=YA(3,MN(I,1))*C75+375.
          CALL PLOTHGL(XP,YPP,3)
          CALL PLOT(P*YA(2,MN(I,1)),P*YA(3,MN(I,1)),3)
          P=PFAKT(MN(J,1))
      IF(NSTOP.EQ.1)RETURN
          XP=YA(2,MN(J,1))*C75+512.
          YPP=YA(3,MN(J,1))*C75+375.
          CALL PLOTHGL(XP,YPP,2)
          CALL PLOT(P*YA(2,MN(J,1)),P*YA(3,MN(J,1)),2)
        ENDIF
  100   CONTINUE
      ENDIF
    5 IF(MV(MG(10)+1,13).EQ.0) GOTO 102
* BER
      IF(MG(1).GE.MI(37)) GOTO 101
      IF (DIST(MV(MG(10)+1,13)).GE.DIST(MN(MG(1)+1,1))) GOTO 102
  101 IF(MG(2).GE.MI( 7)) GOTO 103
      IF(DIST(MV(MG(10)+1,13)).GE.DIST(MO(MG(2)+1))  ) GOTO 102
      GOTO 103
  102 IF (MG(1).GE.MI(37))GO TO 47
      IF (MG(2).GE.MI(7)) GOTO 104
      IF  (DIST(MN(MG(1)+1,1)  ).GT.DIST(MO(MG(2)+1)))GO TO 48
C
C ***********************
C *** DRAW POLYHEDRON ***
C ***********************
C
  104 MG(1)=MG(1)+1
      KP=IYA(1,MN(MG(1),1))/IYAQ
      KL=1
      DO 88 L=1,10
   88 IF(KP.EQ.ML(L,1)) KL=ML(L,2)
      CALL NEWPEN(KL)
      CALL SP(KL)
      IF (MI(34).EQ.0) GOTO 5
      DO 9 I=1,28
    9 MA1(I)=-1
      IF (MN(MG(1),7).LE.0)THEN
        MG(12)=9
        MG(13)=12
      ELSE
        MG(12)=1
        MG(13)=8
      ENDIF
      CALL OBSC
C
C *** CHECK OVERLAPS ***
C
      P=     PFAKT(MN(MG(1),1))
      IF(NSTOP.EQ.1)RETURN
      YB(1)=P*YA(2,MN(MG(1),1))
      YB(2)=P*YA(3,MN(MG(1),1))
      YB(3)= FLOAT(MN(MG(1),8))*1E-3
      CALL CHECK(1)
CVALUES ARE NOT CHANGED IN CHECK:
COBSOLETE P=     PFAKT(MN(MG(1),1))
COBSOLETE YB(1)=P*YA(2,MN(MG(1),1))
COBSOLETE YB(2)=P*YA(3,MN(MG(1),1))
C
C *** SEND LINE PARAMETERS TO INTER ***
C
      K=18
C
C *** COUNT LINES ***
C
      DO 21 I=1,19
      IF (MA(I+1,1).LE.0)GO TO 22
   21 CONTINUE
      I=20
C
C *** SEND LINE PARAMETERS TO INTER AND CHECK OUTLINES ***
C
   22 MG(18)=3
      DO 27 J=1,I
      P=      PFAKT(MA(J,1))
      IF(NSTOP.EQ.1)RETURN
      YB( 9)=P*YA(2,MA(J,1))
      YB(10)=P*YA(3,MA(J,1))
      P=      PFAKT(MA(J,2))
      IF(NSTOP.EQ.1)RETURN
      YB(11)=P*YA(2,MA(J,2))
      YB(12)=P*YA(3,MA(J,2))
      YD(13)=YB(1)
      YD(14)=YB(2)
      YD(15)=0.5*(YB( 9)+YB(11))
      YD(16)=0.5*(YB(10)+YB(12))
      YD(21)=(YD(13)-YD(15))**2+(YD(14)-YD(16))**2
      DO 26 L=1,I
      IF (J.EQ.L)GO TO 26
      P=      PFAKT(MA(L,1))
      IF(NSTOP.EQ.1)RETURN
      YD(17)=P*YA(2,MA(L,1))
      YD(18)=P*YA(3,MA(L,1))
      P=      PFAKT(MA(L,2))
      IF(NSTOP.EQ.1)RETURN
      YD(19)=P*YA(2,MA(L,2))
      YD(20)=P*YA(3,MA(L,2))
      CALL INTLL
      IF (ABS(YD(5)-0.5).GT.0.5)GO TO 26
      IF (YD(6).LT.0.)GO TO 26
      YD(22)=(YD(13)-YD(1))**2+(YD(14)-YD(2))**2
      IF (YD(22).GT.YD(21))GO TO 27
   26 CONTINUE
      K=K+1
      MG(K)=J
   27 CALL INTER
      IF (MI(41).LT.0)GO TO 36
C
C *** SHADE POLYHEDRON ***
C
      MG(18)=1
      DO 35 I=1,6
      IF (MB(I).LE.0)GO TO 36
      MG(7)=MK(MB(I),1)
      MG(8)=MK(MB(I),2)
      MG(9)=MK(MB(I),3)
      YD(25)=0.
      YD(26)=0.
      IF (YA(4,MN(MG(1),MG(7))).LE.YA(4,MN(MG(1),MG(8))))THEN
        J    =MG(7)
        MG(7)=MG(8)
        MG(8)=J
      ENDIF
      IF (YA(4,MN(MG(1),MG(7))).LE.YA(4,MN(MG(1),MG(9))))THEN
        J    =MG(7)
        MG(7)=MG(9)
        MG(9)=J
      ENDIF
      P=      PFAKT(MN(MG(1),MG(7)))
      IF(NSTOP.EQ.1)RETURN
      YD(27)=P*YA(2,MN(MG(1),MG(7)))
      YD(28)=P*YA(3,MN(MG(1),MG(7)))
      P=      PFAKT(MN(MG(1),MG(8)))
      IF(NSTOP.EQ.1)RETURN
      YD(29)=P*YA(2,MN(MG(1),MG(8)))
      YD(30)=P*YA(3,MN(MG(1),MG(8)))
      P=      PFAKT(MN(MG(1),MG(9)))
      IF(NSTOP.EQ.1)RETURN
      YD(31)=P*YA(2,MN(MG(1),MG(9)))
      YD(32)=P*YA(3,MN(MG(1),MG(9)))
      YD(33)=YD(29)-YD(27)
      YD(34)=YD(30)-YD(28)
      YD(35)=YD(31)-YD(27)
      YD(36)=YD(32)-YD(28)
      JN=MN(MG(1),1)
      DO 600 II=1,IS
      IF (JN.GE.NH(II,1) .AND. JN.LE.NH(II,2)) GOTO 601
  600 CONTINUE
C MODIF ALB : DEFAULT NUMBER OF SHADE IS 4 INSTEAD OF 9
      JN=4
C
      STEP=1./FLOAT(JN+1)
      GOTO 602
  601 JN=NH(II,3)
      STEP=1./FLOAT(JN+1)
  602 IF (JN.GT.0)THEN
        DO 34 J=1,JN
        YD(37)=J*STEP
        YB( 9)=YD(27)+YD(37)*YD(33)
        YB(10)=YD(28)+YD(37)*YD(34)
        YB(11)=YD(27)+YD(37)*YD(35)
        YB(12)=YD(28)+YD(37)*YD(36)
   34   CALL INTER
      ENDIF
   35 CONTINUE
C
C *** SORT OUTLINES ***
C
   36 L=3
      IF(MG(1).LE.0)RETURN
      IF(MG(19).LE.0)RETURN
      MN(MG(1),2)=MA(MG(19),1)
      MN(MG(1),3)=MA(MG(19),2)
   37 DO 43 I=20,K
      IF (MG(I).LE.0)GO TO 43
      IF (MA(MG(I),1).EQ.MN(MG(1),L))THEN
        L=L+1
        MN(MG(1),L)=MA(MG(I),2)
      ELSE
        IF (MA(MG(I),2).NE.MN(MG(1),L))GO TO 43
        L=L+1
        MN(MG(1),L)=MA(MG(I),1)
      ENDIF
      MG(I)=-1
      IF (L.LT.7)GO TO 37
      IF (L.EQ.7)GO TO 45
      WRITE (IO2,1000)
      WRITE (IO2,1000)
      RETURN
   43 CONTINUE
      MN(MG(1),L)=-1
   45 IF (MN(MG(1),2).EQ.MN(MG(1),L))MN(MG(1),L)=-1
      GOTO 185
C
C ********************
C ***              ***
C *** DRAW CIRCLES ***
C ***              ***
C ********************
C
   47 IF (MG(2).GE.MI(7)) GOTO 185
   48 MG(2)=MG(2)+1
      KP=ABS(IYA(1,MO(MG(2))))/IYAQ
      KL=1
      DO 89 L=1,10
   89 IF(KP.EQ.ML(L,1)) KL=ML(L,2)
      CALL NEWPEN(KL)
      CALL SP(KL)
      P=     PFAKT(MO(MG(2)))
      IF(NSTOP.EQ.1)RETURN
      YB(1)=P*YA(2,MO(MG(2)))
      YB(2)=P*YA(3,MO(MG(2)))
      YB(3)=P*RAD (MO(MG(2)))
C     IF (M(34).EQ.0) GOTO 58
C
C *** CHECK OVERLAPS ***
C
      CALL CHECK(2)
C
C *** SET STARTING POINT ***
C
      MG(15)=2
      YR(1,1)= YB(3)*.01745
      YR(1,2)=-YB(3)
      YR(1,3)=-181.
      YR(2,1)=-YR(1,1)
      YR(2,2)= YR(1,2)
      YR(2,3)= 181.
      IF (MI(34).EQ.0) GOTO 69
      IF (MG(4).LE.0)GO TO 69
C
C *** CALCULATE OVERLAPS WITH POLYHEDRA ***
C
      YB(19)=YB(1)
      YB(20)=YB(2)
      DO 68 I=1,MG(4)
      LOG=.FALSE.
      MG(16)=6
      DO 67 J=2,7
      IF (MN(MC(I),J+1).LE.0)THEN
        MG(16)=J-1
        LOG=.TRUE.
      ENDIF
      P=      PFAKT(MN(MC(I),J))
      IF(NSTOP.EQ.1)RETURN
      YB( 9)=P*YA(2,MN(MC(I),J))
      YB(10)=P*YA(3,MN(MC(I),J))
      MG(13)=J-1
      MG(14)=MOD(MG(13),MG(16))+2
      P=      PFAKT(MN(MC(I),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YB(11)=P*YA(2,MN(MC(I),MG(14)))
      YB(12)=P*YA(3,MN(MC(I),MG(14)))
      CALL INTLC
      IF (LOG) GOTO 68
   67 CONTINUE
   68 CONTINUE
   69 IF (MG(5).LE.0)GO TO 705
C
C *** CALCULATE OVERLAPS WITH CIRCLES ***
C
      DO 74 I=1,MG(5)
      P=     PFAKT(MO(MD(I)))
      IF(NSTOP.EQ.1)RETURN
      YB(5)=P*YA(2,MO(MD(I)))
      YB(6)=P*YA(3,MO(MD(I)))
      YB(8)=P*RAD (MO(MD(I)))
      CALL INTCC
   74 CONTINUE
C
C *** CALCULATE INTERSECTIONS WITH BIZARRE POLYHEDRA ***
C
  705 IF(MG(11).LE.0)GO TO 75
      YB(19)=YB(1)
      YB(20)=YB(2)
      DO 707 I=1,MG(11)
      LOG=.FALSE.
      MG(16)=12
      DO 708 J=1,12
      IF(MV(ME(I),J).LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)RETURN
      IF(MV(ME(I),J+1).LE.0)THEN
        LOG=.TRUE.
        MG(16)=J
      ENDIF
      P=      PFAKT(MV(ME(I),J))
      IF(NSTOP.EQ.1)RETURN
      YB( 9)=P*YA(2,MV(ME(I),J))
      YB(10)=P*YA(3,MV(ME(I),J))
      MG(13)=J+1
      MG(14)=MOD(MG(13),MG(16))
      IF(MG(14).LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)RETURN
      IF(ME(I).LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)RETURN
      IF(MV(ME(I),MG(14)).LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)RETURN
      P=      PFAKT(MV(ME(I),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YB(11)=P*YA(2,MV(ME(I),MG(14)))
      YB(12)=P*YA(3,MV(ME(I),MG(14)))
      CALL INTLC
      IF(LOG) GOTO 707
  708 CONTINUE
  707 CONTINUE
C
C *** SORT INTERSECTIONS ***
C
   75 CALL SORT
      TEMP=1./YB(3)
      DO 76 I=1,MG(15)
      YR(I,1)=YR(I,1)*TEMP
   76 YR(I,2)=YR(I,2)*TEMP
C
C *** CHECK OBSCURING OF SEGMENTS ***
C
      DO 84 I=2,MG(15)
      YB(12)=0.5*(YR(I,3)+YR(I-1,3))
      YB(23)=YB(3)*SIN(YB(12)*YC35)+YB(1)
      YB(24)=YB(3)*COS(YB(12)*YC35)+YB(2)
      IF (MI(34).NE.0 .AND. MG(4).GT.0)THEN
        DO 77 J=1,MG(4)
        MG(3)=MC(J)
        CALL INOUT
        IF (MG(17).LE.0)GO TO 84
   77   CONTINUE
      ENDIF
      IF (MG(5).GT.0)THEN
        DO 80 J=1,MG(5)
        P=           PFAKT(MO(MD(J)))
      IF(NSTOP.EQ.1)RETURN
        IF ((YB(23)-P*YA(2,MO(MD(J))))**2
     *     +(YB(24)-P*YA(3,MO(MD(J))))**2 .LE.
     *              (P*RAD(MO(MD(J))))**2      )GO TO 84
   80   CONTINUE
      ENDIF
      IF(MG(11).GT.0)THEN
        DO 803 J=1,MG(11)
        MG(3)=ME(J)
        CALL OUTIN
        IF(MG(17).LE.0)GO TO 84
  803   CONTINUE
      ENDIF
C
C *** DRAW SEGMENT ***
C *** CIRCLES ARE DRAWN 3X IN DO LOOP 83; REMOVE C AND ERASE 2. 81 ***
C
      II11=IFIX(.2*YR(I-1,3))+38
      II22=IFIX(.2*YR(I  ,3))+37
      NP=3
      DO 83 J=-NRES,NRES
      YC(6)=YB(3)+J*RES
      IF(MOD(J,2).EQ.0)THEN
        XP=YB(1)+YC(6)*YR(I-1,1)
        YPP=YB(2)+YC(6)*YR(I-1,2)
        CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
        CALL PLOT  (XP,YPP,NP)
        NP=2
        IF(II22.GE.II11)THEN
          DO 81 K=II11,II22
        XP=YB(1)+YC(6)*TBSICO(K)
        YPP=YB(2)+YC(6)*TBSICO(K+18)
        CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
   81     CALL PLOT(XP,YPP,NP)
        ENDIF
        XP=YB(1)+YC(6)*YR(I,1)
        YPP=YB(2)+YC(6)*YR(I,2)
        CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
        CALL PLOT(XP,YPP,NP)
      ELSE
        XP=YB(1)+YC(6)*YR(I,1)
        YPP=YB(2)+YC(6)*YR(I,2)
        CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
        CALL PLOT  (XP,YPP,NP)
        NP=2
        IF(II22.GE.II11)THEN
          DO 82 K=II22,II11,-1
          XP=YB(1)+YC(6)*TBSICO(K)
          YPP=YB(2)+YC(6)*TBSICO(K+18)
          CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
   82     CALL PLOT(XP,YPP,NP)
        ENDIF
        XP=YB(1)+YC(6)*YR(I-1,1)
        YPP=YB(2)+YC(6)*YR(I-1,2)
        CALL PLOTHGL(XP*C75+512.,YPP*C75+375.,NP)
        CALL PLOT(XP,YPP,NP)
      ENDIF
   83 CONTINUE
C
   84 CONTINUE
      IZ=MG(7)
      GOTO 185
C
C ********************
C ***              ***
C *** DRAW BIZARRE ***
C ***  POLYHEDRON  ***
C ***              ***
C ********************
C
  103 MG(10)=MG(10)+1
      KP=IYA(1,MV(MG(10),13))/IYAQ
      KL=1
      DO 107 K=1,10
  107 IF(KP.EQ.ML(K,1)) KL=ML(K,2)
      CALL NEWPEN(KL)
      CALL SP(KL)
C
C *** CHECK OVERLAPS ***
C
      P=       PFAKT(MV(MG(10),13))
      IF(NSTOP.EQ.1)RETURN
      YB(1)=  P*YA(2,MV(MG(10),13))
      YB(2)=  P*YA(3,MV(MG(10),13))
      YB(3)=FLOAT(MU(MV(MG(10),14),1,1))*1E-3
      CALL CHECK(3)
      CALL SCOB
      IF(NSTOP.EQ.1)RETURN
      P=     PFAKT(MV(MG(10),13))
      IF(NSTOP.EQ.1)RETURN
      YB(1)=P*YA(2,MV(MG(10),13))
      YB(2)=P*YA(3,MV(MG(10),13))
      K=18
      DO 125 I=1,19
      IF(MA(I+1,1).LE.0) GOTO 126
  125 CONTINUE
      I=20
  126 MG(18)=3
      DO 127 J=1,I
      IF(MA(J,1).LE.0)GO TO 127
      IF(MA(J,2).LE.0)GO TO 127
      P=      PFAKT(MA(J,1))
      IF(NSTOP.EQ.1)RETURN
      YB( 9)=P*YA(2,MA(J,1))
      YB(10)=P*YA(3,MA(J,1))
      P=      PFAKT(MA(J,2))
      IF(NSTOP.EQ.1)RETURN
      YB(11)=P*YA(2,MA(J,2))
      YB(12)=P*YA(3,MA(J,2))
      YD(13)=YB(1)
      YD(14)=YB(2)
      YD(15)=0.5*(YB(9)+YB(11))
      YD(16)=0.5*(YB(10)+YB(12))
      YD(21)=(YD(13)-YD(15))**2+(YD(14)-YD(16))**2
      DO 128 L=1,I
      IF(J.EQ.L)GO TO 128
      IF(MA(L,1).LE.0)GO TO 128
      IF(MA(L,2).LE.0)GO TO 128
      P=      PFAKT(MA(L,1))
      IF(NSTOP.EQ.1)RETURN
      YD(17)=P*YA(2,MA(L,1))
      YD(18)=P*YA(3,MA(L,1))
      P=      PFAKT(MA(L,2))
      IF(NSTOP.EQ.1)RETURN
      YD(19)=P*YA(2,MA(L,2))
      YD(20)=P*YA(3,MA(L,2))
      CALL INTLL
      IF(ABS(YD(5)-0.5).GT.0.5)GO TO 128
      IF(YD(6).LT.0.)GO TO 128
      YD(22)=(YD(13)-YD(1))**2+(YD(14)-YD(2))**2
      IF(YD(22).GT.YD(21))GO TO 1270
  128 CONTINUE
      K=K+1
      MG(K)=J
 1270 CALL INTER
  127 CONTINUE
C
C *** SHADE BIZARRE POLYHEDRON ***
C
      KK=K
CNOTUSED:LL=1
      DO 166 NN=1,19
  166 IF(MA(NN+1,1).LE.0.) GOTO 167
      NN=20
  167 DO 162 I=3,NN
      IND=.FALSE.
  163 J=1
      MB(1)=I
      DO 164 K=2,6
  164 MB(K)=0
  151 CONTINUE
      P2=      PFAKT(MA(MB(J),2))
      IF(NSTOP.EQ.1)RETURN
      P1=      PFAKT(MA(MB(J),1))
      IF(NSTOP.EQ.1)RETURN
      YD(13)=P2*YA(2,MA(MB(J),2))-P1*YA(2,MA(MB(J),1))
      YD(14)=P2*YA(3,MA(MB(J),2))-P1*YA(3,MA(MB(J),1))
      YD(18)=1
      DO 154 K=1,NN
      DO 152 L=1,J
  152 IF(K.EQ.MB(L)) GOTO 154
      IF(MA(K,1).NE.MA(MB(J),2).AND.MA(K,2).NE.MA(MB(J),2)) GOTO 154
      IF(MA(K,1).NE.MA(MB(J),2))THEN
        LL     =MA(K,2)
        MA(K,2)=MA(K,1)
        MA(K,1)=LL
      ENDIF
      P2=      PFAKT(MA(K,2))
      IF(NSTOP.EQ.1)RETURN
      P1=      PFAKT(MA(K,1))
      IF(NSTOP.EQ.1)RETURN
      YD(15)=P2*YA(2,MA(K,2))-P1*YA(2,MA(K,1))
      YD(16)=P2*YA(3,MA(K,2))-P1*YA(3,MA(K,1))
      YD(17)=YD(13)*YD(16)-YD(14)*YD(15)
      IF(YD(17).LE.0) GOTO 154
      YD(17)=(YD(13)*YD(15)+YD(14)*YD(16))/
     *(SQRT(YD(13)**2+YD(14)**2)*SQRT(YD(15)**2+YD(16)**2))
      IF(YD(17).GE.YD(18)) GOTO 154
      IF(K.GE.I) GOTO 161
      YD(18)=YD(17)
      MB(J+1)=K
  154 CONTINUE
      J=J+1
      IF(MB(J).EQ.0) GOTO 161
      IF(MA(MB(J),2).NE.MA(MB(1),1))THEN
        IF(J.LE.6) GOTO 151
        WRITE (IO3,*)'FACE BIZARRE'
        WRITE (IO2,*)'FACE BIZARRE'
        RETURN
      ENDIF
      DO 156 K=1,J
  156 MB(K)=MA(MB(K),1)
      L=1
      DO 157 K=2,J
  157 IF(YA(4,MB(K)).GE.YA(4,MB(L))) L=K
      MG(12)=L+1
      IF(L.EQ.J) MG(12)=1
      MG(13)=L-1
      IF(L.EQ.1) MG(13)=J
      P1=      PFAKT(MB(L))
      IF(NSTOP.EQ.1)RETURN
      P2=      PFAKT(MB(MG(12)))
      IF(NSTOP.EQ.1)RETURN
      P3=      PFAKT(MB(MG(13)))
      IF(NSTOP.EQ.1)RETURN
      YD(13)=(P2*YA(2,MB(MG(12)))-P1*YA(2,MB(L)))*0.1
      YD(14)=(P2*YA(3,MB(MG(12)))-P1*YA(3,MB(L)))*0.1
      YD(15)=(P3*YA(2,MB(MG(13)))-P1*YA(2,MB(L)))*0.1
      YD(16)=(P3*YA(3,MB(MG(13)))-P1*YA(3,MB(L)))*0.1
      DEN=YD(16)*YD(13)-YD(15)*YD(14)
      IF(ABS(DEN).LT.1E-6)THEN
        WRITE (IO3,*) 'impossible de definir un pavage de cette face'
        WRITE (IO2,*) 'impossible de definir un pavage de cette face'
        GOTO 161
      ENDIF
      IF(J.EQ.3)THEN
        LIMX=8
        LIMY=8
      ELSE
        YD(17)=10.*DEN
        YD(18)=10.*DEN
        DO 158 K=1,J
        IF(K.NE.MG(12) .AND. K.NE.MG(13) .AND. K.NE.L)THEN
          P1=      PFAKT(MB(K))
       IF(NSTOP.EQ.1)RETURN
         P2=      PFAKT(MB(L))
      IF(NSTOP.EQ.1)RETURN
          YD(19)=P1*YA(2,MB(K))-P2*YA(2,MB(L))
          YD(20)=P1*YA(3,MB(K))-P2*YA(3,MB(L))
          VECT1=YD(16)*YD(19)-YD(15)*YD(20)
          VECT2=YD(20)*YD(13)-YD(19)*YD(14)
          IF(VECT1.GT.YD(17)) YD(17)=VECT1
          IF(VECT2.GT.YD(18)) YD(18)=VECT2
        ENDIF
  158   CONTINUE
        LIMX=IFIX(YD(17)/DEN-0.05)
        LIMY=IFIX(YD(18)/DEN-0.05)
      ENDIF
      DO 160 JX=1,LIMX
      DO 159 JY=1,LIMY
      P=      PFAKT(MB(L))
      IF(NSTOP.EQ.1)RETURN
      YB(23)=P*YA(2,MB(L))+JX*YD(13)+JY*YD(15)
      YB(24)=P*YA(3,MB(L))+JX*YD(14)+JY*YD(16)
      CALL RETIN
      IF(J.EQ.3.AND.(JX+JY).EQ.9) GOTO 160
  159 CONTINUE
  160 CONTINUE
      YD(23)=YB(23)
      YD(24)=YB(24)
  161 IF(.NOT.IND)THEN
        LL     =MA(I,1)
        MA(I,1)=MA(I,2)
        MA(I,2)=LL
        IND=.TRUE.
        GOTO 163
      ENDIF
  162 CONTINUE
C
C *** SORT OUTLINES ***
C
      L=2
      IF(MG(10).LE.0)RETURN
      IF(MG(19).LE.0)RETURN
      MV(MG(10),1)=MA(MG(19),1)
      MV(MG(10),2)=MA(MG(19),2)
  137 DO 143 I=20,KK
      IF(MG(I).LE.0)GO TO 143
      IF(MA(MG(I),1).EQ.MV(MG(10),L))THEN
        L=L+1
        MV(MG(10),L)=MA(MG(I),2)
      ELSE
        IF(MA(MG(I),2).NE.MV(MG(10),L))GO TO 143
        L=L+1
        MV(MG(10),L)=MA(MG(I),1)
      ENDIF
      MG(I)=-1
      IF(L.LT.12)GO TO 137
      IF(L.EQ.12)GO TO 145
      WRITE (IO3,1000)
      WRITE (IO2,1000)
      RETURN
  143 CONTINUE
      MV(MG(10),L)=-1
  145 IF(MV(MG(10),1).EQ.MV(MG(10),L))MV(MG(10),L)=-1
  185 IF(MG(1).LT.MI(37)) GOTO 5
      IF(MG(2).LT.MI(7)) GOTO 5
      IF(MV(MG(10)+1,13).NE.0) GOTO 5
C 110 CONTINUE
      RETURN
 1000 FORMAT (///,1X,'AN ERROR OCCURED IN DETERMINATION OF POLYHEDRAL OU
     &TLINES, THE JOB IS TERMINATED')
C1001 FORMAT (1X,'4 ',I1)
      END
C---------------------------------------------------------------------------
      SUBROUTINE DRAWL
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(1428),
     *MG(68),MI(25933),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/MODI/ID,PA1,PA2,AR1,AR2,OR1,OR2,RES,FAC,NRES,IBUR
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
C
C START: YB(9/10) (->YD(23/24)
      IF ((YD(23)-YB(9))**2+(YD(24)-YB(10))**2.GT.1.E-4)THEN
        YD(23)=YB( 9)
        YD(24)=YB(10)
        XP=YD(23)*C75+512.
        YPP=YD(24)*C75+375.
        CALL PLOTHGL(XP,YPP,3)
        CALL PLOT(YD(23),YD(24),3)
      ENDIF
C END: YB(11/12)
        XP=YB(11)*C75+512.
        YPP=YB(12)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
      CALL PLOT(YB(11),YB(12),2)
C RETRACE
      IF (MG(18).GT.1 .AND. NRES.GT.0)THEN
        YD(10)=YD(24)-YB(12)
        YD(11)=YD(23)-YB(11)
        YD(12)=YD(10)*YD(10)+YD(11)*YD(11)
        IF(YD(12).GT.0.)THEN
          YD(12)=RES/SQRT(YD(12))
          YD(10)= YD(10)*YD(12)
          YD(11)=-YD(11)*YD(12)
          DO 1 I=1,NRES
           Y0=YD(10)*I
           Y1=YD(11)*I
        XP=(YB(11)+Y0)*C75+512.
        YPP=(YB(12)+Y1)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
           CALL PLOT(YB(11)+Y0,YB(12)+Y1,2)
        XP=(YD(23)+Y0)*C75+512.
        YPP=(YD(24)+Y1)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
           CALL PLOT(YD(23)+Y0,YD(24)+Y1,2)
        XP=(YD(23)-Y0)*C75+512.
        YPP=(YD(24)-Y1)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
           CALL PLOT(YD(23)-Y0,YD(24)-Y1,2)
        XP=(YB(11)-Y0)*C75+512.
        YPP=(YB(12)-Y1)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
1          CALL PLOT(YB(11)-Y0,YB(12)-Y1,2)
C       CALL PLOT(YB(11)+YD(10),YB(12)+YD(11),2)
C       CALL PLOT(YD(23)+YD(10),YD(24)+YD(11),2)
C       CALL PLOT(YD(23)-YD(10),YD(24)-YD(11),2)
C       CALL PLOT(YB(11)-YD(10),YB(12)-YD(11),2)
        ENDIF
      ENDIF
      YD(23)=YB(11)
      YD(24)=YB(12)
      RETURN
      END
C-------------------------------------------------------------------------
      SUBROUTINE FIND(N,M,L,LOG)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(106),YZ(100,5),
     *YE(200,5),MH(20),MI(11198),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      LOGICAL LOG
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
C
      DO 1 J=1,MIN(M,12)
      IF(YZ(J,5).GT.YP(L,3)) GOTO 2
      IF(IYA(1,IFIX(YZ(J,1))).LE.0.) GOTO 14
      MV(N,J)=IFIX(YZ(J,1))
    1 CONTINUE
    2 J=J-1
      IF(.NOT.LOG) GOTO 13
      DO 3 I=1,N-1
    3 IF(MV(I,14).EQ.MV(N,14)) GOTO 5
      GOTO 13
    5 IF(J.GT.0)THEN
      IF(YZ(J,5).LE.MU(MV(I,14),1,1)/1000.) GOTO 6
      J=J-1
      ENDIF
      IF(J.GE.1) GOTO 5
    6 IF(J.GT.MU(MV(I,14),1,2)) GOTO 8
      IF(J.LT.MU(MV(I,14),1,2)) GOTO 14
      DO 7 K=J+1,MIN(M,12)
    7 MV(N,K)=0
      CALL RCGN(N,I)
      IF(NSTOP.EQ.1)RETURN
      GOTO 16
    8 K=1
    9 IF(MV(K,14).NE.MV(N,14)) GOTO 12
      DO 11 I=1,14
      DO 10 J=K,N-1
   10 MV(J,I)=MV(J+1,I)
   11 MV(N,I)=0
      N=N-1
   12 K=K+1
      IF(K.LT.N) GOTO 9
   13 IF(N.GT.1000.OR.N.LE.0)THEN
      NSTOP=1
      GO TO 16
      ENDIF
      IF(L.GT.10.OR.L.LE.0)THEN
      NSTOP=1
      GO TO 16
      ENDIF
      IF(MV(N,14).GT.5.OR.MV(N,14).LE.0)THEN
      NSTOP=1
      GO TO 16
      ENDIF
      MU(MV(N,14),1,1)=YP(L,3)*1000.
      CALL BZRR(N,M)
      IF(NSTOP.EQ.1)RETURN
      GOTO 16
   14 DO 15 K=1,14
   15 MV(N,K)=0
      N=N-1
   16 RETURN
      END
C--------------------------------------------------------------------------
      SUBROUTINE INOUT
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),MA(1428),
     *MG(68),MI(174),MN(1000,8),MO(3759),MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
C *** DETERMINE NUMBER OF EDGES ***
C
      DO 1 I=3,6
      IF (MN(MG(3),I+2).LE.0)GO TO 2
    1 CONTINUE
      I=6
C
C *** CHECK IF B(23),B(24) IS OBSCURED BY POLYHEDRON NO. F(3) ***
C
    2 K=I+MOD(I,2)-1
      DO 3 J=1,K,2
      MG(12)=MOD(J,I)+2
      MG(13)=MOD(J+1,I)+2
      MG(14)=MOD(J-1,I)+2
      P2=PFAKT(MN(MG(3),MG(12)))
      IF(NSTOP.EQ.1)RETURN
*      write(1,1000) MG(3),(MG(i),MN(MG(3),MG(i)),i=12,14)
* 1000 format(1x,7i10)
      P3=PFAKT(MN(MG(3),MG(13)))
      IF(NSTOP.EQ.1)RETURN
      P4=PFAKT(MN(MG(3),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YB(13)=P3*YA(2,MN(MG(3),MG(13)))-P2*YA(2,MN(MG(3),MG(12)))
      YB(14)=P3*YA(3,MN(MG(3),MG(13)))-P2*YA(3,MN(MG(3),MG(12)))
      YB(15)=P4*YA(2,MN(MG(3),MG(14)))-P2*YA(2,MN(MG(3),MG(12)))
      YB(16)=P4*YA(3,MN(MG(3),MG(14)))-P2*YA(3,MN(MG(3),MG(12)))
      YD(11)=YB(23)-P2*YA(2,MN(MG(3),MG(12)))
      YD(12)=YB(24)-P2*YA(3,MN(MG(3),MG(12)))
C
C   Possibility to divide by zero..... MODIF ALB...
C
      YMANS=YB(15)*YB(14)-YB(16)*YB(13)
      IF(YMANS.EQ.0.)YMANS=0.0000000001
      YD( 5)=(YD(11)*YB(14)-YD(12)*YB(13))/YMANS
      YMANS=YB(13)*YB(16)-YB(14)*YB(15)
      IF(YMANS.EQ.0.)YMANS=0.0000000001
      YD( 6)=(YD(11)*YB(16)-YD(12)*YB(15))/YMANS
C
      YD(10)=AMIN1(YD(5),YD(6))
      IF (YD(10).LE.0.)GO TO 4
    3 CONTINUE
    4 MG(17)=-IFIX(SIGN(1.,YD(10)))
      RETURN
      END
C---------------------------------------------------------------------------
      SUBROUTINE INTCC
      EQUIVALENCE (YC36,YC(36))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      YD( 9)=YB(5)-YB(1)
      YD(10)=YB(6)-YB(2)
      YD(11)=YD(9)**2+YD(10)**2
      YD(12)=(YD(11)+YB(3)**2-YB(8)**2)*0.5
      YD(13)=YB(3)**2*YD(11)-YD(12)**2
      IF(YD(13).GT.0.)THEN
        YD(11)=1./YD(11)
        YD(14)=YD(11)*YD(12)
        YD( 3)=YD(14)*YD( 9)
        YD( 4)=YD(14)*YD(10)
        YD(15)=YD(11)*SQRT(YD(13))
        YD(10)=YD(10)*YD(15)
        YD( 9)=-YD(9)*YD(15)
        YD(1)=YD(3)-YD(10)
        YD(2)=YD(4)-YD( 9)
        YD(3)=YD(3)+YD(10)
        YD(4)=YD(4)+YD( 9)
        MG(15)=MG(15)+1

        YR(MG(15),1)=YD(1)
        YR(MG(15),2)=YD(2)
        YR(MG(15),3)=YC36*ATAN2(YD(1),YD(2))
        MG(15)=MG(15)+1
        YR(MG(15),1)=YD(3)
        YR(MG(15),2)=YD(4)
        YR(MG(15),3)=YC36*ATAN2(YD(3),YD(4))
C     ELSE
C no intersection
C       YD(1)=YB(1)
C       YD(3)=YD(1)
C       YD(2)=YB(2)-YB(3)
C       YD(4)=YD(2)
C       WRITE(6,100)
C       WRITE(6,101)YB(1),YB(2),YB(3)
C       WRITE(6,101)YB(5),YB(6),YB(8)
      ENDIF
      RETURN
C100   FORMAT(' no intersection (x,y,r):')
C101   FORMAT(1X,3F10.6)
      END
C------------------------------------------------------------------------
      SUBROUTINE INTCL
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(1496),
     *MI(25933),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      YD(5)=-1.
      YD(6)=-1.
      YD( 7)=YB( 9)-YB(19)
      YD( 8)=YB(10)-YB(20)
      YD( 9)=YB(11)-YB( 9)
      YD(10)=YB(12)-YB(10)
      YD(11)=YD( 9)*YD( 9)+YD(10)*YD(10)
      YD(12)=YD( 7)*YD( 9)+YD( 8)*YD(10)
      YD(13)=YD(12)*YD(12)-YD(11)*
     *  (YD( 7)*YD( 7)+YD( 8)*YD( 8)-YB(3)*YB(3))
      IF (YD(13).GE.0.)THEN
C
C Possibility to divide by zero, MODIF ALB
C
        IF(YD(11).EQ.0.)YD(11)=0.0000000001
        YD(14)=-YD(12)/YD(11)
        YD(13)=SQRT(YD(13))/YD(11)
C
        YD(5)=YD(14)+YD(13)
        YD(1)=YB( 9)+YD(5)*YD( 9)
        YD(2)=YB(10)+YD(5)*YD(10)
        IF (YD(13).GT.1E-4)THEN
          YD(6)=YD(14)-YD(13)
          YD(3)=YB( 9)+YD(6)*YD( 9)
          YD(4)=YB(10)+YD(6)*YD(10)
        ENDIF
      ENDIF
      RETURN
      END
C------------------------------------------------------------------------
      SUBROUTINE INTLC
      EQUIVALENCE (YC36,YC(36))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(10,2),MM(70),MN(1000,8),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/MODI/ID,PA1,PA2,AR1,AR2,OR1,OR2,RES,FAC,NRES,IBUR
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
C      INTEGER L,KP,KN
      CHARACTER TIT*40,BL*4
C     LOGICAL LOG,IND
CNOUSECOMMON/PLC/XPL1,XPL2,KPL
CNOUSEINTEGERZ(28)
CNOUSEEQUIVALENCE (ME(1,1),Z(1))
C
      CALL INTCL
      IF (ABS(YD(5)-0.5).LE.0.5)THEN
        MG(15)=MG(15)+1
        YR(MG(15),1)=YD(1)-YB(1)
        YR(MG(15),2)=YD(2)-YB(2)
        YR(MG(15),3)=YC36*ATAN2(YR(MG(15),1) , YR(MG(15),2))
      ENDIF
      IF (ABS(YD(6)-0.5).LE.0.5)THEN
        MG(15)=MG(15)+1
        YR(MG(15),1)=YD(3)-YB(1)
        YR(MG(15),2)=YD(4)-YB(2)
        YR(MG(15),3)=YC36*ATAN2(YR(MG(15),1) , YR(MG(15),2))
      ENDIF
      RETURN
      END
C--------------------------------------------------------------------------
      SUBROUTINE INTER
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(90),MN(1000,8),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
      YR(1,1)=YB(9)
      YR(1,2)=YB(10)
      YR(1,3)=0.
      YR(2,1)=YB(11)
      YR(2,2)=YB(12)
      YR(2,3)=SQRT((YB(9)-YB(11))**2+(YB(10)-YB(12))**2)
      MG(15)=2
      IF (MG(4).LE.0)GO TO 10
C
C *** CALCULATE OVERLAPS WITH OTHER POLYHEDRA ***
C
      YD(13)=YB( 9)
      YD(14)=YB(10)
      YD(15)=YB(11)
      YD(16)=YB(12)
      DO 9 I=1,MG(4)

      MG(16)=6
      DO 8 J=2,7
      IF (MN(MC(I),J)  .LT.1) GOTO 9
      IF (MN(MC(I),J+1).LE.0) MG(16)=J-1
      P=      PFAKT(MN(MC(I),J))
      IF(NSTOP.EQ.1)RETURN
      YD(17)=P*YA(2,MN(MC(I),J))
      YD(18)=P*YA(3,MN(MC(I),J))
      MG(13)=J-1
      MG(14)=MOD(MG(13),MG(16))+2
      P=      PFAKT(MN(MC(I),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YD(19)=P*YA(2,MN(MC(I),MG(14)))
      YD(20)=P*YA(3,MN(MC(I),MG(14)))
      CALL INTLL
      IF (ABS(YD(5)-0.5).GT.0.5)GO TO 8
      IF (ABS(YD(6)-0.5).GT.0.5)GO TO 8
      YB(8)=SQRT((YB(9)-YD(1))**2+(YB(10)-YD(2))**2)
      IF (YB(8).LE.1E-3)GO TO 8
      DO 6 K=1,MG(15)
      IF (ABS(YB(8)-YR(K,3)).LE.1E-5)GO TO 8
    6 CONTINUE
      MG(15)=MG(15)+1
      YR(MG(15),1)=YD(1)
      YR(MG(15),2)=YD(2)
      YR(MG(15),3)=YB(8)
    8 CONTINUE
    9 CONTINUE
C
C *** CALCULATE INTERSECTIONS WITH CIRCLES ***
C
   10 IF (MG(5).GT.0)THEN
        DO 15 I=1,MG(5)
        P=      PFAKT(MO(MD(I)))
      IF(NSTOP.EQ.1)RETURN
        YB(3 )=P*RAD (MO(MD(I)))
        YB(19)=P*YA(2,MO(MD(I)))
        YB(20)=P*YA(3,MO(MD(I)))
        CALL INTCL
        IF (ABS(YD(5)-0.5).LE.0.5)THEN
          MG(15)=MG(15)+1
          YR(MG(15),1)=YD(1)
          YR(MG(15),2)=YD(2)
          YR(MG(15),3)=SQRT((YB(9)-YD(1))**2+(YB(10)-YD(2))**2)
        ENDIF
        IF (ABS(YD(6)-0.5).LE.0.5)THEN
          MG(15)=MG(15)+1
          YR(MG(15),1)=YD(3)
          YR(MG(15),2)=YD(4)
          YR(MG(15),3)=SQRT((YB(9)-YD(3))**2+(YB(10)-YD(4))**2)
        ENDIF
   15   CONTINUE
      ENDIF
C
C *** CALCULATE OVERLAPS WITH BIZARRE POLYHEDRA ***
C
      IF(MG(11).LE.0) GOTO 39
      YD(13)=YB(9)
      YD(14)=YB(10)
      YD(15)=YB(11)
      YD(16)=YB(12)
      DO 24 I=1,MG(11)
      MG(16)=12
      DO 25 J=1,12
      IF(MV(ME(I),J)  .LT.1) GOTO 25
      IF(MV(ME(I),J+1).LE.0) MG(16)=J
      P=      PFAKT(MV(ME(I),J))
      IF(NSTOP.EQ.1)RETURN
      YD(17)=P*YA(2,MV(ME(I),J))
      YD(18)=P*YA(3,MV(ME(I),J))
      MG(13)=J
      MG(14)=MOD(MG(13),MG(16))+1
      P=      PFAKT(MV(ME(I),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YD(19)=P*YA(2,MV(ME(I),MG(14)))
      YD(20)=P*YA(3,MV(ME(I),MG(14)))
      CALL INTLL
      IF(ABS(YD(5)-0.5).GT.0.5)GO TO 25
      IF(ABS(YD(6)-0.5).GT.0.5)GO TO 25
      YB(8)=SQRT((YB(9)-YD(1))**2+(YB(10)-YD(2))**2)
      IF(YB(8).LE.1E-3)GO TO 25
      DO 30 K=1,MG(15)
      IF(ABS(YB(8)-YR(K,3)).LE.1E-5)GO TO 25
   30 CONTINUE
      MG(15)=MG(15)+1
      YR(MG(15),1)=YD(1)
      YR(MG(15),2)=YD(2)
      YR(MG(15),3)=YB(8)
   25 CONTINUE
   24 CONTINUE
C
C *** SORT INTERSECTIONS ***
C
   39 IF(MG(15).GT.0)CALL SORT
C
C *** CHECK OBSCURING OF LINE SEGMENTS ***
C
      DO 22 I=1,MG(15)-1
      YB( 9)=YR(I,1)
      YB(10)=YR(I,2)
      YB(11)=YR(I+1,1)
      YB(12)=YR(I+1,2)
      YB(23)=(YB( 9)+YB(11))*0.5
      YB(24)=(YB(10)+YB(12))*0.5
      IF (MG(1).NE.0 .AND. MG(4).NE.0)THEN
        DO 18 J=1,MG(4)
        MG(3)=MC(J)
        CALL INOUT
        IF (MG(17).LE.0)GO TO 22
   18   CONTINUE
      ENDIF
      IF (MG(2).NE.0 .AND. MG(5).NE.0)THEN
        DO 20 J=1,MG(5)
        P=      PFAKT(MO(MD(J)))
      IF(NSTOP.EQ.1)RETURN
        YB( 8)=P*RAD (MO(MD(J)))
        YB(19)=P*YA(2,MO(MD(J)))
        YB(20)=P*YA(3,MO(MD(J)))
        YD(5)=SQRT((YB(23)-YB(19))**2+(YB(24)-YB(20))**2)
        IF (YD(5).LE.YB(8))GO TO 22
   20   CONTINUE
      ENDIF
      IF(MG(10).EQ.0.OR.MG(11).LE.0) GOTO 21
      DO 33 J=1,MG(11)
      MG(3)=ME(J)
      CALL OUTIN
      IF(MG(17).LE.0)GO TO 22
   33 CONTINUE
C
C
C *** DRAW LINE ***
C
   21 CALL DRAWL
   22 CONTINUE
      RETURN
      END
C---------------------------------------------------------------
      SUBROUTINE INTLL
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(1536),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      YD( 9)=YD(15)-YD(13)
      YD(10)=YD(16)-YD(14)
      YD(11)=YD(19)-YD(17)
      YD(12)=YD(20)-YD(18)
      DEN=YD(9)*YD(12)-YD(10)*YD(11)
      IF (ABS(DEN)-1E-8) 1,1,2
    1 YD(5)=10.
      RETURN
    2 YD( 5)=(YD(10)*(YD(17)-YD(13))-YD(9)*(YD(18)-YD(14)))/DEN
      YD( 6)=(YD(12)*(YD(17)-YD(13))-YD(11)*(YD(18)-YD(14)))/DEN
      YD( 1)=YD(17)+YD(5)*YD(11)
      YD( 2)=YD(18)+YD(5)*YD(12)
      RETURN
      END
C------------------------------------------------------------------------
      FUNCTION IPACK(J)
      COMMON/ALL1/YA(4,6000),YP(1721),MI(104),MM(70),MN(8000),
     *MO(3024),MU(14735),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DO 1 I=19,44
    1 IF (MN(1).EQ.MM(I)) GOTO 2
      GO TO 5
    2 MN(1)=I-19
      DO 3 I=19,44
    3 IF (MN(2).EQ.MM(I)) GOTO 4
      IF (MN(2).NE.MM(1)) GOTO 5
      I=45
    4 MN(2)=I-19
      I=MN(1)*27+MN(2)+1
      IPACK=I*1000+MN(3)*10
      MN(3)=I
      RETURN
    5 WRITE (IO2,1000) MN(1),MN(2)
      WRITE (IO3,1000) MN(1),MN(2)
      RETURN
 1000 FORMAT (1X,2A1,' IS NOT A COMMON ATOM NAME, ONLY LETTERS A-Z ARE
     &ACCEPTED, JOB IS TERMINATED')
      END
C-------------------------------------------------------------------------
      SUBROUTINE MERGE(NN)
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(106),YZ(100,5),YE(200,5),
     *MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
CNOTUSED:DIMENSION W(30)
      COMMON/BZR/BX,BY,BZ
      INTEGER BX(10),BY(13),BZ(4)
      LOGICAL LOG
C
      BY(11)=MU(NN,BZ(1),7)
      DO 40 I=1,BY(11)
   40 BY(I)=MU(NN,BZ(1),I)
      BZ(1)=-BZ(1)
      LL=1
   45 II=BY(LL)
      JJ=BY(MOD(LL,BY(11))+1)
   43 LOG=.FALSE.
      DO 3 I=2,4
      IF(BZ(I).LE.0) GOTO 3
      LOG=.TRUE.
      DO 2 J=1,MU(NN,BZ(I),7)
      IF(MU(NN,BZ(I),J).NE.II) GOTO 2
      K=MOD(J,MU(NN,BZ(I),7))+1
      IF(MU(NN,BZ(I),K).NE.JJ) GOTO 1
      BY(12)=BZ(I)
      BY(13)=K
      BZ(I)=-BZ(I)
      GOTO 4
    1 K=MOD(J-2+MU(NN,BZ(I),7),MU(NN,BZ(I),7))+1
      IF(MU(NN,BZ(I),K).NE.JJ) GOTO 3
      BY(12)=-BZ(I)
      BY(13)=J
      BZ(I)=-BZ(I)
      GOTO 4
    2 CONTINUE
    3 CONTINUE
      IF(.NOT.LOG) GOTO 46
      LL=LL+1
      IF(LL.GT.BY(11)) GOTO 46
      GOTO 45
    4 NM=MU(NN,ABS(BY(12)),7)-2
      DO 24 KK=BY(11),LL+1,-1
   24 BY(KK+NM)=BY(KK)
      IF(BY(12).GE.0) GOTO 27
      K=1
      M=LL+1
      N=NM+LL
      GOTO 28
   27 K=-1
      M=NM+LL
      N=LL+1
C rxf
   28 MUNN=MU(NN,ABS(BY(12)),7)
      J=MOD(BY(13),MUNN)+1
C rxf   28 J=MOD(BY(13),MU(NN,ABS(BY(12)),7))+1
      DO 9 L=M,N,K
      BY(L)=MU(NN,ABS(BY(12)),J)
C rxf
      MUNN=MU(NN,ABS(BY(12)),7)
    9 J=MOD(J,MUNN)+1
C rxf    9 J=MOD(J,MU(NN,ABS(BY(12)),7))+1
      BY(11)=BY(11)+NM
   41 LL=1
      GOTO 45
   46 DO 44 I=1,4
   44 BZ(I)=ABS(BZ(I))
      RETURN
      END
*BER**Subroutine Newpen schreibt Stiftnummer auf Datei 9*************
      SUBROUTINE NEWPEN(ISTIFT)
      INTEGER ISTIFT
      WRITE(9,50005) ISTIFT,ISTIFT
50005 FORMAT(' 4',1X,I2,1X,I2)
      RETURN
      END
C------------------------------------------------------------------------
      SUBROUTINE OBSC
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(1382),MG(68),MI(48),MK(12,3),ML(90),MN(1000,8),MO(17759),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
      DO 101 K=1,6
101   MB(K)=0
      K=0
      L=0
      DO 102 N=1,3
102   YB(N)= YA( N+1,MN(MG(1),1) )
C
      DO 130 I=MG(12),MG(13)
C
      DO 111 N=1,3
111   MG(N+6)=MN(MG(1),MK(I,N))
      DO 112 N=1,3
      YB(N+ 4)=YA(N+1,MG(8))-YA(N+1,MG(7))
      YB(N+ 7)=YA(N+1,MG(9))-YA(N+1,MG(7))
112   YB(N+13)=YB(N)        -YA(N+1,MG(7))
      YB(11)=YB(6)*YB(10)-YB(7)*YB( 9)
      YB(12)=YB(7)*YB( 8)-YB(5)*YB(10)
      YB(13)=YB(5)*YB( 9)-YB(6)*YB( 8)
C
      YB(17)=YB(11)*YB(14)+YB(12)*YB(15)+YB(13)*YB(16)
C
      YB(13)=YB(13)*
     *(YB(25)-YA(4,MG(7)))-YB(11)*YA(2,MG(7))-YB(12)*YA(3,MG(7))
C
      IF (YB(17).GE.0.)YB(13)=-YB(13)
      IF (YB(13).LE.0.)GO TO 130
      K=K+1
      MB(K)=I
      IF (L.GT.0)THEN
        DO 121 J=1,L
        IF (MK(I,1).EQ.MA(J,1) .AND. MK(I,2).EQ.MA(J,2)) GOTO 122
121     IF (MK(I,2).EQ.MA(J,1) .AND. MK(I,1).EQ.MA(J,2)) GOTO 122
      ENDIF
      L=L+1
      MA(L,1)=MK(I,1)
      MA(L,2)=MK(I,2)
122   IF (L.GT.1)THEN
        DO 123 J=1,L
        IF (MK(I,1).EQ.MA(J,1) .AND. MK(I,3).EQ.MA(J,2)) GOTO 124
123     IF (MK(I,3).EQ.MA(J,1) .AND. MK(I,1).EQ.MA(J,2)) GOTO 124
      ENDIF
      L=L+1
      MA(L,1)=MK(I,1)
      MA(L,2)=MK(I,3)
124   IF (L.GT.2)THEN
        DO 125 J=1,L
        IF (MK(I,2).EQ.MA(J,1) .AND. MK(I,3).EQ.MA(J,2)) GOTO 130
125     IF (MK(I,3).EQ.MA(J,1) .AND. MK(I,2).EQ.MA(J,2)) GOTO 130
      ENDIF
      L=L+1
      MA(L,1)=MK(I,2)
      MA(L,2)=MK(I,3)
130   CONTINUE
      DO 131 J=1,L
      DO 131 K=1,2
131   MA(J,K)=MN(MG(1),MA(J,K))
      IF(L.NE.20)THEN
        DO 132 J=L+1,20
        MA(J,1)=0
132     MA(J,2)=0
      ENDIF
      RETURN
      END
C-----------------------------------------------------------------------
      SUBROUTINE OUTIN
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),MA(1428),
     *MG(68),MI(174),MN(1000,8),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
      DO 1 I=3,11
      IF(MV(MG(3),I+1).LE.0)GO TO 2
    1 CONTINUE
      I=12
    2 K=I+MOD(I,2)-1
      DO 3 J=1,K,2
      MG(12)=MOD(J,I)+1
      MG(13)=MOD(J+1,I)+1
      MG(14)=MOD(J-1,I)+1
*      if(mg(13).gt.14) write(1,2001) k,j,i,mg(12),mg(13),mg(14)
*2001  format(7(1x,i10))
      P2=PFAKT(MV(MG(3),MG(12)))
      IF(NSTOP.EQ.1)RETURN
      P3=PFAKT(MV(MG(3),MG(13)))
      IF(NSTOP.EQ.1)RETURN
      P4=PFAKT(MV(MG(3),MG(14)))
      IF(NSTOP.EQ.1)RETURN
      IF(MV(MG(3),MG(12)).LE.0)GO TO 3
      IF(MV(MG(3),MG(13)).LE.0)GO TO 3
      IF(MV(MG(3),MG(14)).LE.0)GO TO 3
      YB(13)=P3*YA(2,MV(MG(3),MG(13)))-P2*YA(2,MV(MG(3),MG(12)))
      YB(14)=P3*YA(3,MV(MG(3),MG(13)))-P2*YA(3,MV(MG(3),MG(12)))
      YB(15)=P4*YA(2,MV(MG(3),MG(14)))-P2*YA(2,MV(MG(3),MG(12)))
      YB(16)=P4*YA(3,MV(MG(3),MG(14)))-P2*YA(3,MV(MG(3),MG(12)))
      YD(11)=YB(23)-P2*YA(2,MV(MG(3),MG(12)))
      YD(12)=YB(24)-P2*YA(3,MV(MG(3),MG(12)))
C
C   Possibility to divide by zero..... MODIF ALB...
C
      YMANS=YB(15)*YB(14)-YB(16)*YB(13)
      IF(YMANS.EQ.0.)YMANS=0.0000000001
      YD( 5)=(YD(11)*YB(14)-YD(12)*YB(13))/YMANS
      YMANS=YB(13)*YB(16)-YB(14)*YB(15)
      IF(YMANS.EQ.0.)YMANS=0.0000000001
      YD( 6)=(YD(11)*YB(16)-YD(12)*YB(15))/YMANS
C
      YD(10)=AMIN1(YD(5),YD(6))
      IF (YD(10).LE.0.)GO TO 4
    3 CONTINUE
    4 MG(17)=-IFIX(SIGN(1.,YD(10)))
      RETURN
      END
Cp------------------------------------------------------------------------
      SUBROUTINE OUTLIN
      COMMON/ALL1/YA(4,6000),YP(70),YB(1583),MG(68),MI(174),MN(1000,8),
     *MO(3000),MZ(12,2),MU(5,21,7),MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      MG(18)=1
      DO 1 I=1,12
      P=      PFAKT(MZ(I,1))
      IF(NSTOP.EQ.1)RETURN
      YB( 9)=P*YA(2,MZ(I,1))
      YB(10)=P*YA(3,MZ(I,1))
      P=      PFAKT(MZ(I,2))
      IF(NSTOP.EQ.1)RETURN
      YB(11)=P*YA(2,MZ(I,2))
      YB(12)=P*YA(3,MZ(I,2))
    1 CALL DRAWL
      RETURN
      END
C---------------------------------------------------------------------------
      FUNCTION PFAKT(I)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(1651),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
      DV=ABS(YB(25))
      IF(DV.GT.1000.)THEN
        PFAKT=1.
      ELSE
           IF(I.GT.6000)THEN
           WRITE(20,*)' I > 6000 in Function PFAKT !'
           NSTOP=1
           GO TO 1
           ENDIF
           IF(I.LE.0)THEN
           WRITE(20,*)' I < or = 0 in Function PFAKT !'
           NSTOP=1
           GO TO 1
           ENDIF
        PFAKT=DV/(DV-YA(4,I))
      ENDIF
1     CONTINUE
      RETURN
      END
*BER**Subroutine Plot, schreibt Koordinaten auf Datei 9**************
      SUBROUTINE PLOT(XKOORD,YKOORD,INUMBER)
      REAL XKOORD,YKOORD
      INTEGER INUMBER
      WRITE(9,50004) INUMBER,XKOORD,YKOORD
50004 FORMAT(I2,1X,2(F10.4,1X))
      RETURN
      END
*********************************************************************
      FUNCTION RAD(J)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(1651),MI(25933),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
      PARAMETER (IYAQ=10)

      I=MOD(IYA(1,J),10)
      RAD=YB(I+103)
      RETURN
      END
C--------------------------------------------------------------------------
      SUBROUTINE RCGN(I,J)
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(106),YZ(100,5),YE(200,5),
     *MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DIMENSION W(30),NX(12)
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
C
      DO 11 K=1,12
   11 NX(K)=0
      K=0
   10 IF(K.GE.MU(MV(J,14),1,2)) GOTO 6
      K=K+1
      IF(J.LE.0)THEN
      NSTOP=1
      ENDIF
      IF(K.LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)GO TO 18
      IF(MV(J,K).LE.0) GOTO 6
      W(1)=YA(2,MV(J,K))-YA(2,MV(J,13))
      W(2)=YA(3,MV(J,K))-YA(3,MV(J,13))
      W(3)=YA(4,MV(J,K))-YA(4,MV(J,13))
      W(9)=SQRT(W(1)**2+W(2)**2+W(3)**2)
      DO 4 L=NX(K)+1,MU(MV(J,14),1,2)
      IF(NX(K)+1.LE.MU(MV(J,14),1,2)) GOTO 12
      WRITE (IO3,*)'AN ERROR OCCURED IN THE DETERMINATION OF A B.P.'
      WRITE (IO2,*)'AN ERROR OCCURED IN THE DETERMINATION OF A B.P.'
      DO 22 K=1,14
   22 MV(I,K)=0
      I=I-1
      GOTO 18
   12 IF(K.EQ.1) GOTO 13
      DO 9 M=1,K-1
    9 IF(L.EQ.NX(M)) GOTO 4
   13 IF(ABS(IYA(1,MV(I,L))).NE.ABS(IYA(1,MV(J,K)))) GOTO 4
      W(5)=YA(2,MV(I,L))-YA(2,MV(I,13))
      W(6)=YA(3,MV(I,L))-YA(3,MV(I,13))
      W(7)=YA(4,MV(I,L))-YA(4,MV(I,13))
      W(8)=SQRT(W(5)**2+W(6)**2+W(7)**2)
      IF(ABS(W(9)-W(8)).LE.1E-4) GOTO 21
      IF(W(8)-W(9)) 15,15,20
   21 IF(K.NE.1) GOTO 14
      NX(1)=L
      GOTO 10
   14 DO 5 M=1,K-1
      W(1)=YA(2,MV(J,M))-YA(2,MV(J,K))
      W(2)=YA(3,MV(J,M))-YA(3,MV(J,K))
      W(3)=YA(4,MV(J,M))-YA(4,MV(J,K))
      W(4)=SQRT(W(1)**2+W(2)**2+W(3)**2)
      W(5)=YA(2,MV(I,NX(M)))-YA(2,MV(I,L))
      W(6)=YA(3,MV(I,NX(M)))-YA(3,MV(I,L))
      W(7)=YA(4,MV(I,NX(M)))-YA(4,MV(I,L))
      W(8)=SQRT(W(5)**2+W(6)**2+W(7)**2)
    5 IF(ABS(W(4)-W(8)).GE.1E-4) GOTO 15
      NX(K)=L
      GOTO 3
   15 CONTINUE
    4 CONTINUE
   20 K=K-2
      NX(K+2)=0
    3 IF(K.LE.12) GOTO 10
    6 DO 7 N=1,K
      IF(I.LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NX(N).LE.0)THEN
      NSTOP=1
      ENDIF
      IF(NSTOP.EQ.1)GO TO 18
      NX(N)=MV(I,NX(N))
    7 CONTINUE
      DO 8 K=1,12
    8 MV(I,K)=NX(K)
      DO 19 K=1,I-1
      IF(DIST(MV(K,13)).LE.DIST(MV(I,13))) GOTO 19
      IF(J.GE.K) J=J+1
      DO 16 L=1,14
      DO 17 M=I,K,-1
   17 MV(M+1,L)=MV(M,L)
      MV(K,L)=MV(I+1,L)
   16 MV(I+1,L)=0
      GOTO 18
   19 CONTINUE
   18 RETURN
      END
C-------------------------------------------------------------------------
      SUBROUTINE READ
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(48,3,4),
     *YF(5,15,12),MH(20),MI(84),ML(10,2),MM(70),MN(8000),MO(3024),
     *MU(14735),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/MODI/ID,PM(6),RES,FAC,NRES,IBUR
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
      CHARACTER TIT*40,BL*4
      LOGICAL LOG,LOG3
CNOUSEDIMENSION P(3024)
CNOUSEEQUIVALENCE (MO(1),P(1))
      WRITE (IO2,1000)
      WRITE (IO2,1013)
C
C *** READ ALL OPTIONS ***
C
      READ (IO1,100) TIT
  100 FORMAT (4X,A)
      WRITE (IO2,103) TIT
  103 FORMAT (1X,'* TITL',A40)
      IS=0
    1 READ (IO1,1002) (MN(I),I=1,78)
      WRITE (IO2,1003) (MN(I),I=1,78)
      LOG3=.FALSE.
      DO 2 I=1,22
C
C 10 42 84  5 22 75 46 65  9 66 20 13 47 81 16 70  8  7 18  3 68 38
C AD BO BZ CE CI CO EN FI FO MO OC OM PA PE RO SH SP SY TE TI VI XY
C
C                                  TI CE SY SP FO AD OM RO TE OC CI
    2 IF (MN(1).EQ.MM(I+44)) GOTO ( 3, 5, 7, 8, 9,10,13,16,18,20,22,
     *38,42,46,47,65,66,68,70,75,81,84),I
C     XY BO EN PA FI MO VI SH CO PE BZ
C
      WRITE (IO2,1004) MN(1),MN(2)
      GOTO 1
C
C *** STORE TITLE ***
C
C  3:TI
    3 DO 4 I=3,78
    4 MN(I+76)=MN(I)
      GOTO 1
C
C *** STORE LATTICE CONSTANTS ***
C
C  5:CE
    5 CALL READN
      DO 6 I=1,6
    6 YC(I+22)=YD(I)
      GOTO 1
C
C *** STORE SYMMETRY OPERATORS ***
C
C  7:SY
C rxf if added
    7 if (ibur.ne.2) CALL READS
      IBUR=0
      GOTO 1
C
C *** GENERATE SYMMETRY OPERATORS ***
C
C  8:SP
    8 CALL SPGR
      GOTO 1
C
C *** STORE NUMBER OF ATOMS AND INPUT FORMAT ***
C
C  9:FO
    9 MI(1)=1
      CALL READN
      MI(9)=YD(1)
      GOTO 1
C
C *** STORE ADDED ATOMS ***
C
C 10:AD
   10 CALL READN
      DO 12 I=1,30
      IF (YD(I).LT.0.)GO TO 1
      MI(10)=MI(10)+1
   12 MN(MI(10))=IFIX(YD(I))
      GOTO 1
C
C *** STORE OMITTED ATOMS ***
C
C 13:OM
   13 CALL READN
      DO 15 I=1,30
      IF (YD(I).LE.-1E20)GO TO 1
      MI(11)=MI(11)+1
   15 MN(MI(11))=IFIX(YD(I))
      GOTO 1
C
C *** UPDATE ROTATION MATRIX ***
C
C 16:RO
   16 CALL READN
      MI(6)=1
      DO 17 I=1,3
   17 CALL ROTATE(I,YD(I))
      GOTO 1
C
C *** STORE TETRAHEDRA CONSTANTS ***
C
C 18:TE
   18 CALL READN
      MI(12)=1
      DO 19 I=1,8
   19 YC(I+6)=YD(I)
      GOTO 1
C
C *** STORE OCTAHEDRA CONSTANTS ***
C
C 20:OC
   20 CALL READN
      MI(13)=1
      DO 21 I=1,8
   21 YC(I+14)=YD(I)
      GOTO 1
C
C *** STORE CIRCLE NUMBERS ***
C
C 22:CI
   22 CALL READN
      MI(35)=MI(35)+1
      YC(MI(35)+78)=YD(1)
      I=2
      LOG=.FALSE.
      IF (YD(35).GT.0.)GO TO 34
   23 K=0
   24 I=I+1
      IF (I.GT.78)GO TO 1
      IF (MN(I).EQ.MM(1)) GOTO 24
      DO 26 J=19,44
   26 IF (MN(I).EQ.MM(J)) GOTO 27
      IF (.NOT.LOG) GOTO 24
      WRITE (IO2,1012) MN(I),MI(35)
      RETURN
   27 LOG=.TRUE.
      IF (MN(I+1).EQ.MM(1))THEN
        L=45
        GOTO 31
      ENDIF
      DO 29 L=19,44
   29 IF (MN(I+1).EQ.MM(L)) GOTO 30
      WRITE (IO2,1012) MN(I+1),MI(35)
      RETURN
   30 IF (MN(I+2).EQ.MM(1)) GOTO 31
      WRITE (IO2,1015) MN(I),MN(I+1),MN(I+2)
      RETURN
   31 K=((J-19)*27+(L-18))*100+MI(35)
      MI(38)=MI(38)+1
      IF (MI(38).GE.44)THEN
        WRITE (IO2,1016)
        RETURN
      ENDIF
      MN(MI(38)+556)=K
      I=I+1
      GOTO 23
   34 DO 36 I=2,30
      MI(7)=MI(7)+1
      IF (MI(7).GE.100)THEN
        WRITE (IO2,1017)
        RETURN
      ENDIF
      IF (YD(I).LE.-1E10)GO TO 37
   36 MN(MI(7)+456)=YD(I)
      MI(7)=MI(7)+1
      IF (MI(7).GE.100)THEN
        WRITE (IO2,1017)
        RETURN
      ENDIF
   37 MN(MI(7)+456)=-10000*MI(35)
      GOTO 1
C
C *** STORE XYZ-RANGE ***
C
C 38:XY
   38 CALL READN
      DO 39 I=1,6
   39 YC(I)=YD(I)
      MI(45)=1
      IF (YD(7).LE.-1E20)GO TO 1
      DO 41 I=7,12
   41 YC(I+22)=YD(I)
      MI(45)=2
      GOTO 1
C
C *** INTERPRET AND STORE BONDLENGTH AND TOLERANCE ***
C
C 42:BO
   42 CALL READN
      MN(1)=YD(1)
      MN(2)=IABS(MN(1))
      IF (MN(2).LE.0)GO TO 45
      MI(MN(2)+13)=MI(MN(2)+13)+1
      YF(MN(2),MI(MN(2)+13),1)=SIGN(YD(2),YD(1))
      DO 44 I=3,12
      IF (YD(I).LE.0.)GO TO 1
   44 YF(MN(2),MI(MN(2)+13),I-1)=YD(I)
      GOTO 1
   45 YC(32)=YD(2)
      YC(33)=YD(3)
      YC(34)=YD(4)
C
C *** THE END MARK HAS BEEN READ, RETURN TO MAIN ***
C
C 46:EN
   46 WRITE (IO2,1013)
      WRITE (IO2,1014)
      RETURN
C
C *** STORE PLOT SPECIFICATIONS ***
C
C 47:PA
   47 CALL READN
      DO 48 I=2,7
   48 IF (YD(I).LT.-1E2) YD(I)=-1.
      IF (YD(1).LE.1E-2) MI(46)=0
      FAC=YD(1)
C
      IF (IFIX(YD(2)).EQ.0)MI(44)=0
      IF (IFIX(YD(3)).EQ.0)MI(48)=0
      IF (IFIX(YD(4)).EQ.0)MI(34)=0
      IF (IFIX(YD(5)).EQ.0)MI(39)=0
      IF (IFIX(YD(6)).EQ.0)MI(40)=0
      IF (IFIX(YD(7)).EQ.0)MI(46)=0
      IF (YD(8).GT.0.)YC(89)=YD(8)
      IF (YD(9).GT.0.)NRES=NINT(YD(9))
      IF (YD(10).GT.0.)RES=YD(10)
      GOTO 1
C
C *** STORE FILE NUMBERS ***
C
C 65:FI
   65 CALL READN
CESPOIR      IF (YD(1).GT.0.) IO1=IFIX(YD(1))
CESPOIR      IF (YD(2).GT.0.) IO2=IFIX(YD(2))
CESPOIR      IF (YD(3).GT.0.) IO3=IFIX(YD(3))
CESPOIR      IF (YD(4).GT.0.) IO4=IFIX(YD(4))
      GOTO 1
C
C *** STORE DEVICE AND SCALE PARAMETERS ***
C
C 66:MO
   66 CALL READN
      ID=IFIX(YD(1))
      DO 67 I=2,7
      IF (YD(I).LT.-1E10) GOTO 1
   67 PM(I-1)=YD(I)
      GOTO 1
C
C *** STORE VIEW DIRECTION
C
C 68:VI
   68 CALL READN
      IF(MI(6).EQ.1)WRITE(IO2,1019)
      DO 69 I=1,6
   69 PP(I)=YD(I)
      GOTO 1
C
C *** STORE DO LOOP PARAMETERS FOR SHADING  ***
C
C 70:SH
   70 CALL READN
      I=1
      MI(41)=IFIX(YD(I))
   71 I=I+1
      IF (YD(I).LT.-1E20) GOTO 1
      IF (YD(I+1).LE.-1E20)THEN
        LOG3=.TRUE.
        YD(I+1)=-YD(I+1)
      ENDIF
      IS=IS+1
      NH(IS,1)=IFIX(YD(I))
      NH(IS,3)=MI(41)
      IF (YD(I+1).LT.0.)THEN
        NH(IS,2)=-IFIX(YD(I+1))
        I=I+1
      ELSE
        NH(IS,2)=NH(IS,1)
        IF (LOG3) GOTO 1
      ENDIF
      GOTO 71
C
C *** STORE PENS TO BE USED  ***
C
C 75:CO
   75 I=2
      K=1
   76 I=I+1
      IF(I.GT.76) GOTO 1
      IF(MN(I).EQ.MM(1)) GOTO 76
      DO 77 L=19,44
   77 IF(MN(I).EQ.MM(L)) MN(I)=L
      DO 78 L=19,44
   78 IF(MN(I+1).EQ.MM(L)) GOTO 82
      L=45
   82 MN(I+1)=L
      ML(K,1)=MN(I)*27+MN(I+1)-531
      IF(MN(I+1).NE.45) I=I+1
   79 I=I+2
      IF(MN(I).EQ.MM(1)) GOTO 79
      DO 80 L=4,9
   80 IF(MN(I).EQ.MM(L)) ML(K,2)=L-3
      K=K+1
      GOTO 76
C
C *** STORE STEREOVIEW PARAMETERS ***
C
C 81:PE
   81 CALL READN
      YB(25)=YD(1)
      GOTO 1
C
C *** STORE ATOMS TO BE DRAWN IN BIZARRE POLYHEDRA ***
C
C 84:BZ
   84 CALL READN
      DO 85 I=1,5
      IF(YP(I,1).EQ.0.) GOTO 86
   85 CONTINUE
      WRITE (IO3,1018)
      WRITE (IO2,1018)
   86 YP(I,3)=YD(1)
      K=0
   92 K=K+1
      DO 87 J=19,44
   87 IF(MN(K).EQ.MM(J)) GOTO 91
      GOTO 92
   91 MN(K)=J
      DO 88 J=19,44
   88 IF(MN(K+1).EQ.MM(J)) GOTO 89
      J=45
   89 MN(K+1)=J
      YP(I,1)=MN(K)*27+MN(K+1)-531
      GOTO 1
 1000 FORMAT ('1')
C1001 FORMAT (1X,29('*'),' INPUT CONTROL FROM UNIT',I2,1X,28('*'))
 1002 FORMAT (2A2,76A1)
 1003 FORMAT (' * ',2A2,71A1,' *')
 1004 FORMAT (' * !!! WARNING !!! THE LAST OPTION ',2A2,' IS IGNORED')
C1005 FORMAT (40A2)
C1006 FORMAT ('  !!! WARNING !!! THERE ARE TOO MANY CONSTRAINTS FOR BO
C    &NDS')
C1007 FORMAT (20A4)
C1008 FORMAT (2A1,I2,5X,3F10.5)
C1009 FORMAT (2A1,I2,26X,3F10.6)
C1010 FORMAT (2A1,I2,3F10.6)
C1011 FORMAT (80A1)
 1012 FORMAT ('  !!! ERROR !!! THE CHARACTER ',A1,' IN OPTION CIRC NO.
     & ',I3,' IS ILLEGAL, THE JOB IS TERMINATED')
 1013 FORMAT ('*',78X,'*')
 1014 FORMAT (1X,79('*'))
 1015 FORMAT (1X,3A1,' IS NOT A COMMON ATOM NAME, ONLY 2 LETTERS PER NAM
     &E ARE ACCEPTED, JOB IS TERMINATED')
 1016 FORMAT (' ONLY 44 SPECIFICATIONS ARE ACCEPTED DEFINED IN OPTION
     &CIRC,',/,' USE ATOM NUMBERS INSTEAD OF ATOM NAMES AND TRY AGAIN,',
     */,'JOB IS TERMINATED')
 1017 FORMAT (' THESE ARE TOO MANY CIRCLE SPECIFICATIONS,'/,' USE ATOM N
     *AMES INSTEAD OF NUMBERS AND TRY AGAIN'/'JOB IS TERMINATED')
 1018 FORMAT (' TOO MANY ATOMS IN BIZARRE POLYHEDRA')
 1019 FORMAT (' PREVIOUS ROTATIONS ARE APPLIED AFTER THIS INSTRUCTION')
      END
      SUBROUTINE READN
      COMMON/ALL1/YA(4,6000),YP(185),YD(40),YY(1496),MI(104),MM(70),
     *MN(8000),MO(17759),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      LOGICAL PNT,II
      DO 1 I=3,80
    1 IF (MN(I).EQ.MM(16)) MN(I)=MM(13)
      DO 2 I=1,30
    2 YD(I)=-1E30
      MN(1)=2
      MN(2)=0
      YD(35)=1.
    3 YD(31)=0.
      YD(32)=10.
      YD(33)=1.
      MI(26)=0
      EXP=1.
      PNT=.FALSE.
      II=.FALSE.
    4 MN(1)=MN(1)+1
      IF (MN(1)-76) 5,5,13
    5 IF (II) GOTO 6
      IF (MN(MN(1)).EQ.MM(1).OR.MN(MN(1)).EQ.MM(2)) GOTO 4
      II=.TRUE.
    6 DO 7 J=3,14
    7 IF (MN(MN(1)).EQ.MM(J)) GOTO 9
      IF (MN(MN(1)).EQ.MM(1)) GOTO 8
      YD(35)=-1.
      GOTO 3
    8 MN(2)=MN(2)+1
      YD(MN(2))=YD(31)*YD(33)
      GOTO 3
    9 IF (J-13) 10,11,12
   10 YD(34)=1.*J-3.
      IF (PNT)THEN
        MI(26)=MI(26)+1
        EXP=0.1*EXP
      ENDIF
      YD(31)=YD(31)*YD(32)+YD(34)*EXP
      GOTO 4
   11 PNT=.TRUE.
      YD(32)=1.
      GOTO 4
   12 YD(33)=-1.
      GOTO 4
   13 RETURN
      END
C------------------------------------------------------------------------
      SUBROUTINE READS
      COMMON/ALL1/YA(4,6000),YP(185),YD(40),YY(48,3,4),YF(920),MI(104),
     *MM(70),MN(8000),MO(17759),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      LOGICAL II,XYZ
C
      MN(1)=2
      II=.FALSE.
      XYZ=.FALSE.
C do not count if this was unit matrix
   99 IF(.NOT.XYZ)MI(5)=MI(5)+1
      XYZ=.TRUE.
      MN(2)=1
    1 YD(1)=1.
    2 MN(1)=MN(1)+1
      IF (MN(1).GT.78)GO TO 12
C skip blank or +
      IF (MN(MN(1)).EQ.MM(1).OR.MN(MN(1)).EQ.MM(2)) GOTO 2
C ',': next line of symmetry operations
      IF (MN(MN(1)).EQ.MM(16))THEN
        MN(2)=MN(2)+1
        GOTO 2
      ENDIF
C ';': new symmetry relationship
      IF (MN(MN(1)).EQ.MM(15))GO TO 99
C '-' in rotation matrix
      IF (MN(MN(1)).EQ.MM(14))THEN
        YD(1)=-1.
        GOTO 2
      ENDIF
    6 DO 7 J=4,9
    7 IF (MN(MN(1)).EQ.MM(J)) GOTO 9
      IF (II) WRITE (IO3,1000)
      IF (II) WRITE (IO2,1000)
      DO 8 J=42,44
C rotation (J-40: 2-4 for x-z)
      IF (MN(MN(1)).EQ.MM(J))THEN
        YY(MI(5),MN(2),J-40)=YD(1)
        IF(YD(1).NE.1. .OR. MN(2).NE.J-40)XYZ=.FALSE.
        GOTO 1
      ENDIF
    8 CONTINUE
      WRITE (IO3,1000)
      WRITE (IO2,1000)
    9 IF (II)THEN
        II=.FALSE.
        XYZ=.FALSE.
C shift
        YY(MI(5),MN(2),1)=YD(2)/FLOAT(J-3)
        GOTO 2
      else
        II=.TRUE.
C shift numerator
        YD(2)=FLOAT(J-3)
C skip next and look for denominator
        MN(1)=MN(1)+2
        GOTO 6
      ENDIF
   12 RETURN
 1000 FORMAT (' UNKNOWN SYMBOL')
      END
C-----------------------------------------------------------------------
      SUBROUTINE RETIN
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(90),MN(1000,8),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/PLOTHP/IOK,VFT,LREC,IPENA,NEWLIN,C75
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
C
C *** CHECK IF B(23),B(24) IS IN THE FACE DEFINED BY F(I) ***
C
      DO 1 I=3,5
      IF(MB(I+1).LE.0)GO TO 2
    1 CONTINUE
      I=6
    2 K=I+MOD(I,2)-1
      DO 3 J=1,K,2
      MG(12)=MOD(J,I)+1
      MG(13)=MOD(J+1,I)+1
      MG(14)=MOD(J-1,I)+1
      P2=      PFAKT(MB(MG(12)))
      IF(NSTOP.EQ.1)RETURN
      P3=      PFAKT(MB(MG(13)))
      IF(NSTOP.EQ.1)RETURN
      P4=      PFAKT(MB(MG(14)))
      IF(NSTOP.EQ.1)RETURN
      YB(13)=P3*YA(2,MB(MG(13)))-P2*YA(2,MB(MG(12)))
      YB(14)=P3*YA(3,MB(MG(13)))-P2*YA(3,MB(MG(12)))
      YB(15)=P4*YA(2,MB(MG(14)))-P2*YA(2,MB(MG(12)))
      YB(16)=P4*YA(3,MB(MG(14)))-P2*YA(3,MB(MG(12)))
      YD(11)=YB(23)-P2*YA(2,MB(MG(12)))
      YD(12)=YB(24)-P2*YA(3,MB(MG(12)))
      YD( 5)=(YD(11)*YB(14)-YD(12)*YB(13))/(YB(15)*YB(14)-YB(16)*YB(13))
      YD( 6)=(YD(11)*YB(16)-YD(12)*YB(15))/(YB(13)*YB(16)-YB(14)*YB(15))
      YD(10)=AMIN1(YD(5),YD(6))
      IF (YD(10).LE.0)GO TO 4
    3 CONTINUE
    4 MG(17)=-IFIX(SIGN(1.,YD(10)))
      IF(MG(17).GE.0)GO TO 22
C
C *** CHECK IF B(23),B(24) IS OBSCURED ***
C
      IF (MG(1).NE.0 .AND. MG(4).NE.0)THEN
        DO 18 J=1,MG(4)
        MG(3)=MC(J)
        CALL INOUT
        IF (MG(17).LE.0)GO TO 22
   18   CONTINUE
      ENDIF
      IF (MG(2).NE.0 .AND. MG(5).NE.0)THEN
        DO 20 J=1,MG(5)
        P=      PFAKT(MO(MD(J)))
      IF(NSTOP.EQ.1)RETURN
        YB( 8)=P*RAD (MO(MD(J)))
        YB(19)=P*YA(2,MO(MD(J)))
        YB(20)=P*YA(3,MO(MD(J)))
        YD(5)=SQRT((YB(23)-YB(19))**2+(YB(24)-YB(20))**2)
        IF (YD(5).LE.YB(8))GO TO 22
   20   CONTINUE
      ENDIF
      IF(MG(10).EQ.0.OR.MG(11).LE.0) GOTO 21
      DO 33 J=1,MG(11)
      MG(3)=ME(J)
      CALL OUTIN
      IF(MG(17).LE.0)GO TO 22
   33 CONTINUE
C
C
C *** DRAW LINE ***
C
   21 CALL PLOT(YB(23),YB(24),3)
        XP=YB(23)*C75+512.
        YPP=YB(24)*C75+375.
        CALL PLOTHGL(XP,YPP,3)
      CALL PLOT(YB(23),YB(24),2)
        XP=YB(23)*C75+512.
        YPP=YB(24)*C75+375.
        CALL PLOTHGL(XP,YPP,2)
   22 CONTINUE
      RETURN
      END
C--------------------------------------------------------------------------
      SUBROUTINE ROTATE(NAX,ANGLE)
      EQUIVALENCE (YC35,YC(35))
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(48,3,4),
     *YF(5,15,12),MH(20),MI(84),ML(10,2),MM(70),MN(8000),MO(3024),
     *MU(14735),IO1,IO2,IO3,IO4
C
      IF(ANGLE.EQ.0.)RETURN
      T=ANGLE*YC35
C TO KEEP COMPATIBILITY WITH OLD VERSION
C ROTATION AROUND Y-AXIS IN OPPOSITE DIRECTION
*BER**Da kein I deklariert --> fr I NAX gesetzt*********************
*BER  IF(I.EQ.2)T=-T
      IF(NAX.EQ.2)T=-T
*********************************************************************
      C=COS(T)
      S=SIN(T)
      I1=3*(1+MOD(NAX  ,3))
      I2=3*(1+MOD(NAX+1,3))
      DO 1 I=1,3
      T=YB(I+I1)
      YB(I+I1)=C*T-S*YB(I+I2)
1     YB(I+I2)=S*T+C*YB(I+I2)
C      WRITE(IO2,13579)NAX,ANGLE
C13579 FORMAT(/' ACHSE, WINKEL:',I2,F8.2)
C      WRITE(IO2,97531)(YB(I),I=4,12)
C97531 FORMAT(3F10.6)
      RETURN
      END
C------------------------------------------------------------------------
      SUBROUTINE SCOB
      CHARACTER*28 CR
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(10,2),MM(70),MN(1000,8),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
CNOUSEINTEGER Z(28)
C
      I=0
      M=MV(MG(10),14)
      DO 106 K=2,MU(M,1,3)+1
      DO 200 N=2,4
      IF(MV(MG(10),MU(M,K,1)).LE.0)GO TO 900
      IF(MV(MG(10),MU(M,K,2)).LE.0)GO TO 900
      IF(MV(MG(10),MU(M,K,3)).LE.0)GO TO 900
      YB(N-1)=YA(N,MV(MG(10),MU(M,K,3)))-YA(N,MV(MG(10),MU(M,K,2)))
      YB(N+2)=YA(N,MV(MG(10),MU(M,K,1)))-YA(N,MV(MG(10),MU(M,K,2)))
200   CONTINUE
      YB(7)=YB(2)*YB(6)-YB(3)*YB(5)
      YB(8)=YB(3)*YB(4)-YB(1)*YB(6)
      YB(9)=YB(1)*YB(5)-YB(2)*YB(4)
      DO 201 N=2,4
  201 YB(N+2)=YA(N,MV(MG(10),13))-YA(N,MV(MG(10),MU(M,K,2)))
      YB(4)=YB(4)*YB(7)+YB(5)*YB(8)+YB(6)*YB(9)
      IF(ABS(YB(4)).LE.1E-4.AND.MU(MV(MG(10),14),1,3).EQ.1) GOTO 150
      YB(5)=YB(9)*(YB(25)-YA(4,MV(MG(10),MU(M,K,2))))
     &-YB(7)*YA(2,MV(MG(10),MU(M,K,2)))-YB(8)*YA(3,MV(MG(10),MU(M,K,2)))
      IF(YB(4)*YB(5).GE.-1E-4) GOTO 106
  150 DO 135 L=1,MU(M,K,7)
      JJ=MOD(L,MU(M,K,7))+1
      N1=MU(M,K,L)
      N2=MU(M,K,JJ)
      DO 137 NN=1,I
      IF(MA(NN,1).EQ.MV(MG(10),N1).AND.MA(NN,2).EQ.MV(MG(10),N2))
     * GOTO 135
      IF(MA(NN,2).EQ.MV(MG(10),N1).AND.MA(NN,1).EQ.MV(MG(10),N2))
     * GOTO 135
  137 CONTINUE
      I=I+1
      MA(I,1)=MV(MG(10),N1)
      MA(I,2)=MV(MG(10),N2)
  135 CONTINUE
  106 CONTINUE
  132 IF(I.EQ.20) GOTO 139
      DO 140 J=I+1,20
      MA(J,1)=0
  140 MA(J,2)=0
  139 CONTINUE
      GO TO 1000
900   NSTOP=1
1000  CONTINUE
      RETURN
      END
C-------------------------------------------------------------------------
      SUBROUTINE SEARCH(J,II,MAXPOL)
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(106),YZ(100,5),YE(200,5),
     *MH(20),MI(104),MM(70),MN(1000,8),MO(17759),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      LOGICAL LOG
      MAXPOL=0
      IF (MI(40) .LE. 0)GO TO 20
      YC(91)=YC(15)
      YC(92)=YC(16)
      YC(93)=YC(17)
      YC(94)=YC(18)
      MH(1)=7
    2 LOG=.FALSE.
      YC(95)=1.414*(YC(17)+YC(18))
      MI(37)=MI(37)+1
      IF (MI(37).GT.500)THEN
        WRITE (IO3,1000)
        WRITE (IO2,1000)
        MAXPOL=1
        RETURN
      ENDIF
      MH(4)=3
      DO 10 I=1,II
      IF (ABS(YE(I,3)-YC(91)) .GT. YC(92))GO TO 10
      IF (ABS(YE(I,4)-YC(93)) .GT. YC(94))GO TO 10
      IF (ABS(YE(I,5)-YC(93)) .GT. YC(94))GO TO 10
      MH(2)=IFIX(YE(I,1))
      MH(3)=IFIX(YE(I,2))
      IF (.NOT.LOG)THEN
        MN(MI(37),2)=MH(2)
        MN(MI(37),3)=MH(3)
        LOG=.TRUE.
        GOTO 10
      ENDIF
      MH(5)=MH(4)
      DO 7 K=2,MH(5)
      IF (MH(2) .EQ. MN(MI(37),K))GO TO 8
    7 CONTINUE
      MH(4)=MH(4)+1
        IF(MI(37).GT.500)THEN
        WRITE(20,*)'MI(37) > 500 in subroutine SEARCH'
        MAXPOL=1
        RETURN
        ENDIF 
        IF(MH(4).GT.8)THEN
        WRITE(20,*)'MH(4) > 8 in subroutine SEARCH'
        MAXPOL=1
        RETURN
        ENDIF 
      MN(MI(37),MH(4))=MH(2)
    8 DO 9 K=2,MH(5)
      IF (MH(3) .EQ. MN(MI(37),K))GO TO 10
    9 CONTINUE
      MH(4)=MH(4)+1
        IF(MI(37).GT.500)THEN
        WRITE(20,*)'MI(37) > 500 in subroutine SEARCH'
        MAXPOL=1
        RETURN
        ENDIF 
        IF(MH(4).GT.8)THEN
        WRITE(20,*)'MH(4) > 8 in subroutine SEARCH'
        MAXPOL=1
        RETURN
        ENDIF 
      MN(MI(37),MH(4))=MH(3)
   10 CONTINUE
      IF (MH(4) .LT. MH(1))THEN
        MI(37)=MI(37)-1
        GOTO 18
      ENDIF
      MN(MI(37),1)=J
      YB(1)=0.
      P=PFAKT(MN(MI(37),1))
      IF(NSTOP.EQ.1)RETURN
      YB(2)=P*YA(2,MN(MI(37),1))
      YB(3)=P*YA(3,MN(MI(37),1))
      DO 13 I=2,MH(1)
      P=PFAKT(MN(MI(37),I))
      IF(NSTOP.EQ.1)RETURN
      YB(4)=P*YA(2,MN(MI(37),I))
      YB(5)=P*YA(3,MN(MI(37),I))
      YB(1)=AMAX1(YB(1),SQRT((YB(2)-YB(4))**2+(YB(3)-YB(5))**2))
   13 CONTINUE
      MN(MI(37),8)=IFIX(YB(1)*1E3)
      IF (MH(1) .NE. 7)GO TO 21
      DO 17 I=2,3
      YB(1)=YA(2,MN(MI(37),I))
      YB(2)=YA(3,MN(MI(37),I))
      YB(3)=YA(4,MN(MI(37),I))
      YC(96)=0.
      DO 16 J=I+1,9-I
      YB(4)=YA(2,MN(MI(37),J))
      YB(5)=YA(3,MN(MI(37),J))
      YB(6)=YA(4,MN(MI(37),J))
      YC(95)=SQRT((YB(4)-YB(1))**2+(YB(5)-YB(2))**2+(YB(6)-YB(3))**2)
      IF (YC(95) .GT. YC(96))THEN
        YC(96)=YC(95)
        MH(4)=J
      ENDIF
   16 CONTINUE
      J=MN(MI(37),MH(4))
      MN(MI(37),MH(4))=MN(MI(37),9-I)
      MN(MI(37),9-I)=J
   17 CONTINUE
   18 IF (MH(1) .NE. 7)GO TO 21
      IF (MI(39) .LE. 0)GO TO 21
   20 YC(91)=YC( 7)
      YC(92)=YC( 8)
      YC(93)=YC( 9)
      YC(94)=YC(10)
      MH(1)=5
      GOTO 2
   21 RETURN
 1000 FORMAT (1X,'THE MAXIMUM NUMBER OF POLYHEDRA IS  500, THE JOB IS TE
     &RMINATED')
      END
C------------------------------------------------------------------------
      SUBROUTINE SORT
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(376),YR(194,3),MF(600),
     *MG(68),MI(25933),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DO 5 I=1,MG(15)-1
      K=I
      DO 2 J=I+1,MG(15)
      IF (YR(K,3).GT.YR(J,3))K=J
    2 CONTINUE
      IF (K.NE.I)THEN
        DO 4 J=1,3
        YB(14) =YR(I,J)
        YR(I,J)=YR(K,J)
    4   YR(K,J)=YB(14)
      ENDIF
    5 CONTINUE
      RETURN
      END
C----------------------------------------------------------------------
      SUBROUTINE SORT1(L)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(106),YZ(100,5),
     *YE(1020),MI(25933),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DO 4 I=1,L-1
      K=0
      YB(24)=YZ(I,5)
      DO 1 J=I+1,L
      IF (YB(24).LT.YZ(J,5)) GOTO 1
      YB(24)=YZ(J,5)
      K=J
    1 CONTINUE
      IF (K) 4,4,2
    2 DO 3 J=1,5
      YB(J)  =YZ(I,J)
      YZ(I,J)=YZ(K,J)
    3 YZ(K,J)=YB(J)
    4 CONTINUE
      RETURN
      END
C----------------------------------------------------------------------
      SUBROUTINE SPGR
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),YY(48,3,4),
     *YF(920),MI(104),MM(70),MN(25759),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
C rxf common added
      COMMON/MODI/ID,PM(6),RES,FAC,NRES,IBUR
      CHARACTER*28 CR
      DIMENSION IROTM(24,3,3),TS(24,3)
C rxf
      if (ibur.ne.0) then
      ibur=2
      CALL BURZ (1,ISYSTM,MI(44),IBRAVL,MI(5),IROTM,TS,MN(3))
C
      YB(13)=0.
      YB(14)=0.
      YB(15)=0.
C      WRITE(20,*)'  IBRAVL = ',IBRAVL
      IF(MI(44).EQ.0)THEN
        WRITE(IO2,1001)
      ELSE
        WRITE(IO2,1002)
      ENDIF
      WRITE(IO2,1003)
      DO 1 N=1,MI(5)
        WRITE(IO2,1004)(TS(N,I),(IROTM(N,I,J),J=1,3),I=1,3)
        DO 1 I=1,3
          YY(N,I,1)=TS(N,I)
          DO 1 J=1,3
1           YY(N,I,J+1)=FLOAT(IROTM(N,I,J))
C rxf else inserted
      else
      do 100 i=3,78
      do 100 j=19,25
  100 if (mn(i).eq.mm(j)) ibravl=j-18
      endif
      GOTO (9,7,6,5,4,3,8,2),IBRAVL
    2 WRITE (IO2,1005)
      GO TO 9
    3 MI(4)=2
      YB(16)=0.5
      YB(17)=0.5
      YB(18)=0.5
      GO TO 9
    4 MI(4)=4
      YB(17)=0.5
      YB(18)=0.5
      YB(19)=0.5
      YB(21)=0.5
      YB(22)=0.5
      YB(23)=0.5
      GO TO 9
    5 MI(4)=2
      YB(16)=0.5
      YB(17)=0.5
      GO TO 9
    6 MI(4)=2
      YB(16)=0.5
      YB(18)=0.5
      GO TO 9
    7 MI(4)=2
      YB(17)=0.5
      YB(18)=0.5
      GO TO 9
    8 MI(4)=3
      YB(16)=0.33333333
      YB(17)=0.66666667
      YB(18)=0.66666667
      YB(19)=0.66666667
      YB(20)=0.33333333
      YB(21)=0.33333333
9     WRITE(IO2,1006)
      WRITE(IO2,1007)(YB(I+12),I=1,3*MI(4))
      RETURN
1001  FORMAT(' * -  no symmetry center at origin')
1002  FORMAT(' * -  symmetry center at origin')
1003  FORMAT(' * -  symmetry operations:')
1004  FORMAT(' * - ',3(F10.5,3I3))
1005  FORMAT(' * !!! bravais type not recognized,',
     *  ' default type is primitive !!!')
1006  FORMAT(' * -  translation vectors:')
1007  FORMAT(' * - ',3F10.5)
      END
C-------------------------------------------------------------------------
      SUBROUTINE SPLIT(I,J,LOG)
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(106),YZ(100,5),YE(200,5),
     *MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(5,21,7),MV(1000,14),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/BZR/BX,BY,BZ
      INTEGER BY(13),BZ(4),BX(10)
CNOUSEDIMENSION W(30)
      LOGICAL LOG,LOG1
*BER**LOG1 nicht def., Annahme: Voreinstellung LOG1=FALSE************
      LOG1=.FALSE.
*********************************************************************
C
      I1=1
      I3=2
      I9=0
      I4=1
    7 CONTINUE
      IF(I1.LE.0.OR.I1.GT.4)THEN
      NSTOP=1
      GO TO 13
      ENDIF
      IF(BZ(I1).EQ.0) LOG=.TRUE.
      IF(LOG) GOTO 2
      I2=BZ(I1)
      I9=I9-1
      I1=I1+1
      GOTO 3
    2 I2=MU(I,1,3)+I3
      I3=I3+1
    3 I5=1
      IF(I4.EQ.1.AND.(.NOT.LOG1)) I6=I2
    6 MU(I,I2,I5)=BY(I4)
      I5=I5+1
      IF(LOG1) GOTO 11
      IF(BX(I4).EQ.1) GOTO 5
   11 LOG1=.FALSE.
      IF(I4.EQ.MIN(10,BY(11))) GOTO 9
      I4=I4+1
      GOTO 6
    9 I7=MU(I,I6,7)
      IF(.NOT.LOG) I9=I9+1
      I1=I1-1
      DO 8 I8=1,I5-1
      MU(I,I6,I8+I7)=MU(I,I2,I8)
    8 MU(I,I2,I8)=0
      MU(I,I6,7)=I7+I5-1
      GOTO 16
    5 MU(I,I2,I5)=J
      MU(I,I2,7)=I5
      I9=I9+1
      IF(I4.EQ.MIN(10,BY(11))) GOTO 10
      LOG1=.TRUE.
      GOTO 7
   10 I10=MU(I,I6,7)
      MU(I,I6,I10+1)=BY(I4)
      MU(I,I6,7)=I10+1
   16 MU(I,1,3)=MU(I,1,3)+I9
      IF(LOG) GOTO 13
      DO 12 I5=4,I1,-1
      IF(BZ(I5).LE.0) GOTO 12
      DO 14 I2=1,7
      DO 15 I3=BZ(I5),MU(I,1,3)
   15 MU(I,I3,I2)=MU(I,I3+1,I2)
   14 MU(I,I3,I2)=0
      MU(I,1,3)=MU(I,1,3)-1
   12 CONTINUE
   13 RETURN
      END
C-------------------------------------------------------------------------
*BER  SUBROUTINE TIMER(N,LAST_TIMER)
*BER  INTEGER TIMTAB(10),NUMTAB(10),TIME
*BER  CALL LIB$STAT_TIMER(2,TIME)
*BER  LAST_TIMER=L
*BER  TIMTAB(N)=TIMTAB(N)+TIME
*BER  NUMTAB(N)=NUMTAB(N)+1
*BER  L=N
*BER  ENTRY TINIT
*BER  CALL LIB$INIT_TIMER
*BER  RETURN
*BER
*BER  ENTRY TIMOUT
*BER  DO I=1,10
*BER    IF(NUMTAB(I).GT.0)THEN
*BER      FTIME=0.01*FLOAT(TIMTAB(I))
*BER      WRITE(6,100)I,NUMTAB(I),FTIME,FTIME/NUMTAB(I)
*BER      SUM=SUM+FTIME
*BER    ENDIF
*BER  ENDDO
*BER  WRITE(6,101)SUM
*BER  RETURN
*BER100   FORMAT(I5,I8,F12.3,F12.4)
*BER101   FORMAT(F25.3)
*BER  END
C-----------------------------------------------------------------------------
      SUBROUTINE UNPACK(I)
      COMMON/ALL1/YA(4,6000),YP(1701),MH(20),MI(104),MM(70),MN(25759),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      DIMENSION IYA(4,6000)
      EQUIVALENCE (IYA,YA)
      PARAMETER (IYAQ=1000)
      PARAMETER (IYAQ1=10)
C
      MH(12)=ABS(IYA(1,I))
      MH(11)=MH(12)/IYAQ
      MH(12)=MH(12)-MH(11)*IYAQ
      MH(1)=(MH(11)-1)/27
      MH(2)=MH(11)-MH(1)*27
      MH(3)=MH(12)/IYAQ1
      MH(1)=MM(MOD(MH(1)+18,44)+1)
      MH(2)=MM(MOD(MH(2)+17,44)+1)
      RETURN
      END
C---------------------------------------------------------------------------
      SUBROUTINE VIEW
      EQUIVALENCE (YC36,YC(36))
      COMMON/ALL1/YA(4,6000),YP(70),YB(25),YC(90),YD(40),YY(48,3,4),
     *YF(900),MH(20),MI(104),MM(70),MN(8000),MO(3024),MU(14735),
     *IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      CHARACTER*28 CR
      COMMON/ADD1/NH(50,3),PP(10),IS
      COMMON/ADD2/TIT,BL
      CHARACTER TIT*40,BL*4
      DIMENSION ROT1(3),ROT2(3)
C
C TURN BACKWARD ACCUMULATED ROTATION MATRIX
      ROT2(3)=0.
      IF(YB(4).NE.0. .OR. YB(7).NE.0.)
     *ROT2(3)=YC36*ATAN2(YB(7),YB(4))
      CALL ROTATE(3,-ROT2(3))
      ROT2(2)=YC36*ATAN2(-YB(10),YB(4))
      CALL ROTATE(2,-ROT2(2))
      ROT2(1)=YC36*ATAN2(YB(11),YB(8))
      CALL ROTATE(1,-ROT2(1))
      DO 11 I=4,12
      YB(I)=0.
      IF(MOD(I,4).EQ.0)YB(I)=1.
11    CONTINUE
C
      IF(PP(2).LE.-1E6)PP(2)=1.
      IF(PP(3).LE.-1E6)THEN
        MO(7)=IFIX(PP(1))
        MO(8)=IFIX(PP(2))
        DO 1 I=1,3
        PP(I  )=YA(I+1,MO(7))
    1   PP(I+3)=YA(I+1,MO(8))
      ENDIF
      DO 2 I=1,3
      IF(PP(I+3).LE.-1E6)PP(I+3)=0.
    2 YD(I)=PP(I)-PP(I+3)

      CALL CORTH
      ROT1(1)=0.
      IF(YD(5).NE.0. .OR. YD(6).NE.0.)
     *ROT1(1)=YC36*ATAN2(YD(5),YD(6))
      ROT1(2)=YC36*ATAN2(-YD(4),SQRT(YD(5)*YD(5)+YD(6)*YD(6)))
      CALL ROTATE(1,ROT1(1))
      CALL ROTATE(2,ROT1(2))
C     place axis with longest projection across paper:
C     try a, b and c-axis (origin # 2,3 and 4)
C     with XYZR as factor, when given:
      PP(3)=0.
      DO 5 I=1,3
      IF(MI(45).EQ.0)THEN
        PP(6)=1.
      ELSE
        PP(6)=ABS(YC(I+I)-YC(I+I-1))
      ENDIF
      DO 3 J=1,3
    3 YD(J)=YA(J+1,I+1)*PP(6)
      CALL CORTH
C     projection: (4,5,6 is x,y,r)
      PP(4)=YB( 4)*YD(4)+YB( 5)*YD(5)+YB( 6)*YD(6)
      PP(5)=YB( 7)*YD(4)+YB( 8)*YD(5)+YB( 9)*YD(6)
      PP(6)=SQRT(PP(4)*PP(4)+PP(5)*PP(5))
      IF(PP(6).GT.PP(3))THEN
C     this is longer. memorize!
        DO 4 J=1,3
    4   PP(J)=PP(J+3)
      ENDIF
    5 CONTINUE
      ROT1(3)=YC36*ATAN2(-PP(2),PP(1))
      CALL ROTATE(3,ROT1(3))
      WRITE (IO2,1001) ROT1
      DO 6 I=1,3
    6 CALL ROTATE(I,ROT2(I))
      WRITE (IO2,1002) ROT2
 1001 FORMAT(' THE ROTATION ANGLES =',3F8.3,
     *' ARE CALCULATED FROM THE VIEW DIRECTION IN OPTION VIEW')
 1002 FORMAT(' UPON THIS VIEW ROTATIONS OF',3F8.3,' ARE PERFORMED.')
      RETURN
      END
C
C  MODIF ALB : A LOT............
C
      SUBROUTINE WRITE(LL,TEMPO)
      COMMON/ALL1/YA(4,6000),YP(225),MA(1476),MH(20),MI(104),
     *MM(70),MN(1000,8),MO(3759),MV(1000,14),IO1,IO2,IO3,IO4
      COMMON/ALL2/CR,NSTOP
      COMMON/WRL/AID,NAID,CART,IO5,IO7,IO8,NCIRC,ICOL,NCOL,NPT
      COMMON/WRLOUT/COOR,ICON,IO6,JCOR,JICO,NPOL,NCO,NID
      CHARACTER*28 CR
	CHARACTER*(*) TEMPO
      INTEGER AID(6000,2)
      DIMENSION NA(13,3),ICON(6000,2),NPOL(6000),COLOR(20,3)
      DIMENSION NAID(6000),CART(6000,3),COOR(6000,3),NCIRC(6000)
      DIMENSION ICOL(6000),NCOL(20,2),NCO(6000),NID(6000)
      DIMENSION XM(6)
C
      WRITE (IO2,1000)
      WRITE (IO2,1010)
CESPOIR      WRITE (6,1000)
      JCOR=0
      JICO=0
      NUM=0
      IO6=IO5
      IF(MI(37))150,150,200
200   DO 5 I=1,MI(37)-1
      K=0
      O=DIST(MN(I,1))
      DO 2 J=I+1,MI(37)
      IF(O-DIST(MN(J,1))) 2,1,1
    1 O=DIST(MN(J,1))
C     IF (O-A(H(J,1),4)) 2,1,1
C   1 O=A(H(J,1),4)
      K=J
    2 CONTINUE
      IF (K) 5,5,3
    3 DO 4 J=1,8
      L=MN(I,J)
      MN(I,J)=MN(K,J)
    4 MN(K,J)=L
    5 CONTINUE
      DO 11 I=1,MI(37)
      NUM=NUM+1
      DO 7 J=1,7
      IF (MN(I,J)) 8,8,6
    6 CALL UNPACK(MN(I,J))
      NA(J,1)=MH(1)
      NA(J,2)=MH(2)
    7 NA(J,3)=MH(3)
    8 IF (J-7) 9,10,10
    9 WRITE (IO2,1001)NUM,(MN(I,J),(NA(J,L),L=1,3),J=1,5)
      DO 100 J=1,5
      IF(MN(I,J).EQ.-1) GO TO 11
100   CONTINUE
      JCOR=JCOR+1
      WRITE(IO2,1005)NUM,MN(I,1),AID(MN(I,1),1),
     &AID(MN(I,1),2),(CART(MN(I,1),J),J=1,3)
      COOR(JCOR,1)=CART(MN(I,1),1)
      COOR(JCOR,2)=CART(MN(I,1),2)
      COOR(JCOR,3)=CART(MN(I,1),3)
      NPOL(JCOR)=0
      NID(JCOR)=MN(I,1)
      JT=JCOR
      DO 12,II=1,4
      JCOR=JCOR+1
      NPOL(JCOR)=4
      NID(JCOR)=MN(I,1)
      NCO(JCOR)=ICOL(MN(I,1))
      JJ=II+1
      WRITE(IO2,1006)MN(I,JJ),AID(MN(I,JJ),1),
     &AID(MN(I,JJ),2),(CART(MN(I,JJ),J),J=1,3)
      COOR(JCOR,1)=CART(MN(I,JJ),1)
      COOR(JCOR,2)=CART(MN(I,JJ),2)
      COOR(JCOR,3)=CART(MN(I,JJ),3)
      JICO=JICO+1
      ICON(JICO,1)=JT
      ICON(JICO,2)=JCOR
12    CONTINUE
CESPOIR      WRITE (6,1003) NUM,MN(I,1),(NA(1,L),L=1,3)
      GOTO 11
   10 WRITE (IO2,1002)NUM,(MN(I,J),(NA(J,L),L=1,3),J=1,7)
      DO 101 J=1,7
      IF(MN(I,J).EQ.-1) GO TO 11
101   CONTINUE
      JCOR=JCOR+1
      WRITE(IO2,1005)NUM,MN(I,1),AID(MN(I,1),1),
     &AID(MN(I,1),2),(CART(MN(I,1),J),J=1,3)
      COOR(JCOR,1)=CART(MN(I,1),1)
      COOR(JCOR,2)=CART(MN(I,1),2)
      COOR(JCOR,3)=CART(MN(I,1),3)
      NPOL(JCOR)=0
      NID(JCOR)=MN(I,1)
      JT=JCOR
      DO 13,II=1,6
      JCOR=JCOR+1
      NPOL(JCOR)=6
      NID(JCOR)=MN(I,1)
      NCO(JCOR)=ICOL(MN(I,1))
      JJ=II+1
      WRITE(IO2,1006)MN(I,JJ),AID(MN(I,JJ),1),
     &AID(MN(I,JJ),2),(CART(MN(I,JJ),J),J=1,3)
      COOR(JCOR,1)=CART(MN(I,JJ),1)
      COOR(JCOR,2)=CART(MN(I,JJ),2)
      COOR(JCOR,3)=CART(MN(I,JJ),3)
      JICO=JICO+1
      ICON(JICO,1)=JT
      ICON(JICO,2)=JCOR
13    CONTINUE
CESPOIR      WRITE (6,1004) NUM,MN(I,1),(NA(1,L),L=1,3)
   11 CONTINUE
C
C AND NOW THE ZARBI POLYHEDRA
C
150   IF(LL)300,300,201
201   DO 299 I=1,LL
      NUM=NUM+1
      DO 202 J=1,12
      IF(MV(I,J).EQ.0)GO TO 203
202   CONTINUE
203   NPOLBZ=J-1
      DO 199 J=1,NPOLBZ
      CALL UNPACK(MV(I,J))
      NA(J,1)=MH(1)
      NA(J,2)=MH(2)
199   NA(J,3)=MH(3)
      CALL UNPACK(MV(I,13))
      NA(13,1)=MH(1)
      NA(13,2)=MH(2)
      NA(13,3)=MH(3)
C NPOLBZ is the number of corners in the zarbi polyhedron
      WRITE (IO2,2002)NUM,MV(I,13),(NA(13,L),L=1,3),
     &(MV(I,J),(NA(J,L),L=1,3),J=1,NPOLBZ)
      DO 204 J=1,13
      IF(MV(I,J).EQ.-1) GO TO 299
204   CONTINUE
      JCOR=JCOR+1
C  The central atom of the zarbi
      WRITE(IO2,1005)NUM,MV(I,13),AID(MV(I,13),1),
     &AID(MV(I,13),2),(CART(MV(I,13),J),J=1,3)
      COOR(JCOR,1)=CART(MV(I,13),1)
      COOR(JCOR,2)=CART(MV(I,13),2)
      COOR(JCOR,3)=CART(MV(I,13),3)
      NPOL(JCOR)=0
      NID(JCOR)=MV(I,13)
      JT=JCOR
C  The atoms at the zarbi corners
      DO 205,II=1,NPOLBZ
      JCOR=JCOR+1
      NPOL(JCOR)=NPOLBZ
C  those with 4 and 6 are distinguished by adding 10
      IF(NPOL(JCOR).EQ.4)NPOL(JCOR)=14
      IF(NPOL(JCOR).EQ.6)NPOL(JCOR)=16
      NID(JCOR)=MV(I,13)
      NCO(JCOR)=ICOL(MV(I,13))
      JJ=II
      WRITE(IO2,1006)MV(I,JJ),AID(MV(I,JJ),1),
     &AID(MV(I,JJ),2),(CART(MV(I,JJ),J),J=1,3)
      COOR(JCOR,1)=CART(MV(I,JJ),1)
      COOR(JCOR,2)=CART(MV(I,JJ),2)
      COOR(JCOR,3)=CART(MV(I,JJ),3)
      JICO=JICO+1
      ICON(JICO,1)=JT
      ICON(JICO,2)=JCOR
205   CONTINUE
CESPOIR      WRITE (6,1007) NUM,MV(I,13),(NA(13,L),L=1,3)
299   CONTINUE 
C
C
300   CALL DEFCOL(COLOR)
C
C  FIND MIN AND MAX COORDINATES
C
      XM(1)=COOR(1,1)
      XM(2)=COOR(1,1)
      XM(3)=COOR(1,2)
      XM(4)=COOR(1,2)
      XM(5)=COOR(1,3)
      XM(6)=COOR(1,3)
      DO 301 I=2,JCOR
      IF(COOR(I,1).LT.XM(1))XM(1)=COOR(I,1)
      IF(COOR(I,1).GT.XM(2))XM(2)=COOR(I,1)
      IF(COOR(I,2).LT.XM(3))XM(3)=COOR(I,2)
      IF(COOR(I,2).GT.XM(4))XM(4)=COOR(I,2)
      IF(COOR(I,3).LT.XM(5))XM(5)=COOR(I,3)
      IF(COOR(I,3).GT.XM(6))XM(6)=COOR(I,3)
301   CONTINUE
      X0=XM(1)+(XM(2)-XM(1))/2.
      Y0=XM(3)+(XM(4)-XM(3))/2.
      Z0=10.*ABS(XM(6)-XM(5))
C
C     VRML DRAWING
C
      IF(JCOR.LE.0)THEN
      WRITE(IO2,*)'  JCOR < or = 0 in subroutine Write'
      GO TO 6000
      ENDIF
      IF(JCOR.GT.6000)THEN
      WRITE(IO2,*)'  JCOR > 6000 in subroutine Write'
      GO TO 6000
      ENDIF
      IF(JICO.LE.0)THEN
      WRITE(IO2,*)'  JICO < or = 0 in subroutine Write'
      GO TO 6000
      ENDIF
      IF(JICO.GT.6000)THEN
      WRITE(IO2,*)'  JICOR > 6000 in subroutine Write'
      GO TO 6000
      ENDIF
      CALL WRLWI(COLOR,MI(48),X0,Y0,Z0)
      CALL WRLPO(COLOR,MI(48),MI(37),LL,X0,Y0,Z0)
      CALL WRLPOV(COLOR,MI(48),MI(37),LL,TEMPO)
6000  RETURN
      RETURN
 1000 FORMAT (//,1X,'THE FOLLOWING ATOMS FORM POLYHEDRA:',/)
 1010 FORMAT (/,16X,'CA',12X,'A1',12X,'A2',12X,'A3',12X,'A4',12X,
     &'A5',12X,'A6',/,
     &'--------------------------------------------------------------
     &-----------------------------------------')
 1001 FORMAT (1X,'TET.',I3,2X,5('(',I4,')',1X,2A1,I2,3X))
 1002 FORMAT (1X,'OCT.',I3,2X,7('(',I4,')',1X,2A1,I2,3X))
 2002 FORMAT (1X,'BZR.',I3,2X,7('(',I4,')',1X,2A1,I2,3X)/
     &10X,6('(',I4,')',1X,2A1,I2,3X))
 1003 FORMAT (1X,'TETR.',1X,I3,3X,'(',I4,')',1X,2A1,I2)
 1004 FORMAT (1X,'OCTA.',1X,I3,3X,'(',I4,')',1X,2A1,I2)
 1005 FORMAT (1X,'POLY.',I3,3X,I4,2X,2A1,3F10.5)
 1006 FORMAT (4X,'ENVIR.',3X,I4,2X,2A1,3F10.5)
 1007 FORMAT (1X,'BZRR.',1X,I3,3X,'(',I4,')',1X,2A1,I2)
      END
C----------------------------------------------------------------------------
      SUBROUTINE WRLWI(COLOR,MCELL,X0,Y0,Z0)
C  WIREFRAME MODEL wi.wrl
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      COMMON/WRL/AID,NAID,CART,IO6,IO7,IO8,NCIRC,ICOL,NCOL,NPT
      DIMENSION NAID(6000),CART(6000,3),COOR(6000,3),NCIRC(6000)
      DIMENSION ICOL(6000),NCOL(20,2),NID(6000),COLOR(20,3)
      INTEGER ICON(6000,2),AID(6000,2),NPOL(6000),NCO(6000)
      WRITE(IO5,1010)
      WRITE(IO5,1009)
      WRITE(IO5,1011)
      WRITE(IO5,998)X0,Y0,Z0
      WRITE(IO5,1012)
      WRITE(IO5,999)
      WRITE(IO5,1013)
      WRITE(IO5,1011)
      WRITE(IO5,1011)
      WRITE(IO5,1015)
      DO 1,I=1,JCOR
      WRITE(IO5,1000)(COOR(I,J),J=1,3)
1     CONTINUE
      WRITE(IO5,1016)
      KCO=NCO(ICON(1,2))
      WRITE(IO5,1014)(COLOR(KCO,KK),KK=1,3)
      WRITE(IO5,1017)
      DO 2,I=1,JICO
      IF(NCO(ICON(I,2)).EQ.KCO)GO TO 100
      KCO=NCO(ICON(I,2))
      WRITE(IO5,1016)
      WRITE(IO5,1014)(COLOR(KCO,KK),KK=1,3)
      WRITE(IO5,1017)
100   WRITE(IO5,1001)(ICON(I,J)-1,J=1,2)
2     CONTINUE
      WRITE(IO5,1016)
      WRITE(IO5,1019)
      WRITE(IO5,1019)
C
C  DRAW CELL AXES
C
      IF(MCELL.GT.0)THEN
      WRITE(IO5,1020)
      WRITE(IO5,1011)
      WRITE(IO5,1015)
      DO 3,I=1,8
      WRITE(IO5,1000)(CART(I,J),J=1,3)
3     CONTINUE
      WRITE(IO5,1016)
      WRITE(IO5,1017)
      WRITE(IO5,1021)
      WRITE(IO5,1016)
      WRITE(IO5,1019)
      ENDIF
      WRITE(IO5,1019)
      RETURN
C WRITE VRML header
998   FORMAT('   PerspectiveCamera{'/
     &'       position '3F8.3/
     &'       heightAngle 0.2 }')
999   FORMAT('  ShapeHints { shapeType UNKNOWN_SHAPE_TYPE }')
1000  FORMAT(5X,3F8.3,',')
1001  FORMAT(1X,I4,',',I4,',-1,')
1009  FORMAT('DEF BackgroundColor Info { string "1. 1. 1." }')
1010  FORMAT('#VRML V1.0 ascii')
1011  FORMAT('Separator{')
1012  FORMAT('Info{string"Created by STRUVIR" }')
1013  FORMAT('Material{'/'ambientColor 0.1 0.1 0.1'/
     &'   diffuseColor 0.75 0.75 0.75'/'specularColor 0.9 0.9 0.9'/
     &'   shininess 0.8 }')
1014  FORMAT('Material{diffuseColor '3F6.3/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
1015  FORMAT('Coordinate3{point[')
1016  FORMAT(1X,']}')
1017  FORMAT('IndexedLineSet{coordIndex[')
1019  FORMAT(1X,'}')
1020  FORMAT('Material{diffuseColor 0.00 0.00 0.00'/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
1021  FORMAT(1X,'0,1,4,2,0,3,5,7,6,3,-1,1,5,-1,2,6,-1,4,7,-1')
      END
C------------------------------------------------------------------                                        
      SUBROUTINE WRLPO(COLOR,MCELL,MPOT,LL,X0,Y0,Z0)
C  POLYHEDRA MODEL po.wrl
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      COMMON/WRL/AID,NAID,CART,IO6,IO7,IO8,NCIRC,ICOL,NCOL,NPT
      DIMENSION NAID(6000),CART(6000,3),COOR(6000,3),NCIRC(6000)
      INTEGER ICON(6000,2),AID(6000,2),NPOL(6000),NCO(6000)
      DIMENSION ICOL(6000),NCOL(20,2),COLOR(20,3),NID(6000)
      DIMENSION NI(3,220),MAR(40,2),NQ(4,20)
      WRITE(IO7,1010)
      WRITE(IO7,1009)
      WRITE(IO7,1011)
      WRITE(IO7,998)X0,Y0,Z0
      WRITE(IO7,1012)
      WRITE(IO7,999)
      WRITE(IO7,1013)
      WRITE(IO7,1011)
      WRITE(IO7,1011)
      I=0
      IPOL=1
1     I=I+1
      IF(I.GT.JCOR)GO TO 100
      IF(NPOL(I).EQ.0)GO TO 1
      IF(NPOL(I).EQ.6)GO TO 6
      IF(NPOL(I).EQ.3)GO TO 30
      IF(NPOL(I).EQ.14)GO TO 40
C      IF(NPOL(I).EQ.5)GO TO 50
      IF(NPOL(I).EQ.5)GO TO 60
      IF(NPOL(I).EQ.16)GO TO 60
      IF(NPOL(I).EQ.7)GO TO 60
      IF(NPOL(I).EQ.8)GO TO 60
      IF(NPOL(I).EQ.9)GO TO 60
      IF(NPOL(I).EQ.10)GO TO 60
      IF(NPOL(I).EQ.11)GO TO 60
      IF(NPOL(I).EQ.12)GO TO 60
C  THEN NPOL(I).EQ.4
      IF(NPOL(I).NE.4)GO TO 1
C
C  AX4 tetrahedron
C
      WRITE(IO7,1005)NID(I),AID(NID(I),1),AID(NID(I),2)
      WRITE(IO7,1014)(COLOR(NCO(I),KK),KK=1,3)
      WRITE(IO7,1015)
      DO 2 JJ=I,I+3
      WRITE(IO7,1000)(COOR(JJ,J),J=1,3)
2     CONTINUE
      I=JJ
      IPOL=IPOL+1
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      WRITE(IO7,1040)
      WRITE(IO7,1016)
      WRITE(IO7,2000)
      WRITE(IO7,1017)
      WRITE(IO7,2001)
      GO TO 1
C
C  AX6 octahedron
C
6     WRITE(IO7,1005)NID(I),AID(NID(I),1),AID(NID(I),2)
      WRITE(IO7,1060)(COLOR(NCO(I),KK),KK=1,3)
      WRITE(IO7,1015)
      DO 3 JJ=I,I+5
      WRITE(IO7,1000)(COOR(JJ,J),J=1,3)
3     CONTINUE
      I=JJ
      IPOL=IPOL+1
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      WRITE(IO7,1050)
      WRITE(IO7,1051)
      WRITE(IO7,1016)
      WRITE(IO7,2000)
      WRITE(IO7,1017)
      WRITE(IO7,2002)
      GO TO 1
C
C  AX3 Polyhedron : draw X3 in their plane
C
30    WRITE(IO7,1005)NID(I),AID(NID(I),1),AID(NID(I),2)
      WRITE(IO7,1060)(COLOR(NCO(I),KK),KK=1,3)
      WRITE(IO7,1015)
      DO 31 JJ=I,I+2
      WRITE(IO7,1000)(COOR(JJ,J),J=1,3)
31     CONTINUE
      I=JJ
      IPOL=IPOL+1
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      WRITE(IO7,3050)
      WRITE(IO7,1016)
      WRITE(IO7,2000)
      WRITE(IO7,1017)
      WRITE(IO7,3051)
      GO TO 1
C
C  AX4 Polyhedron not being tetrahedron: presumably square
C               more or less plane
C
40    WRITE(IO7,1005)NID(I),AID(NID(I),1),AID(NID(I),2)
      WRITE(IO7,1060)(COLOR(NCO(I),KK),KK=1,3)
      WRITE(IO7,1015)
      CALL GEOM(I-1,I+3,NS)
      DO 41 JJ=I,I+3
      WRITE(IO7,1000)(COOR(JJ,J),J=1,3)
41     CONTINUE
      I=JJ
      IPOL=IPOL+1
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      WRITE(IO7,4050)
      WRITE(IO7,1016)
      WRITE(IO7,2000)
      WRITE(IO7,1017)
      WRITE(IO7,4051)
      GO TO 1
C
C  AX5, AX6 (not octahedron) to AX12 Polyhedron 
C
60    IA=I-1
      IF(NPOL(I).EQ.16)NPOL(I)=6
      IB=I+NPOL(I)-1
      IC=IPOL-MPOT
      CALL FACED(IA,IB,NI,NT,IC,MAR,NAR,NQ,NS)
      WRITE(IO7,1005)NID(I),AID(NID(I),1),AID(NID(I),2)
      WRITE(IO7,1060)(COLOR(NCO(I),KK),KK=1,3)
      WRITE(IO7,1015)
      DO 61 JJ=I,I+NPOL(I)-1
      WRITE(IO7,1000)(COOR(JJ,J),J=1,3)
61    CONTINUE
      I=JJ
      IPOL=IPOL+1
      IF(NT.EQ.0)GO TO 71
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      IF(NT.EQ.1)GO TO 70
      DO 62 JJ=1,NT-1
      CALL ACLOC3(NI,JJ,IA+1)
      WRITE(IO7,5060)(NI(IN,JJ)-1,IN=1,3)
62    CONTINUE
70    CALL ACLOC3(NI,NT,IA+1)
      WRITE(IO7,5061)(NI(IN,NT)-1,IN=1,3)
71    IF(NS.EQ.0)GO TO 75
      WRITE(IO7,1016)
      WRITE(IO7,1030)
      IF(NS.EQ.1)GO TO 74
      DO 73 JJ=1,NS-1
      CALL ACLOC4(NQ,JJ,IA+1)
      WRITE(IO7,5070)(NQ(IN,JJ)-1,IN=1,4)
73    CONTINUE
74    CALL ACLOC4(NQ,NS,IA+1)
      WRITE(IO7,5071)(NQ(IN,NS)-1,IN=1,4)
75    WRITE(IO7,1016)
      WRITE(IO7,2000)
      WRITE(IO7,1017)
      DO 63 JJ=1,NAR-1
      WRITE(IO7,5062)MAR(JJ,1),MAR(JJ,2)
63    CONTINUE
      WRITE(IO7,5063)MAR(NAR,1),MAR(NAR,2)
      GO TO 1
100   WRITE(IO7,1019)
      WRITE(IO7,1019)
C
C DRAW SPHERES IF ANY
C
      NOCIRC=0
      DO 3000,I=9,NPT
      IF(NCIRC(I).EQ.0) GO TO 3000
      NOCIRC=NOCIRC+1
3000  CONTINUE
      IF(NOCIRC.EQ.0) GO TO 4000
      WRITE(IO7,1011)
      DO 3001,I=9,NPT
      IF(NCIRC(I).EQ.0) GO TO 3001
      WRITE(IO7,1011)
      WRITE(IO7,3002)(COLOR(ICOL(I),KK),KK=1,3)
      WRITE(IO7,3003)(CART(NCIRC(I),J),J=1,3)
      WRITE(IO7,3004)
3001  CONTINUE
      WRITE(IO7,1019)
C
C  DRAW CELL AXES
C
4000  IF(MCELL.GT.0)THEN
      WRITE(IO7,1020)
      WRITE(IO7,1011)
      WRITE(IO7,1015)
      DO 4,I=1,8
      WRITE(IO7,1000)(CART(I,J),J=1,3)
4     CONTINUE
      WRITE(IO7,1016)
      WRITE(IO7,1017)
      WRITE(IO7,1021)
      WRITE(IO7,1016)
      WRITE(IO7,1019)
      ENDIF
      WRITE(IO7,1019)
      RETURN
C---------------------------------------------------------------------------
998   FORMAT('   PerspectiveCamera{'/
     &'       position '3F8.3/
     &'       heightAngle 0.2 }')
999   FORMAT('  ShapeHints { shapeType UNKNOWN_SHAPE_TYPE }')
1000  FORMAT(5X,3F8.3,',')
1001  FORMAT(1X,I4,',',I4,',-1,')
1005  FORMAT ('# POLYHEDRON ',I4,2X,2A1)
1009  FORMAT('DEF BackgroundColor Info { string "1. 1. 1." }')
1010  FORMAT('#VRML V1.0 ascii')
1011  FORMAT('Separator{')
1012  FORMAT('Info{string"Created by STRUVIR" }')
1013  FORMAT('Material{'/'ambientColor 0.1 0.1 0.1'/
     &'   diffuseColor 0.75 0.75 0.75'/'specularColor 0.9 0.9 0.9'/
     &'   shininess 0.8 }')
1014  FORMAT('Material{diffuseColor '3F6.3/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
1015  FORMAT('Coordinate3{point[')
1016  FORMAT(1X,']}')
1017  FORMAT('IndexedLineSet{coordIndex[')
1019  FORMAT(1X,'}')
1020  FORMAT('Material{diffuseColor 0.00 0.00 0.00'/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
1021  FORMAT(1X,'0,1,4,2,0,3,5,7,6,3,-1,1,5,-1,2,6,-1,4,7,-1')
1030  FORMAT('IndexedFaceSet{coordIndex[')
1040  FORMAT('0,1,2,-1,0,1,3,-1,0,2,3,-1,1,2,3,-1,')
1050  FORMAT('0,2,1,-1,0,1,3,-1,0,3,4,-1,0,4,2,-1,')
1051  FORMAT('5,2,1,-1,5,1,3,-1,5,3,4,-1,5,4,2,-1,')
1060  FORMAT('Material{diffuseColor '3F6.3/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
2000  FORMAT('Material{diffuseColor 0.00 0.00 0.00'/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
2001  FORMAT('0,1,2,0,3,1,-1,2,3,-1 ]}')
2002  FORMAT('0,1,5,4,3,1,2,4,0,3,5,2,0,-1 ]}')
3002  FORMAT('Material{diffuseColor '3F6.3/
     &'    specularColor 0.5 0.5 0.5'/'    shininess 0.2 }')
3003  FORMAT('Translation{translation '3F8.3' }')
3004  FORMAT('Sphere{radius 0.5 } }')
3050  FORMAT('0,2,1,-1,')
3051  FORMAT('0,1,2,0,-1 ]}')
4050  FORMAT('0,2,1,-1,0,3,1,-1')
4051  FORMAT('0,2,1,3,0,-1 ]}')
5050  FORMAT('0,2,3,-1,0,3,4,-1,0,4,2,-1,1,2,3,-1,1,3,4,-1,1,4,2,-1')
5051  FORMAT('2,3,4,2,1,4,0,3,1,-1,0,2,-1 ]}')
5060  FORMAT(3(I2',')'-1,')
5061  FORMAT(3(I2',')'-1')
5062  FORMAT(2(I2',')'-1,')
5063  FORMAT(2(I2',')'-1 ]}')
5070  FORMAT(4(I2',')'-1,')
5071  FORMAT(4(I2',')'-1')
      END

C-----------------------------------------------------------------------
C  The ACLOC3 subroutine performs surface tests for triangles
C
      SUBROUTINE ACLOC3(NI,JJ,I)
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      DIMENSION COOR(6000,3),ICON(6000,2),NPOL(6000),NCO(6000)
      DIMENSION NID(6000),NI(3,220),NTEMP(3),X(3),Y(3),Z(3)
      DO 100 J=1,3
      X(J)=COOR(I+NI(J,JJ)-1,1)
      Y(J)=COOR(I+NI(J,JJ)-1,2)
100   Z(J)=COOR(I+NI(J,JJ)-1,3)
      P1=Z(1)*Y(2)-Y(1)*Z(2)
      P2=X(1)*Z(2)-Z(1)*X(2)
      P3=Y(1)*X(2)-X(1)*Y(2)
      C=X(3)*P1+Y(3)*P2+Z(3)*P3
C      WRITE(4,*)'Triangular face JJ,  C  =',JJ,C
C
C OPTION: C must be negative otherwise reverse order
C 
      IF(C.LT.0.) GO TO 200
      DO 150 J=1,3
150   NTEMP(3-J+1)=NI(J,JJ)
      DO 151 J=1,3
151   NI(J,JJ)=NTEMP(J)
200   CONTINUE   
      RETURN
      END
C
C  The ACLOC4 subroutine performs surface tests for squares
C
      SUBROUTINE ACLOC4(NQ,JJ,I)
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      DIMENSION COOR(6000,3),ICON(6000,2),NPOL(6000),NCO(6000)
      DIMENSION NID(6000),NQ(4,20),NTEMP(4),X(3),Y(3),Z(3)
      DO 100 J=1,3
      X(J)=COOR(I+NQ(J,JJ)-1,1)
      Y(J)=COOR(I+NQ(J,JJ)-1,2)
100   Z(J)=COOR(I+NQ(J,JJ)-1,3)
      P1=Z(1)*Y(2)-Y(1)*Z(2)
      P2=X(1)*Z(2)-Z(1)*X(2)
      P3=Y(1)*X(2)-X(1)*Y(2)
      C=X(3)*P1+Y(3)*P2+Z(3)*P3
C      WRITE(4,*)'Square face JJ,  C  =',JJ,C
C
C OPTION: C must be negative otherwise reverse order
C 
      IF(C.LT.0.) GO TO 200
      DO 150 J=1,4
150   NTEMP(4-J+1)=NQ(J,JJ)
      DO 151 J=1,4
151   NQ(J,JJ)=NTEMP(J)
200   CONTINUE   
      RETURN
      END
C
C The GEOM subroutine places first of the list the coordinates
C          of the X1 and X2 atoms having the largest X1-M-X2
C          angle in the MXn polyhedron  (n.LE.12). Then X3 and
C          X4 with the shortest X3-M-X4 angles will be listed,
C          unless X3 or X4 is identical to X1 or X2 so only the
C          third group of coordinates will correspond to either
C          X3-M-X1 or X3-M-X2 with shortest angle.
C
C Moreover : NS may have several values according to the
C            polyhedron type :
C            MX5 : NS=3 if triangular bipyramid defined as
C                  3 angles in the range 90-135 
C
      SUBROUTINE GEOM(I,IJ,NS)
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      DIMENSION COOR(6000,3),ICON(6000,2),NPOL(6000),NCO(6000)
      DIMENSION YZ(100,5),YE(200,5),NID(6000),CO(12,3),MH(20)
C
      YC36=57.29578
      LL=0
      NS=0
      DO 4 J=I+1,IJ
      LL=LL+1
      YZ(LL,1)=FLOAT(J)
      YZ(LL,2)=COOR(I,1)-COOR(J,1)
      YZ(LL,3)=COOR(I,2)-COOR(J,2)
      YZ(LL,4)=COOR(I,3)-COOR(J,3)
      YZ(LL,5)=SQRT(YZ(LL,2)*YZ(LL,2)+
     *             YZ(LL,3)*YZ(LL,3)+
     *             YZ(LL,4)*YZ(LL,4))
    4 CONTINUE
      TEST1=0.
      TEST2=180.
      II=0
      DO 13 J=1,LL-1
      DO 12 K=J+1,LL
      II=II+1
      YE(II,1)=YZ(J,1)
      YE(II,2)=YZ(K,1)
      MH(8)=IFIX(YE(II,2))
      MH(7)=IFIX(YE(II,1))
      IF (ABS(YZ(J,5)*YZ(K,5)).GE.1E-8)THEN
        ARCO=(YZ(J,2)*YZ(K,2)+YZ(J,3)*YZ(K,3)+YZ(J,4)*YZ(K,4))/
     &  (YZ(J,5)*YZ(K,5))
        IF(ARCO.GT. 1.)ARCO= 1.
        IF(ARCO.LT.-1.)ARCO=-1.
      ELSE
        ARCO=1.
      ENDIF
      YE(II,3)=ACOS(ARCO)*YC36
      IF(YE(II,3).LT.135.AND.YE(II,3).GT.90.)NS=NS+1
      IF(YE(II,3).LT.TEST1)GO TO 120
      TEST1=YE(II,3)
      M1=MH(8)
      M2=MH(7)
120   IF(YE(II,3).GT.TEST2)GO TO 12
      TEST2=YE(II,3)
      M3=MH(8)
      M4=MH(7)
   12 CONTINUE
   13 CONTINUE
      CO(1,1)=COOR(M1,1)
      CO(1,2)=COOR(M1,2)
      CO(1,3)=COOR(M1,3)
      CO(2,1)=COOR(M2,1)
      CO(2,2)=COOR(M2,2)
      CO(2,3)=COOR(M2,3)
      LL=2
      IF(M3.EQ.M1.OR.M3.EQ.M2)GO TO 131
      CO(LL+1,1)=COOR(M3,1)
      CO(LL+1,2)=COOR(M3,2)
      CO(LL+1,3)=COOR(M3,3)
      M5=M3
      LL=LL+1
131   IF(M4.EQ.M1.OR.M4.EQ.M2)GO TO 132
      CO(LL+1,1)=COOR(M4,1)
      CO(LL+1,2)=COOR(M4,2)
      CO(LL+1,3)=COOR(M4,3)
      M5=M4
      LL=LL+1
132   IF(LL.NE.4) GO TO 133
      DO 14 J=I+1,IJ
      IF(M1.EQ.J)GO TO 14
      IF(M2.EQ.J)GO TO 14
      IF(M3.EQ.J)GO TO 14
      IF(M4.EQ.J)GO TO 14
      LL=LL+1
      CO(LL,1)=COOR(J,1)
      CO(LL,2)=COOR(J,2)
      CO(LL,3)=COOR(J,3)
   14 CONTINUE
      GO TO 140
133   DO 139 J=I+1,IJ
      IF(M1.EQ.J)GO TO 139
      IF(M2.EQ.J)GO TO 139
      IF(M5.EQ.J)GO TO 139
      LL=LL+1
      CO(LL,1)=COOR(J,1)
      CO(LL,2)=COOR(J,2)
      CO(LL,3)=COOR(J,3)
139   CONTINUE
140   LL=0
      DO 15 J=I+1,IJ
      LL=LL+1
      COOR(J,1)=CO(LL,1)
      COOR(J,2)=CO(LL,2)
      COOR(J,3)=CO(LL,3)
   15 CONTINUE
      RETURN
      END
C
C The FACED subroutine recovers edges and faces for the
C            bizarre polyhedra
C           
      SUBROUTINE FACED(I,IJ,NI,NT,LL,MAR,NAR,NQ,NS)
C
C  I  is the ID number for the central atom coordinates COOR(I,3)
C  from j = I+1 to IJ are the ID numbers for the polyhedron
C                     vertices coordinates
C  NI(3,1 to NT) will contain the triangle vertices numbers in return
C  NQ(3,1 to NS) will contain the square vertices numbers in return
C  MAR(1 to NAR,2) will contain edges
C
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      DIMENSION COOR(6000,3),ICON(6000,2),NPOL(6000),NCO(6000)
      DIMENSION CO(12,3),NID(6000),NQ(4,20),N(15)
      DIMENSION NI(3,220),MAR(40,2),MVV(12),ICODE(220),C(220)
C
      CALL EDGE(LL,MAR,NAR,MVV)
      LL=0
      DO 4 J=I+1,IJ
      LL=LL+1
      CO(LL,1)=COOR(J,1)
      CO(LL,2)=COOR(J,2)
      CO(LL,3)=COOR(J,3)
    4 CONTINUE
C
C      THEN: find all possible triangle XJ-XK-XL
      NT=0
      DO 150 J=1,LL-2
      DO 140 K=J+1,LL-1
      DO 130 L=K+1,LL
      IF(K.EQ.L)GO TO 130
      NT=NT+1
      NI(1,NT)=J
      NI(2,NT)=K
      NI(3,NT)=L
C      WRITE(4,*)'NT N1 N2,N3',NT,J,K,L
130   CONTINUE
140   CONTINUE
150   CONTINUE
C
C Find those triangles at the polyhedron surface
C            by the "brute force" 
C
      DO 160 KK=1,NT
160   ICODE(KK)=0
      DO 300 KK=1,NT
      J=NI(1,KK)
      K=NI(2,KK)
      L=NI(3,KK)
      X1=CO(K,1)-CO(J,1)
      Y1=CO(K,2)-CO(J,2)
      Z1=CO(K,3)-CO(J,3)
      X2=CO(L,1)-CO(J,1)
      Y2=CO(L,2)-CO(J,2)
      Z2=CO(L,3)-CO(J,3)
      P1=Y1*Z2-Y2*Z1
      P2=X2*Z1-X1*Z2
      P3=X1*Y2-X2*Y1
      DO 290 KL=1,NT
      IF(KL.EQ.KK)C(KL)=0.
      IF(KL.EQ.KK)GO TO 290
C
C X3,Y3,Z3 are coordinates of a vector pointing from
C   the barycenter of the face KL toward a vertex of the face KK
C
      X3=(CO(NI(1,KL),1)+CO(NI(2,KL),1)+CO(NI(3,KL),1))/3.
      Y3=(CO(NI(1,KL),2)+CO(NI(2,KL),2)+CO(NI(3,KL),2))/3.
      Z3=(CO(NI(1,KL),3)+CO(NI(2,KL),3)+CO(NI(3,KL),3))/3.
      X3=CO(J,1)-X3
      Y3=CO(J,2)-Y3
      Z3=CO(J,3)-Z3
      C(KL)=X3*P1+Y3*P2+Z3*P3
C      WRITE(4,*)'KK KL C',KK,KL,C
      IF(ABS(C(KL)).LT.0.0001)C(KL)=0.
      DO 280 KKK=1,KL
      IF(C(KKK).EQ.0.)GO TO 280
      IF(C(KL).EQ.0.)GO TO 280
C If signs are different go to 300
      IF(C(KL)*C(KKK).LT.0.)GO TO 300
280   CONTINUE
290   CONTINUE
      ICODE(KK)=1
300   CONTINUE
      IT=0
      DO 310 KK=1,NT
      IF(ICODE(KK).EQ.0)GO TO 310
      IT=IT+1
      ICODE(IT)=1
      NI(1,IT)=NI(1,KK)
      NI(2,IT)=NI(2,KK)
      NI(3,IT)=NI(3,KK)
C     WRITE(4,*)'IT N1 N2 N3',IT,NI(1,IT),NI(2,IT),NI(3,IT)
310   CONTINUE
      NT=IT
C
C  Are there square faces, if yes, reduce 4 triangles to one face 
C
      NS=0
      DO 400 KK=1,NT-3
      N(1)=NI(1,KK)
      N(2)=NI(2,KK)
      N(3)=NI(3,KK)
      DO 400 IZ=1,2
      DO 390 KL=KK+IZ,NT
      N(4)=NI(1,KL)
      N(5)=NI(2,KL)
      N(6)=NI(3,KL)
      IF(N(1).EQ.N(4).AND.N(2).EQ.N(5))GO TO 391
390   CONTINUE
      GO TO 400
391   DO 392 KM=KL+1,NT
      N(7)=NI(1,KM)
      N(8)=NI(2,KM)
      N(9)=NI(3,KM)
      IF(N(7).EQ.N(1).AND.N(8).EQ.N(3).AND.N(9).EQ.N(6))GO TO 393
392   CONTINUE
      GO TO 400
393   DO 394 KN=KM+1,NT
      N(10)=NI(1,KN)
      N(11)=NI(2,KN)
      N(12)=NI(3,KN)
      IF(N(10).EQ.N(2).AND.N(11).EQ.N(3).AND.N(12).EQ.N(9))GO TO 395
394   CONTINUE
      GO TO 400
C
C  A square face has been found
C         but where is the diagonal ? It cannot belong
C         to another triangle
C
395   NN1=0
      NN2=0
      NN3=0
      NN4=0
      NN5=0
      NN6=0
      DO 397 KP=1,NT
      IF(KP.EQ.KK)GO TO 397
      IF(KP.EQ.KL)GO TO 397
      IF(KP.EQ.KM)GO TO 397
      IF(KP.EQ.KN)GO TO 397
      N(13)=NI(1,KP)
      N(14)=NI(2,KP)
      N(15)=NI(3,KP)
      M1=N(1)
      M2=N(2)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN1=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN1=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN1=1
      M1=N(1)
      M2=N(3)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN2=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN2=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN2=1
      M1=N(4)
      M2=N(6)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN3=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN3=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN3=1
      M1=N(2)
      M2=N(3)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN4=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN4=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN4=1
      M1=N(5)
      M2=N(6)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN5=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN5=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN5=1
      M1=N(11)
      M2=N(12)
      IF(M1.EQ.N(13).AND.M2.EQ.N(14))NN6=1
      IF(M1.EQ.N(14).AND.M2.EQ.N(15))NN6=1
      IF(M1.EQ.N(13).AND.M2.EQ.N(15))NN6=1
397   CONTINUE
398   ICODE(KK)=0
      ICODE(KL)=0
      ICODE(KM)=0
      ICODE(KN)=0
      IF(NN1.EQ.0.AND.NN6.EQ.0)GO TO 401
      IF(NN2.EQ.0.AND.NN5.EQ.0)GO TO 402
      IF(NN3.EQ.0.AND.NN4.EQ.0)GO TO 403
      WRITE(4,*)'square face detected but not counted'
      GO TO 400
401   NS=NS+1
      NQ(1,NS)=N(1)
      NQ(2,NS)=N(3)
      NQ(3,NS)=N(2)
      NQ(4,NS)=N(12)
C      WRITE(4,*)'NS N1 N2 N3 N4',NS,(NQ(KQ,NS),KQ=1,4)
      GO TO 400
402   NS=NS+1
      NQ(1,NS)=N(1)
      NQ(2,NS)=N(2)
      NQ(3,NS)=N(3)
      NQ(4,NS)=N(12)
C      WRITE(4,*)'NS N1 N2 N3 N4',NS,(NQ(KQ,NS),KQ=1,4)
      GO TO 400
403   NS=NS+1
      NQ(1,NS)=N(1)
      NQ(2,NS)=N(2)
      NQ(3,NS)=N(12)
      NQ(4,NS)=N(3)
C      WRITE(4,*)'NS N1 N2 N3 N4',NS,(NQ(KQ,NS),KQ=1,4)
400   CONTINUE
      IF(NS.EQ.0)GO TO 501
      IT=0
      DO 500 KK=1,NT
      IF(ICODE(KK).EQ.0)GO TO 500
      IT=IT+1
      NI(1,IT)=NI(1,KK)
      NI(2,IT)=NI(2,KK)
      NI(3,IT)=NI(3,KK)
C      WRITE(4,*)'IT N1 N2 N3',IT,NI(1,IT),NI(2,IT),NI(3,IT)
500   CONTINUE
      NT=IT
501   CONTINUE
C
C  Are some edge lacking ... ?
C
      LL=IJ-I
      DO 1010 JJ=1,NAR
      DO 1010 J=1,LL
      IF(MAR(JJ,1).EQ.MVV(J))MAR(JJ,1)=J
      IF(MAR(JJ,2).EQ.MVV(J))MAR(JJ,2)=J
1010  CONTINUE
      DO 600 KL=1,NT
      N1=NI(1,KL)
      N2=NI(2,KL)
      N3=NI(3,KL)
      DO 590 JJ=1,NAR
      N4=MAR(JJ,1)
      N5=MAR(JJ,2)
      IF(N4.EQ.N1.AND.N5.EQ.N2)GO TO 591
      IF(N4.EQ.N2.AND.N5.EQ.N1)GO TO 591
590   CONTINUE
C      IF(NAR.GT.40)PRINT *,'MORE THAN 40 EDGES =STOP'
      IF(NAR.GT.40)WRITE(20,*)'MORE THAN 40 EDGES =STOP'
      IF(NAR.GT.40)RETURN
      NAR=NAR+1
      MAR(NAR,1)=N1
      MAR(NAR,2)=N2
591   CONTINUE
      DO 592 JJ=1,NAR
      N4=MAR(JJ,1)
      N5=MAR(JJ,2)
      IF(N4.EQ.N2.AND.N5.EQ.N3)GO TO 593
      IF(N4.EQ.N3.AND.N5.EQ.N2)GO TO 593
592   CONTINUE
      NAR=NAR+1
C      IF(NAR.GT.40)PRINT *,'MORE THAN 40 EDGES =STOP'
      IF(NAR.GT.40)WRITE(20,*)'MORE THAN 40 EDGES =STOP'
      IF(NAR.GT.40)RETURN
      MAR(NAR,1)=N2
      MAR(NAR,2)=N3
593   CONTINUE
600   CONTINUE
      DO 1011 L=1,NAR
      MAR(L,1)=MAR(L,1)-1
      MAR(L,2)=MAR(L,2)-1
1011  CONTINUE
      RETURN
      END
C-----------------------------------------------------------------------
      SUBROUTINE EDGE(LL,MAR,NAR,MVV)
      COMMON/ALL1/YA(4,6000),YP(10,7),YB(25),YC(90),YD(40),MA(20,2),
     *MB(6),MC(80),MD(80),ME(40),YR(194,3),MF(600),MG(68),MI(48),
     *MK(12,3),ML(10,2),MM(70),MN(1000,8),MO(3024),MU(5,21,7),
     *MV(1000,14),IO1,IO2,IO3,IO4
      DIMENSION MAR(40,2),MVV(12)
C
      MG(10)=LL
      I=0
      M=MV(MG(10),14)
      DO 106 K=2,MU(M,1,3)+1
      DO 135 L=1,MU(M,K,7)
      JJ=MOD(L,MU(M,K,7))+1
      N1=MU(M,K,L)
      N2=MU(M,K,JJ)
      DO 137 NN=1,I
      IF(MA(NN,1).EQ.MV(MG(10),N1).AND.MA(NN,2).EQ.MV(MG(10),N2))
     * GOTO 135
      IF(MA(NN,2).EQ.MV(MG(10),N1).AND.MA(NN,1).EQ.MV(MG(10),N2))
     * GOTO 135
  137 CONTINUE
      I=I+1
      MA(I,1)=MV(MG(10),N1)
      MA(I,2)=MV(MG(10),N2)
      MAR(I,1)=MV(MG(10),N1)
      MAR(I,2)=MV(MG(10),N2)
  135 CONTINUE
  106 CONTINUE
  132 IF(I.EQ.20) GOTO 139
      DO 140 J=I+1,20
      MA(J,1)=0
  140 MA(J,2)=0
  139 CONTINUE
      NAR=I

      DO 200 J=1,12
200   MVV(J)=MV(MG(10),J)
      RETURN
      END
C-------------------------------------------------------------------------
      SUBROUTINE DEFCOL(C)
      DIMENSION C(20,3)
C  Blue sky
      C(1,1)=0.0
      C(1,2)=1.0
      C(1,3)=1.0
C  Green light
      C(2,1)=0.5
      C(2,2)=1.0
      C(2,3)=0.0
C  Red pyrope
      C(3,1)=1.0
      C(3,2)=0.1
      C(3,3)=0.5
C  Yellow
      C(4,1)=1.0
      C(4,2)=1.0
      C(4,3)=0.0
C  Blue grey
      C(5,1)=0.6
      C(5,2)=0.6
      C(5,3)=1.0
C  Orange
      C(6,1)=1.0
      C(6,2)=0.6
      C(6,3)=0.0
C  Purple
      C(7,1)=0.5
      C(7,2)=0.0
      C(7,3)=0.5
C  Brown
      C(8,1)=0.5
      C(8,2)=0.3
      C(8,3)=0.0
C  Blue-green-dark
      C(9,1)=0.0
      C(9,2)=0.5
      C(9,3)=0.5
C  Red
      C(10,1)=1.0
      C(10,2)=0.0
      C(10,3)=0.0
C  Green
      C(11,1)=0.0
      C(11,2)=1.0
      C(11,3)=0.0
C  Blue
      C(12,1)=0.0
      C(12,2)=0.0
      C(12,3)=1.0
C  Pink
      C(13,1)=1.0
      C(13,2)=0.0
      C(13,3)=1.0
C  Red dark
      C(14,1)=0.5
      C(14,2)=0.0
      C(14,3)=0.0
C  Green dark
      C(15,1)=0.0
      C(15,2)=0.5
      C(15,3)=0.0
C  Blue night
      C(16,1)=0.0
      C(16,2)=0.0
      C(16,3)=0.5
C  Red wine
      C(17,1)=0.5
      C(17,2)=0.05
      C(17,3)=0.25
C  White
      C(18,1)=1.0
      C(18,2)=1.0
      C(18,3)=1.0
C  Grey
      C(19,1)=0.5
      C(19,2)=0.5
      C(19,3)=0.5
C  Black
      C(20,1)=0.0
      C(20,2)=0.0
      C(20,3)=0.0
      RETURN
      END
C
C
C---------------------------------------------------------------------------
C Writes polyhedron model in POVRAY-format to file
C Mark Koennecke July 1996
C no warranties taken 
C Free for non commercial usage 
C----------------------------------------------------------------------------                                        
      SUBROUTINE WRLPOV(COLOR,MCELL,MPOT,LL,TEMPO)
      COMMON/WRLOUT/COOR,ICON,IO5,JCOR,JICO,NPOL,NCO,NID
      COMMON/WRL/AID,NAID,CART,IO6,IO7,IO8,NCIRC,ICOL,NCOL,NPT
      DIMENSION NAID(6000),CART(6000,3),COOR(6000,3),NCIRC(6000)
      INTEGER ICON(6000,2),AID(6000,2),NPOL(6000),NCO(6000)
      DIMENSION ICOL(6000),NCOL(20,2),COLOR(20,3),NID(6000)
      DIMENSION NI(3,220),MAR(40,2),NQ(4,20)
      INTEGER IOPOV, I	
      CHARACTER*20 COLNAM
      CHARACTER*24 TEMP	
      CHARACTER*(*) TEMPO	
      LOGICAL T	
      REAL MIN(3), MAX(3),MIT(3), DIS
      INTEGER U,UU	

C create a file
      IOPOV = 27	
      lfile=len(TEMPO)
      do while (TEMPO(lfile:lfile).eq.' ')
         lfile=lfile-1
      enddo
      TEMP=TEMPO(1:LFILE)//'.pov'
      INQUIRE(FILE=TEMP,EXIST=T)
      IF(T)THEN	
	      CALL FILEDEL(IOPOV,TEMP)
      END IF
      OPEN(IOPOV,FILE=TEMP,STATUS='NEW')      	
C create a header
      WRITE(IOPOV,2001)
 2001 FORMAT('/*------------------------------------------------')
      WRITE(IOPOV,*)' POVRAY input file created by STRUPLO'
      WRITE(IOPOV,*)' STRUPLO is a crystal structure plotting package'
      WRITE(IOPOV,*)' Courtesy by Reinhard Fischer'
      WRITE(IOPOV,*)' Modified to produce POVRAY output by:'
      WRITE(IOPOV,*)' Dr. Mark Koennecke PSI'
      WRITE(IOPOV,*)' July 1996 '
      WRITE(IOPOV,*)'   '
      WRITE(IOPOV,*)'   '
      WRITE(IOPOV,*)'   '
      WRITE(IOPOV,*)' The following declares provide default properties'
      WRITE(IOPOV,*)' for the polygons. The default is stupid color    '
      WRITE(IOPOV,*)' By editing these declares, the properties of '
      WRITE(IOPOV,*)' polygons can easily be adapted for groups of '
      WRITE(IOPOV,*)' Atoms respective polygons '
      WRITE(IOPOV,2002)
 2002 FORMAT('------------------------------------------------------*/')

C write declares with colours
        DO 2010, I = 1, 20			  	 			
        WRITE(COLNAM,2003)I
 2003   FORMAT('PolyTyp',I2.2)
        WRITE(IOPOV,2004)COLNAM
 2004   FORMAT(' #declare ',A20,' = ')
        WRITE(IOPOV,2005)COLOR(I,1),COLOR(I,2),COLOR(I,3)
 2005   FORMAT('  texture { pigment {color rgb < ',3(F6.3,1X),' > }')
        WRITE(IOPOV,*)'      finish {ambient 0.5 } }'
 2010 CONTINUE

C find center of data for positioning the camera sensibly
      DO 6000, U = 1, 3
	MIN(U) = 999999.
        MAX(U) = -999999.
 6000 CONTINUE
      DO 6010, U = 1, JCOR
	DO 6020, UU = 1,3
          IF(COOR(U,UU) .GT. MAX(UU))MAX(UU) = COOR(U,UU)
          IF(COOR(U,UU) .LT. MIN(UU))MIN(UU) = COOR(U,UU)
 6020   CONTINUE
 6010 CONTINUE
      DO 6030, U = 1, 3
        MIT(U) = MAX(U) - (ABS(MAX(U) - MIN(U))/2)
 6030 CONTINUE 

C get distance: find extent and assume 30degree camera angle
      DIS = ABS(MAX(1)-MIN(1))
      IF(ABS(MAX(2)-MIN(2)) .GT. DIS)DIS = ABS(MAX(2)-MIN(2))
      DIS = DIS /.5777 	

C write the default camera
      WRITE(IOPOV,*)' camera {'
      WRITE(IOPOV,6031)MIT(1), MIT(2), DIS
 6031 FORMAT('        location  < ',2(F8.4,',',1X),F8.4,' >')
      WRITE(IOPOV,6032)(MIT(U), U = 1,3)
 6032 FORMAT('        look_at  < ',2(F8.4,',',1X),F8.4,' >')
      WRITE(IOPOV,*)' }'			

C the default light source
      WRITE(IOPOV,*)' light_source { < 80., 0., 0.>'
      WRITE(IOPOV,*)'           color rgb <1.,1.,1.>}'

      I=0
      IPOL=1
C start loop
1     I=I+1
      IF(I.GT.JCOR)GO TO 100
      IF(NPOL(I).EQ.0)GO TO 1
      IF(NPOL(I).EQ.6)GO TO 6
      IF(NPOL(I).EQ.3)GO TO 30
      IF(NPOL(I).EQ.14)GO TO 40
C      IF(NPOL(I).EQ.5)GO TO 50
      IF(NPOL(I).EQ.5)GO TO 60
      IF(NPOL(I).EQ.16)GO TO 60
      IF(NPOL(I).EQ.7)GO TO 60
      IF(NPOL(I).EQ.8)GO TO 60
      IF(NPOL(I).EQ.9)GO TO 60
      IF(NPOL(I).EQ.10)GO TO 60
      IF(NPOL(I).EQ.11)GO TO 60
      IF(NPOL(I).EQ.12)GO TO 60
C  THEN NPOL(I).EQ.4
      IF(NPOL(I).NE.4)GO TO 1
C
C  AX4 tetrahedron
C
      WRITE(IOPOV,2011)NID(I), AID(NID(I),1),AID(NID(I),2)
 2011 FORMAT(' // Tetrahedron ', I4, 2X, 2A1)
C first triangle
      WRITE(COLNAM,2003)NCO(I)
      WRITE(IOPOV,3009)
 3009 FORMAT(' union { ')
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+2,J), J=1,3)
 2021 FORMAT('     < ',2(1X,F8.4,','),F8.4,' > , ')
 3001 FORMAT('     < ',2(1X,F8.4,','),F8.4,' > }')
 2022 FORMAT('    texture { ',A20, ' }   } ')
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I+1,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,2022)COLNAM
      I = I + 3
      IPOL = IPOL + 1 
      GO TO 1
C
C  AX6 octahedron
C
6     WRITE(IOPOV,2006)NID(I), AID(NID(I),1),AID(NID(I),2)
 2006 FORMAT(' // Octahedron ', I4, 2X, 2A1)
      WRITE(COLNAM,2003)NCO(I)
      WRITE(IOPOV,3009)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+1,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+3,J), J=1,3)
 
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+4,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+4,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+2,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I+5,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+1,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I+5,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+3,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I+5,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+4,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I+5,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+4,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,2022)COLNAM
      I = I + 5
      IPOL = IPOL + 1	 
      GO TO 1
C
C  AX3 Polyhedron : draw X3 in their plane
C
30    WRITE(IOPOV,2006)NID(I), AID(NID(I),1),AID(NID(I),2)
 2030 FORMAT(' // AX3 Polyhedron ', I4, 2X, 2A1)
      WRITE(COLNAM,2003)NCO(I)
      WRITE(IOPOV,3009)	
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,2022)COLNAM 
      I = I + 2
      IPOL = IPOL + 1
      GO TO 1
C
C  AX4 Polyhedron not being tetrahedron: presumably square
C               more or less plane
C
40    WRITE(IOPOV,2006)NID(I), AID(NID(I),1),AID(NID(I),2)
 2031 FORMAT(' // AX4 Square-coordination  ', I4, 2X, 2A1)
      WRITE(COLNAM,2003)NCO(I)
      CALL GEOM(I-1,I+3,NS)
      WRITE(IOPOV,3009)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+2,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+1,J), J=1,3)

      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(I,J),J=1,3)
      WRITE(IOPOV,2021)(COOR(I+3,J), J=1,3)
      WRITE(IOPOV,3001)(COOR(I+1,J), J=1,3)
      WRITE(IOPOV,2022)COLNAM
      I = I + 3
      IPOL = IPOL + 1 
      GO TO 1
C
C  AX5, AX6 (not octahedron) to AX12 Polyhedron 
C
60    IA=I-1
      IF(NPOL(I).EQ.16)NPOL(I)=6
      IB=I+NPOL(I)-1
      IC=IPOL-MPOT
      CALL FACED(IA,IB,NI,NT,IC,MAR,NAR,NQ,NS)
      WRITE(IOPOV,2012)NID(I), AID(NID(I),1),AID(NID(I),2)
 2012 FORMAT(' // General polyhedron  ', I4, 2X, 2A1)
      WRITE(COLNAM,2003)NCO(I)
      WRITE(IOPOV,3009)
      I= I + NPOL(I) - 1
      IPOL=IPOL+1
      IF(NT.EQ.0)GO TO 71
      IF(NT.EQ.1)GO TO 70
      DO 62 JJ=1,NT-1
      CALL ACLOC3(NI,JJ,IA+1)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NI(1,JJ),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NI(2,JJ),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NI(3,JJ),J), J=1,3)
62    CONTINUE
70    CALL ACLOC3(NI,NT,IA+1)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NI(1,NT),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NI(2,NT),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NI(3,NT),J), J=1,3)
71    IF(NS.EQ.0)GO TO 75
      IF(NS.EQ.1)GO TO 74
      DO 73 JJ=1,NS-1
      CALL ACLOC4(NQ,JJ,IA+1)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NQ(1,JJ),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NQ(2,JJ),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NQ(3,JJ),J), J=1,3)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NQ(1,JJ),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NQ(3,JJ),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NQ(4,JJ),J), J=1,3)
73    CONTINUE
74    CALL ACLOC4(NQ,NS,IA+1)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NQ(1,NS),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NQ(2,NS),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NQ(3,NS),J), J=1,3)
      WRITE(IOPOV,*)' triangle {'
      WRITE(IOPOV,2021)(COOR(NQ(1,NS),J),J=1,3)
      WRITE(IOPOV,2021)(COOR(NQ(3,NS),J), J=1,3)
      WRITE(IOPOV,3001)(COOR(NQ(4,NS),J), J=1,3)
75    CONTINUE
      WRITE(IOPOV,2022)COLNAM 
      GO TO 1
100   WRITE(IOPOV,*)' '
      WRITE(IOPOV,*)' '
C
C DRAW SPHERES IF ANY
C
      WRITE(IOPOV,*)' // Drawing spheres ........'
      NOCIRC=0
      DO 3000,I=9,NPT
      IF(NCIRC(I).EQ.0) GO TO 3000
      NOCIRC=NOCIRC+1
3000  CONTINUE
      IF(NOCIRC.EQ.0) GO TO 4000
      DO 3002,I=9,NPT
      IF(NCIRC(I).EQ.0) GO TO 3002
      WRITE(IOPOV,2040)
 2040 FORMAT(' sphere { ')
      WRITE(COLNAM,2003)ICOL(I)
      WRITE(IOPOV,2041)(CART(NCIRC(I),J),J=1,3)
 2041 FORMAT('         < ',2(1X,F8.4,','),1X,F8.4, ' >, 0.7 ')
      WRITE(IOPOV,2042)COLNAM
 2042 FORMAT('       texture { ',A20,'} } ')	
 3002 CONTINUE
 4000 CONTINUE
      WRITE(IOPOV,*)' // end of file '
      CLOSE(IOPOV)	
      RETURN
      END
      SUBROUTINE PSPLOT(TEMPO)
C
C     Program to read the generic plot file STRUPLO.XYZ output by 
C     Reinhard X. Fischer's program STRUPL91 and to generate the 
C     PostScript file STRUPLO.PS containing the plot.
C
C     Written by: Donald L. Ward
C                 Department of Chemistry
C                 Michigan State University
C                 East Lansing, MI 48824-1322
C
C     Version:    O4 November 1993
C
C     modified by A. Le Bail     March 1996
C                 for PC machine : BACKSPACE was not running
C
C
C     ------------------------------------------------------------------
C
C     The STRUPLO.XYZ input data file has the following structure and 
C     definitions:    ----- ---- ---- 
C
C     record 1 - (F10.4)
C      col.  1-10  scale factor
C
C     record 2 - (40X,A40) 
C      col. 41-80  title
C
C     record 3 - (F10.4,F11.4)
C      col.  1-10  X origin shift (in cm.)
C           11-21  Y origin shift (in cm.)
C      NOTE: the plot view frame has X,Y dimensions (in cm.) twice the 
C            origin shift values.
C
C     The first three records contain information common to the entire
C     drawing.   Thereafter, blocks of records deal with individual
C     features of the drawing.   Each block consists of:
C
C     record 1 - (I2,I3)
C      col.  1- 2  code number (= 4)
C      col.  3- 5  pen number (= 1, 2, 3, 4, 5, 6, 7, or 8)
C
C     records 2-n - (I2,2F11.4)
C      col.  1- 2  code number (= 2 for move with pen down)
C                              (= 3 for move with pen up)
C      col.  3-13  X coordinate to move to
C      col. 14-24  Y coordinate to move to
C
C     The file just ends without any special terminator record.
C
C     ------------------------------------------------------------------
C
C     The STRUPLO.PS output plot file has the following structure and 
C     definitions:   ------ ---- ----
C
C     record 1 - (Adobe PS level)
C      %!PS-Adobe-1.0
C
C     record 2 - (creator of this file)
C      %%Creator: PSPLOT version 04-Nov-93 (by Donald L. Ward)
C
C     record 3 - (title of this file)
C      %%Title: Plot file from STRUPL91 (by Reinhard X. Fischer)
C
C     record 4 - (date/time this program was run)
C      %%CreationDate: dd-mmm-yy hh:mm:ss
C
C     record 5 -
C      %%Pages: (atend)
C
C     record 6 -
C      %%DocumentFonts: (atend)
C
C     record 7 - (define, in points, 8.5 by 11 view frame)
C      %%BoundingBox: 0 0 612 792
C
C     record 8 -
C      %%EndComments
C
C     record 9 -
C      %%EndProlog
C
C     record 10 - (start first page)
C      %%Page: 1 1
C
C     record 11 -
C      %%PageResources: (atend)
C
C     record 12 - (select font)
C      /Times-Roman findfont
C      15 scalefont
C      setfont
C
C     record 13 - (move coordinates origin to center of page)
C      nnnn.nn nnnn.nn translate
C
C     record 13a - (rotate image 90-degrees for landscape mode)
C      90 rotate                      (ONLY if transformed to landscape)
C
C     record 13b - (move coordinates origin to center of page)
C      nnnn.nn nnnn.nn translate      (ONLY if transformed to landscape)
C
C     record 14 - (set position for title)
C      nnnn.nn nnnn.nn moveto
C
C     record 15 - (print title)
C      (aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa) show
C
C     records 16 to n - (block of records for one plot feature)
C      %Start new feature
C      nnnn.nn setlinewidth             (ONLY if Black & White selected)
C      nnnn.nn nnnn.nn nnnn.nn setrgbcolor      (ONLY if Color selected)
C      newpath
C       nnnn.nn  nnnn.nn moveto
C       nnnn.nn  nnnn.nn lineto
C          "        "       "
C      stroke
C
C    (repeat "records 16 to n" for each feature in plot)
C
C     record last-8
C      statusdict begin /manualfeed false store end
C
C     record last-7
C      showpage 
C
C     record last-6
C      %%PageTrailer
C
C     record last-5
C      %%PageResources: font Times-Roman
C
C     record last-4
C      %%Trailer
C
C     record last-3
C      %%DocumentFonts: Times-Roman
C
C     record last-2
C      %%Pages: 1
C
C     record last-1
C      %%DocumentNeededResources: font Times-Roman
C
C     record last
C      %%EOF
C
C     ------------------------------------------------------------------
C
C      CHARACTER  CHAR*1, IDATE*9, ITIME*8, TITLE*40
      CHARACTER  CHAR*1, TITLE*40
      CHARACTER*24 TEMP
      CHARACTER*(*) TEMPO
      LOGICAL QEX
      INTEGER  ICODE, ICOLOR, ICRT, IKBD, INFILE, IPASS, IPEN, OUTFILE
      REAL     COLOR(3,8), NIBS(8)
      REAL     SCALE, XCORD, XORGSHFT, YCORD, YORGSHFT
      DATA NIBS  / 0.10, 0.15, 0.23, 0.34, 0.51, 0.76, 1.14, 1.71/
      DATA COLOR / 0.00, 0.00, 0.00,   ! BLACK
     *             0.25, 0.25, 0.25,   ! GREY
     *             1.00, 0.00, 0.00,   ! RED
     *             0.00, 1.00, 0.00,   ! GREEN
     *             0.00, 0.00, 1.00,   ! BLUE
     *             1.00, 0.00, 1.00,   ! MAGENTA
     *             1.00, 1.00, 0.00,   ! YELLOW
     *             1.00, 0.25, 0.00/   ! ORANGE
C
C     Open input and output plot files
C
      INFILE = 31
      OUTFILE = 32
      IKBD = 5
      ICRT = 6
      lfile=len(TEMPO)
      do while (TEMPO(lfile:lfile).eq.' ')
         lfile=lfile-1
      enddo
      TEMP=TEMPO(1:LFILE)//'ps.scr'
      OPEN (UNIT=INFILE,  FILE=TEMP, STATUS='OLD')
      TEMP=TEMPO(1:LFILE)//'stru.ps'
      INQUIRE(FILE=TEMP,EXIST=QEX)
      IF(QEX.EQ..FALSE.)GO TO 6000
      OPEN(UNIT=OUTFILE,FILE=TEMP,STATUS='OLD')
      CLOSE(UNIT=OUTFILE,STATUS='DELETE')
6000  OPEN(UNIT=OUTFILE,FILE=TEMP,STATUS='NEW')
C     *      CARRIAGECONTROL='LIST')
C
C     Get time/date information, write welcome message
C
C      CALL DATE (IDATE)
C      CALL TIME (ITIME)
C      WRITE (ICRT, 1000) IDATE, ITIME
C
C     Query whether plot is to be Black & White (default) or Color
C
      ICOLOR = 0
CESPOIR      WRITE (ICRT, 1001)
CESPOIR      READ (IKBD, 1002) CHAR
CESPOIR      IF (CHAR .EQ. 'Y' .OR. CHAR .EQ. 'y') ICOLOR = 1
C
C     Write PostScript comment records 
C
      WRITE (OUTFILE, 1003)
      WRITE (OUTFILE, 1004)
      WRITE (OUTFILE, 1005)
C      WRITE (OUTFILE, 1006) IDATE, ITIME
      WRITE (OUTFILE, 1007)
      WRITE (OUTFILE, 1008)
      WRITE (OUTFILE, 1009)
      WRITE (OUTFILE, 1010)
      WRITE (OUTFILE, 1011)
      WRITE (OUTFILE, 1012)
      WRITE (OUTFILE, 1013)
C
C     Process three input header records
C
      READ (INFILE, 1014) SCALE
      READ (INFILE, 1015) TITLE
      READ (INFILE, 1014) XORGSHFT, YORGSHFT
C
C     Translate origin to center the drawing on the sheet of paper, 
C     scale, and change to landscape mode (if necessary); plot title 
C
      CALL SCALEROT (SCALE, ICRT, INFILE, OUTFILE, TITLE)
C
C     Process block of records for one feature of plot
C     ------------------------------------------------
C
      READ (INFILE,*) ICODE, IPEN
10    IF (ICOLOR .EQ. 0) WRITE (OUTFILE, 1017) NIBS(IPEN)
      IF (ICOLOR .EQ. 1) WRITE (OUTFILE, 1018) (COLOR(I,IPEN),I=1,3)
      IPASS = 0
20    READ (INFILE,*, END=60) ICODE, XCORD, YCORD
      GO TO (30, 40, 50), ICODE - 1
      GO TO 999
C
C     Code number = 2, move with "pen down"
C
30    XCORD = SCALE * XCORD
      YCORD = SCALE * YCORD
      WRITE (OUTFILE, 1020) XCORD, YCORD 
      IPASS = 1
      GO TO 20
C
C     Code number = 3, move with "pen up"
C
40    IF (IPASS .EQ. 1) WRITE (OUTFILE, 1021)
      IPASS = 0
      XCORD = SCALE * XCORD
      YCORD = SCALE * YCORD
      WRITE (OUTFILE, 1022) XCORD, YCORD
      GO TO 20
C
C     Code number = 4, backspace input file, start new block
C
50    IPEN=(XCORD+0.5)
      IF (IPASS .EQ. 1) WRITE (OUTFILE, 1021)
      GO TO 10
C
C     End-of-file encountered on input plot file, finish output file
C
60    WRITE (OUTFILE, 1023)
      WRITE (OUTFILE, 1024)
      WRITE (OUTFILE, 1025)
      WRITE (OUTFILE, 1026)
      WRITE (OUTFILE, 1027)
      WRITE (OUTFILE, 1028)
      WRITE (OUTFILE, 1029)
      WRITE (OUTFILE, 1030)
      WRITE (OUTFILE, 1031)
      CLOSE(UNIT=INFILE,STATUS='DELETE')
      CLOSE (OUTFILE)
C      WRITE (ICRT, 1032)
      RETURN
C
      WRITE(4,*)'Invalid code number encountered on input plot file'
C
999   RETURN
C
C     ------------------------------------------------------------------
C
1000  FORMAT (/,' PSPLOT - (version 04-Nov-93)',/,
C     *        '  current date: ',A9,/,
C     *        '      and time: ',1X,A8,//,
     *        ' Program to read the STRUPLO.XYZ generic plot file',
     *        ' produced by STRUPL91',/,
     *        '  and to generate the PostScript plot file STRUPLO.PS',/)
1001  FORMAT (' The default is a Black & White drawing.',/,
     *        '$Do you, instead, want a COLOR drawing?  (Y/N):')
1002  FORMAT (A1)
1003  FORMAT ('%!PS-Adobe-1.0')
1004  FORMAT ('%%Creator: PSPLOT version 04-Nov-93 (by Donald L.',
     *        ' Ward)')
1005  FORMAT ('%%Title: Plot file from STRUPL91 (by Reinhard X.',
     *        ' Fischer)')
C1006  FORMAT ('%%CreationDate:',A9,1X,A8)
1007  FORMAT ('%%Pages: (atend)')
1008  FORMAT ('%%DocumentFonts: (atend)')
1009  FORMAT ('%%BoundingBox: 0 0 612 792')
1010  FORMAT ('%%EndComments')
1011  FORMAT ('%%EndProlog')
1012  FORMAT ('%%Page: 1 1')
1013  FORMAT ('%%PageResources: (atend)')
1014  FORMAT (F10.4,F11.4)
1015  FORMAT (40X,A40)
1016  FORMAT (I2,I3)
1017  FORMAT ('%Start new feature',/,
     *        F8.2,' setlinewidth')
1018  FORMAT ('%Start new feature',/,
     *        3F8.2,' setrgbcolor')
1019  FORMAT (I2,2F11.4)
1020  FORMAT (2F8.2,' lineto')
1021  FORMAT ('stroke')
1022  FORMAT ('newpath',/,2F8.2,' moveto')
1023  FORMAT ('statusdict begin /manualfeed false store end')
1024  FORMAT ('showpage')
1025  FORMAT ('%%PageTrailer')
1026  FORMAT ('%%PageResources: font Times-Roman')
1027  FORMAT ('%%Trailer')
1028  FORMAT ('%%DocumentFonts: Times-Roman')
1029  FORMAT ('%%Pages: 1')
1030  FORMAT ('%%DocumentNeededResources: font Times-Roman')
1031  FORMAT ('%%EOF')
1032  FORMAT (' PSPLOT Completed',/)
      END
      SUBROUTINE SCALEROT (SCALE, ICRT, INFILE, OUTFILE, TITLE)
C
C     Routine to scale plot coordinates as "points" (and rotate image by 
C     90 degrees, if necessary) to best place plot in 8.0 x 10.5 area on
C     an 8.5 x 11 sheet of paper.   Translate the origin by x=306, y=396
C     "points" to put plot origin in the middle of the sheet.
C
C     ------------------------------------------------------------------
C
      CHARACTER  TITLE*40
      INTEGER  ICODE, ICRT, INFILE, IPEN, IPORT, OUTFILE
      REAL     DX, DY, SCALE, SCALEX, SCALEY, TX, TY, XCORD, XMAX
      REAL     XMIN, YCORD, YMAX, YMIN
C
C     Examine entire plot file to determine +-X and +-Y extremes
C
      XMAX = -10000.
      XMIN =  10000.
      YMAX = -10000.
      YMIN =  10000.
10    READ (INFILE, *) ICODE, IPEN
20    READ (INFILE, *, END=50) ICODE, XCORD, YCORD
      GO TO (30, 30, 40), ICODE - 1
      WRITE(4,*)'Invalid code number encountered on input plot file'
      RETURN
30    IF (XCORD .LT. XMIN) XMIN = XCORD
      IF (XCORD .GT. XMAX) XMAX = XCORD
      IF (YCORD .LT. YMIN) YMIN = YCORD
      IF (YCORD .GT. YMAX) YMAX = YCORD
40    GO TO 20
C
C     Reposition input file for use by main program
C
50    REWIND INFILE
      READ (INFILE, 1000) ICODE
      READ (INFILE, 1000) ICODE
      READ (INFILE, 1000) ICODE
C
C     Determine scale factor (assuming 8 X 10.5 area) and whether to
C     remain in portrait mode or rotate image into landscape mode.
C
      IF ((XMAX - XMIN) .LT. (YMAX - YMIN)) THEN 
        IPORT = 0  ! Will remain in portrait mode
        SCALEX = 576./(XMAX - XMIN)
        SCALEY = 756./(YMAX - YMIN)
        SCALE = SCALEX
        IF (SCALEY .LT. SCALE) SCALE = SCALEY
      ELSE
        IPORT = 1  ! Will rotate into landscape mode
        SCALEX = 756./(XMAX - XMIN)
        SCALEY = 576./(YMAX - YMIN)
        SCALE = SCALEX
        IF (SCALEY .LT. SCALE) SCALE = SCALEY
      ENDIF
C
C     Determine origin off-set from center of the 8.5 x 11 sheet of paper
C
      DX = (SCALE * (XMAX + XMIN) / 2.)
      DY = (SCALE * (YMAX + YMIN) / 2.)
C
C     Reduce scale by 10%
C
      SCALE = SCALE * 0.9
C
C     Translate origin, (rotate into landscape,) plot title at top of page
C
      WRITE (OUTFILE, 1002)
C     -- portrait mode
      IF (IPORT .EQ. 0) THEN
        WRITE (OUTFILE, 1003) 306. - DX, 396. - DY
        TX = -290. + DX
        TY =  369. + DY
        WRITE (OUTFILE, 1004) TX, TY 
      ENDIF
C     -- landscape mode
      IF (IPORT .EQ. 1) THEN
        WRITE (OUTFILE, 1003) 306., 396.
        WRITE (OUTFILE, 1005) ! rotate image by 90 degrees 
        WRITE (OUTFILE, 1003) - DX, - DY
        TX = -380. + DX
        TY =  279. + DY
        WRITE (OUTFILE, 1004) TX, TY
      ENDIF
C
c *** NEED TO CHECK (and correct) whether the 40 characters of title contain
c *** UNBALANCED parentheses (this will crash the PostScript printing!!)
c
      CALL CHKTITLE (TITLE)
      WRITE (OUTFILE, 1006) TITLE
      RETURN
C
1000  FORMAT (I2,I3)
1001  FORMAT (I2,2F11.4)
1002  FORMAT ('/Times-Roman findfont',/,
     *        '15 scalefont',/,
     *        'setfont')
1003  FORMAT (2F7.2,' translate')
1004  FORMAT (2F7.2,' moveto')
1005  FORMAT ('90 rotate')
1006  FORMAT ('(',A40,') show')
      END
      SUBROUTINE CHKTITLE (TITLE)
C
C     CHECK (and correct) the 40 characters of title so that it does NOT
C     contain UNBALANCED parentheses (otherwise, the PostScript printing
C     will crash!!)
C
      CHARACTER  ATITLE(40)*1, BTITLE*40, IREPL*1, ISYMB*1, TITLE*40
      INTEGER*4  ICL, ICR, IDIF
      EQUIVALENCE (ATITLE(1), BTITLE)
C
      BTITLE = TITLE
C
C     count numbers of left and right parentheses
C
      DO 10 I=1,40
        IF (ATITLE(I) .EQ. '(') ICL = ICL + 1
        IF (ATITLE(I) .EQ. ')') ICR = ICR + 1
10    CONTINUE
C
C     if parentheses are balanced, all's well, return
C
      IF (ICL .EQ. ICR) RETURN
C
C     replace offending parenthesis(es) with bracket(s)
C
      ISYMB = ')'
      IREPL = '}'
      IDIF = ICL - ICR
      IF (IDIF .GE. 1) ISYMB = '('
      IF (IDIF .GE. 1) IREPL = '{'
      IDIF = IABS(IDIF)
      DO 20 I=40,1,-1
        IF (ATITLE(I) .NE. ISYMB) GO TO 20
        ATITLE(I) = IREPL
        IDIF = IDIF - 1
        IF (IDIF .LE. 0) GO TO 30
20    CONTINUE
      WRITE(4,*)'Invalid code number encountered on input plot file'
      RETURN
C
C     return
C
30    TITLE = BTITLE
      RETURN
C
      END
*
************************************************************* 
*
      BLOCK DATA
      CHARACTER*4 TABNC(128),TBXC(256)
      COMMON/TABCHR/TABNC,TBXC
      COMMON/TABLES/TBX(256,10),TBD(128,10),TABN(128) 
C
C  Scattering factor tables
C     V. F. Sears, Neutron News 3, 26 (1992)
      DATA TABNC(1),TABN(1)/'h   ',-0.3739/
      DATA TABNC(2),TABN(2)/'he  ',0.326/
      DATA TABNC(3),TABN(3)/'li  ',-0.190/
      DATA TABNC(4),TABN(4)/'be  ',0.779/
      DATA TABNC(5),TABN(5)/'b   ',0.665/
      DATA TABNC(6),TABN(6)/'c   ',0.6646/
      DATA TABNC(7),TABN(7)/'n   ',0.936/
      DATA TABNC(8),TABN(8)/'o   ',0.5803/
      DATA TABNC(9),TABN(9)/'f   ',0.565/
      DATA TABNC(10),TABN(10)/'ne  ',0.4566/
      DATA TABNC(11),TABN(11)/'na  ',0.363/
      DATA TABNC(12),TABN(12)/'mg  ',0.5375/
      DATA TABNC(13),TABN(13)/'al  ',0.3449/
      DATA TABNC(14),TABN(14)/'si  ',0.4149/
      DATA TABNC(15),TABN(15)/'p   ',0.513/
      DATA TABNC(16),TABN(16)/'s   ',0.2847/
      DATA TABNC(17),TABN(17)/'cl  ',0.95770/
      DATA TABNC(18),TABN(18)/'ar  ',0.1909/
      DATA TABNC(19),TABN(19)/'k   ',0.367/
      DATA TABNC(20),TABN(20)/'ca  ',0.470/
      DATA TABNC(21),TABN(21)/'sc  ',1.23/
      DATA TABNC(22),TABN(22)/'ti  ',-0.3438/
      DATA TABNC(23),TABN(23)/'v   ',-0.0382/
      DATA TABNC(24),TABN(24)/'cr  ',0.3635/
      DATA TABNC(25),TABN(25)/'mn  ',-0.373/
      DATA TABNC(26),TABN(26)/'fe  ',0.945/
      DATA TABNC(27),TABN(27)/'co  ',0.249/
      DATA TABNC(28),TABN(28)/'ni  ',1.03/
      DATA TABNC(29),TABN(29)/'cu  ',0.7718/
      DATA TABNC(30),TABN(30)/'zn  ',0.5680/
      DATA TABNC(31),TABN(31)/'ga  ',0.7288/
      DATA TABNC(32),TABN(32)/'ge  ',0.819/
      DATA TABNC(33),TABN(33)/'as  ',0.658/
      DATA TABNC(34),TABN(34)/'se  ',0.7970/
      DATA TABNC(35),TABN(35)/'br  ',0.680/
      DATA TABNC(36),TABN(36)/'kr  ',0.781/
      DATA TABNC(37),TABN(37)/'rb  ',0.709/
      DATA TABNC(38),TABN(38)/'sr  ',0.702/
      DATA TABNC(39),TABN(39)/'y   ',0.775/
      DATA TABNC(40),TABN(40)/'zr  ',0.716/
      DATA TABNC(41),TABN(41)/'nb  ',0.7054/
      DATA TABNC(42),TABN(42)/'mo  ',0.672/
      DATA TABNC(43),TABN(43)/'tc  ',0.68/
      DATA TABNC(44),TABN(44)/'ru  ',0.703/
      DATA TABNC(45),TABN(45)/'rh  ',0.588/
      DATA TABNC(46),TABN(46)/'pd  ',0.591/
      DATA TABNC(47),TABN(47)/'ag  ',0.5922/
      DATA TABNC(48),TABN(48)/'cd  ',0.487/
      DATA TABNC(49),TABN(49)/'in  ',0.407/
      DATA TABNC(50),TABN(50)/'sn  ',0.6225/
      DATA TABNC(51),TABN(51)/'sb  ',0.557/
      DATA TABNC(52),TABN(52)/'te  ',0.580/
      DATA TABNC(53),TABN(53)/'i   ',0.528/
      DATA TABNC(54),TABN(54)/'xe  ',0.492/
      DATA TABNC(55),TABN(55)/'cs  ',0.542/
      DATA TABNC(56),TABN(56)/'ba  ',0.507/
      DATA TABNC(57),TABN(57)/'la  ',0.824/
      DATA TABNC(58),TABN(58)/'ce  ',0.484/
      DATA TABNC(59),TABN(59)/'pr  ',0.458/
      DATA TABNC(60),TABN(60)/'nd  ',0.769/
      DATA TABNC(61),TABN(61)/'eu  ',0.722/
      DATA TABNC(62),TABN(62)/'gd  ',0.65/
      DATA TABNC(63),TABN(63)/'tb  ',0.738/
      DATA TABNC(64),TABN(64)/'dy  ',1.69/
      DATA TABNC(65),TABN(65)/'ho  ',0.801/
      DATA TABNC(66),TABN(66)/'er  ',0.779/
      DATA TABNC(67),TABN(67)/'tm  ',0.707/
      DATA TABNC(68),TABN(68)/'yb  ',1.243/
      DATA TABNC(69),TABN(69)/'lu  ',0.721/
      DATA TABNC(70),TABN(70)/'hf  ',0.78/
      DATA TABNC(71),TABN(71)/'ta  ',0.691/
      DATA TABNC(72),TABN(72)/'w   ',0.486/
      DATA TABNC(73),TABN(73)/'re  ',0.92/
      DATA TABNC(74),TABN(74)/'os  ',1.07/
      DATA TABNC(75),TABN(75)/'ir  ',1.06/
      DATA TABNC(76),TABN(76)/'pt  ',0.960/
      DATA TABNC(77),TABN(77)/'au  ',0.763/
      DATA TABNC(78),TABN(78)/'hg  ',1.269/
      DATA TABNC(79),TABN(79)/'tl  ',0.8776/
      DATA TABNC(80),TABN(80)/'pb  ',0.9405/
      DATA TABNC(81),TABN(81)/'bi  ',0.8532/
      DATA TABNC(82),TABN(82)/'th  ',1.031/
      DATA TABNC(83),TABN(83)/'u   ',0.8417/
      DATA TABNC(84),TABN(84)/'np  ',1.06/
      DATA TABNC(85),TABN(85)/'pu  ',0.77/
      DATA TABNC(86),TABN(86)/'d   ',0.6671/
      DATA TABNC(87),TABN(87)/'pm  ',1.26/
      DATA TABNC(88),TABN(88)/'sm  ',0.80/
      DATA TABNC(89),TABN(89)/'ra  ',1.0/
      DATA TABNC(90),TABN(90)/'pa  ',0.91/
      DATA TABNC(91),TABN(91)/'am  ',0.83/
      DATA TABNC(92),TABN(92)/'cm  ',0.95/
      DATA TBD(  2, 1),TBD(  2, 2),TBD(  2, 3),TBD(  2, 4),TBD(  2, 5)
     *    /   .002    ,   .002    ,   .001    ,  -.000    ,  -.000    /
      DATA TBD(  2, 6),TBD(  2, 7),TBD(  2, 8),TBD(  2, 9),TBD(  2,10)
     *    /   .001    ,   .001    ,   .000    ,   .000    ,   .000    /
      DATA TBD(  3, 1),TBD(  3, 2),TBD(  3, 3),TBD(  3, 4),TBD(  3, 5)
     *    /   .008    ,   .005    ,   .003    ,  -.000    ,  -.001    /
      DATA TBD(  3, 6),TBD(  3, 7),TBD(  3, 8),TBD(  3, 9),TBD(  3,10)
     *    /   .003    ,   .002    ,   .001    ,   .000    ,   .000    /
      DATA TBD(  4, 1),TBD(  4, 2),TBD(  4, 3),TBD(  4, 4),TBD(  4, 5)
     *    /   .018    ,   .013    ,   .008    ,   .000    ,  -.000    /
      DATA TBD(  4, 6),TBD(  4, 7),TBD(  4, 8),TBD(  4, 9),TBD(  4,10)
     *    /   .009    ,   .007    ,   .004    ,   .001    ,   .000    /
      DATA TBD(  5, 1),TBD(  5, 2),TBD(  5, 3),TBD(  5, 4),TBD(  5, 5)
     *    /   .035    ,   .026    ,   .017    ,   .002    ,   .000    /
      DATA TBD(  5, 6),TBD(  5, 7),TBD(  5, 8),TBD(  5, 9),TBD(  5,10)
     *    /   .021    ,   .015    ,   .009    ,   .002    ,   .001    /
      DATA TBD(  6, 1),TBD(  6, 2),TBD(  6, 3),TBD(  6, 4),TBD(  6, 5)
     *    /   .059    ,   .044    ,   .029    ,   .004    ,   .001    /
      DATA TBD(  6, 6),TBD(  6, 7),TBD(  6, 8),TBD(  6, 9),TBD(  6,10)
     *    /   .042    ,   .029    ,   .018    ,   .003    ,   .002    /
      DATA TBD(  7, 1),TBD(  7, 2),TBD(  7, 3),TBD(  7, 4),TBD(  7, 5)
     *    /   .090    ,   .069    ,   .047    ,   .008    ,   .003    /
      DATA TBD(  7, 6),TBD(  7, 7),TBD(  7, 8),TBD(  7, 9),TBD(  7,10)
     *    /   .073    ,   .052    ,   .032    ,   .006    ,   .004    /
      DATA TBD(  8, 1),TBD(  8, 2),TBD(  8, 3),TBD(  8, 4),TBD(  8, 5)
     *    /   .129    ,   .100    ,   .069    ,   .014    ,   .006    /
      DATA TBD(  8, 6),TBD(  8, 7),TBD(  8, 8),TBD(  8, 9),TBD(  8,10)
     *    /   .119    ,   .085    ,   .053    ,   .010    ,   .006    /
      DATA TBD(  9, 1),TBD(  9, 2),TBD(  9, 3),TBD(  9, 4),TBD(  9, 5)
     *    /   .174    ,   .138    ,   .097    ,   .021    ,   .011    /
      DATA TBD(  9, 6),TBD(  9, 7),TBD(  9, 8),TBD(  9, 9),TBD(  9,10)
     *    /   .184    ,   .132    ,   .083    ,   .016    ,   .010    /
      DATA TBD( 10, 1),TBD( 10, 2),TBD( 10, 3),TBD( 10, 4),TBD( 10, 5)
     *    /   .223    ,   .180    ,   .129    ,   .030    ,   .016    /
      DATA TBD( 10, 6),TBD( 10, 7),TBD( 10, 8),TBD( 10, 9),TBD( 10,10)
     *    /   .270    ,   .195    ,   .124    ,   .025    ,   .015    /
      DATA TBD( 11, 1),TBD( 11, 2),TBD( 11, 3),TBD( 11, 4),TBD( 11, 5)
     *    /   .272    ,   .224    ,   .165    ,   .042    ,   .023    /
      DATA TBD( 11, 6),TBD( 11, 7),TBD( 11, 8),TBD( 11, 9),TBD( 11,10)
     *    /   .381    ,   .277    ,   .177    ,   .036    ,   .022    /
      DATA TBD( 12, 1),TBD( 12, 2),TBD( 12, 3),TBD( 12, 4),TBD( 12, 5)
     *    /   .318    ,   .269    ,   .204    ,   .056    ,   .032    /
      DATA TBD( 12, 6),TBD( 12, 7),TBD( 12, 8),TBD( 12, 9),TBD( 12,10)
     *    /   .522    ,   .381    ,   .246    ,   .052    ,   .031    /
      DATA TBD( 13, 1),TBD( 13, 2),TBD( 13, 3),TBD( 13, 4),TBD( 13, 5)
     *    /   .355    ,   .311    ,   .244    ,   .072    ,   .042    /
      DATA TBD( 13, 6),TBD( 13, 7),TBD( 13, 8),TBD( 13, 9),TBD( 13,10)
     *    /   .693    ,   .509    ,   .330    ,   .071    ,   .043    /
      DATA TBD( 14, 1),TBD( 14, 2),TBD( 14, 3),TBD( 14, 4),TBD( 14, 5)
     *    /   .377    ,   .347    ,   .283    ,   .090    ,   .055    /
      DATA TBD( 14, 6),TBD( 14, 7),TBD( 14, 8),TBD( 14, 9),TBD( 14,10)
     *    /   .900    ,   .664    ,   .434    ,   .095    ,   .058    /
      DATA TBD( 15, 1),TBD( 15, 2),TBD( 15, 3),TBD( 15, 4),TBD( 15, 5)
     *    /   .374    ,   .370    ,   .319    ,   .110    ,   .068    /
      DATA TBD( 15, 6),TBD( 15, 7),TBD( 15, 8),TBD( 15, 9),TBD( 15,10)
     *    /  1.142    ,   .847    ,   .557    ,   .124    ,   .076    /
      DATA TBD( 16, 1),TBD( 16, 2),TBD( 16, 3),TBD( 16, 4),TBD( 16, 5)
     *    /   .335    ,   .375    ,   .348    ,   .132    ,   .084    /
      DATA TBD( 16, 6),TBD( 16, 7),TBD( 16, 8),TBD( 16, 9),TBD( 16,10)
     *    /  1.423    ,  1.061    ,   .702    ,   .159    ,   .099    /
      DATA TBD( 17, 1),TBD( 17, 2),TBD( 17, 3),TBD( 17, 4),TBD( 17, 5)
     *    /   .243    ,   .352    ,   .366    ,   .155    ,   .101    /
      DATA TBD( 17, 6),TBD( 17, 7),TBD( 17, 8),TBD( 17, 9),TBD( 17,10)
     *    /  1.747    ,  1.309    ,   .872    ,   .201    ,   .125    /
      DATA TBD( 18, 1),TBD( 18, 2),TBD( 18, 3),TBD( 18, 4),TBD( 18, 5)
     *    /   .070    ,   .286    ,   .365    ,   .179    ,   .118    /
      DATA TBD( 18, 6),TBD( 18, 7),TBD( 18, 8),TBD( 18, 9),TBD( 18,10)
     *    /  2.110    ,  1.589    ,  1.066    ,   .250    ,   .156    /
      DATA TBD( 19, 1),TBD( 19, 2),TBD( 19, 3),TBD( 19, 4),TBD( 19, 5)
     *    /  -.221    ,   .163    ,   .341    ,   .203    ,   .137    /
      DATA TBD( 19, 6),TBD( 19, 7),TBD( 19, 8),TBD( 19, 9),TBD( 19,10)
     *    /  2.514    ,  1.904    ,  1.286    ,   .306    ,   .193    /
      DATA TBD( 20, 1),TBD( 20, 2),TBD( 20, 3),TBD( 20, 4),TBD( 20, 5)
     *    /  -.717    ,  -.038    ,   .285    ,   .226    ,   .156    /
      DATA TBD( 20, 6),TBD( 20, 7),TBD( 20, 8),TBD( 20, 9),TBD( 20,10)
     *    /  2.968    ,  2.256    ,  1.533    ,   .372    ,   .235    /
      DATA TBD( 21, 1),TBD( 21, 2),TBD( 21, 3),TBD( 21, 4),TBD( 21, 5)
     *    / -1.683    ,  -.357    ,   .189    ,   .248    ,   .175    /
      DATA TBD( 21, 6),TBD( 21, 7),TBD( 21, 8),TBD( 21, 9),TBD( 21,10)
     *    /  3.470    ,  2.643    ,  1.807    ,   .446    ,   .283    /
      DATA TBD( 22, 1),TBD( 22, 2),TBD( 22, 3),TBD( 22, 4),TBD( 22, 5)
     *    / -3.841    ,  -.896    ,   .035    ,   .267    ,   .194    /
      DATA TBD( 22, 6),TBD( 22, 7),TBD( 22, 8),TBD( 22, 9),TBD( 22,10)
     *    /   .459    ,  3.070    ,  2.110    ,   .530    ,   .338    /
      DATA TBD( 23, 1),TBD( 23, 2),TBD( 23, 3),TBD( 23, 4),TBD( 23, 5)
     *    / -2.161    , -1.973    ,  -.198    ,   .284    ,   .213    /
      DATA TBD( 23, 6),TBD( 23, 7),TBD( 23, 8),TBD( 23, 9),TBD( 23,10)
     *    /   .548    ,  3.533    ,  2.443    ,   .624    ,   .399    /
      DATA TBD( 24, 1),TBD( 24, 2),TBD( 24, 3),TBD( 24, 4),TBD( 24, 5)
     *    / -1.639    , -3.367    ,  -.568    ,   .295    ,   .229    /
      DATA TBD( 24, 6),TBD( 24, 7),TBD( 24, 8),TBD( 24, 9),TBD( 24,10)
     *    /   .650    ,   .481    ,  2.808    ,   .729    ,   .468    /
      DATA TBD( 25, 1),TBD( 25, 2),TBD( 25, 3),TBD( 25, 4),TBD( 25, 5)
     *    / -1.339    , -2.095    , -1.179    ,   .301    ,   .244    /
      DATA TBD( 25, 6),TBD( 25, 7),TBD( 25, 8),TBD( 25, 9),TBD( 25,10)
     *    /   .764    ,   .566    ,  3.204    ,   .845    ,   .545    /
      DATA TBD( 26, 1),TBD( 26, 2),TBD( 26, 3),TBD( 26, 4),TBD( 26, 5)
     *    / -1.124    , -1.623    , -2.464    ,   .299    ,   .256    /
      DATA TBD( 26, 6),TBD( 26, 7),TBD( 26, 8),TBD( 26, 9),TBD( 26,10)
     *    /   .893    ,   .662    ,  3.608    ,   .973    ,   .630    /
      DATA TBD( 27, 1),TBD( 27, 2),TBD( 27, 3),TBD( 27, 4),TBD( 27, 5)
     *    /  -.956    , -1.343    , -2.956    ,   .285    ,   .261    /
      DATA TBD( 27, 6),TBD( 27, 7),TBD( 27, 8),TBD( 27, 9),TBD( 27,10)
     *    /  1.036    ,   .769    ,   .509    ,  1.113    ,   .724    /
      DATA TBD( 28, 1),TBD( 28, 2),TBD( 28, 3),TBD( 28, 4),TBD( 28, 5)
     *    /  -.795    , -1.129    , -2.019    ,   .263    ,   .265    /
      DATA TBD( 28, 6),TBD( 28, 7),TBD( 28, 8),TBD( 28, 9),TBD( 28,10)
     *    /  1.196    ,   .888    ,   .589    ,  1.266    ,   .826    /
      DATA TBD( 29, 1),TBD( 29, 2),TBD( 29, 3),TBD( 29, 4),TBD( 29, 5)
     *    /  -.684    ,  -.978    , -1.612    ,   .222    ,   .260    /
      DATA TBD( 29, 6),TBD( 29, 7),TBD( 29, 8),TBD( 29, 9),TBD( 29,10)
     *    /  1.373    ,  1.021    ,   .678    ,  1.431    ,   .938    /
      DATA TBD( 30, 1),TBD( 30, 2),TBD( 30, 3),TBD( 30, 4),TBD( 30, 5)
     *    /  -.570    ,  -.841    , -1.354    ,   .163    ,   .249    /
      DATA TBD( 30, 6),TBD( 30, 7),TBD( 30, 8),TBD( 30, 9),TBD( 30,10)
     *    /  1.569    ,  1.168    ,   .777    ,  1.609    ,  1.059    /
      DATA TBD( 31, 1),TBD( 31, 2),TBD( 31, 3),TBD( 31, 4),TBD( 31, 5)
     *    /  -.462    ,  -.717    , -1.163    ,   .081    ,   .228    /
      DATA TBD( 31, 6),TBD( 31, 7),TBD( 31, 8),TBD( 31, 9),TBD( 31,10)
     *    /  1.786    ,  1.331    ,   .886    ,  1.801    ,  1.190    /
      DATA TBD( 32, 1),TBD( 32, 2),TBD( 32, 3),TBD( 32, 4),TBD( 32, 5)
     *    /  -.365    ,  -.607    , -1.011    ,  -.030    ,   .196    /
      DATA TBD( 32, 6),TBD( 32, 7),TBD( 32, 8),TBD( 32, 9),TBD( 32,10)
     *    /  2.022    ,  1.508    ,  1.006    ,  2.007    ,  1.332    /
      DATA TBD( 33, 1),TBD( 33, 2),TBD( 33, 3),TBD( 33, 4),TBD( 33, 5)
     *    /  -.273    ,  -.503    ,  -.879    ,  -.178    ,   .152    /
      DATA TBD( 33, 6),TBD( 33, 7),TBD( 33, 8),TBD( 33, 9),TBD( 33,10)
     *    /  2.283    ,  1.704    ,  1.139    ,  2.223    ,  1.481    /
      DATA TBD( 34, 1),TBD( 34, 2),TBD( 34, 3),TBD( 34, 4),TBD( 34, 5)
     *    /  -.198    ,  -.413    ,  -.767    ,  -.374    ,   .090    /
      DATA TBD( 34, 6),TBD( 34, 7),TBD( 34, 8),TBD( 34, 9),TBD( 34,10)
     *    /  2.563    ,  1.916    ,  1.283    ,  2.456    ,  1.643    /
      DATA TBD( 35, 1),TBD( 35, 2),TBD( 35, 3),TBD( 35, 4),TBD( 35, 5)
     *    /  -.130    ,  -.328    ,  -.665    ,  -.652    ,   .008    /
      DATA TBD( 35, 6),TBD( 35, 7),TBD( 35, 8),TBD( 35, 9),TBD( 35,10)
     *    /  2.872    ,  2.149    ,  1.439    ,  2.713    ,  1.820    /
      DATA TBD( 36, 1),TBD( 36, 2),TBD( 36, 3),TBD( 36, 4),TBD( 36, 5)
     *    /  -.082    ,  -.256    ,  -.574    , -1.044    ,  -.099    /
      DATA TBD( 36, 6),TBD( 36, 7),TBD( 36, 8),TBD( 36, 9),TBD( 36,10)
     *    /  3.201    ,  2.398    ,  1.608    ,  2.973    ,  2.003    /
      DATA TBD( 37, 1),TBD( 37, 2),TBD( 37, 3),TBD( 37, 4),TBD( 37, 5)
     *    /  -.012    ,  -.161    ,  -.465    , -1.657    ,  -.230    /
      DATA TBD( 37, 6),TBD( 37, 7),TBD( 37, 8),TBD( 37, 9),TBD( 37,10)
     *    /  3.608    ,  2.709    ,  1.820    ,  3.264    ,  2.203    /
      DATA TBD( 38, 1),TBD( 38, 2),TBD( 38, 3),TBD( 38, 4),TBD( 38, 5)
     *    /   .006    ,  -.106    ,  -.386    , -2.951    ,  -.406    /
      DATA TBD( 38, 6),TBD( 38, 7),TBD( 38, 8),TBD( 38, 9),TBD( 38,10)
     *    /  4.002    ,  3.009    ,  2.025    ,  3.542    ,  2.411    /
      DATA TBD( 39, 1),TBD( 39, 2),TBD( 39, 3),TBD( 39, 4),TBD( 39, 5)
     *    /   .007    ,  -.061    ,  -.314    , -2.965    ,  -.639    /
      DATA TBD( 39, 6),TBD( 39, 7),TBD( 39, 8),TBD( 39, 9),TBD( 39,10)
     *    /  4.422    ,  3.329    ,  2.245    ,   .560    ,  2.630    /
      DATA TBD( 40, 1),TBD( 40, 2),TBD( 40, 3),TBD( 40, 4),TBD( 40, 5)
     *    /  -.013    ,  -.028    ,  -.248    , -2.197    ,  -.957    /
      DATA TBD( 40, 6),TBD( 40, 7),TBD( 40, 8),TBD( 40, 9),TBD( 40,10)
     *    /  4.876    ,  3.676    ,  2.482    ,   .621    ,  2.860    /
      DATA TBD( 41, 1),TBD( 41, 2),TBD( 41, 3),TBD( 41, 4),TBD( 41, 5)
     *    /  -.063    ,  -.012    ,  -.191    , -1.825    , -1.416    /
      DATA TBD( 41, 6),TBD( 41, 7),TBD( 41, 8),TBD( 41, 9),TBD( 41,10)
     *    /  5.353    ,  4.043    ,  2.735    ,   .688    ,  3.103    /
      DATA TBD( 42, 1),TBD( 42, 2),TBD( 42, 3),TBD( 42, 4),TBD( 42, 5)
     *    /  -.153    ,  -.017    ,  -.145    , -1.590    , -2.205    /
      DATA TBD( 42, 6),TBD( 42, 7),TBD( 42, 8),TBD( 42, 9),TBD( 42,10)
     *    /  5.862    ,  4.434    ,  3.005    ,   .759    ,  3.353    /
      DATA TBD( 43, 1),TBD( 43, 2),TBD( 43, 3),TBD( 43, 4),TBD( 43, 5)
     *    /  -.270    ,  -.039    ,  -.105    , -1.420    , -5.524    /
      DATA TBD( 43, 6),TBD( 43, 7),TBD( 43, 8),TBD( 43, 9),TBD( 43,10)
     *    /  6.406    ,  4.854    ,  3.296    ,   .836    ,  3.651    /
      DATA TBD( 44, 1),TBD( 44, 2),TBD( 44, 3),TBD( 44, 4),TBD( 44, 5)
     *    /  -.424    ,  -.083    ,  -.077    , -1.287    , -2.649    /
      DATA TBD( 44, 6),TBD( 44, 7),TBD( 44, 8),TBD( 44, 9),TBD( 44,10)
     *    /  6.984    ,  5.300    ,  3.605    ,   .919    ,   .596    /
      DATA TBD( 45, 1),TBD( 45, 2),TBD( 45, 3),TBD( 45, 4),TBD( 45, 5)
     *    /  -.639    ,  -.157    ,  0.1215   , -1.177    , -2.128    /
      DATA TBD( 45, 6),TBD( 45, 7),TBD( 45, 8),TBD( 45, 9),TBD( 45,10)
     *    /  7.594    ,  5.773    ,  3.9337   ,  1.007    ,   .654    /
      DATA TBD( 46, 1),TBD( 46, 2),TBD( 46, 3),TBD( 46, 4),TBD( 46, 5)
     *    /  -.924    ,  -.259    ,  -.060    , -1.085    , -1.834    /
      DATA TBD( 46, 6),TBD( 46, 7),TBD( 46, 8),TBD( 46, 9),TBD( 46,10)
     *    /  8.235    ,  6.271    ,  4.282    ,  1.101    ,   .717    /
      DATA TBD( 47, 1),TBD( 47, 2),TBD( 47, 3),TBD( 47, 4),TBD( 47, 5)
     *    / -1.303    ,  -.416    ,  -.079    , -1.005    , -1.637    /
      DATA TBD( 47, 6),TBD( 47, 7),TBD( 47, 8),TBD( 47, 9),TBD( 47,10)
     *    /  8.912    ,  6.800    ,  4.653    ,  1.202    ,   .783    /
      DATA TBD( 48, 1),TBD( 48, 2),TBD( 48, 3),TBD( 48, 4),TBD( 48, 5)
     *    / -1.788    ,  -.626    ,  -.126    ,  -.936    , -1.493    /
      DATA TBD( 48, 6),TBD( 48, 7),TBD( 48, 8),TBD( 48, 9),TBD( 48,10)
     *    /  9.627    ,  7.356    ,  5.045    ,  1.310    ,   .854    /
      DATA TBD( 49, 1),TBD( 49, 2),TBD( 49, 3),TBD( 49, 4),TBD( 49, 5)
     *    / -2.401    ,  -.888    ,  -.194    ,  -.873    , -1.378    /
      DATA TBD( 49, 6),TBD( 49, 7),TBD( 49, 8),TBD( 49, 9),TBD( 49,10)
     *    / 10.380    ,  7.943    ,  5.459    ,  1.424    ,   .930    /
      DATA TBD( 50, 1),TBD( 50, 2),TBD( 50, 3),TBD( 50, 4),TBD( 50, 5)
     *    / -3.194    , -1.214    ,  -.287    ,  -.816    , -1.284    /
      DATA TBD( 50, 6),TBD( 50, 7),TBD( 50, 8),TBD( 50, 9),TBD( 50,10)
     *    / 11.166    ,  8.557    ,  5.894    ,  1.546    ,  1.010    /
      DATA TBD( 51, 1),TBD( 51, 2),TBD( 51, 3),TBD( 51, 4),TBD( 51, 5)
     *    / -4.267    , -1.630    ,  -.418    ,  -.772    , -1.212    /
      DATA TBD( 51, 6),TBD( 51, 7),TBD( 51, 8),TBD( 51, 9),TBD( 51,10)
     *    / 11.995    ,  9.203    ,  6.352    ,  1.675    ,  1.096    /
      DATA TBD( 52, 1),TBD( 52, 2),TBD( 52, 3),TBD( 52, 4),TBD( 52, 5)
     *    / -5.852    , -2.147    ,  -.579    ,  -.726    , -1.144    /
      DATA TBD( 52, 6),TBD( 52, 7),TBD( 52, 8),TBD( 52, 9),TBD( 52,10)
     *    / 12.850    ,  9.885    ,  6.835    ,  1.812    ,  1.187    /
      DATA TBD( 53, 1),TBD( 53, 2),TBD( 53, 3),TBD( 53, 4),TBD( 53, 5)
     *    / -8.133    , -2.812    ,  -.783    ,  -.684    , -1.084    /
      DATA TBD( 53, 6),TBD( 53, 7),TBD( 53, 8),TBD( 53, 9),TBD( 53,10)
     *    / 11.933    , 10.608    ,  7.348    ,  1.958    ,  1.284    /
      DATA TBD( 54, 1),TBD( 54, 2),TBD( 54, 3),TBD( 54, 4),TBD( 54, 5)
     *    /-10.742    , -3.652    , -1.022    ,  -.644    , -1.029    /
      DATA TBD( 54, 6),TBD( 54, 7),TBD( 54, 8),TBD( 54, 9),TBD( 54,10)
     *    / 12.919    , 11.382    ,  7.904    ,  2.119    ,  1.391    /
      DATA TBD( 55, 1),TBD( 55, 2),TBD( 55, 3),TBD( 55, 4),TBD( 55, 5)
     *    /-11.460    , -4.832    , -1.334    ,  -.613    ,  -.983    /
      DATA TBD( 55, 6),TBD( 55, 7),TBD( 55, 8),TBD( 55, 9),TBD( 55,10)
     *    /  9.981    , 12.164    ,  8.460    ,  2.282    ,  1.500    /
      DATA TBD( 56, 1),TBD( 56, 2),TBD( 56, 3),TBD( 56, 4),TBD( 56, 5)
     *    /-12.135    , -6.683    , -1.716    ,  -.588    ,  -.942    /
      DATA TBD( 56, 6),TBD( 56, 7),TBD( 56, 8),TBD( 56, 9),TBD( 56,10)
     *    /  3.565    , 12.937    ,  9.036    ,  2.452    ,  1.615    /
      DATA TBD( 57, 1),TBD( 57, 2),TBD( 57, 3),TBD( 57, 4),TBD( 57, 5)
     *    / -9.574    , -8.388    , -2.170    ,  -.564    ,  -.904    /
      DATA TBD( 57, 6),TBD( 57, 7),TBD( 57, 8),TBD( 57, 9),TBD( 57,10)
     *    /  3.843    , 11.953    ,  9.648    ,  2.632    ,  1.735    /
      DATA TBD( 58, 1),TBD( 58, 2),TBD( 58, 3),TBD( 58, 4),TBD( 58, 5)
     *    / -7.817    ,-12.457    , -2.939    ,  -.530    ,  -.859    /
      DATA TBD( 58, 6),TBD( 58, 7),TBD( 58, 8),TBD( 58, 9),TBD( 58,10)
     *    /  4.130    ,  6.285    , 10.535    ,  2.845    ,  1.873    /
      DATA TBD( 59, 1),TBD( 59, 2),TBD( 59, 3),TBD( 59, 4),TBD( 59, 5)
     *    / -7.486    ,-11.016    , -3.431    ,  -.535    ,  -.842    /
      DATA TBD( 59, 6),TBD( 59, 7),TBD( 59, 8),TBD( 59, 9),TBD( 59,10)
     *    /  4.427    ,  9.874    , 10.933    ,  3.018    ,  1.995    /
      DATA TBD( 60, 1),TBD( 60, 2),TBD( 60, 3),TBD( 60, 4),TBD( 60, 5)
     *    / -6.891    ,-12.122    , -4.357    ,  -.530    ,  -.818    /
      DATA TBD( 60, 6),TBD( 60, 7),TBD( 60, 8),TBD( 60, 9),TBD( 60,10)
     *    /  4.741    ,  3.627    , 11.614    ,  3.225    ,  2.135    /
      DATA TBD( 61, 1),TBD( 61, 2),TBD( 61, 3),TBD( 61, 4),TBD( 61, 5)
     *    / -6.429    , -9.616    , -5.696    ,  -.533    ,  -.798    /
      DATA TBD( 61, 6),TBD( 61, 7),TBD( 61, 8),TBD( 61, 9),TBD( 61,10)
     *    /  5.073    ,  3.883    , 12.320    ,  3.442    ,  2.281    /
      DATA TBD( 62, 1),TBD( 62, 2),TBD( 62, 3),TBD( 62, 4),TBD( 62, 5)
     *    / -6.050    , -8.352    , -7.718    ,  -.542    ,  -.782    /
      DATA TBD( 62, 6),TBD( 62, 7),TBD( 62, 8),TBD( 62, 9),TBD( 62,10)
     *    /  5.416    ,  4.149    , 11.276    ,  3.669    ,  2.435    /
      DATA TBD( 63, 1),TBD( 63, 2),TBD( 63, 3),TBD( 63, 4),TBD( 63, 5)
     *    / -5.779    , -7.565    , -9.242    ,  -.564    ,  -.774    /
      DATA TBD( 63, 6),TBD( 63, 7),TBD( 63, 8),TBD( 63, 9),TBD( 63,10)
     *    /  5.773    ,  4.427    , 11.946    ,  3.904    ,  2.595    /
      DATA TBD( 64, 1),TBD( 64, 2),TBD( 64, 3),TBD( 64, 4),TBD( 64, 5)
     *    / -5.525    , -6.980    , -9.498    ,  -.591    ,  -.767    /
      DATA TBD( 64, 6),TBD( 64, 7),TBD( 64, 8),TBD( 64, 9),TBD( 64,10)
     *    /  6.153    ,  4.721    ,  9.242    ,  4.151    ,  2.764    /
      DATA TBD( 65, 1),TBD( 65, 2),TBD( 65, 3),TBD( 65, 4),TBD( 65, 5)
     *    / -5.250    , -6.492    ,-10.423    ,  -.619    ,  -.761    /
      DATA TBD( 65, 6),TBD( 65, 7),TBD( 65, 8),TBD( 65, 9),TBD( 65,10)
     *    /  6.549    ,  5.026    ,  9.748    ,  4.410    ,  2.940    /
      DATA TBD( 66, 1),TBD( 66, 2),TBD( 66, 3),TBD( 66, 4),TBD( 66, 5)
     *    / -5.040    , -6.112    ,-12.255    ,  -.666    ,  -.765    /
      DATA TBD( 66, 6),TBD( 66, 7),TBD( 66, 8),TBD( 66, 9),TBD( 66,10)
     *    /  6.958    ,  5.343    ,  3.704    ,  4.678    ,  3.124    /
      DATA TBD( 67, 1),TBD( 67, 2),TBD( 67, 3),TBD( 67, 4),TBD( 67, 5)
     *    / -4.878    , -5.810    , -9.733    ,  -.723    ,  -.773    /
      DATA TBD( 67, 6),TBD( 67, 7),TBD( 67, 8),TBD( 67, 9),TBD( 67,10)
     *    /  7.387    ,  5.675    ,  3.937    ,  4.958    ,  3.316    /
      DATA TBD( 68, 1),TBD( 68, 2),TBD( 68, 3),TBD( 68, 4),TBD( 68, 5)
     *    / -4.753    , -5.565    , -8.488    ,  -.795    ,  -.790    /
      DATA TBD( 68, 6),TBD( 68, 7),TBD( 68, 8),TBD( 68, 9),TBD( 68,10)
     *    /  7.833    ,  6.022    ,  4.181    ,  5.248    ,  3.515    /
      DATA TBD( 69, 1),TBD( 69, 2),TBD( 69, 3),TBD( 69, 4),TBD( 69, 5)
     *    / -4.652    , -5.361    , -7.701    ,  -.884    ,  -.815    /
      DATA TBD( 69, 6),TBD( 69, 7),TBD( 69, 8),TBD( 69, 9),TBD( 69,10)
     *    /  8.291    ,  6.378    ,  4.432    ,  5.548    ,  3.723    /
      DATA TBD( 70, 1),TBD( 70, 2),TBD( 70, 3),TBD( 70, 4),TBD( 70, 5)
     *    / -4.580    , -5.190    , -7.133    ,  -.988    ,  -.847    /
      DATA TBD( 70, 6),TBD( 70, 7),TBD( 70, 8),TBD( 70, 9),TBD( 70,10)
     *    /  8.759    ,  6.745    ,  4.693    ,  5.858    ,  3.937    /
      DATA TBD( 71, 1),TBD( 71, 2),TBD( 71, 3),TBD( 71, 4),TBD( 71, 5)
     *    / -4.592    , -5.088    , -6.715    , -1.118    ,  -.890    /
      DATA TBD( 71, 6),TBD( 71, 7),TBD( 71, 8),TBD( 71, 9),TBD( 71,10)
     *    /  9.277    ,  7.148    ,  4.977    ,  6.185    ,  4.164    /
      DATA TBD( 72, 1),TBD( 72, 2),TBD( 72, 3),TBD( 72, 4),TBD( 72, 5)
     *    / -4.540    , -4.948    , -6.351    , -1.258    ,  -.937    /
      DATA TBD( 72, 6),TBD( 72, 7),TBD( 72, 8),TBD( 72, 9),TBD( 72,10)
     *    /  9.811    ,  7.565    ,  5.271    ,  6.523    ,  4.399    /
      DATA TBD( 73, 1),TBD( 73, 2),TBD( 73, 3),TBD( 73, 4),TBD( 73, 5)
     *    / -4.499    , -4.823    , -6.048    , -1.421    ,  -.993    /
      DATA TBD( 73, 6),TBD( 73, 7),TBD( 73, 8),TBD( 73, 9),TBD( 73,10)
     *    / 10.364    ,  7.996    ,  5.577    ,  6.872    ,  4.643    /
      DATA TBD( 74, 1),TBD( 74, 2),TBD( 74, 3),TBD( 74, 4),TBD( 74, 5)
     *    / -4.483    , -4.719    , -5.790    , -1.598    , -1.048    /
      DATA TBD( 74, 6),TBD( 74, 7),TBD( 74, 8),TBD( 74, 9),TBD( 74,10)
     *    / 10.929    ,  8.439    ,  5.891    ,  7.232    ,  4.894    /
      DATA TBD( 75, 1),TBD( 75, 2),TBD( 75, 3),TBD( 75, 4),TBD( 75, 5)
     *    / -4.503    , -4.647    , -5.581    , -1.816    , -1.127    /
      DATA TBD( 75, 6),TBD( 75, 7),TBD( 75, 8),TBD( 75, 9),TBD( 75,10)
     *    / 11.520    ,  8.903    ,  6.221    ,  7.605    ,  5.156    /
      DATA TBD( 76, 1),TBD( 76, 2),TBD( 76, 3),TBD( 76, 4),TBD( 76, 5)
     *    / -4.527    , -4.578    , -5.391    , -2.066    , -1.216    /
      DATA TBD( 76, 6),TBD( 76, 7),TBD( 76, 8),TBD( 76, 9),TBD( 76,10)
     *    / 12.140    ,  9.389    ,  6.566    ,  7.990    ,  5.427    /
      DATA TBD( 77, 1),TBD( 77, 2),TBD( 77, 3),TBD( 77, 4),TBD( 77, 5)
     *    / -4.584    , -4.535    , -5.233    , -2.352    , -1.319    /
      DATA TBD( 77, 6),TBD( 77, 7),TBD( 77, 8),TBD( 77, 9),TBD( 77,10)
     *    / 12.787    ,  9.895    ,  6.925    ,  8.388    ,  5.708    /
      DATA TBD( 78, 1),TBD( 78, 2),TBD( 78, 3),TBD( 78, 4),TBD( 78, 5)
     *    / -4.668    , -4.510    , -5.096    , -2.688    , -1.438    /
      DATA TBD( 78, 6),TBD( 78, 7),TBD( 78, 8),TBD( 78, 9),TBD( 78,10)
     *    / 13.451    , 10.418    ,  7.297    ,  8.798    ,  5.998    /
      DATA TBD( 79, 1),TBD( 79, 2),TBD( 79, 3),TBD( 79, 4),TBD( 79, 5)
     *    / -4.803    , -4.523    , -4.990    , -3.084    , -1.576    /
      DATA TBD( 79, 6),TBD( 79, 7),TBD( 79, 8),TBD( 79, 9),TBD( 79,10)
     *    / 14.143    , 10.963    ,  7.686    ,  9.223    ,  6.299    /
      DATA TBD( 80, 1),TBD( 80, 2),TBD( 80, 3),TBD( 80, 4),TBD( 80, 5)
     *    / -4.945    , -4.532    , -4.883    , -3.556    , -1.730    /
      DATA TBD( 80, 6),TBD( 80, 7),TBD( 80, 8),TBD( 80, 9),TBD( 80,10)
     *    / 14.860    , 11.528    ,  8.089    ,  9.659    ,  6.610    /
      DATA TBD( 81, 1),TBD( 81, 2),TBD( 81, 3),TBD( 81, 4),TBD( 81, 5)
     *    / -5.161    , -4.596    , -4.818    , -4.133    , -1.910    /
      DATA TBD( 81, 6),TBD( 81, 7),TBD( 81, 8),TBD( 81, 9),TBD( 81,10)
     *    / 15.595    , 12.108    ,  8.505    , 10.102    ,  6.930    /
      DATA TBD( 82, 1),TBD( 82, 2),TBD( 82, 3),TBD( 82, 4),TBD( 82, 5)
     *    / -5.420    , -4.688    , -4.776    , -4.861    , -2.116    /
      DATA TBD( 82, 6),TBD( 82, 7),TBD( 82, 8),TBD( 82, 9),TBD( 82,10)
     *    / 16.341    , 12.700    ,  8.930    , 10.559    ,  7.258    /
      DATA TBD( 83, 1),TBD( 83, 2),TBD( 83, 3),TBD( 83, 4),TBD( 83, 5)
     *    / -5.742    , -4.817    , -4.756    , -5.924    , -2.353    /
      DATA TBD( 83, 6),TBD( 83, 7),TBD( 83, 8),TBD( 83, 9),TBD( 83,10)
     *    / 17.139    , 13.331    ,  9.383    , 11.042    ,  7.600    /
      DATA TBD( 84, 1),TBD( 84, 2),TBD( 84, 3),TBD( 84, 4),TBD( 84, 5)
     *    / -6.132    , -4.992    , -4.772    , -7.444    , -2.630    /
      DATA TBD( 84, 6),TBD( 84, 7),TBD( 84, 8),TBD( 84, 9),TBD( 84,10)
     *    / 17.942    , 13.969    ,  9.843    ,  9.961    ,  7.949    /
      DATA TBD( 85, 1),TBD( 85, 2),TBD( 85, 3),TBD( 85, 4),TBD( 85, 5)
     *    / -6.545    , -5.173    , -4.787    , -8.862    , -2.932    /
      DATA TBD( 85, 6),TBD( 85, 7),TBD( 85, 8),TBD( 85, 9),TBD( 85,10)
     *    / 18.775    , 14.629    , 10.317    , 10.403    ,  8.307    /
      DATA TBD( 86, 1),TBD( 86, 2),TBD( 86, 3),TBD( 86, 4),TBD( 86, 5)
     *    / -7.052    , -5.402    , -4.833    , -7.912    , -3.285    /
      DATA TBD( 86, 6),TBD( 86, 7),TBD( 86, 8),TBD( 86, 9),TBD( 86,10)
     *    / 19.615    , 15.299    , 10.803    ,  7.754    ,  8.674    /
      DATA TBD( 87, 1),TBD( 87, 2),TBD( 87, 3),TBD( 87, 4),TBD( 87, 5)
     *    / -7.614    , -5.659    , -4.898    , -7.620    , -3.702    /
      DATA TBD( 87, 6),TBD( 87, 7),TBD( 87, 8),TBD( 87, 9),TBD( 87,10)
     *    / 20.461    , 15.977    , 11.296    ,  8.105    ,  9.047    /
      DATA TBD( 88, 1),TBD( 88, 2),TBD( 88, 3),TBD( 88, 4),TBD( 88, 5)
     *    / -8.318    , -5.976    , -4.994    , -7.725    , -4.192    /
      DATA TBD( 88, 6),TBD( 88, 7),TBD( 88, 8),TBD( 88, 9),TBD( 88,10)
     *    / 21.327    , 16.668    , 11.799    ,  8.472    ,  9.428    /
      DATA TBD( 89, 1),TBD( 89, 2),TBD( 89, 3),TBD( 89, 4),TBD( 89, 5)
     *    / -9.150    , -6.313    , -5.091    , -8.127    , -4.784    /
      DATA TBD( 89, 6),TBD( 89, 7),TBD( 89, 8),TBD( 89, 9),TBD( 89,10)
     *    / 22.240    , 17.397    , 12.330    ,  8.870    ,  9.819    /
      DATA TBD( 90, 1),TBD( 90, 2),TBD( 90, 3),TBD( 90, 4),TBD( 90, 5)
     *    /-10.382    , -6.695    , -5.216    , -8.960    , -5.555    /
      DATA TBD( 90, 6),TBD( 90, 7),TBD( 90, 8),TBD( 90, 9),TBD( 90,10)
     *    / 23.161    , 18.140    , 12.868    ,  9.284    , 10.227    /
      DATA TBD( 91, 1),TBD( 91, 2),TBD( 91, 3),TBD( 91, 4),TBD( 91, 5)
     *    /-10.930    , -7.126    , -5.359    ,-10.673    , -6.735    /
      DATA TBD( 91, 6),TBD( 91, 7),TBD( 91, 8),TBD( 91, 9),TBD( 91,10)
     *    / 23.121    , 18.879    , 13.409    ,  9.654    , 10.637    /
      DATA TBD( 92, 1),TBD( 92, 2),TBD( 92, 3),TBD( 92, 4),TBD( 92, 5)
     *    /-12.152    , -7.624    , -5.529    ,-11.158    , -7.842    /
      DATA TBD( 92, 6),TBD( 92, 7),TBD( 92, 8),TBD( 92, 9),TBD( 92,10)
     *    / 24.097    , 19.642    , 13.967    ,  4.148    ,  9.570    /
      DATA TBD( 93, 1),TBD( 93, 2),TBD( 93, 3),TBD( 93, 4),TBD( 93, 5)
     *    /-12.280    , -8.187    , -5.712    , -9.725    , -8.473    /
      DATA TBD( 93, 6),TBD( 93, 7),TBD( 93, 8),TBD( 93, 9),TBD( 93,10)
     *    / 23.658    , 20.425    , 14.536    ,  4.330    ,  6.999    /
      DATA TBD( 94, 1),TBD( 94, 2),TBD( 94, 3),TBD( 94, 4),TBD( 94, 5)
     *    /-12.771    , -8.872    , -5.930    , -8.926    , -7.701    /
      DATA TBD( 94, 6),TBD( 94, 7),TBD( 94, 8),TBD( 94, 9),TBD( 94,10)
     *    / 24.607    , 21.173    , 15.087    ,  4.511    ,  7.296    /
      DATA TBD( 95, 1),TBD( 95, 2),TBD( 95, 3),TBD( 95, 4),TBD( 95, 5)
     *    /-13.513    , -9.743    , -6.176    , -8.416    , -7.388    /
      DATA TBD( 95, 6),TBD( 95, 7),TBD( 95, 8),TBD( 95, 9),TBD( 95,10)
     *    / 25.540    , 21.896    , 15.634    ,  4.697    ,  7.589    /
      DATA TBD( 96, 1),TBD( 96, 2),TBD( 96, 3),TBD( 96, 4),TBD( 96, 5)
     *    /-14.827    ,-10.539    , -6.498    , -7.990    , -7.485    /
      DATA TBD( 96, 6),TBD( 96, 7),TBD( 96, 8),TBD( 96, 9),TBD( 96,10)
     *    / 26.801    , 21.942    , 16.317    ,  4.908    ,  7.931    /
      DATA TBD( 97, 1),TBD( 97, 2),TBD( 97, 3),TBD( 97, 4),TBD( 97, 5)
     *    /-16.272    ,-11.641    , -6.798    , -7.683    , -7.638    /
      DATA TBD( 97, 6),TBD( 97, 7),TBD( 97, 8),TBD( 97, 9),TBD( 97,10)
     *    / 27.898    , 22.785    , 16.930    ,  5.107    ,  8.246    /
      data tbxc( 1),tbx(  1,1),tbx(  1,2),tbx(  1,3),tbx(  1,4)
     *   /'h   '   ,.493002   ,10.5109   ,.322912   ,26.1257   /
      data tbx(  1,5),tbx(  1,6),tbx(  1,7),tbx(  1,8),tbx(  1,9)
     *   / .140191   ,3.14236   ,.040810   ,57.7997   ,.003038   /
      data tbx(  1,10)/1./
      data tbxc(  2),tbx(  2,1),tbx(  2,2),tbx(  2,3),tbx(  2,4)
     *   /'h   '   ,.489918   ,20.6593   ,.262003   ,7.74039   /
      data tbx(  2,5),tbx(  2,6),tbx(  2,7),tbx(  2,8),tbx(  2,9)
     *   / .196767   ,49.5519   ,.049879   ,2.20159   ,.001305   /
      data tbx(  2,10)/1./
      data tbxc(  3),tbx(  3,1),tbx(  3,2),tbx(  3,3),tbx(  3,4)
     *   /'h-1 '   ,.897661   ,53.1368   ,.565616   ,15.1870   /
      data tbx(  3,5),tbx(  3,6),tbx(  3,7),tbx(  3,8),tbx(  3,9)
     *   / .415815   ,186.576   ,.116973   ,3.56709   ,.002389   /
      data tbx(  3,10)/1./
      data tbxc(  4),tbx(  4,1),tbx(  4,2),tbx(  4,3),tbx(  4,4)
     *   /'he  '   ,  .8734   , 9.1037   ,  .6309   , 3.3568   /
      data tbx(  4,5),tbx(  4,6),tbx(  4,7),tbx(  4,8),tbx(  4,9)
     *   /   .3112   ,22.9276   ,  .1780   ,  .9821   ,  .0064   /
      data tbx(  4,10)/1./
      data tbxc(  5),tbx(  5,1),tbx(  5,2),tbx(  5,3),tbx(  5,4)
     *   /'li  '   , 1.1282   , 3.9546   ,  .7508   , 1.0524   /
      data tbx(  5,5),tbx(  5,6),tbx(  5,7),tbx(  5,8),tbx(  5,9)
     *   /   .6175   ,85.3905   ,  .4653   ,168.261   ,  .0377   /
      data tbx(  5,10)/2./
      data tbxc(  6),tbx(  6,1),tbx(  6,2),tbx(  6,3),tbx(  6,4)
     *   /'li+1'   ,  .6968   , 4.6237   ,  .7888   , 1.9557   /
      data tbx(  6,5),tbx(  6,6),tbx(  6,7),tbx(  6,8),tbx(  6,9)
     *   /   .3414   ,  .6316   ,  .1563   ,10.0953   ,  .0167   /
      data tbx(  6,10)/2./
      data tbxc(  7),tbx(  7,1),tbx(  7,2),tbx(  7,3),tbx(  7,4)
     *   /'be  '   , 1.5919   ,43.6427   , 1.1278   , 1.8623   /
      data tbx(  7,5),tbx(  7,6),tbx(  7,7),tbx(  7,8),tbx(  7,9)
     *   /   .5391   ,103.483   ,  .7029   ,  .5420   ,  .0385   /
      data tbx(  7,10)/3./
      data tbxc(  8),tbx(  8,1),tbx(  8,2),tbx(  8,3),tbx(  8,4)
     *   /'be+2'   , 6.2603   ,  .0027   ,  .8849   ,  .8313   /
      data tbx(  8,5),tbx(  8,6),tbx(  8,7),tbx(  8,8),tbx(  8,9)
     *   /   .7993   , 2.2758   ,  .1647   , 5.1146   ,-6.1092   /
      data tbx(  8,10)/3./
      data tbxc(  9),tbx(  9,1),tbx(  9,2),tbx(  9,3),tbx(  9,4)
     *   /'b   '   , 2.0545   ,23.2185   , 1.3326   , 1.0210   /
      data tbx(  9,5),tbx(  9,6),tbx(  9,7),tbx(  9,8),tbx(  9,9)
     *   /  1.0979   ,60.3498   ,  .7068   ,  .1403   , -.1932   /
      data tbx(  9,10)/4./
      data tbxc( 10),tbx( 10,1),tbx( 10,2),tbx( 10,3),tbx( 10,4)
     *   /'c   '   , 2.3100   ,20.8439   , 1.0200   ,10.2075   /
      data tbx( 10,5),tbx( 10,6),tbx( 10,7),tbx( 10,8),tbx( 10,9)
     *   /  1.5886   ,  .5687   ,  .8650   ,51.6512   ,  .2156   /
      data tbx( 10,10)/5./
      data tbxc( 11),tbx( 11,1),tbx( 11,2),tbx( 11,3),tbx( 11,4)
     *   /'cval'   ,2.26069   ,22.6907   ,1.56165   ,.656665   /
      data tbx( 11,5),tbx( 11,6),tbx( 11,7),tbx( 11,8),tbx( 11,9)
     *   / 1.05075   ,9.75618   ,.839259   ,55.5949   ,.286977   /
      data tbx( 11,10)/5./
      data tbxc( 12),tbx( 12,1),tbx( 12,2),tbx( 12,3),tbx( 12,4)
     *   /'n   '   ,12.2126   ,  .0057   , 3.1322   , 9.8933   /
      data tbx( 12,5),tbx( 12,6),tbx( 12,7),tbx( 12,8),tbx( 12,9)
     *   /  2.0125   ,28.9975   , 1.1663   ,  .5826   ,-11.529   /
      data tbx( 12,10)/6./
      data tbxc( 13),tbx( 13,1),tbx( 13,2),tbx( 13,3),tbx( 13,4)
     *   /'o   '   , 3.0485   ,13.2771   , 2.2868   , 5.7011   /
      data tbx( 13,5),tbx( 13,6),tbx( 13,7),tbx( 13,8),tbx( 13,9)
     *   /  1.5463   ,  .3239   ,  .8670   ,32.9089   ,  .2508   /
      data tbx( 13,10)/7./
      data tbxc( 14),tbx( 14,1),tbx( 14,2),tbx( 14,3),tbx( 14,4)
     *   /'o-1 '   ,4.19160   ,12.8573   ,1.63969   ,4.17236   /
      data tbx( 14,5),tbx( 14,6),tbx( 14,7),tbx( 14,8),tbx( 14,9)
     *   / 1.52673   ,47.0179   ,-20.307   ,-.01404   ,21.9412   /
      data tbx( 14,10)/7./
      data tbxc(213),tbx(213,1),tbx(213,2),tbx(213,3),tbx(213,4)
     *   /'o-2 '   ,4.758   ,7.831   ,3.637   ,30.05   /
      data tbx(213,5),tbx(213,6),tbx(213,7),tbx(213,8),tbx(213,9)
     *   / 0.000   ,0.000   ,0.0000   ,0.000   ,1.594   /
      data tbx(213,10)/7./
      data tbxc( 15),tbx( 15,1),tbx( 15,2),tbx( 15,3),tbx( 15,4)
     *   /'f   '   , 3.5392   ,10.2825   , 2.6412   , 4.2944   /
      data tbx( 15,5),tbx( 15,6),tbx( 15,7),tbx( 15,8),tbx( 15,9)
     *   /  1.5170   ,  .2615   , 1.0243   ,26.1476   ,  .2776   /
      data tbx( 15,10)/8./
      data tbxc( 16),tbx( 16,1),tbx( 16,2),tbx( 16,3),tbx( 16,4)
     *   /'f-1 '   ,3.63220   ,5.27756   ,3.51057   ,14.7353   /
      data tbx( 16,5),tbx( 16,6),tbx( 16,7),tbx( 16,8),tbx( 16,9)
     *   / 1.26064   ,.442258   ,.940706   ,47.3437   ,.653396   /
      data tbx( 16,10)/8./
      data tbxc( 17),tbx( 17,1),tbx( 17,2),tbx( 17,3),tbx( 17,4)
     *   /'ne  '   , 3.9553   , 8.4042   , 3.1125   , 3.4262   /
      data tbx( 17,5),tbx( 17,6),tbx( 17,7),tbx( 17,8),tbx( 17,9)
     *   /  1.4546   ,  .2306   , 1.1251   ,21.7184   ,  .3515   /
      data tbx( 17,10)/9./
      data tbxc( 18),tbx( 18,1),tbx( 18,2),tbx( 18,3),tbx( 18,4)
     *   /'na  '   , 4.7626   , 3.2850   , 3.1736   , 8.8422   /
      data tbx( 18,5),tbx( 18,6),tbx( 18,7),tbx( 18,8),tbx( 18,9)
     *   /  1.2674   ,  .3136   , 1.1128   ,129.424   ,  .6760   /
      data tbx( 18,10)/10./
      data tbxc( 19),tbx( 19,1),tbx( 19,2),tbx( 19,3),tbx( 19,4)
     *   /'na+1'   , 3.2565   , 2.6671   , 3.9362   , 6.1153   /
      data tbx( 19,5),tbx( 19,6),tbx( 19,7),tbx( 19,8),tbx( 19,9)
     *   /  1.3998   ,  .2001   , 1.0032   ,14.0390   ,  .4040   /
      data tbx( 19,10)/10./
      data tbxc( 20),tbx( 20,1),tbx( 20,2),tbx( 20,3),tbx( 20,4)
     *   /'mg  '   , 5.4204   , 2.8275   , 2.1735   ,79.2611   /
      data tbx( 20,5),tbx( 20,6),tbx( 20,7),tbx( 20,8),tbx( 20,9)
     *   /  1.2269   ,  .3808   , 2.3073   , 7.1937   ,  .8584   /
      data tbx( 20,10)/11./
      data tbxc( 21),tbx( 21,1),tbx( 21,2),tbx( 21,3),tbx( 21,4)
     *   /'mg+2'   , 3.4988   , 2.1676   , 3.8378   , 4.7542   /
      data tbx( 21,5),tbx( 21,6),tbx( 21,7),tbx( 21,8),tbx( 21,9)
     *   /  1.3284   ,  .1850   ,  .8497   ,10.1411   ,  .4853   /
      data tbx( 21,10)/11./
      data tbxc( 22),tbx( 22,1),tbx( 22,2),tbx( 22,3),tbx( 22,4)
     *   /'al  '   , 6.4202   , 3.0387   , 1.9002   ,  .7426   /
      data tbx( 22,5),tbx( 22,6),tbx( 22,7),tbx( 22,8),tbx( 22,9)
     *   /  1.5936   ,31.5472   , 1.9646   ,85.0886   , 1.1151   /
      data tbx( 22,10)/12./
      data tbxc( 23),tbx( 23,1),tbx( 23,2),tbx( 23,3),tbx( 23,4)
     *   /'al+3'   ,4.17448   ,1.93816   ,3.38760   ,4.14553   /
      data tbx( 23,5),tbx( 23,6),tbx( 23,7),tbx( 23,8),tbx( 23,9)
     *   / 1.20296   ,.228753   ,.528137   ,8.28524   ,.706786   /
      data tbx( 23,10)/12./
      data tbxc( 24),tbx( 24,1),tbx( 24,2),tbx( 24,3),tbx( 24,4)
     *   /'si  '   , 6.2915   , 2.4386   , 3.0353   ,32.3337   /
      data tbx( 24,5),tbx( 24,6),tbx( 24,7),tbx( 24,8),tbx( 24,9)
     *   /  1.9891   ,  .6785   , 1.5410   ,81.6937   , 1.1407   /
      data tbx( 24,10)/13./
      data tbxc( 25),tbx( 25,1),tbx( 25,2),tbx( 25,3),tbx( 25,4)
     *   /'siv '   ,5.66269   ,2.66520   ,3.07164   ,38.6634   /
      data tbx( 25,5),tbx( 25,6),tbx( 25,7),tbx( 25,8),tbx( 25,9)
     *   / 2.62446   ,.916946   ,1.39320   ,93.5458   ,1.24707   /
      data tbx( 25,10)/13./
      data tbxc( 26),tbx( 26,1),tbx( 26,2),tbx( 26,3),tbx( 26,4)
     *   /'si+4'   ,4.43918   ,1.64167   ,3.20345   ,3.43757   /
      data tbx( 26,5),tbx( 26,6),tbx( 26,7),tbx( 26,8),tbx( 26,9)
     *   / 1.19453   ,.214900   ,.416530   ,6.65365   ,.746297   /
      data tbx( 26,10)/13./
      data tbxc( 27),tbx( 27,1),tbx( 27,2),tbx( 27,3),tbx( 27,4)
     *   /'s   '   , 6.9053   , 1.4679   , 5.2034   ,22.2151   /
      data tbx( 27,5),tbx( 27,6),tbx( 27,7),tbx( 27,8),tbx( 27,9)
     *   /  1.4379   ,  .2536   , 1.5863   ,56.1720   ,  .8669   /
      data tbx( 27,10)/15./
      data tbxc( 28),tbx( 28,1),tbx( 28,2),tbx( 28,3),tbx( 28,4)
     *   /'p   '   , 6.4345   , 1.9067   , 4.1791   ,27.1570   /
      data tbx( 28,5),tbx( 28,6),tbx( 28,7),tbx( 28,8),tbx( 28,9)
     *   /  1.7800   ,  .5260   , 1.4908   ,68.1645   , 1.1149   /
      data tbx( 28,10)/14./
      data tbxc( 29),tbx( 29,1),tbx( 29,2),tbx( 29,3),tbx( 29,4)
     *   /'cl  '   ,11.4604   ,  .0104   , 7.1964   , 1.1662   /
      data tbx( 29,5),tbx( 29,6),tbx( 29,7),tbx( 29,8),tbx( 29,9)
     *   /  6.2556   ,18.5194   , 1.6455   ,47.7784   ,-9.5574   /
      data tbx( 29,10)/16./
      data tbxc( 30),tbx( 30,1),tbx( 30,2),tbx( 30,3),tbx( 30,4)
     *   /'cl-1'   ,18.2915   ,  .0066   , 7.2084   , 1.1717   /
      data tbx( 30,5),tbx( 30,6),tbx( 30,7),tbx( 30,8),tbx( 30,9)
     *   /  6.5337   ,19.5424   , 2.3386   ,60.4486   ,-16.378   /
      data tbx( 30,10)/16./
      data tbxc( 31),tbx( 31,1),tbx( 31,2),tbx( 31,3),tbx( 31,4)
     *   /'ar  '   , 7.4845   ,  .9072   , 6.7723   ,14.8407   /
      data tbx( 31,5),tbx( 31,6),tbx( 31,7),tbx( 31,8),tbx( 31,9)
     *   /   .6539   ,43.8983   , 1.6442   ,33.3929   , 1.4445   /
      data tbx( 31,10)/17./
      data tbxc( 32),tbx( 32,1),tbx( 32,2),tbx( 32,3),tbx( 32,4)
     *   /'k   '   , 8.2186   ,12.7949   , 7.4398   ,  .7748   /
      data tbx( 32,5),tbx( 32,6),tbx( 32,7),tbx( 32,8),tbx( 32,9)
     *   /  1.0519   ,213.187   ,  .8659   ,41.6841   , 1.4228   /
      data tbx( 32,10)/18./
      data tbxc( 33),tbx( 33,1),tbx( 33,2),tbx( 33,3),tbx( 33,4)
     *   /'k+1 '   , 7.9578   ,12.6331   , 7.4917   ,  .7674   /
      data tbx( 33,5),tbx( 33,6),tbx( 33,7),tbx( 33,8),tbx( 33,9)
     *   /  6.3590   , -.0020   , 1.1915   ,31.9128   ,-4.9978   /
      data tbx( 33,10)/18./
      data tbxc( 34),tbx( 34,1),tbx( 34,2),tbx( 34,3),tbx( 34,4)
     *   /'ca  '   , 8.6266   ,10.4421   , 7.3873   ,  .6599   /
      data tbx( 34,5),tbx( 34,6),tbx( 34,7),tbx( 34,8),tbx( 34,9)
     *   /  1.5899   ,85.7484   , 1.0211   ,178.437   , 1.3751   /
      data tbx( 34,10)/19./
      data tbxc( 35),tbx( 35,1),tbx( 35,2),tbx( 35,3),tbx( 35,4)
     *   /'ca+2'   ,15.6348   , -.0074   , 7.9518   ,  .6089   /
      data tbx( 35,5),tbx( 35,6),tbx( 35,7),tbx( 35,8),tbx( 35,9)
     *   /  8.4372   ,10.3116   ,  .8537   ,25.9905   ,-14.875   /
      data tbx( 35,10)/19./
      data tbxc( 36),tbx( 36,1),tbx( 36,2),tbx( 36,3),tbx( 36,4)
     *   /'sc  '   , 9.1890   , 9.0213   , 7.3679   ,  .5729   /
      data tbx( 36,5),tbx( 36,6),tbx( 36,7),tbx( 36,8),tbx( 36,9)
     *   /  1.6409   ,136.108   , 1.4680   ,51.3531   , 1.3329   /
      data tbx( 36,10)/20./
      data tbxc( 37),tbx( 37,1),tbx( 37,2),tbx( 37,3),tbx( 37,4)
     *   /'sc+3'   ,13.4008   ,.298540   ,8.02730   ,7.96290   /
      data tbx( 37,5),tbx( 37,6),tbx( 37,7),tbx( 37,8),tbx( 37,9)
     *   / 1.65943   ,-.28604   ,1.57936   ,16.0662   ,-6.6667   /
      data tbx( 37,10)/20./
      data tbxc( 38),tbx( 38,1),tbx( 38,2),tbx( 38,3),tbx( 38,4)
     *   /'ti  '   , 9.7595   , 7.8508   , 7.3558   ,  .5000   /
      data tbx( 38,5),tbx( 38,6),tbx( 38,7),tbx( 38,8),tbx( 38,9)
     *   /  1.6991   ,35.6338   , 1.9021   ,116.105   , 1.2807   /
      data tbx( 38,10)/21./
      data tbxc( 39),tbx( 39,1),tbx( 39,2),tbx( 39,3),tbx( 39,4)
     *   /'ti+2'   ,9.11423   ,7.52430   ,7.62174   ,.457585   /
      data tbx( 39,5),tbx( 39,6),tbx( 39,7),tbx( 39,8),tbx( 39,9)
     *   / 2.27930   ,19.5361   ,.087899   ,61.6558   ,.897155   /
      data tbx( 39,10)/21./
      data tbxc( 40),tbx( 40,1),tbx( 40,2),tbx( 40,3),tbx( 40,4)
     *   /'ti+3'   ,17.7344   ,.220610   ,8.73816   ,7.04716   /
      data tbx( 40,5),tbx( 40,6),tbx( 40,7),tbx( 40,8),tbx( 40,9)
     *   / 5.25691   ,-.15762   ,1.92134   ,15.9768   ,-14.652   /
      data tbx( 40,10)/21./
      data tbxc( 41),tbx( 41,1),tbx( 41,2),tbx( 41,3),tbx( 41,4)
     *   /'ti+4'   ,19.5114   ,.178847   ,8.23473   ,6.67018   /
      data tbx( 41,5),tbx( 41,6),tbx( 41,7),tbx( 41,8),tbx( 41,9)
     *   / 2.01341   ,-.29263   ,1.52080   ,12.9464   ,-13.280   /
      data tbx( 41,10)/21./
      data tbxc( 42),tbx( 42,1),tbx( 42,2),tbx( 42,3),tbx( 42,4)
     *   /'v   '   ,10.2971   , 6.8657   , 7.3511   ,  .4385   /
      data tbx( 42,5),tbx( 42,6),tbx( 42,7),tbx( 42,8),tbx( 42,9)
     *   /  2.0703   ,26.8938   , 2.0571   ,102.478   , 1.2199   /
      data tbx( 42,10)/22./
      data tbxc( 43),tbx( 43,1),tbx( 43,2),tbx( 43,3),tbx( 43,4)
     *   /'v+2 '   ,10.1060   , 6.8818   , 7.3541   ,  .4409   /
      data tbx( 43,5),tbx( 43,6),tbx( 43,7),tbx( 43,8),tbx( 43,9)
     *   /  2.2884   ,20.3004   ,  .0223   ,115.122   , 1.2298   /
      data tbx( 43,10)/22./
      data tbxc( 44),tbx( 44,1),tbx( 44,2),tbx( 44,3),tbx( 44,4)
     *   /'v+3 '   ,9.43141   ,6.39535   ,7.74190   ,.383349   /
      data tbx( 44,5),tbx( 44,6),tbx( 44,7),tbx( 44,8),tbx( 44,9)
     *   / 2.15343   ,15.1908   ,.016865   ,63.9690   ,.656565   /
      data tbx( 44,10)/22./
      data tbxc( 45),tbx( 45,1),tbx( 45,2),tbx( 45,3),tbx( 45,4)
     *   /'v+5 '   ,15.6887   ,.679003   ,8.14208   ,5.40135   /
      data tbx( 45,5),tbx( 45,6),tbx( 45,7),tbx( 45,8),tbx( 45,9)
     *   / 2.03081   ,9.97278   ,-9.5760   ,.940464   ,1.71430   /
      data tbx( 45,10)/22./
      data tbxc( 46),tbx( 46,1),tbx( 46,2),tbx( 46,3),tbx( 46,4)
     *   /'cr  '   ,10.6406   , 6.1038   , 7.3537   ,  .3920   /
      data tbx( 46,5),tbx( 46,6),tbx( 46,7),tbx( 46,8),tbx( 46,9)
     *   /  3.3240   ,20.2626   , 1.4922   ,98.7399   , 1.1832   /
      data tbx( 46,10)/23./
      data tbxc( 47),tbx( 47,1),tbx( 47,2),tbx( 47,3),tbx( 47,4)
     *   /'cr+2'   ,9.54034   ,5.66078   ,7.75090   ,.344261   /
      data tbx( 47,5),tbx( 47,6),tbx( 47,7),tbx( 47,8),tbx( 47,9)
     *   / 3.58274   ,13.3075   ,.509107   ,32.4224   ,.616898   /
      data tbx( 47,10)/23./
      data tbxc( 48),tbx( 48,1),tbx( 48,2),tbx( 48,3),tbx( 48,4)
     *   /'cr+3'   ,9.68090   ,5.59463   ,7.81136   ,.334393   /
      data tbx( 48,5),tbx( 48,6),tbx( 48,7),tbx( 48,8),tbx( 48,9)
     *   / 2.87603   ,12.8288   ,.113575   ,32.8761   ,.518275   /
      data tbx( 48,10)/23./
      data tbxc( 49),tbx( 49,1),tbx( 49,2),tbx( 49,3),tbx( 49,4)
     *   /'mn  '   ,11.2819   , 5.3409   , 7.3573   ,  .3432   /
      data tbx( 49,5),tbx( 49,6),tbx( 49,7),tbx( 49,8),tbx( 49,9)
     *   /  3.0193   ,17.8674   , 2.2441   ,83.7543   , 1.0896   /
      data tbx( 49,10)/24./
      data tbxc( 50),tbx( 50,1),tbx( 50,2),tbx( 50,3),tbx( 50,4)
     *   /'mn+2'   ,10.8061   , 5.2796   , 7.3620   ,  .3435   /
      data tbx( 50,5),tbx( 50,6),tbx( 50,7),tbx( 50,8),tbx( 50,9)
     *   /  3.5268   ,14.3430   ,  .2184   ,41.3235   , 1.0874   /
      data tbx( 50,10)/24./
      data tbxc( 51),tbx( 51,1),tbx( 51,2),tbx( 51,3),tbx( 51,4)
     *   /'mn+3'   ,9.84521   ,4.91797   ,7.87194   ,.294393   /
      data tbx( 51,5),tbx( 51,6),tbx( 51,7),tbx( 51,8),tbx( 51,9)
     *   / 3.56531   ,10.8171   ,.323613   ,24.1281   ,.393974   /
      data tbx( 51,10)/24./
      data tbxc( 52),tbx( 52,1),tbx( 52,2),tbx( 52,3),tbx( 52,4)
     *   /'mn+4'   ,9.96253   ,4.84850   ,7.97057   ,.283303   /
      data tbx( 52,5),tbx( 52,6),tbx( 52,7),tbx( 52,8),tbx( 52,9)
     *   / 2.76067   ,10.4852   ,.054447   ,27.5730   ,.251877   /
      data tbx( 52,10)/24./
      data tbxc( 53),tbx( 53,1),tbx( 53,2),tbx( 53,3),tbx( 53,4)
     *   /'fe  '   ,11.7695   , 4.7611   , 7.3573   ,  .3072   /
      data tbx( 53,5),tbx( 53,6),tbx( 53,7),tbx( 53,8),tbx( 53,9)
     *   /  3.5222   ,15.3535   , 2.3045   ,76.8805   , 1.0369   /
      data tbx( 53,10)/25./
      data tbxc( 54),tbx( 54,1),tbx( 54,2),tbx( 54,3),tbx( 54,4)
     *   /'fe+2'   ,11.0424   , 4.6538   , 7.3740   ,  .3053   /
      data tbx( 54,5),tbx( 54,6),tbx( 54,7),tbx( 54,8),tbx( 54,9)
     *   /  4.1346   ,12.0546   ,  .4399   ,31.2809   , 1.0097   /
      data tbx( 54,10)/25./
      data tbxc( 55),tbx( 55,1),tbx( 55,2),tbx( 55,3),tbx( 55,4)
     *   /'fe+3'   ,11.1764   , 4.6147   , 7.3863   ,  .3005   /
      data tbx( 55,5),tbx( 55,6),tbx( 55,7),tbx( 55,8),tbx( 55,9)
     *   /  3.3948   ,11.6729   ,  .0724   ,38.5566   ,  .9707   /
      data tbx( 55,10)/25./
      data tbxc( 56),tbx( 56,1),tbx( 56,2),tbx( 56,3),tbx( 56,4)
     *   /'co  '   ,12.2841   , 4.2791   , 7.3409   ,  .2784   /
      data tbx( 56,5),tbx( 56,6),tbx( 56,7),tbx( 56,8),tbx( 56,9)
     *   /  4.0034   ,13.5359   , 2.3488   ,71.1692   , 1.0118   /
      data tbx( 56,10)/26./
      data tbxc( 57),tbx( 57,1),tbx( 57,2),tbx( 57,3),tbx( 57,4)
     *   /'co+2'   ,11.2296   , 4.1231   , 7.3883   ,  .2726   /
      data tbx( 57,5),tbx( 57,6),tbx( 57,7),tbx( 57,8),tbx( 57,9)
     *   /  4.7393   ,10.2443   ,  .7108   ,25.6466   ,  .9324   /
      data tbx( 57,10)/26./
      data tbxc( 58),tbx( 58,1),tbx( 58,2),tbx( 58,3),tbx( 58,4)
     *   /'co+3'   ,10.3380   ,3.90969   ,7.88173   ,.238668   /
      data tbx( 58,5),tbx( 58,6),tbx( 58,7),tbx( 58,8),tbx( 58,9)
     *   / 4.76795   ,8.35583   ,.725591   ,18.3491   ,.286667   /
      data tbx( 58,10)/26./
      data tbxc( 59),tbx( 59,1),tbx( 59,2),tbx( 59,3),tbx( 59,4)
     *   /'ni  '   ,12.8376   , 3.8785   , 7.2920   ,  .2565   /
      data tbx( 59,5),tbx( 59,6),tbx( 59,7),tbx( 59,8),tbx( 59,9)
     *   /  4.4438   ,12.1763   , 2.3800   ,66.3421   , 1.0341   /
      data tbx( 59,10)/27./
      data tbxc( 60),tbx( 60,1),tbx( 60,2),tbx( 60,3),tbx( 60,4)
     *   /'ni+2'   ,11.4166   , 3.6766   , 7.4005   ,  .2449   /
      data tbx( 60,5),tbx( 60,6),tbx( 60,7),tbx( 60,8),tbx( 60,9)
     *   /  5.3442   , 8.8730   ,  .9773   ,22.1626   ,  .8614   /
      data tbx( 60,10)/27./
      data tbxc( 61),tbx( 61,1),tbx( 61,2),tbx( 61,3),tbx( 61,4)
     *   /'ni+3'   ,10.7806   ,3.54770   ,7.75868   ,.223140   /
      data tbx( 61,5),tbx( 61,6),tbx( 61,7),tbx( 61,8),tbx( 61,9)
     *   / 5.22746   ,7.64468   ,.847114   ,16.9673   ,.386044   /
      data tbx( 61,10)/27./
      data tbxc( 62),tbx( 62,1),tbx( 62,2),tbx( 62,3),tbx( 62,4)
     *   /'cu  '   ,13.3380   , 3.5828   , 7.1676   ,  .2470   /
      data tbx( 62,5),tbx( 62,6),tbx( 62,7),tbx( 62,8),tbx( 62,9)
     *   /  5.6158   ,11.3966   , 1.6735   ,64.8126   , 1.1910   /
      data tbx( 62,10)/28./
      data tbxc( 63),tbx( 63,1),tbx( 63,2),tbx( 63,3),tbx( 63,4)
     *   /'cu+1'   ,11.9475   , 3.3669   , 7.3573   ,  .2274   /
      data tbx( 63,5),tbx( 63,6),tbx( 63,7),tbx( 63,8),tbx( 63,9)
     *   /  6.2455   , 8.6625   , 1.5578   ,25.8487   ,  .8900   /
      data tbx( 63,10)/28./
      data tbxc( 64),tbx( 64,1),tbx( 64,2),tbx( 64,3),tbx( 64,4)
     *   /'cu+2'   ,11.8168   ,3.37484   ,7.11181   ,.244078   /
      data tbx( 64,5),tbx( 64,6),tbx( 64,7),tbx( 64,8),tbx( 64,9)
     *   / 5.78135   ,7.98760   ,1.14523   ,19.8970   ,1.14431   /
      data tbx( 64,10)/28./
      data tbxc( 65),tbx( 65,1),tbx( 65,2),tbx( 65,3),tbx( 65,4)
     *   /'zn  '   ,14.0743   , 3.2655   , 7.0318   ,  .2333   /
      data tbx( 65,5),tbx( 65,6),tbx( 65,7),tbx( 65,8),tbx( 65,9)
     *   /  5.1652   ,10.3163   , 2.4100   ,58.7097   , 1.3041   /
      data tbx( 65,10)/29./
      data tbxc( 66),tbx( 66,1),tbx( 66,2),tbx( 66,3),tbx( 66,4)
     *   /'zn+2'   ,11.9719   , 2.9946   , 7.3862   ,  .2031   /
      data tbx( 66,5),tbx( 66,6),tbx( 66,7),tbx( 66,8),tbx( 66,9)
     *   /  6.4668   , 7.0826   , 1.3940   ,18.0995   ,  .7807   /
      data tbx( 66,10)/29./
      data tbxc( 67),tbx( 67,1),tbx( 67,2),tbx( 67,3),tbx( 67,4)
     *   /'ga  '   ,15.2354   , 3.0669   , 6.7006   ,  .2412   /
      data tbx( 67,5),tbx( 67,6),tbx( 67,7),tbx( 67,8),tbx( 67,9)
     *   /  4.3591   ,10.7805   , 2.9623   ,61.4135   , 1.7189   /
      data tbx( 67,10)/30./
      data tbxc( 68),tbx( 68,1),tbx( 68,2),tbx( 68,3),tbx( 68,4)
     *   /'ga+3'   ,12.6920   ,2.81262   ,6.69883   ,.227890   /
      data tbx( 68,5),tbx( 68,6),tbx( 68,7),tbx( 68,8),tbx( 68,9)
     *   / 6.06692   ,6.36441   ,1.00660   ,14.4122   ,1.53545   /
      data tbx( 68,10)/30./
      data tbxc( 69),tbx( 69,1),tbx( 69,2),tbx( 69,3),tbx( 69,4)
     *   /'ge  '   ,16.0816   , 2.8509   , 6.3747   ,  .2516   /
      data tbx( 69,5),tbx( 69,6),tbx( 69,7),tbx( 69,8),tbx( 69,9)
     *   /  3.7068   ,11.4468   , 3.6830   ,54.7625   , 2.1313   /
      data tbx( 69,10)/31./
      data tbxc( 70),tbx( 70,1),tbx( 70,2),tbx( 70,3),tbx( 70,4)
     *   /'ge+4'   ,12.9172   ,2.53718   ,6.70003   ,.205855   /
      data tbx( 70,5),tbx( 70,6),tbx( 70,7),tbx( 70,8),tbx( 70,9)
     *   / 6.06791   ,5.47913   ,.859041   ,11.6030   ,1.45572   /
      data tbx( 70,10)/31./
      data tbxc( 71),tbx( 71,1),tbx( 71,2),tbx( 71,3),tbx( 71,4)
     *   /'as  '   ,16.6723   , 2.6345   , 6.0701   ,  .2647   /
      data tbx( 71,5),tbx( 71,6),tbx( 71,7),tbx( 71,8),tbx( 71,9)
     *   /  3.4313   ,12.9479   , 4.2779   ,47.7972   , 2.5310   /
      data tbx( 71,10)/32./
      data tbxc( 72),tbx( 72,1),tbx( 72,2),tbx( 72,3),tbx( 72,4)
     *   /'se  '   ,17.0006   , 2.4098   , 5.8196   ,  .2726   /
      data tbx( 72,5),tbx( 72,6),tbx( 72,7),tbx( 72,8),tbx( 72,9)
     *   /  3.9731   ,15.2372   , 4.3543   ,43.8163   , 2.8409   /
      data tbx( 72,10)/33./
      data tbxc( 73),tbx( 73,1),tbx( 73,2),tbx( 73,3),tbx( 73,4)
     *   /'br  '   ,17.1789   , 2.1723   , 5.2358   ,16.5796   /
      data tbx( 73,5),tbx( 73,6),tbx( 73,7),tbx( 73,8),tbx( 73,9)
     *   /  5.6377   ,  .2609   , 3.9851   ,41.4328   , 2.9557   /
      data tbx( 73,10)/34./
      data tbxc( 74),tbx( 74,1),tbx( 74,2),tbx( 74,3),tbx( 74,4)
     *   /'br-1'   ,17.1718   , 2.2059   , 6.3338   ,19.3345   /
      data tbx( 74,5),tbx( 74,6),tbx( 74,7),tbx( 74,8),tbx( 74,9)
     *   /  5.5754   ,  .2871   , 3.7272   ,58.1535   , 3.1776   /
      data tbx( 74,10)/34./
      data tbxc( 75),tbx( 75,1),tbx( 75,2),tbx( 75,3),tbx( 75,4)
     *   /'kr  '   ,17.3555   , 1.9384   , 6.7286   ,16.5623   /
      data tbx( 75,5),tbx( 75,6),tbx( 75,7),tbx( 75,8),tbx( 75,9)
     *   /  5.5493   ,  .2261   , 3.5375   ,39.3972   , 2.8250   /
      data tbx( 75,10)/35./
      data tbxc( 76),tbx( 76,1),tbx( 76,2),tbx( 76,3),tbx( 76,4)
     *   /'rb  '   ,17.1784   , 1.7888   , 9.6435   ,17.3151   /
      data tbx( 76,5),tbx( 76,6),tbx( 76,7),tbx( 76,8),tbx( 76,9)
     *   /  5.1399   ,  .2748   , 1.5292   ,164.934   , 3.4873   /
      data tbx( 76,10)/36./
      data tbxc( 77),tbx( 77,1),tbx( 77,2),tbx( 77,3),tbx( 77,4)
     *   /'rb+1'   ,17.5816   , 1.7139   , 7.6598   ,14.7957   /
      data tbx( 77,5),tbx( 77,6),tbx( 77,7),tbx( 77,8),tbx( 77,9)
     *   /  5.8981   ,  .1603   , 2.7817   ,31.2087   , 2.0782   /
      data tbx( 77,10)/36./
      data tbxc( 78),tbx( 78,1),tbx( 78,2),tbx( 78,3),tbx( 78,4)
     *   /'sr  '   ,17.5663   , 1.5564   , 9.8184   ,14.0988   /
      data tbx( 78,5),tbx( 78,6),tbx( 78,7),tbx( 78,8),tbx( 78,9)
     *   /  5.4220   ,  .1664   , 2.6694   ,132.376   , 2.5064   /
      data tbx( 78,10)/37./
      data tbxc( 79),tbx( 79,1),tbx( 79,2),tbx( 79,3),tbx( 79,4)
     *   /'sr+2'   ,18.0874   , 1.4907   , 8.1373   ,12.6963   /
      data tbx( 79,5),tbx( 79,6),tbx( 79,7),tbx( 79,8),tbx( 79,9)
     *   /  2.5654   ,24.5651   ,-34.193   , -.0138   ,41.4025   /
      data tbx( 79,10)/37./
      data tbxc( 80),tbx( 80,1),tbx( 80,2),tbx( 80,3),tbx( 80,4)
     *   /'y   '   ,17.7760   ,1.40290   ,10.2946   ,12.8006   /
      data tbx( 80,5),tbx( 80,6),tbx( 80,7),tbx( 80,8),tbx( 80,9)
     *   / 5.72629   ,.125599   ,3.26588   ,104.354   ,1.91213   /
      data tbx( 80,10)/38./
      data tbxc( 81),tbx( 81,1),tbx( 81,2),tbx( 81,3),tbx( 81,4)
     *   /'y+3 '   ,17.9268   ,1.35417   ,9.15310   ,11.2145   /
      data tbx( 81,5),tbx( 81,6),tbx( 81,7),tbx( 81,8),tbx( 81,9)
     *   / 1.76795   ,22.6599   ,-33.108   ,-.01319   ,40.2602   /
      data tbx( 81,10)/38./
      data tbxc( 82),tbx( 82,1),tbx( 82,2),tbx( 82,3),tbx( 82,4)
     *   /'zr  '   ,17.8765   ,1.27618   ,10.9480   ,11.9160   /
      data tbx( 82,5),tbx( 82,6),tbx( 82,7),tbx( 82,8),tbx( 82,9)
     *   / 5.41732   ,.117622   ,3.65721   ,87.6627   ,2.06929   /
      data tbx( 82,10)/39./
      data tbxc( 83),tbx( 83,1),tbx( 83,2),tbx( 83,3),tbx( 83,4)
     *   /'zr+4'   ,18.1668   ,1.21480   ,10.0562   ,10.1483   /
      data tbx( 83,5),tbx( 83,6),tbx( 83,7),tbx( 83,8),tbx( 83,9)
     *   / 1.01118   ,21.6054   ,-2.6479   ,-.10276   ,9.41454   /
      data tbx( 83,10)/39./
      data tbxc( 84),tbx( 84,1),tbx( 84,2),tbx( 84,3),tbx( 84,4)
     *   /'nb  '   ,17.6142   ,1.18865   ,12.0144   ,11.7660   /
      data tbx( 84,5),tbx( 84,6),tbx( 84,7),tbx( 84,8),tbx( 84,9)
     *   / 4.04183   ,.204785   ,3.53346   ,69.7957   ,3.75591   /
      data tbx( 84,10)/40./
      data tbxc( 85),tbx( 85,1),tbx( 85,2),tbx( 85,3),tbx( 85,4)
     *   /'nb+3'   ,19.8812   ,.019175   ,18.0653   ,1.13305   /
      data tbx( 85,5),tbx( 85,6),tbx( 85,7),tbx( 85,8),tbx( 85,9)
     *   / 11.0177   ,10.1621   ,1.94715   ,28.3389   ,-12.912   /
      data tbx( 85,10)/40./
      data tbxc( 86),tbx( 86,1),tbx( 86,2),tbx( 86,3),tbx( 86,4)
     *   /'nb+5'   ,17.9163   ,1.12446   ,13.3417   ,.028781   /
      data tbx( 86,5),tbx( 86,6),tbx( 86,7),tbx( 86,8),tbx( 86,9)
     *   / 10.7990   ,9.28206   ,.337905   ,25.7228   ,-6.3934   /
      data tbx( 86,10)/40./
      data tbxc( 87),tbx( 87,1),tbx( 87,2),tbx( 87,3),tbx( 87,4)
     *   /'mo  '   , 3.7025   ,  .2772   ,17.2356   , 1.0958   /
      data tbx( 87,5),tbx( 87,6),tbx( 87,7),tbx( 87,8),tbx( 87,9)
     *   / 12.8876   ,11.0040   , 3.7429   ,61.6584   , 4.3875   /
      data tbx( 87,10)/41./
      data tbxc( 88),tbx( 88,1),tbx( 88,2),tbx( 88,3),tbx( 88,4)
     *   /'mo+3'   ,21.1664   ,.014734   ,18.2017   ,1.03031   /
      data tbx( 88,5),tbx( 88,6),tbx( 88,7),tbx( 88,8),tbx( 88,9)
     *   / 11.7423   ,9.53659   ,2.30951   ,26.6307   ,-14.421   /
      data tbx( 88,10)/41./
      data tbxc( 89),tbx( 89,1),tbx( 89,2),tbx( 89,3),tbx( 89,4)
     *   /'mo+5'   ,21.0149   ,.014345   ,18.0992   ,1.02238   /
      data tbx( 89,5),tbx( 89,6),tbx( 89,7),tbx( 89,8),tbx( 89,9)
     *   / 11.4632   ,8.78809   ,.740625   ,23.3452   ,-14.316   /
      data tbx( 89,10)/41./
      data tbxc( 90),tbx( 90,1),tbx( 90,2),tbx( 90,3),tbx( 90,4)
     *   /'mo+6'   ,17.8871   ,1.03649   ,11.1750   ,8.48061   /
      data tbx( 90,5),tbx( 90,6),tbx( 90,7),tbx( 90,8),tbx( 90,9)
     *   / 6.57891   ,.058881   ,.0000   ,.0000    ,.344941   /
      data tbx( 90,10)/41./
      data tbxc( 91),tbx( 91,1),tbx( 91,2),tbx( 91,3),tbx( 91,4)
     *   /'tc  '   ,19.1301   ,.864132   ,11.0948   ,8.14487   /
      data tbx( 91,5),tbx( 91,6),tbx( 91,7),tbx( 91,8),tbx( 91,9)
     *   / 4.64901   ,21.5707   ,2.71263   ,86.8472   ,5.40428   /
      data tbx( 91,10)/42./
      data tbxc( 92),tbx( 92,1),tbx( 92,2),tbx( 92,3),tbx( 92,4)
     *   /'ru  '   ,19.2674   ,.808520   ,12.9182   ,8.43467   /
      data tbx( 92,5),tbx( 92,6),tbx( 92,7),tbx( 92,8),tbx( 92,9)
     *   / 4.86337   ,24.7997   ,1.56756   ,94.2928   ,5.37874   /
      data tbx( 92,10)/43./
      data tbxc( 93),tbx( 93,1),tbx( 93,2),tbx( 93,3),tbx( 93,4)
     *   /'ru+3'   ,18.5638   ,.847329   ,13.2885   ,8.37164   /
      data tbx( 93,5),tbx( 93,6),tbx( 93,7),tbx( 93,8),tbx( 93,9)
     *   / 9.32602   ,.017662   ,3.00964   ,22.8870   ,-3.1892   /
      data tbx( 93,10)/43./
      data tbxc( 94),tbx( 94,1),tbx( 94,2),tbx( 94,3),tbx( 94,4)
     *   /'ru+4'   ,18.5003   ,.844582   ,13.1787   ,8.12534   /
      data tbx( 94,5),tbx( 94,6),tbx( 94,7),tbx( 94,8),tbx( 94,9)
     *   / 4.71304   ,.036495   ,2.18535   ,20.8504   ,1.42357   /
      data tbx( 94,10)/43./
      data tbxc( 95),tbx( 95,1),tbx( 95,2),tbx( 95,3),tbx( 95,4)
     *   /'rh  '   ,19.2957   ,.751536   ,14.3501   ,8.21758   /
      data tbx( 95,5),tbx( 95,6),tbx( 95,7),tbx( 95,8),tbx( 95,9)
     *   / 4.73425   ,25.8749   ,1.28918   ,98.6062   ,5.32800   /
      data tbx( 95,10)/44./
      data tbxc( 96),tbx( 96,1),tbx( 96,2),tbx( 96,3),tbx( 96,4)
     *   /'rh+3'   ,18.8785   ,.764252   ,14.1259   ,7.84438   /
      data tbx( 96,5),tbx( 96,6),tbx( 96,7),tbx( 96,8),tbx( 96,9)
     *   / 3.32515   ,21.2487   ,-6.1989   ,-.01036   ,11.8678   /
      data tbx( 96,10)/44./
      data tbxc( 97),tbx( 97,1),tbx( 97,2),tbx( 97,3),tbx( 97,4)
     *   /'rh+4'   ,18.8545   ,.760825   ,13.9806   ,7.62436   /
      data tbx( 97,5),tbx( 97,6),tbx( 97,7),tbx( 97,8),tbx( 97,9)
     *   / 2.53464   ,19.3317   ,-5.6526   ,-.01020   ,11.2835   /
      data tbx( 97,10)/44./
      data tbxc( 98),tbx( 98,1),tbx( 98,2),tbx( 98,3),tbx( 98,4)
     *   /'pd  '   ,19.3319   ,.698655   ,15.5017   ,7.98929   /
      data tbx( 98,5),tbx( 98,6),tbx( 98,7),tbx( 98,8),tbx( 98,9)
     *   / 5.29537   ,25.2052   ,.605844   ,76.8986   ,5.26593   /
      data tbx( 98,10)/45./
      data tbxc( 99),tbx( 99,1),tbx( 99,2),tbx( 99,3),tbx( 99,4)
     *   /'pd+2'   ,19.1701   ,.696219   ,15.2096   ,7.55573   /
      data tbx( 99,5),tbx( 99,6),tbx( 99,7),tbx( 99,8),tbx( 99,9)
     *   / 4.32234   ,22.5057   ,.0000   ,.0000   ,5.29160   /
      data tbx( 99,10)/45./
      data tbxc(100),tbx(100,1),tbx(100,2),tbx(100,3),tbx(100,4)
     *   /'pd+4'   ,19.2493   ,.683839   ,14.7900   ,7.14833   /
      data tbx(100,5),tbx(100,6),tbx(100,7),tbx(100,8),tbx(100,9)
     *   / 2.89289   ,17.9144   ,-7.9492   ,.005127   ,13.0174   /
      data tbx(100,10)/45./
      data tbxc(101),tbx(101,1),tbx(101,2),tbx(101,3),tbx(101,4)
     *   /'ag  '   ,19.2808   ,  .6446   ,16.6885   , 7.4726   /
      data tbx(101,5),tbx(101,6),tbx(101,7),tbx(101,8),tbx(101,9)
     *   /  4.8045   ,24.6605   , 1.0463   ,99.8156   , 5.1790   /
      data tbx(101,10)/46./
      data tbxc(102),tbx(102,1),tbx(102,2),tbx(102,3),tbx(102,4)
     *   /'ag+1'   ,19.1812   ,.646179   ,15.9719   ,7.19123   /
      data tbx(102,5),tbx(102,6),tbx(102,7),tbx(102,8),tbx(102,9)
     *   / 5.27475   ,21.7326   ,.357534   ,66.1147   ,5.21572   /
      data tbx(102,10)/46./
      data tbxc(103),tbx(103,1),tbx(103,2),tbx(103,3),tbx(103,4)
     *   /'ag+2'   ,19.1643   ,.645643   ,16.2456   ,7.18544   /
      data tbx(103,5),tbx(103,6),tbx(103,7),tbx(103,8),tbx(103,9)
     *   / 4.37090   ,21.4072   ,.0000   ,.0000   ,5.21404   /
      data tbx(103,10)/46./
      data tbxc(104),tbx(104,1),tbx(104,2),tbx(104,3),tbx(104,4)
     *   /'cd  '   ,19.2214   ,  .5946   ,17.6444   , 6.9089   /
      data tbx(104,5),tbx(104,6),tbx(104,7),tbx(104,8),tbx(104,9)
     *   /  4.4610   ,24.7008   , 1.6029   ,87.4825   , 5.0694   /
      data tbx(104,10)/47./
      data tbxc(105),tbx(105,1),tbx(105,2),tbx(105,3),tbx(105,4)
     *   /'cd+2'   ,19.1514   ,.597922   ,17.2535   ,6.80639   /
      data tbx(105,5),tbx(105,6),tbx(105,7),tbx(105,8),tbx(105,9)
     *   / 4.47128   ,20.2521   ,.0000   ,.0000   ,5.11937   /
      data tbx(105,10)/47./
      data tbxc(106),tbx(106,1),tbx(106,2),tbx(106,3),tbx(106,4)
     *   /'in  '   ,19.1624   ,  .5476   ,18.5596   , 6.3776   /
      data tbx(106,5),tbx(106,6),tbx(106,7),tbx(106,8),tbx(106,9)
     *   /  4.2948   ,25.8499   , 2.0396   ,92.8029   , 4.9391   /
      data tbx(106,10)/48./
      data tbxc(107),tbx(107,1),tbx(107,2),tbx(107,3),tbx(107,4)
     *   /'in+3'   ,19.1045   ,.551522   ,18.1108   ,6.32470   /
      data tbx(107,5),tbx(107,6),tbx(107,7),tbx(107,8),tbx(107,9)
     *   / 3.78897   ,17.3595   ,.0000   ,.0000   ,4.99635   /
      data tbx(107,10)/48./
      data tbxc(108),tbx(108,1),tbx(108,2),tbx(108,3),tbx(108,4)
     *   /'sn  '   ,19.1889   , 5.8303   ,19.1005   ,  .5031   /
      data tbx(108,5),tbx(108,6),tbx(108,7),tbx(108,8),tbx(108,9)
     *   /  4.4585   ,26.8909   , 2.4663   ,83.9571   , 4.7821   /
      data tbx(108,10)/49./
      data tbxc(109),tbx(109,1),tbx(109,2),tbx(109,3),tbx(109,4)
     *   /'sn+2'   ,19.1094   ,  .5036   ,19.0548   , 5.8378   /
      data tbx(109,5),tbx(109,6),tbx(109,7),tbx(109,8),tbx(109,9)
     *   /  4.5648   ,23.3752   ,  .4870   ,62.2061   , 4.7861   /
      data tbx(109,10)/49./
      data tbxc(110),tbx(110,1),tbx(110,2),tbx(110,3),tbx(110,4)
     *   /'sn+4'   ,18.9333   , 5.7640   ,19.7131   ,  .4655   /
      data tbx(110,5),tbx(110,6),tbx(110,7),tbx(110,8),tbx(110,9)
     *   /  3.4182   ,14.0049   ,  .0193   , -.7583   , 3.9182   /
      data tbx(110,10)/49./
      data tbxc(111),tbx(111,1),tbx(111,2),tbx(111,3),tbx(111,4)
     *   /'sb  '   ,19.6418   , 5.3034   ,19.0455   ,  .4607   /
      data tbx(111,5),tbx(111,6),tbx(111,7),tbx(111,8),tbx(111,9)
     *   /  5.0371   ,27.9074   , 2.6827   ,75.2825   , 4.5909   /
      data tbx(111,10)/50./
      data tbxc(112),tbx(112,1),tbx(112,2),tbx(112,3),tbx(112,4)
     *   /'sb+3'   ,18.9755   ,.467196   ,18.9330   ,5.22126   /
      data tbx(112,5),tbx(112,6),tbx(112,7),tbx(112,8),tbx(112,9)
     *   / 5.10789   ,19.5902   ,.288753   ,55.5113   ,4.69626   /
      data tbx(112,10)/50./
      data tbxc(113),tbx(113,1),tbx(113,2),tbx(113,3),tbx(113,4)
     *   /'sb+5'   ,19.8685   ,5.44853   ,19.0302   ,.467973   /
      data tbx(113,5),tbx(113,6),tbx(113,7),tbx(113,8),tbx(113,9)
     *   / 2.41253   ,14.1259   ,.0000   ,.0000   ,4.69263   /
      data tbx(113,10)/50./
      data tbxc(114),tbx(114,1),tbx(114,2),tbx(114,3),tbx(114,4)
     *   /'te  '   ,19.9644   ,4.81742   ,19.0138   ,.420885   /
      data tbx(114,5),tbx(114,6),tbx(114,7),tbx(114,8),tbx(114,9)
     *   / 6.14487   ,28.5284   ,2.52390   ,70.8403   ,4.35200   /
      data tbx(114,10)/51./
      data tbxc(115),tbx(115,1),tbx(115,2),tbx(115,3),tbx(115,4)
     *   /'i   '   ,20.1472   , 4.3470   ,18.9949   ,  .3814   /
      data tbx(115,5),tbx(115,6),tbx(115,7),tbx(115,8),tbx(115,9)
     *   /  7.5138   ,27.7660   , 2.2735   ,66.8776   , 4.0712   /
      data tbx(115,10)/52./
      data tbxc(116),tbx(116,1),tbx(116,2),tbx(116,3),tbx(116,4)
     *   /'i-1 '   ,20.2332   , 4.3579   ,18.9970   ,  .3815   /
      data tbx(116,5),tbx(116,6),tbx(116,7),tbx(116,8),tbx(116,9)
     *   /  7.8069   ,29.5259   , 2.8868   ,84.9304   , 4.0714   /
      data tbx(116,10)/52./
      data tbxc(117),tbx(117,1),tbx(117,2),tbx(117,3),tbx(117,4)
     *   /'xe  '   ,20.2933   , 3.9282   ,19.0298   ,  .3440   /
      data tbx(117,5),tbx(117,6),tbx(117,7),tbx(117,8),tbx(117,9)
     *   /  8.9767   ,26.4659   , 1.9900   ,64.2658   , 3.7118   /
      data tbx(117,10)/53./
      data tbxc(118),tbx(118,1),tbx(118,2),tbx(118,3),tbx(118,4)
     *   /'cs  '   ,20.3892   , 3.5690   ,19.1062   ,  .3107   /
      data tbx(118,5),tbx(118,6),tbx(118,7),tbx(118,8),tbx(118,9)
     *   / 10.6620   ,24.3879   , 1.4953   ,213.904   , 3.3352   /
      data tbx(118,10)/54./
      data tbxc(119),tbx(119,1),tbx(119,2),tbx(119,3),tbx(119,4)
     *   /'cs+1'   ,20.3524   , 3.5520   ,19.1278   ,  .3086   /
      data tbx(119,5),tbx(119,6),tbx(119,7),tbx(119,8),tbx(119,9)
     *   / 10.2821   ,23.7128   ,  .9615   ,59.4565   , 3.2791   /
      data tbx(119,10)/54./
      data tbxc(120),tbx(120,1),tbx(120,2),tbx(120,3),tbx(120,4)
     *   /'ba  '   ,20.3361   , 3.2160   ,19.2970   ,  .2756   /
      data tbx(120,5),tbx(120,6),tbx(120,7),tbx(120,8),tbx(120,9)
     *   / 10.8880   ,20.2073   , 2.6959   ,167.202   , 2.7731   /
      data tbx(120,10)/55./
      data tbxc(121),tbx(121,1),tbx(121,2),tbx(121,3),tbx(121,4)
     *   /'ba+2'   ,20.1807   ,3.21367   ,19.1136   ,.283310   /
      data tbx(121,5),tbx(121,6),tbx(121,7),tbx(121,8),tbx(121,9)
     *   / 10.9054   ,20.0558   ,.773634   ,51.7460   ,3.02902   /
      data tbx(121,10)/55./
      data tbxc(122),tbx(122,1),tbx(122,2),tbx(122,3),tbx(122,4)
     *   /'la  '   ,20.5780   ,2.94817   ,19.5990   ,.244475   /
      data tbx(122,5),tbx(122,6),tbx(122,7),tbx(122,8),tbx(122,9)
     *   / 11.3727   ,18.7726   ,3.28719   ,133.124   ,2.14678   /
      data tbx(122,10)/56./
      data tbxc(123),tbx(123,1),tbx(123,2),tbx(123,3),tbx(123,4)
     *   /'la+3'   ,20.2489   ,2.92070   ,19.3763   ,.250698   /
      data tbx(123,5),tbx(123,6),tbx(123,7),tbx(123,8),tbx(123,9)
     *   / 11.6323   ,17.8211   ,.336048   ,54.9453   ,2.40860   /
      data tbx(123,10)/56./
      data tbxc(124),tbx(124,1),tbx(124,2),tbx(124,3),tbx(124,4)
     *   /'ce  '   ,21.1671   ,2.81219   ,19.7695   ,.226836   /
      data tbx(124,5),tbx(124,6),tbx(124,7),tbx(124,8),tbx(124,9)
     *   / 11.8513   ,17.6083   ,3.33049   ,127.113   ,1.86264   /
      data tbx(124,10)/57./
      data tbxc(125),tbx(125,1),tbx(125,2),tbx(125,3),tbx(125,4)
     *   /'ce+3'   ,20.8036   ,2.77691   ,19.5590   ,.231540   /
      data tbx(125,5),tbx(125,6),tbx(125,7),tbx(125,8),tbx(125,9)
     *   / 11.9369   ,16.5408   ,.612376   ,43.1692   ,2.09013   /
      data tbx(125,10)/57./
      data tbxc(126),tbx(126,1),tbx(126,2),tbx(126,3),tbx(126,4)
     *   /'ce+4'   ,20.3235   ,2.65941   ,19.8186   ,.218850   /
      data tbx(126,5),tbx(126,6),tbx(126,7),tbx(126,8),tbx(126,9)
     *   / 12.1233   ,15.7992   ,.144583   ,62.2355   ,1.59180   /
      data tbx(126,10)/57./
      data tbxc(127),tbx(127,1),tbx(127,2),tbx(127,3),tbx(127,4)
     *   /'pr  '   ,22.0440   ,2.77393   ,19.6697   ,.222087   /
      data tbx(127,5),tbx(127,6),tbx(127,7),tbx(127,8),tbx(127,9)
     *   / 12.3856   ,16.7669   ,2.82428   ,143.644   ,2.05830   /
      data tbx(127,10)/58./
      data tbxc(128),tbx(128,1),tbx(128,2),tbx(128,3),tbx(128,4)
     *   /'pr+3'   ,21.3727   ,2.64520   ,19.7491   ,.214299   /
      data tbx(128,5),tbx(128,6),tbx(128,7),tbx(128,8),tbx(128,9)
     *   / 12.1329   ,15.3230   ,.975180   ,36.4065   ,1.77132   /
      data tbx(128,10)/58./
      data tbxc(129),tbx(129,1),tbx(129,2),tbx(129,3),tbx(129,4)
     *   /'pr+4'   ,20.9413   ,2.54467   ,20.0539   ,.202481   /
      data tbx(129,5),tbx(129,6),tbx(129,7),tbx(129,8),tbx(129,9)
     *   / 12.4668   ,14.8137   ,.296689   ,45.4643   ,1.24285   /
      data tbx(129,10)/58./
      data tbxc(130),tbx(130,1),tbx(130,2),tbx(130,3),tbx(130,4)
     *   /'nd  '   ,22.6845   ,2.66248   ,19.6847   ,.210628   /
      data tbx(130,5),tbx(130,6),tbx(130,7),tbx(130,8),tbx(130,9)
     *   / 12.7740   ,15.8850   ,2.85137   ,137.903   ,1.98486   /
      data tbx(130,10)/59./
      data tbxc(131),tbx(131,1),tbx(131,2),tbx(131,3),tbx(131,4)
     *   /'nd+3'   ,21.9610   ,2.52722   ,19.9339   ,.199237   /
      data tbx(131,5),tbx(131,6),tbx(131,7),tbx(131,8),tbx(131,9)
     *   / 12.1200   ,14.1783   ,1.51031   ,30.8717   ,1.47588   /
      data tbx(131,10)/59./
      data tbxc(132),tbx(132,1),tbx(132,2),tbx(132,3),tbx(132,4)
     *   /'pm  '   ,23.3405   ,2.56270   ,19.6095   ,.202088   /
      data tbx(132,5),tbx(132,6),tbx(132,7),tbx(132,8),tbx(132,9)
     *   / 13.1235   ,15.1009   ,2.87516   ,132.721   ,2.02876   /
      data tbx(132,10)/60./
      data tbxc(133),tbx(133,1),tbx(133,2),tbx(133,3),tbx(133,4)
     *   /'pm+3'   ,22.5527   ,2.41740   ,20.1108   ,.185769   /
      data tbx(133,5),tbx(133,6),tbx(133,7),tbx(133,8),tbx(133,9)
     *   / 12.0671   ,13.1275   ,2.07492   ,27.4491   ,1.19499   /
      data tbx(133,10)/60./
      data tbxc(134),tbx(134,1),tbx(134,2),tbx(134,3),tbx(134,4)
     *   /'sm  '   ,24.0042   ,2.47274   ,19.4258   ,.196451   /
      data tbx(134,5),tbx(134,6),tbx(134,7),tbx(134,8),tbx(134,9)
     *   / 13.4396   ,14.3996   ,2.89604   ,128.007   ,2.20963   /
      data tbx(134,10)/61./
      data tbxc(135),tbx(135,1),tbx(135,2),tbx(135,3),tbx(135,4)
     *   /'sm+3'   ,23.1504   ,2.31641   ,20.2599   ,.174081   /
      data tbx(135,5),tbx(135,6),tbx(135,7),tbx(135,8),tbx(135,9)
     *   / 11.9202   ,12.1571   ,2.71488   ,24.8242   ,.954586   /
      data tbx(135,10)/61./
      data tbxc(136),tbx(136,1),tbx(136,2),tbx(136,3),tbx(136,4)
     *   /'eu  '   ,24.6274   , 2.3879   ,19.0886   ,  .1942   /
      data tbx(136,5),tbx(136,6),tbx(136,7),tbx(136,8),tbx(136,9)
     *   / 13.7603   ,13.7546   , 2.9227   ,123.174   , 2.5745   /
      data tbx(136,10)/62./
      data tbxc(137),tbx(137,1),tbx(137,2),tbx(137,3),tbx(137,4)
     *   /'eu+2'   ,24.0063   ,2.27783   ,19.9504   ,.173530   /
      data tbx(137,5),tbx(137,6),tbx(137,7),tbx(137,8),tbx(137,9)
     *   / 11.8034   ,11.6096   ,3.87243   ,26.5156   ,1.36389   /
      data tbx(137,10)/62./
      data tbxc(138),tbx(138,1),tbx(138,2),tbx(138,3),tbx(138,4)
     *   /'eu+3'   ,23.7497   ,2.22258   ,20.3745   ,.163940   /
      data tbx(138,5),tbx(138,6),tbx(138,7),tbx(138,8),tbx(138,9)
     *   / 11.8509   ,11.3110   ,3.26503   ,22.9966   ,.759344   /
      data tbx(138,10)/62./
      data tbxc(139),tbx(139,1),tbx(139,2),tbx(139,3),tbx(139,4)
     *   /'gd  '   ,25.0709   ,2.25341   ,19.0798   ,.181951   /
      data tbx(139,5),tbx(139,6),tbx(139,7),tbx(139,8),tbx(139,9)
     *   / 13.8518   ,12.9331   ,3.54545   ,101.398   ,2.41960   /
      data tbx(139,10)/63./
      data tbxc(140),tbx(140,1),tbx(140,2),tbx(140,3),tbx(140,4)
     *   /'gd+3'   ,24.3466   ,2.13553   ,20.4208   ,.155525   /
      data tbx(140,5),tbx(140,6),tbx(140,7),tbx(140,8),tbx(140,9)
     *   / 11.8708   ,10.5782   ,3.71490   ,21.7029   ,.645089   /
      data tbx(140,10)/63./
      data tbxc(141),tbx(141,1),tbx(141,2),tbx(141,3),tbx(141,4)
     *   /'tb  '   ,25.8976   ,2.24256   ,18.2185   ,.196143   /
      data tbx(141,5),tbx(141,6),tbx(141,7),tbx(141,8),tbx(141,9)
     *   / 14.3167   ,12.6648   ,2.95354   ,115.362   ,3.58324   /
      data tbx(141,10)/64./
      data tbxc(142),tbx(142,1),tbx(142,2),tbx(142,3),tbx(142,4)
     *   /'tb+3'   ,24.9559   ,2.05601   ,20.3271   ,.149525   /
      data tbx(142,5),tbx(142,6),tbx(142,7),tbx(142,8),tbx(142,9)
     *   / 12.2471   ,10.0499   ,3.77300   ,21.2773   ,.691967   /
      data tbx(142,10)/64./
      data tbxc(143),tbx(143,1),tbx(143,2),tbx(143,3),tbx(143,4)
     *   /'dy  '   ,26.5070   ,2.18020   ,17.6383   ,.202172   /
      data tbx(143,5),tbx(143,6),tbx(143,7),tbx(143,8),tbx(143,9)
     *   / 14.5596   ,12.1899   ,2.96577   ,111.874   ,4.29728   /
      data tbx(143,10)/65./
      data tbxc(144),tbx(144,1),tbx(144,2),tbx(144,3),tbx(144,4)
     *   /'dy+3'   ,25.5395   ,1.98040   ,20.2861   ,.143384   /
      data tbx(144,5),tbx(144,6),tbx(144,7),tbx(144,8),tbx(144,9)
     *   / 11.9812   ,9.34972   ,4.50073   ,19.5810   ,.689690   /
      data tbx(144,10)/65./
      data tbxc(145),tbx(145,1),tbx(145,2),tbx(145,3),tbx(145,4)
     *   /'ho  '   ,26.9049   ,2.07051   ,17.2940   ,.197940   /
      data tbx(145,5),tbx(145,6),tbx(145,7),tbx(145,8),tbx(145,9)
     *   / 14.5583   ,11.4407   ,3.63837   ,92.6566   ,4.56796   /
      data tbx(145,10)/66./
      data tbxc(146),tbx(146,1),tbx(146,2),tbx(146,3),tbx(146,4)
     *   /'ho+3'   ,26.1296   ,1.91072   ,20.0994   ,.139358   /
      data tbx(146,5),tbx(146,6),tbx(146,7),tbx(146,8),tbx(146,9)
     *   / 11.9788   ,8.80018   ,4.93676   ,18.5908   ,.852795   /
      data tbx(146,10)/66./
      data tbxc(147),tbx(147,1),tbx(147,2),tbx(147,3),tbx(147,4)
     *   /'er  '   ,27.6563   ,2.07356   ,16.4285   ,.223545   /
      data tbx(147,5),tbx(147,6),tbx(147,7),tbx(147,8),tbx(147,9)
     *   / 14.9779   ,11.3604   ,2.98233   ,105.703   ,5.92046   /
      data tbx(147,10)/67./
      data tbxc(148),tbx(148,1),tbx(148,2),tbx(148,3),tbx(148,4)
     *   /'er+3'   ,26.7220   ,1.84659   ,19.7748   ,.137290   /
      data tbx(148,5),tbx(148,6),tbx(148,7),tbx(148,8),tbx(148,9)
     *   / 12.1506   ,8.36225   ,5.17379   ,17.8974   ,1.17613   /
      data tbx(148,10)/67./
      data tbxc(149),tbx(149,1),tbx(149,2),tbx(149,3),tbx(149,4)
     *   /'tm  '   ,28.1819   ,2.02859   ,15.8851   ,.238849   /
      data tbx(149,5),tbx(149,6),tbx(149,7),tbx(149,8),tbx(149,9)
     *   / 15.1542   ,10.9975   ,2.98706   ,102.961   ,6.75621   /
      data tbx(149,10)/68./
      data tbxc(150),tbx(150,1),tbx(150,2),tbx(150,3),tbx(150,4)
     *   /'tm+3'   ,27.3083   ,1.78711   ,19.3320   ,.136974   /
      data tbx(150,5),tbx(150,6),tbx(150,7),tbx(150,8),tbx(150,9)
     *   / 12.3339   ,7.96778   ,5.38348   ,17.2922   ,1.63929   /
      data tbx(150,10)/68./
      data tbxc(151),tbx(151,1),tbx(151,2),tbx(151,3),tbx(151,4)
     *   /'yb  '   ,28.6641   ,1.98890   ,15.4345   ,.257119   /
      data tbx(151,5),tbx(151,6),tbx(151,7),tbx(151,8),tbx(151,9)
     *   / 15.3087   ,10.6647   ,2.98963   ,100.417   ,7.56672   /
      data tbx(151,10)/69./
      data tbxc(152),tbx(152,1),tbx(152,2),tbx(152,3),tbx(152,4)
     *   /'yb+2'   ,28.1209   ,1.78503   ,17.6817   ,.159970   /
      data tbx(152,5),tbx(152,6),tbx(152,7),tbx(152,8),tbx(152,9)
     *   / 13.3335   ,8.18304   ,5.14657   ,20.3900   ,3.70983   /
      data tbx(152,10)/69./
      data tbxc(153),tbx(153,1),tbx(153,2),tbx(153,3),tbx(153,4)
     *   /'yb+3'   ,27.8917   ,1.73272   ,18.7614   ,.138790   /
      data tbx(153,5),tbx(153,6),tbx(153,7),tbx(153,8),tbx(153,9)
     *   / 12.6072   ,7.64412   ,5.47647   ,16.8153   ,2.26001   /
      data tbx(153,10)/69./
      data tbxc(154),tbx(154,1),tbx(154,2),tbx(154,3),tbx(154,4)
     *   /'lu  '   ,28.9476   ,1.90182   ,15.2208   ,9.98519   /
      data tbx(154,5),tbx(154,6),tbx(154,7),tbx(154,8),tbx(154,9)
     *   / 15.1000   ,.261033   ,3.71601   ,84.3298   ,7.97628   /
      data tbx(154,10)/70./
      data tbxc(155),tbx(155,1),tbx(155,2),tbx(155,3),tbx(155,4)
     *   /'lu+3'   ,28.4628   ,1.68216   ,18.1210   ,.142292   /
      data tbx(155,5),tbx(155,6),tbx(155,7),tbx(155,8),tbx(155,9)
     *   / 12.8429   ,7.33727   ,5.59415   ,16.3535   ,2.97573   /
      data tbx(155,10)/70./
      data tbxc(156),tbx(156,1),tbx(156,2),tbx(156,3),tbx(156,4)
     *   /'hf  '   ,29.1440   ,1.83262   ,15.1726   ,9.59990   /
      data tbx(156,5),tbx(156,6),tbx(156,7),tbx(156,8),tbx(156,9)
     *   / 14.7586   ,.275116   ,4.30013   ,72.0290   ,8.58154   /
      data tbx(156,10)/71./
      data tbxc(157),tbx(157,1),tbx(157,2),tbx(157,3),tbx(157,4)
     *   /'hf+4'   ,28.8131   ,1.59136   ,18.4601   ,.128903   /
      data tbx(157,5),tbx(157,6),tbx(157,7),tbx(157,8),tbx(157,9)
     *   / 12.7285   ,6.76232   ,5.59927   ,14.0366   ,2.39699   /
      data tbx(157,10)/71./
      data tbxc(158),tbx(158,1),tbx(158,2),tbx(158,3),tbx(158,4)
     *   /'ta  '   ,29.2024   ,1.77333   ,15.2293   ,9.37046   /
      data tbx(158,5),tbx(158,6),tbx(158,7),tbx(158,8),tbx(158,9)
     *   / 14.5135   ,.295977   ,4.76492   ,63.3644   ,9.24354   /
      data tbx(158,10)/72./
      data tbxc(159),tbx(159,1),tbx(159,2),tbx(159,3),tbx(159,4)
     *   /'ta+5'   ,29.1587   ,1.50711   ,18.8407   ,.116741   /
      data tbx(159,5),tbx(159,6),tbx(159,7),tbx(159,8),tbx(159,9)
     *   / 12.8268   ,6.31524   ,5.38695   ,12.4244   ,1.78555   /
      data tbx(159,10)/72./
      data tbxc(160),tbx(160,1),tbx(160,2),tbx(160,3),tbx(160,4)
     *   /'w   '   ,29.0818   ,1.72029   ,15.4300   ,9.22590   /
      data tbx(160,5),tbx(160,6),tbx(160,7),tbx(160,8),tbx(160,9)
     *   / 14.4327   ,.321703   ,5.11982   ,57.0560   ,9.88750   /
      data tbx(160,10)/73./
      data tbxc(161),tbx(161,1),tbx(161,2),tbx(161,3),tbx(161,4)
     *   /'w+6 '   ,29.4936   ,1.42755   ,19.3763   ,.104621   /
      data tbx(161,5),tbx(161,6),tbx(161,7),tbx(161,8),tbx(161,9)
     *   / 13.0544   ,5.93667   ,5.06412   ,11.1972   ,1.01074   /
      data tbx(161,10)/73./
      data tbxc(162),tbx(162,1),tbx(162,2),tbx(162,3),tbx(162,4)
     *   /'re  '   ,28.7621   ,1.67191   ,15.7189   ,9.09227   /
      data tbx(162,5),tbx(162,6),tbx(162,7),tbx(162,8),tbx(162,9)
     *   / 14.5564   ,.350500   ,5.44174   ,52.0861   ,10.4720   /
      data tbx(162,10)/74./
      data tbxc(163),tbx(163,1),tbx(163,2),tbx(163,3),tbx(163,4)
     *   /'os  '   ,28.1894   ,1.62903   ,16.1550   ,8.97948   /
      data tbx(163,5),tbx(163,6),tbx(163,7),tbx(163,8),tbx(163,9)
     *   / 14.9305   ,.382661   ,5.67589   ,48.1647   ,11.0005   /
      data tbx(163,10)/75./
      data tbxc(164),tbx(164,1),tbx(164,2),tbx(164,3),tbx(164,4)
     *   /'os+4'   ,30.4190   ,1.37113   ,15.2637   ,6.84706   /
      data tbx(164,5),tbx(164,6),tbx(164,7),tbx(164,8),tbx(164,9)
     *   / 14.7458   ,.165191   ,5.06795   ,18.0030   ,6.49804   /
      data tbx(164,10)/75./
      data tbxc(165),tbx(165,1),tbx(165,2),tbx(165,3),tbx(165,4)
     *   /'ir  '   ,27.3049   ,1.59279   ,16.7296   ,8.86553   /
      data tbx(165,5),tbx(165,6),tbx(165,7),tbx(165,8),tbx(165,9)
     *   / 15.6115   ,.417916   ,5.83377   ,45.0011   ,11.4722   /
      data tbx(165,10)/76./
      data tbxc(166),tbx(166,1),tbx(166,2),tbx(166,3),tbx(166,4)
     *   /'ir+3'   ,30.4156   ,1.34323   ,15.8620   ,7.10909   /
      data tbx(166,5),tbx(166,6),tbx(166,7),tbx(166,8),tbx(166,9)
     *   / 13.6145   ,.204633   ,5.82008   ,20.3254   ,8.27903   /
      data tbx(166,10)/76./
      data tbxc(167),tbx(167,1),tbx(167,2),tbx(167,3),tbx(167,4)
     *   /'ir+4'   ,30.7058   ,1.30923   ,15.5512   ,6.71983   /
      data tbx(167,5),tbx(167,6),tbx(167,7),tbx(167,8),tbx(167,9)
     *   / 14.2326   ,.167252   ,5.53672   ,17.4911   ,6.96824   /
      data tbx(167,10)/76./
      data tbxc(168),tbx(168,1),tbx(168,2),tbx(168,3),tbx(168,4)
     *   /'pt  '   ,27.0059   ,1.51293   ,17.7639   ,8.81174   /
      data tbx(168,5),tbx(168,6),tbx(168,7),tbx(168,8),tbx(168,9)
     *   / 15.7131   ,.424593   ,5.78370   ,38.6103   ,11.6883   /
      data tbx(168,10)/77./
      data tbxc(169),tbx(169,1),tbx(169,2),tbx(169,3),tbx(169,4)
     *   /'pt+2'   ,29.8429   ,1.32927   ,16.7224   ,7.38979   /
      data tbx(169,5),tbx(169,6),tbx(169,7),tbx(169,8),tbx(169,9)
     *   / 13.2153   ,.263297   ,6.35234   ,22.9426   ,9.85329   /
      data tbx(169,10)/77./
      data tbxc(170),tbx(170,1),tbx(170,2),tbx(170,3),tbx(170,4)
     *   /'pt+4'   ,30.9612   ,1.24813   ,15.9829   ,6.60834   /
      data tbx(170,5),tbx(170,6),tbx(170,7),tbx(170,8),tbx(170,9)
     *   / 13.7348   ,.168640   ,5.92034   ,16.9392   ,7.39534   /
      data tbx(170,10)/77./
      data tbxc(171),tbx(171,1),tbx(171,2),tbx(171,3),tbx(171,4)
     *   /'au  '   ,16.8819   ,  .4611   ,18.5913   , 8.6216   /
      data tbx(171,5),tbx(171,6),tbx(171,7),tbx(171,8),tbx(171,9)
     *   / 25.5582   , 1.4826   , 5.8600   ,36.3956   ,12.0658   /
      data tbx(171,10)/78./
      data tbxc(172),tbx(172,1),tbx(172,2),tbx(172,3),tbx(172,4)
     *   /'au+1'   ,28.0109   ,1.35321   ,17.8204   ,7.73950   /
      data tbx(172,5),tbx(172,6),tbx(172,7),tbx(172,8),tbx(172,9)
     *   / 14.3359   ,.356752   ,6.58077   ,26.4043   ,11.2299   /
      data tbx(172,10)/78./
      data tbxc(173),tbx(173,1),tbx(173,2),tbx(173,3),tbx(173,4)
     *   /'au+3'   ,30.6886   ,1.21990   ,16.9029   ,6.82872   /
      data tbx(173,5),tbx(173,6),tbx(173,7),tbx(173,8),tbx(173,9)
     *   / 12.7801   ,.212867   ,6.52354   ,18.6590   ,9.09680   /
      data tbx(173,10)/78./
      data tbxc(174),tbx(174,1),tbx(174,2),tbx(174,3),tbx(174,4)
     *   /'hg  '   ,20.6809   ,  .5450   ,19.0417   , 8.4484   /
      data tbx(174,5),tbx(174,6),tbx(174,7),tbx(174,8),tbx(174,9)
     *   / 21.6575   , 1.5729   , 5.9676   ,38.3246   ,12.6089   /
      data tbx(174,10)/79./
      data tbxc(175),tbx(175,1),tbx(175,2),tbx(175,3),tbx(175,4)
     *   /'hg+1'   ,25.0853   ,1.39507   ,18.4973   ,7.65105   /
      data tbx(175,5),tbx(175,6),tbx(175,7),tbx(175,8),tbx(175,9)
     *   / 16.8883   ,.443378   ,6.48216   ,28.2262   ,12.0205   /
      data tbx(175,10)/79./
      data tbxc(176),tbx(176,1),tbx(176,2),tbx(176,3),tbx(176,4)
     *   /'hg+2'   ,29.5641   ,1.21152   ,18.0600   ,7.05639   /
      data tbx(176,5),tbx(176,6),tbx(176,7),tbx(176,8),tbx(176,9)
     *   / 12.8374   ,.284738   ,6.89912   ,20.7482   ,10.6268   /
      data tbx(176,10)/79./
      data tbxc(177),tbx(177,1),tbx(177,2),tbx(177,3),tbx(177,4)
     *   /'tl  '   ,27.5446   ,.655150   ,19.1584   ,8.70751   /
      data tbx(177,5),tbx(177,6),tbx(177,7),tbx(177,8),tbx(177,9)
     *   / 15.5380   ,1.96347   ,5.52593   ,45.8149   ,13.1746   /
      data tbx(177,10)/80./
      data tbxc(178),tbx(178,1),tbx(178,2),tbx(178,3),tbx(178,4)
     *   /'tl+1'   ,21.3985   ,1.47110   ,20.4723   ,.517394   /
      data tbx(178,5),tbx(178,6),tbx(178,7),tbx(178,8),tbx(178,9)
     *   / 18.7478   ,7.43463   ,6.82847   ,28.8482   ,12.5258   /
      data tbx(178,10)/80./
      data tbxc(179),tbx(179,1),tbx(179,2),tbx(179,3),tbx(179,4)
     *   /'tl+3'   ,30.8695   ,1.10080   ,18.3841   ,6.53852   /
      data tbx(179,5),tbx(179,6),tbx(179,7),tbx(179,8),tbx(179,9)
     *   / 11.9328   ,.219074   ,7.00574   ,17.2114   ,9.80270   /
      data tbx(179,10)/80./
      data tbxc(180),tbx(180,1),tbx(180,2),tbx(180,3),tbx(180,4)
     *   /'pb  '   ,31.0617   ,  .6902   ,13.0637   , 2.3576   /
      data tbx(180,5),tbx(180,6),tbx(180,7),tbx(180,8),tbx(180,9)
     *   / 18.4420   , 8.6180   , 5.9696   ,47.2579   ,13.4118   /
      data tbx(180,10)/81./
      data tbxc(181),tbx(181,1),tbx(181,2),tbx(181,3),tbx(181,4)
     *   /'pb+2'   ,21.7886   ,1.33660   ,19.5682   ,.488383   /
      data tbx(181,5),tbx(181,6),tbx(181,7),tbx(181,8),tbx(181,9)
     *   / 19.1406   ,6.77270   ,7.01107   ,23.8132   ,12.4734   /
      data tbx(181,10)/81./
      data tbxc(182),tbx(182,1),tbx(182,2),tbx(182,3),tbx(182,4)
     *   /'pb+4'   ,32.1244   ,1.00566   ,18.8003   ,6.10926   /
      data tbx(182,5),tbx(182,6),tbx(182,7),tbx(182,8),tbx(182,9)
     *   / 12.0175   ,.147041   ,6.96886   ,14.7140   ,8.08428   /
      data tbx(182,10)/81./
      data tbxc(183),tbx(183,1),tbx(183,2),tbx(183,3),tbx(183,4)
     *   /'bi  '   ,33.3689   ,  .7040   ,12.9510   , 2.9238   /
      data tbx(183,5),tbx(183,6),tbx(183,7),tbx(183,8),tbx(183,9)
     *   / 16.5877   , 8.7937   , 6.4692   ,48.0093   ,13.5782   /
      data tbx(183,10)/82./
      data tbxc(184),tbx(184,1),tbx(184,2),tbx(184,3),tbx(184,4)
     *   /'bi+3'   ,21.8053   ,1.23560   ,19.5026   ,6.24149   /
      data tbx(184,5),tbx(184,6),tbx(184,7),tbx(184,8),tbx(184,9)
     *   / 19.1053   ,.469999   ,7.10295   ,20.3185   ,12.4711   /
      data tbx(184,10)/82./
      data tbxc(185),tbx(185,1),tbx(185,2),tbx(185,3),tbx(185,4)
     *   /'bi+5'   ,33.5364   ,.916540   ,25.0946   ,.039042   /
      data tbx(185,5),tbx(185,6),tbx(185,7),tbx(185,8),tbx(185,9)
     *   / 19.2497   ,5.71414   ,6.91555   ,12.8285   ,-6.7994   /
      data tbx(185,10)/82./
      data tbxc(186),tbx(186,1),tbx(186,2),tbx(186,3),tbx(186,4)
     *   /'po  '   ,34.6726   ,.700999   ,15.4733   ,3.55078   /
      data tbx(186,5),tbx(186,6),tbx(186,7),tbx(186,8),tbx(186,9)
     *   / 13.1138   ,9.55642   ,7.02588   ,47.0045   ,13.6770   /
      data tbx(186,10)/83./
      data tbxc(187),tbx(187,1),tbx(187,2),tbx(187,3),tbx(187,4)
     *   /'at  '   ,35.3163   ,.685870   ,19.0211   ,3.97458   /
      data tbx(187,5),tbx(187,6),tbx(187,7),tbx(187,8),tbx(187,9)
     *   / 9.49887   ,11.3824   ,7.42518   ,45.4715   ,13.7108   /
      data tbx(187,10)/84./
      data tbxc(188),tbx(188,1),tbx(188,2),tbx(188,3),tbx(188,4)
     *   /'rn  '   ,35.5631   ,  .6631   ,21.2816   , 4.0691   /
      data tbx(188,5),tbx(188,6),tbx(188,7),tbx(188,8),tbx(188,9)
     *   /  8.0037   ,14.0422   , 7.4433   ,44.2473   ,13.6905   /
      data tbx(188,10)/85./
      data tbxc(189),tbx(189,1),tbx(189,2),tbx(189,3),tbx(189,4)
     *   /'fr  '   ,35.9299   ,.646453   ,23.0547   ,4.17619   /
      data tbx(189,5),tbx(189,6),tbx(189,7),tbx(189,8),tbx(189,9)
     *   / 12.1439   ,23.1052   ,2.11253   ,150.645   ,13.7247   /
      data tbx(189,10)/86./
      data tbxc(190),tbx(190,1),tbx(190,2),tbx(190,3),tbx(190,4)
     *   /'ra  '   ,35.7630   ,.616341   ,22.9064   ,3.87135   /
      data tbx(190,5),tbx(190,6),tbx(190,7),tbx(190,8),tbx(190,9)
     *   / 12.4739   ,19.9887   ,3.21097   ,142.325   ,13.6211   /
      data tbx(190,10)/87./
      data tbxc(191),tbx(191,1),tbx(191,2),tbx(191,3),tbx(191,4)
     *   /'ra+2'   ,35.2150   ,.604909   ,21.6700   ,3.57670   /
      data tbx(191,5),tbx(191,6),tbx(191,7),tbx(191,8),tbx(191,9)
     *   / 7.91342   ,12.6010   ,7.65078   ,29.8436   ,13.5431   /
      data tbx(191,10)/87./
      data tbxc(192),tbx(192,1),tbx(192,2),tbx(192,3),tbx(192,4)
     *   /'ac  '   ,35.6597   ,.589092   ,23.1032   ,3.65155   /
      data tbx(192,5),tbx(192,6),tbx(192,7),tbx(192,8),tbx(192,9)
     *   / 12.5977   ,18.5990   ,4.08655   ,117.020   ,13.5266   /
      data tbx(192,10)/88./
      data tbxc(193),tbx(193,1),tbx(193,2),tbx(193,3),tbx(193,4)
     *   /'ac+3'   ,35.1736   ,.579689   ,22.1112   ,3.41437   /
      data tbx(193,5),tbx(193,6),tbx(193,7),tbx(193,8),tbx(193,9)
     *   / 8.19216   ,12.9187   ,7.05545   ,25.9443   ,13.4637   /
      data tbx(193,10)/88./
      data tbxc(194),tbx(194,1),tbx(194,2),tbx(194,3),tbx(194,4)
     *   /'th  '   ,35.5645   ,.563359   ,23.4219   ,3.46204   /
      data tbx(194,5),tbx(194,6),tbx(194,7),tbx(194,8),tbx(194,9)
     *   / 12.7473   ,17.8309   ,4.80703   ,99.1722   ,13.4314   /
      data tbx(194,10)/89./
      data tbxc(195),tbx(195,1),tbx(195,2),tbx(195,3),tbx(195,4)
     *   /'th+4'   ,35.1007   ,.555054   ,22.4418   ,3.24498   /
      data tbx(195,5),tbx(195,6),tbx(195,7),tbx(195,8),tbx(195,9)
     *   / 9.78554   ,13.4661   ,5.29444   ,23.9533   ,13.3760   /
      data tbx(195,10)/89./
      data tbxc(196),tbx(196,1),tbx(196,2),tbx(196,3),tbx(196,4)
     *   /'pa  '   ,35.8847   ,.547751   ,23.2948   ,3.41519   /
      data tbx(196,5),tbx(196,6),tbx(196,7),tbx(196,8),tbx(196,9)
     *   / 14.1891   ,16.9235   ,4.17287   ,105.251   ,13.4287   /
      data tbx(196,10)/90./
      data tbxc(197),tbx(197,1),tbx(197,2),tbx(197,3),tbx(197,4)
     *   /'u   '   ,36.0228   ,  .5293   ,23.4128   , 3.3253   /
      data tbx(197,5),tbx(197,6),tbx(197,7),tbx(197,8),tbx(197,9)
     *   / 14.9491   ,16.0927   , 4.1880   ,100.613   ,13.3966   /
      data tbx(197,10)/91./
      data tbxc(198),tbx(198,1),tbx(198,2),tbx(198,3),tbx(198,4)
     *   /'u+3 '  ,35.5747   ,.520480   ,22.5259   ,3.12293   /
      data tbx(198,5),tbx(198,6),tbx(198,7),tbx(198,8),tbx(198,9)
     *   / 12.2165   ,12.7148   ,5.37073   ,26.3394   ,13.3092   /
      data tbx(198,10)/91./
      data tbxc(199),tbx(199,1),tbx(199,2),tbx(199,3),tbx(199,4)
     *   /'u+4 '  ,35.3715   ,.516598   ,22.5326   ,3.05053   /
      data tbx(199,5),tbx(199,6),tbx(199,7),tbx(199,8),tbx(199,9)
     *   / 12.0291   ,12.5723   ,4.79840   ,23.4582   ,13.2671   /
      data tbx(199,10)/91./
      data tbxc(200),tbx(200,1),tbx(200,2),tbx(200,3),tbx(200,4)
     *   /'u+6 '  ,34.8509   ,.507079   ,22.7584   ,2.89030   /
      data tbx(200,5),tbx(200,6),tbx(200,7),tbx(200,8),tbx(200,9)
     *   / 14.0099   ,13.1767   ,1.21457   ,25.2017   ,13.1665   /
      data tbx(200,10)/91./
      data tbxc(201),tbx(201,1),tbx(201,2),tbx(201,3),tbx(201,4)
     *   /'np  '   ,36.1874   ,.511929   ,23.5964   ,3.25396   /
      data tbx(201,5),tbx(201,6),tbx(201,7),tbx(201,8),tbx(201,9)
     *   / 15.6402   ,15.3622   ,4.18550   ,97.4908   ,13.3573   /
      data tbx(201,10)/92./
      data tbxc(202),tbx(202,1),tbx(202,2),tbx(202,3),tbx(202,4)
     *   /'np+3'   ,35.7074   ,.502322   ,22.6130   ,3.03807   /
      data tbx(202,5),tbx(202,6),tbx(202,7),tbx(202,8),tbx(202,9)
     *   / 12.9898   ,12.1449   ,5.43227   ,25.4928   ,13.2544   /
      data tbx(202,10)/92./
      data tbxc(203),tbx(203,1),tbx(203,2),tbx(203,3),tbx(203,4)
     *   /'np+4'   ,35.5103   ,.498626   ,22.5787   ,2.96627   /
      data tbx(203,5),tbx(203,6),tbx(203,7),tbx(203,8),tbx(203,9)
     *   / 12.7766   ,11.9484   ,4.92159   ,22.7502   ,13.2116   /
      data tbx(203,10)/92./
      data tbxc(204),tbx(204,1),tbx(204,2),tbx(204,3),tbx(204,4)
     *   /'np+6'   ,35.0136   ,.489810   ,22.7286   ,2.81099   /
      data tbx(204,5),tbx(204,6),tbx(204,7),tbx(204,8),tbx(204,9)
     *   / 14.3884   ,12.3300   ,1.75669   ,22.6581   ,13.1130   /
      data tbx(204,10)/92./
      data tbxc(205),tbx(205,1),tbx(205,2),tbx(205,3),tbx(205,4)
     *   /'pu  '   ,36.5254   ,.499384   ,23.8083   ,3.26371   /
      data tbx(205,5),tbx(205,6),tbx(205,7),tbx(205,8),tbx(205,9)
     *   / 16.7707   ,14.9455   ,3.47947   ,105.980   ,13.3812   /
      data tbx(205,10)/93./
      data tbxc(206),tbx(206,1),tbx(206,2),tbx(206,3),tbx(206,4)
     *   /'pu+3'   ,35.8400   ,.484938   ,22.7169   ,2.96118   /
      data tbx(206,5),tbx(206,6),tbx(206,7),tbx(206,8),tbx(206,9)
     *   / 13.5807   ,11.5331   ,5.66016   ,24.3992   ,13.1991   /
      data tbx(206,10)/93./
      data tbxc(207),tbx(207,1),tbx(207,2),tbx(207,3),tbx(207,4)
     *   /'pu+4'   ,35.6493   ,.481422   ,22.6460   ,2.89020   /
      data tbx(207,5),tbx(207,6),tbx(207,7),tbx(207,8),tbx(207,9)
     *   / 13.3595   ,11.3160   ,5.18831   ,21.8301   ,13.1555   /
      data tbx(207,10)/93./
      data tbxc(208),tbx(208,1),tbx(208,2),tbx(208,3),tbx(208,4)
     *   /'pu+6'   ,35.1736   ,.473204   ,22.7181   ,2.73848   /
      data tbx(208,5),tbx(208,6),tbx(208,7),tbx(208,8),tbx(208,9)
     *   / 14.7635   ,11.5530   ,2.28678   ,20.9303   ,13.0582   /
      data tbx(208,10)/93./
      data tbxc(209),tbx(209,1),tbx(209,2),tbx(209,3),tbx(209,4)
     *   /'am  '   ,36.6706   ,.483629   ,24.0992   ,3.20647   /
      data tbx(209,5),tbx(209,6),tbx(209,7),tbx(209,8),tbx(209,9)
     *   / 17.3415   ,14.3136   ,3.49331   ,102.273   ,13.3592   /
      data tbx(209,10)/94./
      data tbxc(210),tbx(210,1),tbx(210,2),tbx(210,3),tbx(210,4)
     *   /'cm  '   ,36.6488   ,.465154   ,24.4096   ,3.08997   /
      data tbx(210,5),tbx(210,6),tbx(210,7),tbx(210,8),tbx(210,9)
     *   / 17.3990   ,13.4346   ,4.21665   ,88.4834   ,13.2887   /
      data tbx(210,10)/95./
      data tbxc(211),tbx(211,1),tbx(211,2),tbx(211,3),tbx(211,4)
     *   /'bk  '   ,36.7881   ,.451018   ,24.7736   ,3.04619   /
      data tbx(211,5),tbx(211,6),tbx(211,7),tbx(211,8),tbx(211,9)
     *   / 17.8919   ,12.8946   ,4.23284   ,86.0030   ,13.2754   /
      data tbx(211,10)/96./
      data tbxc(212),tbx(212,1),tbx(212,2),tbx(212,3),tbx(212,4)
     *   /'cf  '   ,36.9185   ,.437533   ,25.1995   ,3.00775   /
      data tbx(212,5),tbx(212,6),tbx(212,7),tbx(212,8),tbx(212,9)
     *   / 18.3317   ,12.4044   ,4.24391   ,83.7881   ,13.2674   /
      data tbx(212,10)/97./
      END
C

